(* clap.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

(* Toplevel *)

let xTRACE_toplevel = ref true

open Form

(* Headers of Saved File (16 chars). *)

let ofile_header = "LOLIGO-EDULIS-O\n"
let afile_header = "LOLIGO-EDULIS-A\n"
let efile_header = "LOLIGO-EDULIS-E\n"

let string_equal = Misc.string_equal

let debug = Cexpr.debug
let info = Cexpr.info
let warning = Cexpr.warning
let error = Cexpr.error
let emerg = Cexpr.emerg

let parser_error_code = 100
let specs_error_code = 101
let timeout_error_code = 102
let uncaught_exception_error_code = 103

(* ================================================================ *)

let save_ofile (f : string) (defs : Translate.ofile_t) : unit = (
  (debug "saving to file...") ;
  let c : out_channel = (open_out_bin f) in
  (output_string c ofile_header) ;
  (Marshal.to_channel c defs [(*flags*)]) ;
  (flush c) ;
  (close_out c) ;
  (debug "saving to file...done"))

let load_ofile (f : string) : Translate.ofile_t = (
  (debug "loading from file...") ;
  let c : in_channel = (open_in_bin f) in
  let h = "----------------" in
  let n = (input c h 0 16) in
  if (n != 16) then
    (failwith "no header") ;
  if (not (string_equal h ofile_header)) then
    (failwith "bad header") ;
  let defs : Translate.ofile_t = (Marshal.from_channel c) in
  (close_in c) ;
  (debug "loading from file...done") ;
  defs)

let save_afile (f : string) (defs : Translate.afile_t) : unit = (
  (debug "saving to file...") ;
  let c : out_channel = (open_out_bin f) in
  (output_string c afile_header) ;
  (Marshal.to_channel c defs [(*flags*)]) ;
  (flush c) ;
  (close_out c) ;
  (debug "saving to file...done"))

let load_afile (f : string) : Translate.afile_t = (
  (debug "loading from file...") ;
  let c : in_channel = (open_in_bin f) in
  let h = "----------------" in
  let n = (input c h 0 16) in
  if (n != 16) then
    (failwith "no header") ;
  if (not (string_equal h afile_header)) then
    (failwith "bad header") ;
  let defs : Translate.afile_t = (Marshal.from_channel c) in
  (close_in c) ;
  (debug "loading from file...done") ;
  defs)

(* ================================================================ *)

(* Quotes a string; improper, not (Unix) shells but for OCaml. *)

let shell_escape (s : string) : string = (
  ("'" ^ (String.escaped s) ^ "'"))

let badoption (command : string) (opt : string list) : 'a = (
  (failwith ("Bad option to " ^ command ^ " '"
          ^ (String.concat " " opt) ^ "'")))

(* Scans argv and splits strings with/without "-".  It returns a list
   of arguments and a list of option sequences.  {longoptions} lists
   options taking an argument.  Not work like getopt. *)

let rec scan_options (command : string) (longoptions : string list) (argv : string list) (options : string list list) : (string list * string list list) = (
  match argv with
    [] ->
      ([], options)
  | (k :: rest1) ->
      let longkey = (Misc.list_find_safe "no"
            (fun o -> (Misc.string_starts_with o k))
            longoptions) in
      if (not (Misc.string_starts_with "-" k)) then
    (argv, options)
      else if (Misc.string_equal "-" k) then
    (rest1, options)
      else if (Misc.string_equal "--" k) then
    (rest1, options)
      else if ((String.length k) == 2) then
    (begin
      let takearg = (not (string_equal longkey "no")) in
      match rest1 with
        [] ->
          if (not takearg) then
        ([], (options @ [[k]]))
          else
        (badoption command argv)
      | (a :: rest2) ->
          if (not takearg) then
        (scan_options command longoptions rest1 (options @ [[k]]))
          else if (Misc.string_starts_with "-" a) then
        (badoption command argv)
          else
        (scan_options command longoptions rest2 (options @ [[k ; a]]))
    end)
      else if (not (string_equal longkey "no")) then
    (begin
      let l = (String.length longkey) in
      let a = (String.sub k l ((String.length k) - l)) in
      if (string_equal longkey k) then
        (begin
          (* Options like "-D XXXX" *)
          match rest1 with
        [] ->
          (badoption command argv)
          | (a :: rest2) ->
          if (Misc.string_starts_with "-" a) then
            (badoption command argv)
          else
            (scan_options command longoptions rest2
            (options @ [[longkey ; a]]))
        end)
      else if ((String.get k l) == ',') then
        (begin
          (* Options like "-Wp,-MD,XXX" *)
          let a1 = (List.tl (Misc.string_tokens [','] a)) in
          (scan_options command longoptions rest1
          (options @ [(longkey :: a1)]))
        end)
      else
        (begin
          (* Options like "-DXXXX" *)
          (scan_options command longoptions rest1
          (options @ [[longkey ; a]]))
        end)
    end)
      else
    (scan_options command longoptions rest1 (options @ [[k]])))

type cc_options_t = {
  mutable ccopt_output : string ;
  ccopt_cppdefines : string Vector.t ;
  ccopt_cppoptions : string Vector.t ;
  ccopt_cppundefines : string Vector.t ;
  ccopt_incdirectories : string Vector.t ;
  ccopt_libdirectories : string Vector.t ;
  ccopt_flags : string Vector.t ;
  ccopt_includes : string Vector.t
}

(* Quotes a string; improper, not (Unix) shells but for OCaml. *)

let shell_escape = String.escaped

(* Runs CPP on a file and returns an output file name.  String quotes
   are improper; not (Unix) shells but for OCaml.  NOTE: It runs "gcc"
   with "-E" instead of "/usr/bin/cpp", because "-MD" option is
   interpreted differently by "gcc" and "cpp". *)

let cpp ?(keep_comments = false) (options : cc_options_t) (file : string) : string = (
  (* Run cpp. *)
  let run_cpp input optionstring output =
  (begin
    if (!xTRACE_toplevel) then
      (debug (";; cpp command: "
               ^ ("gcc -E" ^ " " ^ optionstring
               ^ " " ^ input ^ " > " ^ output) ^ "")) ;
    let exitcode : int =
      (Sys.command ("gcc -E" ^ " " ^ optionstring
             ^ " " ^ input ^ " > " ^ output)) in
    if (exitcode != 0) then
      (Cexpr.generalerror
      ("Bad exit from gcc -E: code=" ^ (string_of_int exitcode)))
  end) in
  let run_sed input optionstring output =
  (begin
    let exitcode : int =
      (Sys.command ("sed" ^ " " ^ optionstring ^ " "
             ^ " < " ^ input ^ " >> " ^ output)) in
    if (exitcode != 0) then
      (Cexpr.generalerror
      ("Bad exit from sed: code=" ^ (string_of_int exitcode)))
  end) in
  (* Add current to include directory, for file is copied in tmp. *)
  let filedirectory = (Filename.dirname file) in
  (Vector.add_first options.ccopt_incdirectories filedirectory) ;
  let temporary0 = (Filename.temp_file "cp0" ".c") in
  let temporary1 = (Filename.temp_file "cp1" ".c") in
  let temporary2 = (Filename.temp_file "cp2" ".c") in
  let temporary3 = (Filename.temp_file "cp3" ".c") in
  let temporary4 = (Filename.temp_file "cp4" ".c") in
  let temp = if keep_comments then "-C " else "" in
  let optionstring : string =
    temp ^ 
    (String.concat " "
    ([]
      @ (Vector.map
        (fun s -> ("-I" ^ (shell_escape s)))
        options.ccopt_incdirectories)
      @ (Vector.map
        (fun s -> ("-L" ^ (shell_escape s)))
        options.ccopt_libdirectories)
      @ (Vector.map
        (fun s -> ("-D" ^ (shell_escape s)))
        options.ccopt_cppdefines)
      @ (Vector.map
        (fun s -> ("-U" ^ (shell_escape s)))
        options.ccopt_cppundefines)
      @ (Vector.map
           (fun s -> (shell_escape s))
           options.ccopt_cppoptions)
      @ (Vector.map
        (fun s -> let s = match s with
                          | "-c"
                          | "-g" -> ""		(* add nothing *)		
                          | "-nostdinc" -> s
                          | _ -> (badoption "cc" [s])
	          in (shell_escape s))
	options.ccopt_flags))) in
  let includestring : string =
      String.concat " "
      ([] @ Vector.map (fun s -> (shell_escape s)) options.ccopt_includes)
  in

  (* Run echo to add "# 1 <filename>" line. *)
  (begin
    let exitcode0 : int =
      (Sys.command ("echo '# 1 \"" ^ (shell_escape file) ^ "\"' > "
             ^ temporary0)) in
    if (exitcode0 != 0) then
      (Cexpr.generalerror
      ("Bad exit from echo: code=" ^ (string_of_int exitcode0)))
  end) ;
  (* Run sed to remove "/*@" and "@*/". *)
  if (not keep_comments) then 
  (begin
    let sedcommand : string = "-e 's|/\\*@| |' -e 's|@\\*/| |'" in
    let exitcode1 : int =
      (Sys.command ("sed" ^ " " ^ sedcommand ^ " "
             ^ " < " ^ file ^ " >> " ^ temporary0)) in
    if (exitcode1 != 0) then
      (Cexpr.generalerror
      ("Bad exit from sed: code=" ^ (string_of_int exitcode1)));
    (* Run cpp. *)
    (run_cpp temporary0 optionstring temporary1) ;
    temporary1
  end)
  else
  (begin
    (* Include header files *)
    let optionstring1 = optionstring ^ " " ^ includestring
                        ^ " -fdirectives-only "
    in
    (run_cpp file optionstring1 temporary1) ;
    (* Temporary hide /*@, /*!, //@, and #{ from cpp *)
    let sedcommand1 : string = " -e 's|/\\*@|__MC_ANNOT_BEGIN|g'"
                               ^ " -e 's|//@|__MC_ANNOT_SINGLELINE|g'"
                               ^ " -e 's|/\\*!|__MC_DOXYANNOT_BEGIN|g'"
                               ^ " -e 's|#{|__MC_ANNOT_COMM_BEGIN|g'"
    in
    (run_sed temporary1 sedcommand1 temporary2) ;
    (* Run cpp. *)
    (run_cpp temporary2 optionstring temporary3) ;
    (* Bring back /*@, /*!, and #{*)
    let sedcommand2 : string = " -e 's|__MC_ANNOT_BEGIN|/\\*@|g'"
                               ^ " -e 's|__MC_ANNOT_SINGLELINE|//@|g'"
                               ^ " -e 's|__MC_DOXYANNOT_BEGIN|/\\*!|g'"
                               ^ " -e 's|__MC_ANNOT_COMM_BEGIN|#{|g'"
    in
    (run_sed temporary3 sedcommand2 temporary4) ;
    temporary4
  end))

let rec make_cc_options ?(harsh_handling = true) (opts : string list list) : cc_options_t = (
  let options = {
    ccopt_output = "" ;
    ccopt_cppdefines = (Vector.create 4 "") ;
    ccopt_cppoptions = (Vector.create 4 "") ;
    ccopt_cppundefines = (Vector.create 4 "") ;
    ccopt_incdirectories = (Vector.create 4 "") ;
    ccopt_libdirectories = (Vector.create 4 "") ;
    ccopt_flags = (Vector.create 4 "") ;
    ccopt_includes = (Vector.create 4 "")
  } in
  (List.iter (set_cc_options ~harsh_handling:harsh_handling options) opts) ;
  options)

and set_cc_options ?(harsh_handling = true) (options : cc_options_t) (opt : string list) : unit = (
  let pmsg (l : string list) : unit = (
    if (!xTRACE_toplevel) then
      (debug ("ignore cc option: '"
               ^ (String.concat " " l) ^ "'"))) in
  match opt with
    [] -> (assert false)
  | ["-c"] ->
      (Vector.add options.ccopt_flags "-c") ;
  | ["-g"] ->
      (Vector.add options.ccopt_flags "-g") ;
  | ["-s"] ->
      (begin (pmsg opt) ; () end)
  | ["-o" ; a] ->
      (options.ccopt_output <- a)
  | ["-D" ; a] ->
      (Vector.add options.ccopt_cppdefines a)
  | ("-E" :: _) ->
      (badoption "cc" opt)
  | ["-I" ; a] ->
      (Vector.add options.ccopt_incdirectories a)
  | ["-L" ; a] ->
      (Vector.add options.ccopt_libdirectories a)
  | ["-U" ; a] ->
      (Vector.add options.ccopt_cppundefines a)
  | ["-nostdinc"] ->
      (Vector.add options.ccopt_flags "-nostdinc")
  | ["-O"] ->
      (begin (pmsg opt) ; () end)
  | ["-Os"] ->
      (begin (pmsg opt) ; () end)
  | ["-include" ; a] ->
      (List.iter (Vector.add options.ccopt_includes) opt)
  | ["-isystem" ; a] ->
      (List.iter (Vector.add options.ccopt_cppoptions) opt)
  | ["-imacros" ; a] ->
      (List.iter (Vector.add options.ccopt_includes) opt)
  | ["-MD" ; a] ->
      (begin (pmsg opt) ; () end)
  | ["-pipe"] ->
      (begin (pmsg opt) ; () end)
  | ("-Wp" :: l) ->
      (*
      let longoptions = [ "-MD" ] in
      let (args, cppopts) = (scan_options "cc/cpp" longoptions l []) in
      if (args != []) then
    (badoption "cc/cpp" opt) ;
      (List.iter (set_cc_options options) cppopts)
      *)
      let wp = (String.concat "," opt) in
      (Vector.add options.ccopt_cppoptions wp)
  | [o] ->
      if (Misc.string_starts_with "-W" o) then
    (begin (pmsg opt) ; () end)
      else if (Misc.string_starts_with "-f" o) then
    (begin (pmsg opt) ; () end)
      else if (Misc.string_starts_with "-m" o) then
    (begin (pmsg opt) ; () end)
      else
    (if harsh_handling then
      (badoption "cc" opt)
    else ())
  | _ ->
    (if harsh_handling then
      (badoption "cc" opt)
    else ()))

let cc (argv : string list) : unit = (
  let longoptions = [ "-D" ; "-I" ; "-L" ; "-U" ; "-o" ;
              "-Wp" ; "-include"; "-isystem" ] in
  let (args, opts) = (scan_options "cc" longoptions argv []) in
  let options : cc_options_t = (make_cc_options opts) in
  let flag_c = (None != (Vector.find_if
                (string_equal "-c")
                options.ccopt_flags)) in
  if ((List.length args) == 0) then
    (begin
      (failwith "no arguments")
    end)
  else
    (begin
      let ofiles =
    (List.map
        (fun s ->
          (* Check suffix. *)
          let n = (String.length s) in
          if (n <= 2) then
        (failwith ("file name too short: " ^ s)) ;
          let p = (String.sub s 0 (n - 2)) in
          let e = (String.sub s (n - 2) 2) in
          match e with
        ".c" ->
          (begin
            let nocpp = false in
            let filename = (p ^ ".c") in
            let temporary =
              if (nocpp) then
            filename
              else
            (cpp options filename) in
            (debug (";; running cpp input='"
                     ^ filename ^ "'")) ;
            (debug (";; running cpp output='"
                     ^ temporary ^ "'")) ;
            let defs = (Interp.parse_source_file temporary filename) in
            if (flag_c) then
              (save_ofile (p ^ ".o") defs) ;
            defs
          end)
          | ".o" ->
          (begin
            let filename = (p ^ ".o") in
            let defs : Translate.ofile_t = (load_ofile filename) in
            defs
          end)
          | _ ->
          (failwith ("unsupported file type: " ^ e)))
        args) in
      let afile = (Interp.package_definitions None ofiles) in
      if (not flag_c) then
    (begin
      let start_symbol = "main" in
      let afile2 : Translate.afile_t = {
        afile with Translate.afile_start = start_symbol
      } in
      let ctx : Interp.program_context_t = {
        Interp.ctx_afile = afile2 ;
        Interp.ctx_global_abstpreds = (Hashtbl.create 10) ;
        Interp.ctx_function_memoizers = (Hashtbl.create 10)
      } in
      (Interp.interpret_program ctx)
    end)
    end))

type ar_options_t = {
  aropt_flags : string Vector.t
}

let rec make_ar_options (opts : string list list) : ar_options_t = (
  let options = {
    aropt_flags = (Vector.create 4 "")
  } in
  (List.iter (set_ar_options options) opts) ;
  options)

and set_ar_options (options : ar_options_t) (opt : string list) : unit = (
  match opt with
    [] -> (assert false)
  | _ ->
      (badoption "cc" opt))

let ar (argv : string list) : unit = (
  let longoptions = [] in
  let (args, opts) = (scan_options "ar" longoptions argv []) in
  (*let options : ar_options_t = (make_ar_options opts) in*)
  if ((List.length args) == 0) then
    (begin
      (failwith "no arguments")
    end)
  else
    (begin
      ()
    end))

let ld (argv : string list) : unit = (
)

(*GOMI*)(*
let clap_usage = [
  "--lex: print lexical elements for trace.\n" ;
  "--cg: print grammar elements (print declaration combination).\n" ;
  "--tree: print parse trees for trace.\n" ;
  "--bb: print basic blocks (after SSA) for trace.\n" ;
  "--form: print translated forms of basic blocks for trace.\n" ;
  "--smt: print SAT forms for trace.\n"
]
*)

let rec check_clap_options (rest : string list) (args : string list) : string list = (
  match rest with
    [] ->
      args
  | ("--lex" :: tl) ->
      (Clex.yyTRACE := true) ;
      (check_clap_options tl args)
  | ("--cg" :: tl) ->
      (Cexpr.cgTRACE := true) ;
      (check_clap_options tl args)
  | ("--tree" :: tl) ->
      (Csyntax.bbTRACE_dump_parsed := true) ;
      (check_clap_options tl args)
  | ("--flatten" :: tl) ->
      (Csyntax.bbTRACE_dump_flatten := true) ;
      (check_clap_options tl args)
  | ("--bbbreakup" :: tl) ->
      (Csyntax.bbTRACE_dump_breakup := true) ;
      (check_clap_options tl args)
  | ("--bbcontrol" :: tl) ->
      (Csyntax.bbTRACE_dump_collapse := true) ;
      (check_clap_options tl args)
  | ("--ssavariables" :: tl) ->
      (Csyntax.bbTRACE_dump_variables := true) ;
      (check_clap_options tl args)
  | ("--ssaflow" :: tl) ->
      (Csyntax.bbTRACE_dump_ssa_flow := true) ;
      (check_clap_options tl args)
  | ("--block" :: tl) ->
      (Csyntax.bbTRACE_dump_ssa_binding := true) ;
      (check_clap_options tl args)
  | ("--bbequalset" :: tl) ->
      (Csyntax.bbTRACE_dump_ssa_equalset := true) ;
      (check_clap_options tl args)
  | ("--form" :: tl) ->
      (Interp.aiTRACE_dump_form := true) ;
      (check_clap_options tl args)
  | ("--smt" :: tl) ->
      (Smt.smtTRACE_sat := true) ;
      (check_clap_options tl args)
  | ("--cnf" :: tl) ->
      (Smt.smtTRACE_cnf := true) ;
      (check_clap_options tl args)
  | ("--trace" :: tl) ->
      (Interp.aiTRACE_interp := true) ;
      (check_clap_options tl args)
  | ("--notrace" :: tl) ->
      (Interp.aiTRACE_interp := false) ;
      (check_clap_options tl args)
  | ("--tracesmt" :: tl) ->
      (Interp.aiTRACE_interp_smt := true) ;
      (check_clap_options tl args)
  | ("--merge" :: tl) ->
      (Interp.aiOPTION_merge_cubes := true) ;
      (check_clap_options tl args)
  | ("--nomerge" :: tl) ->
      (Interp.aiOPTION_merge_cubes := false) ;
      (check_clap_options tl args)
  | (any :: tl) ->
      (check_clap_options tl (args @ [any])))

let check_ (argv : string list) : unit = (
  (*
    let f = (PRED (EQ, (VAR "ptr%3"), (ARITH (NUM (num_of_int 0))))) in
  (Translate.interpret_test "main" defs.ofile_defs f)
  *)
)

let write_unsafe_subtyping (msg : string)
                           (memo : string * string * string * string * string)
                           (xml_location : string)
: unit = (
    (Art.save_result_as_xml_file xml_location 
                                 false
                                 (Art.create_new_art_path ())
                                 memo
                                 msg);
    let la_error : Lazyabstraction.la_error_context_t =
      {
        Lazyabstraction.ctx_error_path = Art.create_new_art_path ();
        Lazyabstraction.ctx_error_location = memo;
        Lazyabstraction.ctx_error_message = msg
      } in
    Lazyabstraction.print_checking_result false la_error 0.0;
)

let do_parse_and_check (args : string list)
                       (cc_options : cc_options_t)
                       (ps_options : Checker.ps_options_t)
                       : unit = (
  let clap_dir = Filename.dirname Sys.executable_name in
  let path_to_assert = clap_dir ^ "/include/assert.h" in
  if not (Sys.file_exists path_to_assert) then
      (warning ("Cannot find a file " ^ path_to_assert))
  else if not (ps_options.Checker.psopt_no_assert) then
      (List.iter (Vector.add cc_options.ccopt_includes)
                 ["-include"; path_to_assert]) ;
  let flag_c =
      (None != (Vector.find_if (string_equal "-c") cc_options.ccopt_flags))
  in

  (* Set cut functions *)
  Sexpr.cut_functions := ps_options.Checker.psopt_cutlist;
  Checker.init_builtin_ofile ();
  LogicUtil.Logic.init (ps_options.Checker.psopt_round_coeffs);

  (* enable / disable feature: insert assert(true) to check all pointers *)
  if (ps_options.Checker.psopt_intensive_pointer_checking) then
      Interp.intensive_pointer_checking := true
  else
      Interp.intensive_pointer_checking := false;

  Sinstrument.insert_func_contract_callee :=
      ps_options.Checker.psopt_check_in_callees;
  Csyntax.insert_checking_array_index :=
      ps_options.Checker.psopt_check_array_index;

  if ((List.length args) == 0) then
    (begin
      (failwith "no arguments")
    end)
  else
    (begin try
      let ((cfa_ofiles_1 : (Cfa.ofile_t list)), (stmt_ofiles : (Cexpr.ofile_t list))) =
        (List.fold_left
	  (fun ((cfa_ofiles : (Cfa.ofile_t list)), (stmt_ofiles : (Cexpr.ofile_t list))) (s : string) ->
            (* Check suffix. *)
            let n = (String.length s) in
            if (n <= 2) then
              (failwith ("file name too short: " ^ s)) ;
            let p = (String.sub s 0 (n - 2)) in
            let e = (String.sub s (n - 2) 2) in
            match e with
            | ".c" ->
              (begin
                let nocpp = false in
                let filename = (p ^ ".c") in
                let temporary =
                  if (nocpp) then
                    filename
                  else
                    (cpp ~keep_comments:true cc_options filename) in
                (debug (";; running cpp input='" ^ filename ^ "'")) ;
                (debug (";; running cpp output='" ^ temporary ^ "'")) ;
                let defs =
                    if ps_options.Checker.psopt_bmc then
                        (Checker.parse_source_file
                             temporary
                             filename
                             (Some ps_options.Checker.psopt_iter_max)
                             (not ps_options.Checker.psopt_no_subtyping_check))
                    else
                        (Checker.parse_source_file
                             temporary
                             filename
                             None
                             (not ps_options.Checker.psopt_no_subtyping_check))
                in
                (cfa_ofiles, (stmt_ofiles @ [defs])) 
              end)
            | ".o" ->
              (begin
                let filename = (p ^ ".o") in
                let defs : Cfa.ofile_t = (Checker.load_ofile filename) in
                ((cfa_ofiles @ [defs]), stmt_ofiles)
              end)
            | _ ->
                (failwith ("unsupported file type: " ^ e)))
          ([], []) args)
      in
      Hashtbl.clear Sinstrument.f_pointer_field_hashtbl;
      Hashtbl.clear Ccommon.contract_of_field_hashtbl;
      let cfa_ofiles_2, ps_options =
          Checker.create_cfa_ofiles stmt_ofiles
                                    ps_options
                                    flag_c
      in
      let ofiles : (Cfa.ofile_t list) = cfa_ofiles_1 @ cfa_ofiles_2 in
      let afile =
          try
              (Checker.package_definitions None
                                           (Checker.builtin_ofile :: ofiles))
          with
          | Cexpr.Parser_error error_msg -> (
              LogUtil.log LogUtil.PARSER LogUtil.ERROR error_msg;
              exit parser_error_code)
          | e -> raise e
      in
      if (not flag_c) then
        (begin
	  let do_check (entry : Sexpr.function_entry_t) = (
            let start_symbol = entry in
            let afile2 : Cfa.afile_t = {
              afile with Cfa.afile_start = start_symbol
            } in
            let ctx : Checker.check_context_t = 
	        Checker.create_check_context afile2 ps_options in
            (!Checker.do_check ctx))
	  in Vector.iter do_check ps_options.Checker.psopt_entries
        end)
      else (
        Cfa.dump_all_symbols afile)
    with
    | Sexpr.Specs_error (error_message, pos_opt) -> (
        let (filename, linenumber, cnumber) = (
            match pos_opt with
            | None ->
                ("", -1,-1)
            | Some pos ->
                pos
        ) in
        let error_msg = (
            "\n    Error message: " ^ error_message
          ^ "\n    In file: " ^ filename
          ^ ", line: " ^ (string_of_int linenumber)
          ^ ", position: " ^ (string_of_int cnumber) ^"\n"
        ) in
        let msg = "Specification processing error" in
        LogUtil.log LogUtil.SPECS LogUtil.ERROR msg;
        LogUtil.log LogUtil.SPECS LogUtil.ERROR error_msg; 
        exit specs_error_code)
    | Cexpr.Parser_error error_msg -> (
        LogUtil.log LogUtil.PARSER LogUtil.ERROR error_msg;
        exit parser_error_code)
    | Smain.Unsafe_subtyping (msg, memo) -> (
        write_unsafe_subtyping msg memo ps_options.Checker.psopt_xml_location)
    | e -> (
        LogUtil.log LogUtil.GENERAL LogUtil.ERROR ("Uncaught exception : " ^ (Printexc.to_string e));
        exit uncaught_exception_error_code)
    end))

(* Main entry for processing specification *)
let ps (argv : string list) : unit = (
  let longoptions = [ "-D" ; "-I" ; "-L" ; "-U" ; "-o" ;
                      "-Wp" ; "-include"; "-isystem"; 
	                    "--main"; "--cut-list";
		                  "--slice-path"; "--output-xml";
                      "--recursive-max"; "--iter-max";
                      "--verbose"; "--bmc"; "--strategy";
                      "--timeout"; "--solver";
                      "--intensive-pointer-checking"; "--check-in-callees";
                      "--check-array-index";
                      "--temp-dir"]
  in
  let initoptions = [["-D"; "__ONLY_FOR_MODELCHECKING"]] in
  (* Split option=value into [option;value] *)
  let argv = List.flatten (
                 List.map (fun str ->
                     ExtString.String.nsplit str "="
                 ) argv
             ) in
  let (args, opts) = (scan_options "ps" longoptions argv initoptions) in
  let cc_options : cc_options_t = (make_cc_options ~harsh_handling:false opts) in
  let ps_options : Checker.ps_options_t = (Checker.make_ps_options opts) in

  (* Set verbose components in the LogUtil *)
  Vector.iter LogUtil.add_verbose_component
              ps_options.Checker.psopt_verbose_components;
  if (ps_options.Checker.psopt_debug_mode) then (
      LogUtil.enable_debug_mode ();
      match ps_options.Checker.psopt_solver with
      | LogicUtil.MATHSAT ->
          LogicUtil.enable_mathsat_debug_mode ()
      | _ -> ()
  );

  let timeout =
      if (ps_options.Checker.psopt_timeout > 0.) then
          ps_options.Checker.psopt_timeout *. 60.
      else
          -1.
  in
  (MathsatUtil.temp_dir := ps_options.Checker.psopt_temp_dir);
  let cvar = Condition.create () in
  let _ =
      if (timeout > 0.) then
          let timer_thread =
              Thread.create
                  (fun c ->
                       (debug ("Start the timer thread : "
                               ^ (string_of_float (Unix.time ()))));
                       (Thread.delay timeout);
                       (debug ("Timeout : "
                               ^ (string_of_float (Unix.time ()))));
                       (StatisticUtil.exit_by StatisticUtil.TIMEOUT);
                       (Condition.signal c);
                       (debug "Finish the timer thread"))
                  cvar
          in ()
  in
  let checker_thread =
      Thread.create (fun c ->
                         (Thread.delay 0.5);
                         (debug ("Start the checker thread : "
                                 ^ (string_of_float (Unix.time ()))));
                         (do_parse_and_check args cc_options ps_options);
                         (StatisticUtil.exit_by StatisticUtil.NORMAL);
                         (Condition.signal c);
                         (debug "Finish the checker thread"))
                    cvar
  in
  (Mutex.lock StatisticUtil.mutex);
  (Condition.wait cvar StatisticUtil.mutex);
  (Mutex.unlock StatisticUtil.mutex);
  if (StatisticUtil.timeout_exit ()) then (
      LogUtil.log LogUtil.GENERAL LogUtil.ERROR "Timeout error";
      exit timeout_error_code))

let help () = (
  let usage_msg = "
Usage: clap <subcommand> [options] targetfile ...
     
Subcommands:
  cc      orginary predicate abstraction
  ar
  ld
  check
  ps      check consistency of program specifications written as ACSL annotations in the target file.


Global options:
  --help    display this information

  supported gcc options (to be passed to the cpp preprocessor, please refer to gcc's usage for explanation): -c, -g, -o, -D, -I, -L, -U, -nostdinc, -include, -isystem, -imacros, -Wp


Options of the subcommand 'ps':
  --main NAME        a function which is to be regarded as the entry point for checking (by default, function \"main\" is taken)
  --cut-list FILE    a file which consists of functions whose details are to be pruned (skipped) while checking
  --iter-max NUM     the maximum times that the checker unrolls a loop (the default number is 5)
  --recursive-max NUM
                     the maximum times that the checker unrolls a recursive function (the default number is 5)
  --timeout NUM     the maximum times
  --output-xml FILE  the output XML file
  --slice-path NUM   the path slicing algorithm which is to be used, the options are 1 or 2 (the default value is 2)
  --dot              enable dumping intermediate data such as CFAs, ARTs, etc to DOT files
  --verbose <general/parser/specs/la>
                     cause the checker to print debugging messages in the specified module
  -d, --debug        enable the debug mode
  --intensive-pointer-checking=[y/n]
                     cause the checker to aggressively check consistency of pointer dereferences (could yield state-explosion; the default value is `y')
  --solver=<csisat/mathsat/hybrid>
                     select the integrated SAT/Interpolation solver (by default, csisat is chosen)
  --check-in-callees=[y/n]
                     cause the checker to check for validity of function contracts in the callee side besides in the caller side (default value is `n')
  --check-array-index=[y/n]
                     cause the checker to check for out-of-bound array accesses (default value is `y')
  --strategy=[dfs/hrbc]
                     indicate which searching strategy is in use, either depth-first-search or branch-counting-heuristic (default value is `hrbc')
  --bmc=[y/n]        enable/disable the bounded model checking mechanism (default value is `y')
  --noasrt           cause the checker not to automatically include the file assert.h
  --nostpcheck       cause the checker not to do subtyping check
  --temp-dir=DIRNAME select the temporary directory in use for MathSat (default value is `/tmp')
  --round            cause the solver Csisat to round coefficients of the interpolants to get integers"
  in
  info usage_msg
)
  
let main () = (
  let argv = (List.tl (Array.to_list Sys.argv)) in
  let usage = "Usage: clap [ps/cc/ar/ld/check] ...\n\nUse \"clap --help\" for details.\n" in
  let argv1 = (check_clap_options argv []) in
  if ((List.length argv1) == 0) then
    (begin
      (prerr_endline usage) ;
      (failwith "no argument")
    end)
  else
    (begin
      let need_help = 
        try let _ = List.find (fun a -> (String.compare a "--help") == 0) argv1
            in true
        with Not_found -> false
      in
      if need_help then
        help ()
      else (
        match (List.hd argv1) with
          "cc" ->
            (cc (List.tl argv1))
        | "ar" ->
            (ar (List.tl argv1))
        | "ld" ->
           (ld (List.tl argv1))
        | "check" ->
           (check_ (List.tl argv1))
        | "ps" ->
           (ps (List.tl argv1))
        | _ ->
           (prerr_endline usage) ;
           (failwith "unknown command"))
    end))

let _ = (Checker.do_check := Lazyabstraction.do_lazy_check)
let _ = (Stranslate.scan_string_constant_loop := Clex.scan_string_constant_loop)

let _ = (begin 
             LogUtil.init (); 
             try (main ();
                 LogUtil.close ())
             with 
             e -> (LogUtil.close ();
                 raise e)
        end)
