(* utils/art_dumper.ml                                           *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

(* Define art graph & functions to dump an art on a graph                                    *)

open Art;;
open ConfigUtil;;
open LogicUtil;;
open LogicUtil.Logic;;
open La_common;;

(* ~~~~~~~~~~~~~~~~~~~~~~~~ vertex of art ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
module ART_V = struct
    (* each vertex is an art_node *)
    type t = art_node
    (* overwrite compare-function *)
    let compare a_node1 a_node2 = compare (a_node1) (a_node2)
    (* overwrite equal-function *)
    let equal a_node1 a_node2 = (a_node1 == a_node2)
    (* hash contains vertices *)
    let hash = Hashtbl.hash
end


(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ edge of art ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
module ART_E = struct
    (* each edge is a predicate (assignment or condition) *)
    type t = predicate
    (* overwrites compare-function *)
    let compare pred1 pred2 = compare (pred1) (pred2)
    (* overwrites equal-function *)
    let equal pred1 pred2 = (pred1 == pred2)
    (* hash contains edges of art *)
    let hash = Hashtbl.hash 
    let default = True
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~ Graph of ART ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
module ART_G = struct
    include Graph.Imperative.Digraph.ConcreteLabeled (ART_V) (ART_E)
    let of_art (an_art : art) =
    begin
        let t = create () in
        let art_node_hash = an_art.art_node_hash in
        
        (* add vertices to Graph *)
        Hashtbl.iter (fun key_in_hashtbl v -> add_vertex t v) art_node_hash;
        
        (* add edges to Graph *)
        Hashtbl.iter
            (fun key_in_hashtbl a_node_in_art -> 
	                let outgoing_labels = a_node_in_art.art_outgoing_labels in
	                let outgoing_edges = a_node_in_art.art_outgoing_edges in
	                List.iter2
	                    (fun label_of_child edge_to_child ->
                          try ( 
		                          let child_node : art_node = 
		                              Hashtbl.find (art_node_hash) (label_of_child) 
     		                      in
		                          add_edge_e t (a_node_in_art, edge_to_child, child_node);
                          ) with Not_found -> ()
                      )
	                    outgoing_labels
	                    outgoing_edges
            )
            art_node_hash;
        
        (* return *)
        t    
    end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
module Viz = Agraphviz.Make(ART_G)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let output (visualizer) (an_art) (filename) =
begin
    let file = File_handler.open_file filename in
    Viz.output file visualizer (ART_G.of_art an_art)
end 

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let dump_one_art ?name (an_art) ?filename () =
begin
    let visualizer = 
        object inherit Viz.defaultVisualizer as super
        method graph = 
            match name with 
	            | None -> super#graph
	            | Some s -> (`Label s) :: super#graph
        ;
        (* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)     
        (* get string of a vertex *)
        method vertex_name a_node = begin
            (* get label of art_node *)
	          let label = "LB: " ^ string_of_art_label (a_node.art_label) in
            (* get string of reachable_region of art_node *)
	          let reachable_region_string = 
                "\\nRR : " ^ (string_of_predicate a_node.art_reachable_region) 
            in
            (* get string of support_predicates of art_node *)
	          let art_support_predicates_string = ref "\\nsps: " in
	          let art_support_predicates = a_node.art_support_predicates in
	          List.iter
	              (fun a_support_predicate -> 
                     let new_string = string_of_predicate a_support_predicate in
	                       art_support_predicates_string := 
	                           !art_support_predicates_string ^ "    &    " ^ new_string 
	              )
	              art_support_predicates;
	          (* get all string to display in vertex of graph *)
	          let final_string : string = 
	              label ^ reachable_region_string ^ !art_support_predicates_string  
	          in
	          (* return *)
	          final_string
	      end
            
        (* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)                
        (* get style for vertex of graph *)
        method vertex a_node = begin
            (* form of vertex: form is Diamond if a_node is ART_COND *)
            let new_attributes = 
                if a_node.art_type = ART_COND then [`Shape(`Diamond)]
                else [`Shape(`Box)] 
            in
            (* colour of vertex + form of vertex + default of super *)
            if a_node.art_type = ART_ERROR then 
                [`Color 0xafffaa] @ new_attributes @ (super#vertex a_node)
            else if a_node.art_is_pivot_node = true then
                [`Color 0xa0700d] @ new_attributes @ (super#vertex a_node)
            else if a_node.art_is_cover_node = true then 
                [`Color 0xfff000] @ new_attributes @ (super#vertex a_node)
            else if a_node.art_reachable_region = False then 
                [`Color 0x40e0d0] @ new_attributes @ (super#vertex a_node)
            else (* colour is default of super in other cases *)
                new_attributes @ (super#vertex a_node) 
        end
                
        (* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)        
        (* get string & style of an edge *)
        method edge ((vertex1, an_edge, vertex2) as ed) = begin
	          (* if edge is in Error-path, then color of the edge is red *)
	          let is_in_ep_flag = is_in_error_path (an_art) (vertex1) (vertex2) in 
	          let color = 
	              if is_in_ep_flag = true then 0xff0000
	              else 0x000000
	          in
	          (* get string of this edge *)
	          let string_of_edge = string_of_predicate an_edge in     
	          let final_label = string_of_edge in
	          [`Label final_label;`Color color] @ (super#edge ed) 
	      end
        (* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
    end
    (* save to dot-file *)
    in
    let name_of_outfile = 
        match filename with
         | None -> "art.dot"
         | Some n -> n
    in
    output visualizer an_art name_of_outfile;
end


(*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)
let dump_arts ?source art_list =
begin
    let (source_info : Graph_dot.source_info_t) = (begin
        let name = (match source with
                   | Some n -> n
                   | None -> "")
        in
        {
	          Graph_dot.path_to_source = name;
		        Graph_dot.dot_file = None;
			      Graph_dot.functions = [];
        }    
    end) in
    let count = ref 0 in
    List.iter
        (fun an_art ->
             count := !count + 1; 
             let fun_name = "art" ^ string_of_int !count ^ "_" in
             let filename = (*Filename.temp_file*) 
                 Filename.temp_dir_name ^ ("/mc_" ^ fun_name) ^ ".dot" in
             dump_one_art ~name:fun_name (an_art) ~filename:filename ();
             debug ("Writing an ART to file `" ^ filename ^ "' ... done");
             source_info.Graph_dot.functions <-
                 source_info.Graph_dot.functions
                 @ [{
                      Graph_dot.name = fun_name;
                      Graph_dot.path_to_dot = Some filename
                    }]
        )
        art_list;
    config.graphs := !(config.graphs)
                     @ [Graph_dot.Graph_per_function ("art", [source_info])]
end

