(* utils/cfa_dumper.ml                                           *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

(* Define cfa graph & functions to dump a cfa on a graph                   *)

open Cfa;;
open ConfigUtil;;
open Agraphviz;;
open LogicUtil;;
open LogicUtil.Logic;;
open La_common;;

(* ~~~~~~~~~~~~~~~~~~~~~~~~ vertex of cfa ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
module CFA_V = struct
    (* each vertex is an cfa_node *)
    type t = cfa_node
    (* overwrite compare-function *)
    let compare c_node1 c_node2 = compare (c_node1) (c_node2)
    (* overwrite equal-function *)
    let equal c_node1 c_node2 = (c_node1 == c_node2)
    (* hash contains vertices *)
    let hash = Hashtbl.hash
end


(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ edge of art ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
module CFA_E = struct
    (* each edge is a string *)
    type t = string
    (* overwrites compare-function *)
    let compare s1 s2 = compare (s1) (s2)
    (* overwrites equal-function *)
    let equal s1 s2 = (s1 == s2)
    (* hash contains edges of cfa *)
    let hash = Hashtbl.hash 
    let default = "True"
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~ Graph of CFA ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
module CFA_G = struct
    include Graph.Imperative.Digraph.ConcreteLabeled (CFA_V) (CFA_E)
    let of_cfa (a_cfa : cfa) =
    begin
        let t = create () in
        (match a_cfa with
        | CFA_FUNDEF cfa_fundef -> begin
            let cfa_node_hash = cfa_fundef.cfa_node_hashtbl in
            (* add vertices to Graph *)
            Hashtbl.iter 
                (fun a_label a_cfa_node -> add_vertex t a_cfa_node)
                cfa_node_hash;
            (* add edges to Graph *)
            Hashtbl.iter
                (fun a_label a_cnode -> 
	                  let outgoing_labels = ref a_cnode.cfa_outgoing_labels in
	                  let outgoing_edges =
	                      (match a_cnode.cfa_node_stmt with 
                        | CFA_RESET exprs -> ["Reset" ^ (string_of_expression (Sum exprs))]
                        | CFA_ASSUME pred -> [string_of_predicate pred]
                        | CFA_ENTRY pred | CFA_FORM pred -> [string_of_predicate pred] 
	                      | CFA_COND (pred, s1, s2) -> 
	                          [string_of_predicate pred; string_of_predicate (Not pred)]
	                      | CFA_CALL (expr_list, _) ->
	                          let string_of_expr_list = ref "" in
	                          let name_of_fun = 
                                try List.hd expr_list 
                                with Failure _ -> assert false
                            in
	                          string_of_expr_list := 
	                              !string_of_expr_list ^ string_of_expression name_of_fun 
	                              ^ " (";
	                          List.iter
	                          (fun a_expr -> 
	                              string_of_expr_list := 
	                                  !string_of_expr_list ^ string_of_expression a_expr 
	                                  ^ ", ")
	                          (List.tl expr_list); 
	                          string_of_expr_list := !string_of_expr_list ^ ")";
	                          [!string_of_expr_list] (* temporary *) 
	                      | CFA_RETURN -> 
	                          outgoing_labels := [];
	                          []
	                      | CFA_ERROR -> ["True"]
	                      | CFA_FAIL -> ["True"])
	                  in
	                  List.iter2 
	                      (fun label_of_child edge_to_child-> 
	                          if (label_of_child = "##elsewhere") then ()
	                          else (
	                              let child_node : cfa_node =
	                                  try Hashtbl.find (cfa_node_hash)
	                                                   (label_of_child)
	                                  with Not_found -> assert false
	                              in
	                              add_edge_e t (a_cnode, edge_to_child, child_node);)
                        )
	                      !outgoing_labels
	                      outgoing_edges;
                )
                cfa_node_hash
            end
        (* do not anything if a_cfa is CFA_VARDEF or CFA_OTHERDEF *)
        | anything_else -> ());
        (* return *)
        t  
    end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
module Viz = Agraphviz.Make(CFA_G)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let output (visualizer) (abstract_reachability_tree) (filename) =
begin
    let file = File_handler.open_file filename in
    Viz.output file visualizer (CFA_G.of_cfa abstract_reachability_tree)
end 

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let dump_one_cfa ?name cfa ?filename () =
(begin
    let visualizer = 
        object inherit Viz.defaultVisualizer as super
        method graph = (begin
	          match name with 
		        | None -> super#graph
		        | Some s -> (`Label s) :: super#graph
        end)
        (* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
        (* get string of a vertex *)
        method vertex_name c_node = (begin
	          c_node.cfa_label
	      end)
        (* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)        
        (* get style for vertex of graph *)            
        method vertex c_node = (begin
            (* form of vertex: form is Diamond if a_node is ART_COND *)
            let new_attributes = 
                (match c_node.cfa_node_stmt with
                | CFA_COND cond -> [`Shape(`Diamond)]
                | anything_else -> [`Shape(`Box)]) 
            in
            (* colour of vertex + form of vertex + default of super *)
            (match c_node.cfa_node_stmt with
            | CFA_FORM _ | CFA_ENTRY _ -> 
                new_attributes @ (super#vertex c_node)
            | CFA_RESET _ ->
                [`Color 0xeeeeee] @ new_attributes @ (super#vertex c_node) 
            | CFA_ASSUME _ ->
                [`Color 0xaaaaaa] @ new_attributes @ (super#vertex c_node)
            | CFA_COND _ -> 
                [`Color 0xafffaa] @ new_attributes @ (super#vertex c_node)
            | CFA_CALL _ -> 
                [`Color 0xa0700d] @ new_attributes @ (super#vertex c_node)
            | CFA_RETURN -> 
                [`Color 0xff0000] @ new_attributes @ (super#vertex c_node)
            | CFA_ERROR -> 
                [`Color 0x40e0d0] @ new_attributes @ (super#vertex c_node)
            | CFA_FAIL -> 
                [`Color 0x40e0d0] @ new_attributes @ (super#vertex c_node)) 
        end)
        (* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)    
        (* get string & style of an edge *)
        method edge ((_, a_edge, _) as ed) = (begin
	          let label = a_edge in   
    	      let final_label = label in
	          [`Label final_label] @ (super#edge ed) 
	      end)
        (* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
    end
    (* save to dot-file *)
    in
    let outfile = 
       (match filename with
       | None -> "cfa.dot"
       | Some name -> name)
    in
    output visualizer cfa outfile;
end)

(*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)
let dump_cfas ?source (afile : Cfa.afile_t) =
begin
    let (source_info : Graph_dot.source_info_t) = (begin
        let name = (match source with
                    | Some n -> n
                    | None -> "")
        in
        {
          Graph_dot.path_to_source = name;
          Graph_dot.dot_file = None;
          Graph_dot.functions = [];	
        }
    end)
    in
    Hashtbl.iter
        (fun _ ofile ->
            let srcfilename = ofile.ofile_file_name in
            Hashtbl.iter
                (fun _ a_cfa ->
                     match a_cfa with
                     | CFA_FUNDEF fundef -> (begin
                         let fun_name = fundef.cfa_f_meta_data.cfa_id in
                         let name = get_string_of_meta_data_of_cfa (a_cfa) in
                         let normalize_filename (fn : string) = (
                             ExtString.String.map (
                                 fun c ->
                                     if c = '/' then '_'
                                     else c
                             ) fn
                         ) in
                         let filename = 
                             Filename.temp_dir_name ^ ("/mc_"
                             ^ (normalize_filename (String.escaped srcfilename))
                             ^ "_" ^ fun_name) ^ ".dot"
                         in
                         dump_one_cfa ~name:name a_cfa ~filename:filename ();
                         debug ("Writing a CFA to file `"
                                ^ filename
                                ^ "' ... done");
                         source_info.Graph_dot.functions <- 
                             source_info.Graph_dot.functions 
                             @ [{
                                  Graph_dot.name = fun_name; 
                                  Graph_dot.path_to_dot = Some filename
                                }];
                         end)
                     | anything_else -> (*print_endline "isn't CFA_FUNDEF"*) ())
                ofile.ofile_cfas)
         afile.afile_libs;
    config.graphs := !(config.graphs)
                     @ [Graph_dot.Graph_per_function ("cfa", [source_info])]
end
  

