(* la/la_common.mli *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

open LogicUtil.Logic;;

val warning : string -> unit
val debug : string -> unit
val error : string -> unit
val info : string -> unit
val emerg : string -> unit

val concat_level_for_gvars_in_expression : expression -> expression
(** Concate level (1) for prime gvars in expression. 
    Example: prime gvar x -> concated level - gvar x%1. 
    Return: expression contains concated level - gvars
 *)

val concat_level_for_gvars_in_predicate : predicate -> predicate
(** Concate level (1) for prime gvars in predicate. 
    Example: prime gvar x -> concated level - gvar x%1.
    Return: predicate contains concated level - gvars.
 *)

val var_to_based_var : string -> string
(** Transform variable <varname%level%index> to <varname%level>. 
    Example: prime var x%3%1 -> based var x%3.
    Return: sub string of var.
 *)

val vars_to_based_vars : string list -> string list
(** Transform variable <varname%level%index> to <varname%level>. 
    Example: prime var x%3%1 -> based var x%3.
    Return: based vars of vars (string list type).
 *)

val create_based_var_in_application : expression -> expression
(** Transform variables by removing index of Variable or of Atom - Application
in an application.
    Example: prime application <*x%3, 1> -> based application *x%3.
    Return: based application of an app (expression type).
 *)

val create_based_vars_in_expression : expression -> expression
(** Transform variables by removing index of Variables or Applications in 
in an expression.
    Example: prime application <*x%3, 1> -> based application *x%3.
    Return: based expression (expression type).
 *)

val create_based_vars_in_expression_list : expression list -> expression list
(** Transform variables by removing index of Variable or of Atom - Application
in expressions.
    Example: prime application <*x%3, 1> -> based application *x%3.
    Return: based expr_list (expression list type).
 *)

val create_based_vars_in_predicate : predicate -> predicate
(** Transform variables by removing index of Variable or of Atom - Application
in a predicate.
    Example: prime application <*x%3, 1> -> based application *x%3.
    Return: based predicate (predicate type).
 *)

val create_based_vars_in_predicate_list : predicate list -> predicate list
(** Transform variables by removing index of Variable or of Atom - Application
in a list of predicates.
    Example: prime application <*x%3, 1> -> based application *x%3.
    Return: based predicate list (predicate list type).
 *)

val predlist_of_formlist : Form.form_t list -> predicate list
(** Transform a list of form_ts to a list of predicates.
    Return: a list of predicates (predicate list type).
 *)

val exprlist_of_termlist : Form.term_t list -> expression list
(** Transform a list of term_ts to a list of expressions.
 *)

val compare_expr_with_expr : expression -> expression -> int
(** Compare an expression with an other expression.
    Example: (compare x x) = 0; (compare x y) = 1
    Return: 0 if the exprs are equal, 1 otherwise.
 *)

val is_equivalent_predicates : predicate -> predicate -> int
(** Check if 2 predicates are equivalent or not: create 2 predicates: 
{pred1 => pred2}, {pred2 => pred1}; then check valid the predicates to 
judge the predicates are equivalent or not.
    Return: 0 if the pred1 are equivalent to pred2, 1 otherwise.
 *)

val is_identical_expressions : expression -> expression -> bool
(** Check if 2 expressions are identical or not.
    Return: false if can not match 2 expressions, true otherwise.
 *)

val is_identical_expression_lists : expression list -> expression list -> bool
(** Check if 2 expression list are identical or not.
    Return: false if can not match 2 expression list, or exist a tuple of 
(expr of list1, expr of list2) not identical, true otherwise.
 *)

val is_identical_predicates : predicate -> predicate -> bool
(** Check if 2 predicates are identical or not.
    Return: false if can not match 2 predicates, true otherwise.
 *)

val is_identical_predicate_lists : predicate list -> predicate list -> bool
(** Check if 2 predicate list are identical or not.
    Return: false if can not match 2 predicate list, or exist a tuple of 
(pred of list1, pred of list2) not identical, true otherwise.
 *)

val is_mem_of_expr_list : expression -> expression list -> bool
(** Check if a expression is member of a list of expressions.
    Return: If expr_list has an expression same with the expr then return true, 
else otherwise.
 *)

val member_of_predicate_list : predicate -> predicate list -> bool
(** Check if a predicate is a member of a list of predicates *)

val remove_duplicated_exprs : expression list -> expression list

val is_identical_element_of_pred_list : predicate -> predicate list -> bool
(** Check if a predicate is identical with an element of a predicate list
    Return: if pred_list has a predicate identical with the pred then return true, 
else otherwise.
 *)

val remove_expr_from_expr_list : expression -> expression list -> expression list
(** Remove an expression form a list of expressions.
 *)

val merge_two_predicate_lists : predicate list -> predicate list -> predicate list
(** Merge two predicate lists. For each predicate in pred_list1, check if it is
identical with any element of pred_list2 or not. If not, then add it to
pred_list2 to return.
 *)

val an_element_of_exprs_is_mem : expression list -> expression list -> bool
(** Check if a predicate list has an element which is mem of another 
predicate list.
 *)

val find_all_vars_in_application : expression -> expression list
(** Find all vars & atom applications in an application. Note that atom 
application is pref, or addrof, sref, srefp.
  *)

val find_all_vars_in_expression : expression -> expression list
(** Find all vars & atom applications in an expression. Note that atom 
application is pref, or addrof, sref, srefp.
 *)

val find_all_vars_in_predicate_list : predicate list -> expression list
(** Find all vars & atom applications in a list of predicates. Note that atom 
application is pref, or addrof, sref, srefp.
 *)

val find_all_vars_in_predicate : predicate -> expression list
(** Find all vars & atom applications in a predicate. Note that atom 
application is pref, or addrof, sref, srefp.
 *)

val get_atom_predicates_from_predicate : predicate -> predicate list
(** Get atom_predicates from one predicate. Atom predicate is (reference 
[http://en.wikipedia.org/wiki/First-order_logic]):
 1. Non-logical predicates If P is an n-ary predicate symbol and t1, ..., tn 
are terms then P(t1,...,tn) is a well-formed formula.
 2. Equality: If the equality symbol is considered part of logic, and t1 and 
t2 are terms, then t1 = t2 is a well-formed formula.
 This function is used to simlify support predicates after call "interpolate".  
 *)

val get_strongest_post_eq_condition_for_assignment :
  predicate -> predicate -> predicate list
(** Compute strongest-post-equality-condition for an assignment. Note that 
pre-condition must be an equality predicate. 
    Example: {pre-cond: x = 1} y = x + 1 {post-cond: x = 1 & y = 2}.
    If pre-cond is True then return the assignment. If pre-cond isn't True then 
replace left of pre_eq_condition in right of assignment by right of 
pre_eq_condition. If left of pre_eq_condition is identical with left of 
assignment, then return the replaced assignment, else return the replaced 
assignment & the pre_eq_condition.
 *)

val get_strongest_post_eq_conditions_for_assignment :
  predicate list -> predicate -> predicate list
(** Compute strongest-post-equality-conditions for an assignment. Pre-conditions 
must be equality predicates. 
    Example: {pre-cond: x = 1} y = x + 1 {post-cond: x = 1 & y = 2}.
    If pre-conds is null then raise an error. Else, foreach pre-cond of 
pre_eq_conditions, replace left of the pre-cond in right of assignment by 
right of the pre-cond. If left of pre_eq_condition is identical with left of 
assignment, then save replaced assignment, else save the replaced assignment & 
the pre_eq_condition. After all, add the assignment.
 *)

val get_strongest_post_eq_conditions_for_assignments :
  predicate list -> predicate list -> predicate list
(** Compute strongest post conditions for assignments. Pre-conditions must be
equality predicates. 
    Example: {pre-cond: x = 1} y = x + 1 {post-cond: x = 1 & y = 2}.
    If assignments are null, then return pre_eq_conditions. Else, compute 
forward. Foreach assignment, compute strongest-post-eq-conditions with 
pre-eq-conditions are strongest-post-eq-conditions of the before assignment. 
Note that compute strongest-post-eq-condition for first assignment with 
pre-eq-conditions are the input.
 *)

val get_strongest_post_eq_condition_for_assignments_edge :
  predicate -> predicate -> predicate list
(** Compute strongest post conditions for edge of assignments with a 
pre_eq_condition. The pre-condition must be an equality predicate. 
    Example: {pre-cond: x = 1} y = x + 1 {post-cond: x = 1 & y = 2}.
    If the edge is True then return pre_eq_condition. Else, if the edge 
contains an assignment then call get_strongest_post_eq_conditions_for_assignment.
Else, if the edge contains a list of assignments then call
get_strongest_post_eq_conditions_for_assignments.
 *)

val get_strongest_post_eq_conditions_for_assignments_edge :
  predicate list -> predicate -> predicate list
(** Compute strongest post conditions for edge of assignments with a list of
pre_eq_conditions. The pre-eq-conditions must be equality predicates. 
    Example: {pre-cond: x = 1} y = x + 1 {post-cond: x = 1 & y = 2}
If the edge is True then return pre_eq_conditions. Else, if the edge contains 
an assignment then call get_strongest_post_eq_conditions_for_assignment. Else, 
if the edge contains a list of assignments then call
get_strongest_post_eq_conditions_for_assignments.
 *)

val choose_eq_predicate_with_criteria :
  predicate -> expression list -> predicate option
(** Choose eq-predicate which involves to value of var in var_critera. Note
that eq_predicate is a static single assignment (ssa). 
    Example: 
[pred = {x = 1}; var_criteria = {x}; eq-predicate = {x = 1}];
[pred = {x = 1}; var_criteria = {y}; eq-predicate = null];         
    If the pred is an assignment, and left of the assignment is member of
var_critera, then return Some pred, else return None.
 *)

val choose_eq_predicates_with_criteria :
  predicate list -> expression list -> predicate list
(** Choose eq-predicates which involve to value of vars in var_critera. Note
that eq_predicates is static single assignments (ssa). 
    Example: 
[pred = {x = 1}; var_criteria = {x}; eq-predicate = {x = 1}];
[pred = {x = 1}; var_criteria = {y}; eq-predicate = null];         
    Foreach pred in pred_list, check if it is involve to value of vars in
var_criteria then save it to list to return, else ignore it. 
 *)

val find_eq_predicates : predicate -> predicate list
(** Find all equality-predicates in a predicate. 
 *)

val get_strongest_post_eq_condition_for_art_assume :
  predicate -> predicate -> expression list -> predicate list
(** Compute strongest-post-eq-condition for an assume edge. Note that an assume
egde maybe contains a predicate of assume, or 2 predicates of assume & reset.  
If the edge contains only pred_of_assume, then find all equality
predicates of the pred_of_assume, then compute strongest-post-eq-condition 
traditionally. If the edge contains pred_of_assume & pred_of_reset, first we 
find all equality predicates of the pred_of_assumes & pred_of_reset, then 
concate them and compute strongest-post-eq-conditions traditionally. 
 *)

val get_strongest_post_eq_conditions_for_art_assume :
  predicate -> predicate list -> expression list -> predicate list
(** Compute strongest-post-eq-conditions for an assume edge. Note that an assume
egde maybe contains a predicate of assume, or 2 predicates of assume & reset.  
Foreach pre_eq_condition in pre_eq_conditions, compute 
strongest-post-eq-condition for it with the edge & var_criteria, then save the 
result to list to return.
 *)

val replace_exprs_in_expression :
  expression list -> expression list -> expression -> expression
(** Replace sub-expressions in an expression by new sub-expressions, 
respectively.  
 *)

val replace_exprs_in_expressions :
  expression list -> expression list -> expression list -> expression list
(** Replace sub-expressions in expressions by new sub-expressions, respectively.  
 *)

val replace_exprs_in_predicate :
  expression list -> expression list -> predicate -> predicate
(** Replace sub-expressions in a predicate by new sub-expressions, respectively.  
 *)

val replace_exprs_in_predicates :
  expression list -> expression list -> predicate list -> predicate list
(** Replace sub-expressions in predicates by new sub-expressions, respectively.  
Foreach predicate in current_preds, call replace_exprs_in_predicate.
 *)

val concate_vars_with_tail_in_expression :
  expression -> expression list -> string -> expression
(** Concate variables of an expression & in the demand with a tail.
    Example: [{expression: x%3 + y + 1}; {variables needed concate: y}; 
{tail: %3}; ==> {new expression: x%3 + y%3 + 1}]
 *)

val concate_vars_with_tail_in_predicate :
  predicate -> expression list -> string -> predicate
(** Concate variables of a predicate & in the demand with a tail.
    Example: {predicate: x%3 = y + 1}; {variables needed concate: y}; 
{tail: %3}; ==> {new predicate: x%3 = y%3 + 1}.
 *)

val string_vars_to_exprs : symbol list -> expression list
(** Transform strings to variables.
    Example: [{string: x%3} ==> {variable: Variable x%3}]
 *)

val normalize_expr_with_theories : expression -> expression
(** Temporarily nomalize an expression with one theory: {Pref (Addrof (x)) = x}. 
Replace any sub-expression has form *(&varname) with varname.
 *)

val normalize_pred_with_theories : predicate -> predicate
(** Temporarily nomalize an predicate with one theory: {Pref (Addrof (x)) = x}. 
Replace any sub-expression has form *(&varname) with varname.
 *)

val normalize_exprs_with_theories : expression list -> expression list
(** Temporarily nomalize a list of expressions with one theory:
{Pref (Addrof (x)) = x}.
Replace any sub-expression has form *(&varname) with varname.
 *)

val normalize_preds_with_theories : predicate list -> predicate list
(** Temporarily nomalize a list of predicates with one theory:  
{Pref (Addrof (x)) = x} 
Replace any sub-expression has form *(&varname) with varname.
 *)

val expr_contains_arrays : expression -> bool

val pred_contains_arrays : predicate -> bool

val find_arrays_in_expr : expression -> expression list

val find_arrays_in_pred : predicate -> expression list

val find_indexes_of_arrays_in_expr : expression -> expression list

val find_indexes_of_arrays_in_pred : predicate -> expression list

val get_format_string_of_predicate : string -> predicate -> string

val flatten_implications : predicate -> predicate list

val find_addrofs_in_expression : expression -> expression list

val find_addrofs_in_predicate : predicate -> expression list 

val create_constraint_for_addrof : expression -> predicate 

val remove_mem : 'a -> 'a list -> 'a list

val find_left_of_addrof_assignment : predicate -> expression list

val find_value_of_var_in_pred : predicate -> expression -> string option

val expr_is_in_alias_expr : expression -> expression -> bool

val expr_is_in_alias_pred : expression -> predicate -> bool

val insert_pred_to_where_expr_appears_in_pred : expression -> 
    predicate -> predicate -> predicate

val multiply_an_element_by_n : 'a -> int -> 'a list

val multiply_string_by_n : string -> int -> string

val hashtbl_flatten_to_list : ('a, 'b) Hashtbl.t -> ('a * 'b) list

val hashtbl_import_list : ('a, 'b) Hashtbl.t -> ('a * 'b) list -> unit

val subtract_exprs_from_exprs : expression list -> expression list -> expression list

val remove_elements_in_hashtbl : ('k, 'v) Hashtbl.t -> 'k -> unit

val member_of_list : 'a -> 'a list -> ('a -> 'a -> int) -> bool

val subtract_list_from_list : 'a list -> 'a list -> ('a -> 'a list -> bool) -> 'a list

val merge_two_lists : 'a list -> 'a list -> ('a -> 'a list -> bool) -> 'a list

val replace_pointer_by_memory_in_pred : (int, (int list) * int) Hashtbl.t ->
                                        (expression, int) Hashtbl.t ->
                                        (int, expression) Hashtbl.t ->
                                        predicate ->
                                        predicate