(* la/la_interpolate.ml                                          *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

(** Define functions to analysis counterexamples & find interpolations of
an error path.                                                           
 *)

open LogicUtil.Logic;;
open Art;;
open Cfa;;
open Slicing_path;;
open La_common;;

let debug_checking_null : bool ref = ref false;;
let check_null_algorithm : int ref = ref 1;;

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(**  From an error-path with has changed indexs of vars and enconded, and from
an encoded hash table, compute interpolation for each node in the path
(except error node). Then find pivot-node which is the last True interpolant 
node before a Not True interpolant node. If we can't find pivot-node, default 
label of pivot-node is ("", -1).
 - Example: 
{v
node(i)     edge(i -> i + 1)         interpolation      
1.          z = 100                  True
2.          x = ctr                  x = ctr
3.          ctr = ctr + 1            x = ctr - 1
4.          y = ctr                  x = y - 1
5.          assume(x = m)            y = m + 1
6.          assume(y != m + 1)
The pivot node is node 2.
v}
 - Param: {ul {li {b encoded_hash} [(expression, expression) Hashtbl.t]}
              {li {b changed_index_path} [art_path]}}
 - Pseudo Code: 
{ul {li [Get decoded interpolations from input-path.]}
    {li [Decode the interpolations.]}
    {li [Backward analysis: If we meet a Not True interpolant node, then add
the node & the interpolant of it to list to return. Else, we meet a True 
interpolant node, then the node is pivot node, ignore all remaining nodes. ]}
}
 - Return: 
{ul {li pivot label of pivot node ((string * int) type)}
    {li art_labels & interpolants, respectively ((string * predicate) list type)}
} 
 - Since: 20090611
 *)
let find_pivot_labels_and_interpolants 
      (*(encoded_edges_in_path : predicate list)*) 
      (encoded_hash : (expression, expression) Hashtbl.t)
      (changed_index_path : art_path) 
: (string * int) * (string * int) * (((string * int) * predicate) list) = 
begin
    (* debug *)
    debug "Getting interpolants ...";
    let encoded_interpolants : predicate list = 
        get_interpolants (changed_index_path.art_edges_in_path) in
    (* debug *)
    debug ("Decoding interpolants ...");
    let interpolants : predicate list =
        decode_vars_in_path (encoded_interpolants) (encoded_hash) in
    (* debug *)
    (*let string_of_num_of_inters = string_of_int (List.length interpolants) in
    debug ("Number of interpolants: " ^ string_of_num_of_inters);
    List.iter
        (fun a_in -> debug (string_of_predicate (a_in)))
        interpolants;
    let _ = read_int() in*) 

    (* Find started_pivot_node & stop_pivot_node: *)
    let labels_and_interpolants = ref [] in
    let started_pivot_label = ref ("", - 1) in
    let stop_pivot_label = ref ("", -1) in
    let rev_art_nodes_in_path = List.rev changed_index_path.art_nodes_in_path in
    let rev_interpolants = List.rev (True :: interpolants) in
    let found_started_pivot = ref false in
    let found_stop_pivot = ref false in
    (try
        (List.iter2
            (fun a_node an_itp ->
                (* Find stop_pivot_label *) 
                if !found_stop_pivot = false then begin
                    (* If an_itp isn't False and the previous-itp is False,
                    then previous node is stop-pivot-node *)
                    (if (not (is_sat an_itp)) then begin
                        stop_pivot_label := a_node.art_label;
                    end else begin
                        found_stop_pivot := true;
                    end);
                end;
                (* Find started-pivot-label *)
                if !found_started_pivot = false then begin
                    (* if an_itp isn't True -> this isn't pivot_node *)
                    if (is_valid an_itp) then begin
                        started_pivot_label := a_node.art_label;
                        found_started_pivot := true;
                    end
                end;
                (* Add the an_itp with a_node.art_label to labels_and_interpolants *)
                let (name, version) = a_node.art_label in
                let a_based_itp : predicate =
                    create_based_vars_in_predicate (*false*) (an_itp) 
                in
                labels_and_interpolants := 
                    !labels_and_interpolants @ [(name, version), a_based_itp];
            )
            rev_art_nodes_in_path
            rev_interpolants
        )
    with
    | Invalid_argument _ -> assert false);
        
    (* debug *)
    let started_pivot_string = string_of_art_label !started_pivot_label in 
    let stop_pivot_string = string_of_art_label !stop_pivot_label in
    debug ("---> started_pivot_label: " ^ started_pivot_string);
    debug ("---> stop_pivot_label: " ^ stop_pivot_string);
    debug ("---> Labels_and_interpolants:");
    let rev_labels_and_interpolants = List.rev !labels_and_interpolants in
    List.iter
        (fun ((a_label, _), an_itp) ->
            if (not (is_valid an_itp)) & (is_sat an_itp) then
            begin
                let string_of_an_itp = string_of_predicate an_itp in
                debug (a_label ^ " : " ^ string_of_an_itp)
            end;  
        )
        rev_labels_and_interpolants;
    debug "Counterexample_analysis is done!!!";
    (*debug "\nRefining...";*)
    (*if (!la_debug) then let _ = read_int() in ()
    else ();*)
        
    (* return *)
    (!started_pivot_label, !stop_pivot_label, !labels_and_interpolants)
end


(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let check_null_pointer_in_path (changed_index_path : art_path) 
      (null_pointer_exceptions : (expression * expression * (expression list)) list list)
: unit =
begin
    if !debug_checking_null = true then begin
        debug "check_null_pointer_in_path";
    end;
    (* Encode path *)
    let (encoded_edges_in_path, encoded_hash) = 
        encode_vars_in_path (changed_index_path.art_edges_in_path) in
    let encoded_path : art_path = {
        art_nodes_in_path = changed_index_path.art_nodes_in_path;
        art_edges_in_path = encoded_edges_in_path;
    } in
    (* Foreach null-pointer-exception, check if it is real-error or not *)
    List.iter2
        (fun anode exceptions_of_edge ->
            (* Check if exceptions_of_edge is null then do nothing *) 
            if exceptions_of_edge != [] then begin
                List.iter
                    (fun (prime_expr, replaced_expr, need_addrof_exprs) ->
                        (* get edge at anode.art_label *)
                        let null_edge : predicate = 
                            get_edge (changed_index_path) (anode.art_label)
                        in
                        (* Add constraint of null_pointer *)
                        let null_constraint : predicate = 
                            Eq (replaced_expr, Constant zero_mix_const) in
                        (* Insert constraint to null-edge *)
                        let new_null_edge : predicate = 
                            insert_pred_to_where_expr_appears_in_pred (replaced_expr)
                              (null_constraint) (null_edge)
                        in 
                        (* START-ENCODE *)
                        let encoded_null_edge : predicate = 
                            encode_vars_in_predicate (new_null_edge) (encoded_hash) in
                        let encoded_replaced_expr : expression = 
                            encode_vars_in_expression (replaced_expr) (encoded_hash) in
                        (* STOP-ENCODE *)

                        (* remain only edges relevant to replaced_expr *)
                        let sliced_path : art_path = 
                            slice_path_backward_from_label_with_criterion_var 
                              (encoded_path) (encoded_replaced_expr) 
                              (anode.art_label)
                        in
                        (* Get path-formula *)
                        (*let pred_of_path : predicate = 
                            normalize (And sliced_path.art_edges_in_path) in*)
                        (* Update path-formula *)
                        let new_path_formula : predicate = 
                            And (encoded_null_edge :: sliced_path.art_edges_in_path)
                            (*And [pred_of_path; null_constraint] in*)
                        in
                        (* Debug *)
                        if !debug_checking_null = true then begin
                            debug ("path-formula to check null:");
                            debug (string_of_predicate new_path_formula);
                        end;
                        (* Check if we can reach to the null-error or not: *)
                        let is_feasible : bool = is_sat (new_path_formula) in
                        if is_feasible = true then begin
                            raise (La_exception.Not_point_to (anode.art_label, 
                                  prime_expr, replaced_expr, need_addrof_exprs));
                        end
                    )
                    exceptions_of_edge;
            end;
        )
        changed_index_path.art_nodes_in_path
        null_pointer_exceptions;
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let check_null_pointer_in_path_with_WeiBenbacher_paradigm (changed_index_path : art_path)
      (null_pointer_exceptions : (expression * expression * (expression list)) list list)
: unit =
begin
    if !debug_checking_null = true then begin
        debug "check_null_pointer_in_path";
    end;
    (* Encode path *)
    let (encoded_edges_in_path, encoded_hash) = 
        encode_vars_in_path (changed_index_path.art_edges_in_path) in
    let encoded_path : art_path = {
        art_nodes_in_path = changed_index_path.art_nodes_in_path;
        art_edges_in_path = encoded_edges_in_path;
    } in
    (* Foreach null-pointer-exception, check if it is real-error or not *)
    List.iter2
        (fun anode exceptions_of_edge ->
            (* Check if exceptions_of_edge is null then do nothing *) 
            if exceptions_of_edge != [] then begin
                List.iter
                    (fun (prime_expr, replaced_expr, need_addrof_exprs) ->
                        (* Create null-constraint *)
                        let null_constraint =
                            Eq (replaced_expr, Constant zero_mix_const) in
                        (* get edge at anode.art_label *)
                        let null_edge : predicate = 
                            get_edge (changed_index_path) (anode.art_label)
                        in
                        (* Insert constraint to null-edge *)
                        let new_null_edge : predicate = 
                            insert_pred_to_where_expr_appears_in_pred (replaced_expr)
                              (null_constraint) (null_edge)
                        in 
                        (* START-ENCODE *)
                        let encoded_null_edge : predicate = 
                            encode_vars_in_predicate (new_null_edge) (encoded_hash) in
                       (* STOP-ENCODE *)

                        (* Find criterion-labels to slice path *)
                        let criterion_labels : (string * int) list = 
                            find_criterion_labels_in_path_with_criterion_edge (encoded_path)
                              (encoded_null_edge) (anode.art_label)
                        in
                        (* remain only edges relevant to replaced_expr *)
                        let sliced_path : art_path = 
                            slice_path_with_criteria_labels_from_criterion 
                              (encoded_path) (criterion_labels) 
                              (encoded_null_edge) (anode.art_label)
                        in
                        (* Get path-formula *)
                        (*let pred_of_path : predicate = 
                            normalize (And sliced_path.art_edges_in_path) in*)
                        (* Update path-formula *)
                        let new_path_formula : predicate = 
                            And (encoded_null_edge :: sliced_path.art_edges_in_path)
                            (*And [pred_of_path; null_constraint] in*)
                        in
                        (* Debug *)
                        if !debug_checking_null = true then begin
                            debug ("path-formula to check null:");
                            debug (string_of_predicate new_path_formula);
                        end;
                        (* Check if we can reach to the null-error or not: *)
                        let is_feasible : bool = is_sat (new_path_formula) in
                        if is_feasible = true then begin
                            raise (La_exception.Not_point_to (anode.art_label, 
                                  prime_expr, replaced_expr, need_addrof_exprs));
                        end
                    )
                    exceptions_of_edge;
            end;
        )
        changed_index_path.art_nodes_in_path
        null_pointer_exceptions;
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let check_null_pointer_from_entry_to_pivot_node (changed_index_path : art_path) 
     (null_pointer_exceptions : (expression * expression * (expression list)) list list)
     (stop_pivot_label : string * int) 
: unit =
begin
    (* debug *)
    if !debug_checking_null = true then begin
        debug "check_null_pointer_from_entry_to_pivot_node";
    end;
    (* Encode path *)
    let (encoded_edges_in_path, encoded_hash) = 
        encode_vars_in_path (changed_index_path.art_edges_in_path) in
    let encoded_path : art_path = {
        art_nodes_in_path = changed_index_path.art_nodes_in_path;
        art_edges_in_path = encoded_edges_in_path;
    } in
    let is_check : bool ref = ref true in
    (* Foreach null-pointer-exception, check if it is real-error or not *)
    List.iter2
        (fun anode exceptions_of_edge ->
            if !is_check = true then begin
                (* Check if current-node is pivot-node or not *)
                if (compare_art_labels stop_pivot_label anode.art_label) = 0 then begin
                    is_check := false
                end else (if exceptions_of_edge != [] then begin 
                (* Check if exceptions_of_edge is null then do nothing *) 
                    List.iter
                        (fun (prime_expr, replaced_expr, need_addrof_exprs) ->
                            (* get edge at anode.art_label *)
                            let null_edge : predicate = 
                                get_edge (changed_index_path) (anode.art_label)
                            in
                            (* Add constraint of null_pointer *)
                            let null_constraint : predicate = 
                                Eq (replaced_expr, Constant zero_mix_const) in
                            (* Insert constraint to null-edge *)
                            let new_null_edge : predicate = 
                                insert_pred_to_where_expr_appears_in_pred (replaced_expr)
                                  (null_constraint) (null_edge)
                            in 
                            (* START-ENCODE *)
                            let encoded_null_edge : predicate = 
                                encode_vars_in_predicate (new_null_edge) (encoded_hash) in
                            let encoded_replaced_expr : expression = 
                                encode_vars_in_expression (replaced_expr) (encoded_hash) in
                            (* STOP-ENCODE *)
                            (* remain only edges relevant to replaced_expr *)
                            let sliced_path : art_path = 
                                slice_path_backward_from_label_with_criterion_var 
                                  (encoded_path) (encoded_replaced_expr) 
                                  (anode.art_label)
                            in
                            (* Get path-formula *)
                            let pred_of_path : predicate = 
                                (And sliced_path.art_edges_in_path) in
                            (* Update path-formula *)
                            let new_path_formula : predicate = 
                                And [pred_of_path; encoded_null_edge(*null_constraint*)] in
                            (* debug *)
                            if !debug_checking_null = true then begin
                                debug ("path-formula to check null:");
                                debug (string_of_predicate new_path_formula);
                            end;
                            (* Check if we can reach to the null-error or not: *)
                            let is_feasible : bool = is_sat (new_path_formula) in
                            if is_feasible = true then begin
                                raise (La_exception.Not_point_to (anode.art_label, 
                                      prime_expr, replaced_expr, need_addrof_exprs));
                            end
                        )
                        exceptions_of_edge;
                end);
            end;
        )
        changed_index_path.art_nodes_in_path
        null_pointer_exceptions;
end


(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let check_null_pointer_from_entry_to_pnode_with_WeiBenbacher_prd (changed_index_path : art_path) 
     (null_pointer_exceptions : (expression * expression * (expression list)) list list)
     (stop_pivot_label : string * int) 
: unit =
begin
    (* debug *)
    if !debug_checking_null = true then begin
        debug "check_null_pointer_from_entry_to_pnode_with_WeiBenbacher_prd";
    end;
    (* Encode path *)
    let (encoded_edges_in_path, encoded_hash) = 
        encode_vars_in_path (changed_index_path.art_edges_in_path) in
    let encoded_path : art_path = {
        art_nodes_in_path = changed_index_path.art_nodes_in_path;
        art_edges_in_path = encoded_edges_in_path;
    } in

    let is_check : bool ref = ref true in
    (* Foreach null-pointer-exception, check if it is real-error or not *)
    List.iter2
        (fun anode exceptions_of_edge ->
            if !is_check = true then begin
                (* Check if current-node is pivot-node or not *)
                if (compare_art_labels stop_pivot_label anode.art_label) = 0 then begin
                    is_check := false
                end else (if exceptions_of_edge != [] then begin 
                (* Check if exceptions_of_edge is null then do nothing *) 
                    List.iter
                        (fun (prime_expr, replaced_expr, need_addrof_exprs) ->
                            (* Create null-constraint *)
                            let null_constraint : predicate = 
                                Eq (replaced_expr, Constant zero_mix_const) in
                            (* get edge at anode.art_label *)
                            let null_edge : predicate = 
                                get_edge (changed_index_path) (anode.art_label)
                            in
                            (* Insert constraint to null-edge *)
                            let new_null_edge : predicate = 
                                insert_pred_to_where_expr_appears_in_pred (replaced_expr)
                                  (null_constraint) (null_edge)
                            in
                            (* START-ENCODE *)
                            let encoded_null_edge : predicate = 
                                encode_vars_in_predicate (new_null_edge) (encoded_hash) in
                            (* STOP-ENCODE *)
                            (* Find criterion-labels to slice path *)
                            let criterion_labels : (string * int) list = 
                                find_criterion_labels_in_path_with_criterion_edge (encoded_path)
                                  (encoded_null_edge) (anode.art_label)                            in
                            (* remain only edges relevant to replaced_expr *)
                            let sliced_path : art_path = 
                                slice_path_with_criteria_labels_from_criterion 
                                  (encoded_path) (criterion_labels) 
                                  (encoded_null_edge) (anode.art_label)
                            in
                            (* Get path-formula *)
                            let pred_of_path : predicate = 
                                (And sliced_path.art_edges_in_path) in
                            (* Update path-formula *)
                            let new_path_formula : predicate = 
                                And [pred_of_path; encoded_null_edge(*null_constraint*)] in
                            (* debug *)
                            if !debug_checking_null = true then begin
                                debug ("path-formula to check null:");
                                debug (string_of_predicate new_path_formula);
                            end;
                            (* Check if we can reach to the null-error or not: *)
                            let is_feasible : bool = is_sat (new_path_formula) in
                            if is_feasible = true then begin
                                raise (La_exception.Not_point_to (anode.art_label, 
                                      prime_expr, replaced_expr, need_addrof_exprs));
                            end
                        )
                        exceptions_of_edge;
                end);
            end;
        )
        changed_index_path.art_nodes_in_path
        null_pointer_exceptions;
end
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let encode_null_pointer_exceptions 
      (null_pointer_exceptions : (expression * expression * (expression list)) list list)
: (expression * expression * (expression list)) list list = 
begin
    let encoding_hash : (expression, expression) Hashtbl.t = Hashtbl.create (0) in
    let encoded_exceptions : 
        (expression * expression * (expression list)) list list ref = ref [] in
    (* Foreach pointer_exceptions of an edge, encode it *)
    List.iter
        (fun (edge_exceptions : (expression * expression * (expression list)) list) ->
            (* Foreach pointer-exception-tuple in the edge, encode it *)
            let encoded_edge_exceptions : 
                (expression * expression * (expression list)) list ref = ref [] in
            List.iter
                (fun (expr1, expr2, exprs) ->
                    let encoded_expr1 = encode_vars_in_expression (expr1) (encoding_hash)
                    and encoded_expr2 = encode_vars_in_expression (expr2) (encoding_hash)
                    and encoded_exprs = encode_vars_in_expressions (exprs) (encoding_hash)
                    in
                    (* Update encoded_edge_exceptions *)
                    let encoded_tuple = (encoded_expr1, encoded_expr2, encoded_exprs) in
                    encoded_edge_exceptions := !encoded_edge_exceptions @ [encoded_tuple];
                )
                edge_exceptions;
            (* Update encoded_exceptions *)
            encoded_exceptions := !encoded_exceptions @ [!encoded_edge_exceptions];
        )
        null_pointer_exceptions;
    (* Return *)
    !encoded_exceptions; 
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(**  Analyse an error-path. Check if the error-path reachs to Error-node or  
not. If the error-path is reach to error-node (fomula of constraints of the
path is sat) then raise error notification. Else, with each node in the path, 
compute interpolation to return to refine.
 - Example 1: The error-path can reach to error-node: 
{v Input-path:
node(i)     edge(i -> i + 1)         constraint                    
1.          z = 100                  <z, 0> = 100
2.          x = ctr                  <x, 0> = <ctr, 0>
3.          z = ctr + 1              <z, 1> = <ctr, 0> + 1
4.          y = ctr                  <y, 0> = <ctr, 0>
5.          assume(x = m)            <x, 0> = <m, 0>
6.          assume(z = ctr + 1)      <z, 1> = <ctr, 0> + 1
7.          assume(y = m )           <y, 0> = <m, 0>
v}
{v Sliced-path:
node(i)     edge(i -> i + 1)         constraint                    
1.          True                     True
2.          True                     True    
3.          True                     True
4.          True                     True
5.          True                     True
6.          True                     True
7.          True                     True
v}
 - Example 2: The error-path can not reach to error-node:
{v Input-path:
node(i)     edge(i -> i + 1)         constraint                    
1.          z = 100                  <z, 0> = 100
2.          x = ctr                  <x, 0> = <ctr, 0>
3.          z = ctr + 1              <z, 1> = <ctr, 0> + 1
4.          y = ctr                  <y, 0> = <ctr, 0>
5.          assume(x = m)            <x, 0> = <m, 0>
6.          assume(z = ctr + 1)      <z, 1> = <ctr, 0> + 1
7.          assume(y = m )           <y, 0> = <m, 0>
v}
{v Sliced-path:
node(i)     edge(i -> i + 1)         constraint                 interpolations                    
1.          True                     True                       True
2.          x = ctr                  <x, 0> = <ctr, 0>          x = ctr
3.          z = ctr + 1              <z, 1> = <ctr, 0> + 1      x = ctr
4.          y = ctr                  <y, 0> = <ctr, 0>          y = x
5.          assume(x = m)            <x, 0> = <m, 0>            y = m
6.          assume(z = ctr + 1)      <z, 1> = <ctr, 0> + 1      y = m
7.          assume(y != m)           <y, 0> != <m, 0>
v}
 - Param: {ul {li {b path} [art_path] maybe error path}
              {li {b gvars} [expression list] list of global variables 
(gvars is used in change_index functions, to remove local variables when 
change_index_functions meet a return_node).}}
 - Pseudo Code: 
{ul {li [Change index of variables in path (transform path to static
sigle assignment path).]}
    {li [Encode variables of changed index path before call slicing_path 
(to reduce complexity to reduce runtime)]}
    {li [Slice encoded path (remain only edges relevant to prove the 
infeasible of the path)]}
    {li [Check if the sliced path is feasible (formula of the path is sat), then
the path can reach to the error node, then print message and return]}
    {li [Else, the sliced path is infeasible (formula of the path is unsat), 
then call function "find_pivot_label_and_interpolants" to find pivot label &
interpolantions of nodes in the path]}
}
 - Return: 
{ul {li pivot label of pivot node ((string * int) type)}
    {li art_labels & interpolants, respectively ((string * predicate) list type)}
} 
 - Since: 20090611
 *)
let analysis_path_normally (path : art_path) 
                           (gvars: expression list)
: (string * int) * (((string * int) * predicate) list)=
begin
    (* debug *)
    debug "Analyzing counterexample:";
    debug (get_string_of_art_path path);
    (*debug (get_string_of_labels_in_art_path path);
    debug ("\n" ^ string_of_predicate (And path.art_edges_in_path));
    debug "--------------------------------\n";*)
    if (!la_debug) then pause ()
    else ();

    (* First, change index of variables in path *)
    let null_pointer_exceptions = ref [] in
    let changed_index_path : art_path = 
        change_index_of_variables_in_path (path) (gvars) (false) (false) (true) 
          (null_pointer_exceptions) in
    (* debug *)
    (if (!la_debug) then begin 
        debug "\n\n before slicing path-formulae:\n";
        (*debug (string_of_predicate (And changed_index_path.art_edges_in_path));*)
        debug (get_string_of_path_formulae changed_index_path.art_edges_in_path);  
    end);
    (* encode variables of path before slicing path *)
    let (encoded_edges_in_path, encoded_hash) =
        encode_vars_in_path (changed_index_path.art_edges_in_path) in
    let encoded_path : art_path = {
        art_nodes_in_path = changed_index_path.art_nodes_in_path;
        art_edges_in_path = encoded_edges_in_path;
    } in
    (* Slice path *)
    let sliced_path : art_path = slice_path (encoded_path) in
    (* debug *)
    (if (!la_debug) then begin 
        debug "\n\n after slicing path-formulae:\n";
        (*let string_of_edges =
            string_of_predicate (And sliced_path.art_edges_in_path) in 
        debug (string_of_edges);*)
        debug (get_string_of_path_formulae sliced_path.art_edges_in_path);
        debug "--------------------------------\n";
        (*if (!la_debug) then pause ()
        else ();*)
    end);
    
    (*debug (string_of_predicate (And encoded_edges_in_path));*)
    
    (* check if the error-path is feasible or not *)
    let pred_of_path = (And sliced_path.art_edges_in_path) in
    let is_feasible = is_sat pred_of_path in
    (* if is_feasible is false, then this error_path is a *)
    (* counterexample. We compute interpolants to refine: *)
    if is_feasible = false then begin
        (* Find pivot_labels & interpolants *)
        try (
            let (started_pivot_label, stop_pivot_label, itp_tuples) = 
                find_pivot_labels_and_interpolants
                    (*(sliced_path.art_edges_in_path)*)
                    (encoded_hash)
                    (sliced_path)
            in
            (* Check if sliced_pash contains weak-constraints *)
            if (compare_art_labels started_pivot_label ("", -1)) = 0 then begin
                raise (La_exception.Weak_constraints sliced_path.art_edges_in_path)
            end;
            (* Check null-pointer *)
            debug (string_of_art_label stop_pivot_label);
            check_null_pointer_from_entry_to_pivot_node
                (changed_index_path)
                (!null_pointer_exceptions) (stop_pivot_label);
            (* Return *)
            (started_pivot_label, itp_tuples))
        with
          LogicUtil.Logic.SAT_FORMULAS _ -> (
            check_null_pointer_in_path (changed_index_path)
                                       (!null_pointer_exceptions);
            (* If checking null-pointer is successful, then return *)
            (* debug *)
            (debug "Error: reach to an error node");
            (*debug ("Preds_of_path: \n" ^ string_of_predicate pred_of_path);*)
            (("", - 1), []))
    (* else, print error notification & return *)
    end else begin
        (* Try to check null-pointer *)
        check_null_pointer_in_path (changed_index_path) (!null_pointer_exceptions);

        (* If checking null-pointer is successful, then return *)
        (* debug *)
        (debug "Error: reach to an error node");
        (*debug ("Preds_of_path: \n" ^ string_of_predicate pred_of_path);*)
        (* return *)
        (("", - 1), [])
    end
end
  
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** {i Reference: An abstraction/refinement Scheme for model checking C 
programs (Georg WeiBenbacher). Refinement Algorism of WeiBenbacher:                                                    
{ul {li Create list Condition Thread LC and sort the list by length of    
condition thread.}                                                        
    {li Get list <Head of Condition Thread> LH.}                              
    {li Start with an empty set of predicates and call it L.}      
    {li Consider the next head in the LH and add it to L. Construct the 
conjunction of all heads in LH & check if it yealds a contradiction. If 
this is the case, continue with step 5, otherwise reiterate step 4.}     
    {li Find smallest-subset of LH which yields a contradiction.} 
    {li Add the elements of threads whose heads are in the smallest-subset.} 
    {li Remove all predicates that are false from LH, then return LH.}        
}}
    Analyse an error-path with an idea "slice error-path with criteria 
which are found by applying WeiBenbacher idea". Then, check if the error-path 
reachs to error-node or not. If the error-path is reach to error-node (fomula of 
constraints of the path is sat) then raise error notification. Else, with each 
node in the path, compute interpolation to return to refine.
 - Example 1: The error-path can reach to error-node: 
{v Input-path:
node(i)     edge(i -> i + 1)         constraint                    
1.          z = 100                  <z, 0> = 100
2.          x = ctr                  <x, 0> = <ctr, 0>
3.          z = ctr + 1              <z, 1> = <ctr, 0> + 1
4.          y = ctr                  <y, 0> = <ctr, 0>
5.          assume(x = m)            <x, 0> = <m, 0>
6.          assume(z = ctr + 1)      <z, 1> = <ctr, 0> + 1
7.          assume(y = m )           <y, 0> = <m, 0>
v}
{v Sliced-path:
node(i)     edge(i -> i + 1)         constraint                    
1.          True                     True
2.          True                     True    
3.          True                     True
4.          True                     True
5.          True                     True
6.          True                     True
7.          True                     True
v}
 - Example 2: The error-path can not reach to error-node:
{v Input-path:
node(i)     edge(i -> i + 1)         constraint                    
1.          z = 100                  <z, 0> = 100
2.          x = ctr                  <x, 0> = <ctr, 0>
3.          z = ctr + 1              <z, 1> = <ctr, 0> + 1
4.          y = ctr                  <y, 0> = <ctr, 0>
5.          assume(x = m)            <x, 0> = <m, 0>
6.          assume(z = ctr + 1)      <z, 1> = <ctr, 0> + 1
7.          assume(y = m )           <y, 0> = <m, 0>
v}
{v Sliced-path:
node(i)     edge(i -> i + 1)         constraint                 interpolations                    
1.          True                     True                       True
2.          x = ctr                  <x, 0> = <ctr, 0>          x = ctr
3.          True                     True                       x = ctr
4.          y = ctr                  <y, 0> = <ctr, 0>          y = x
5.          assume(x = m)            <x, 0> = <m, 0>            y = m
6.          True                     True                       y = m
7.          assume(y != m)           <y, 0> != <m, 0>
v}
 - Param: {ul {li {b path} [art_path] maybe error path}
              {li {b gvars} [expression list] list of global variables 
(gvars is used in change_index functions, to remove local variables when 
change_index_functions meet a return_node).}}
 - Pseudo Code: 
{ul {li [Change index of variables in path with flatten constraints 
(transform path to static sigle assignment path). Example:]
{v Assume x = &y;
Edge                not-flatten constraint               flatten constraint
*x = 1;             *x = 1 /\ (x = &y ==> y = 1))        *x = 1 /\ y = 1
v}}
    {li [Encode variables of changed index path before call slicing_path 
(to reduce complexity to reduce runtime).]}
    {li [Find labels of assume nodes in the path which are criteria to 
slice the path.]}
    {li [Change index of variables in path with not flatten constraints.]}
    {li [Encoded the path which has not flatten constraints.]}
    {li [Slice the encoded path with criteria (remain only edges relevant to 
variables of criteria set.)]}
    {li [Check if the sliced path is feasible (formula of the path is sat), then
the path can reach to the error node, then print message and return]}
    {li [Else, the sliced path is infeasible (formula of the path is unsat), 
then call function "find_pivot_label_and_interpolants" to find pivot label &
interpolantions of nodes in the path]}
}
 - Return: 
{ul {li pivot label of pivot node ((string * int) type)}
    {li art_labels & interpolants, respectively ((string * predicate) list type)}
} 
 - Since: 20090611
 *)
let analysis_path_with_WeiBenbacher_paradigm (path : art_path) 
                                             (gvars: expression list)
: (string * int) * (((string * int) * predicate) list) = 
begin
    (* debug *)
    debug "Analyzing path with WeiBenbacher paradigm: ";
    (*debug (get_string_of_art_path path);*)
    (*debug (get_string_of_labels_in_art_path path);
    debug ("\n" ^ string_of_predicate (And path.art_edges_in_path));
    debug "--------------------------------\n";*)
    if (!la_debug) then pause ()
    else ();
        
    (* Step1. Change index of vars in path with flatten = true *)
    let flatten_path : art_path = 
        change_index_of_variables_in_path (path) (gvars) (true) (false) (false) 
          (ref []) in
    (* Debug *)
    (if (!la_debug) then begin 
        debug "\n\n before slicing, flatten path-formulae:\n";
        (*debug (string_of_predicate (And flatten_path.art_edges_in_path));*)
        debug (get_string_of_path_formulae flatten_path.art_edges_in_path);
    end);
    let (encoded_edges_of_flatten_path, encoded_hash_of_flatten_path) = 
        encode_vars_in_path (flatten_path.art_edges_in_path) in
    let encoded_flatten_path : art_path = {
        art_nodes_in_path = flatten_path.art_nodes_in_path;
        art_edges_in_path = encoded_edges_of_flatten_path;
    } in
    (* Step2. Find criteria to slice path *)
    let criteria_labels : (string * int) list = 
        find_criterion_labels_to_slice_path (encoded_flatten_path) in
    (* Step3. Change index of vars in path with flatten = false *)
    let null_pointer_exceptions = ref [] in
    let not_flatten_path : art_path = 
        change_index_of_variables_in_path (path) (gvars) (false) (false) (true) 
          (null_pointer_exceptions) in
    (* debug *)
    (if (!la_debug) then begin 
        debug "\n\n before slicing, not-flatten path-formulae:\n";
        (*debug (string_of_predicate (And not_flatten_path.art_edges_in_path));*)
        debug (get_string_of_path_formulae not_flatten_path.art_edges_in_path);  
    end);
    
    (* Step4. Encode variables of path before call "get_interpolants" *)
    let (encoded_edges_in_path, encoded_hash) = 
        encode_vars_in_path (not_flatten_path.art_edges_in_path) in
    let encoded_not_flatten_path : art_path = {
        art_nodes_in_path = not_flatten_path.art_nodes_in_path;
        art_edges_in_path = encoded_edges_in_path;
    } in
    (* Step5. Slice path with criteria labels *)
    let sliced_path : art_path = (*not_flatten_path*)
        slice_path_with_criteria_labels (encoded_not_flatten_path) (criteria_labels)
    in
    (* debug *)
    (if (!la_debug) then begin 
        debug "\n\n after slicing path-formulae:\n";
        (*let string_of_edges =
            string_of_predicate (And sliced_path.art_edges_in_path) in 
        debug (string_of_edges);*)
        debug (get_string_of_path_formulae sliced_path.art_edges_in_path);
        debug "--------------------------------\n";
        (*if (!la_debug) then pause ()
        else ();*)
    end);
    
    (* Step6. Check if the error_path is feasible or not *)
    let pred_of_path : predicate = (And sliced_path.art_edges_in_path) in
    let is_feasible : bool = is_sat (pred_of_path) in
    (* if is_feasible is false, then this error_path is a *)
    (* counterexample. We compute interpolants to refine: *)
    if is_feasible = false then begin
        (* Find pivot_labels & interpolants *)
        try (
            let (started_pivot_label, stop_pivot_label, itp_tuples) =
                find_pivot_labels_and_interpolants
                    (*(sliced_path.art_edges_in_path)*)
                    (encoded_hash)
                    (sliced_path)
            in
            (* Check if sliced_pash contains weak-constraints *)
            if (compare_art_labels started_pivot_label ("", -1)) = 0 then begin
                raise (La_exception.Weak_constraints sliced_path.art_edges_in_path)
            end;
            (* Check null-pointer *)
            if !check_null_algorithm = 1 then begin
	        debug (string_of_art_label stop_pivot_label);
                check_null_pointer_from_entry_to_pivot_node
                    (not_flatten_path)
                    (!null_pointer_exceptions)
                    (stop_pivot_label);
            end else if !check_null_algorithm = 2 then begin
                check_null_pointer_from_entry_to_pnode_with_WeiBenbacher_prd
                    (flatten_path)
                    (!null_pointer_exceptions)
                    (stop_pivot_label);
            end;
            (* Return *)
            (started_pivot_label, itp_tuples))
        with
          LogicUtil.Logic.SAT_FORMULAS _ -> (
          (* Try to check null-pointer *)
            if !check_null_algorithm = 1 then begin
                check_null_pointer_in_path (not_flatten_path)
                                           (!null_pointer_exceptions);
            end else if !check_null_algorithm = 2 then begin
                check_null_pointer_in_path_with_WeiBenbacher_paradigm
                    (flatten_path)
                    (!null_pointer_exceptions);
            end;
            (* If checking null-pointer is successful, then return *)
            (* debug *)
            (debug "Error: reach to an error node");
            (*debug ("Preds_of_path: \n" ^ string_of_predicate pred_of_path);*)
            (*if (!la_debug) then  pause ()
            else ();*)
            (* return *)
            (("", - 1), []))
    (* else, print error notification & return *)
    end else begin
        (* Try to check null-pointer *)
        if !check_null_algorithm = 1 then begin
            check_null_pointer_in_path (not_flatten_path) (!null_pointer_exceptions);
        end else if !check_null_algorithm = 2 then begin
            check_null_pointer_in_path_with_WeiBenbacher_paradigm (flatten_path)
              (!null_pointer_exceptions);
        end;
        (* If checking null-pointer is successful, then return *)
        (* debug *)
        (debug "Error: reach to an error node");
        (*debug ("Preds_of_path: \n" ^ string_of_predicate pred_of_path);*)
        (*if (!la_debug) then  pause ()
        else ();*)
        (* return *)
        (("", - 1), [])
    end
end     

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(**  
 - Example: 
 - Param: {ul {li {b } []}}
 - Pseudo Code: 
{ul {li []}} 
 - Return: 
 - Since: 20090611
 *)
let find_pivot_label_and_sps_with_fp_criteria (flatten_path : art_path)
    (fp_criteria : ((string * int) * expression list) list)
: (string * int) * (((string * int) * predicate) list) = 
begin
    (* Debug *)
    (*debug ("find_pivot_label_and_sps_with_fp_criteria");*)    

    (* Step1. Create node & edge tuples (because we only have List.iter2 *)
    (* functions, therefor we have to merge node & its outgoing_edge     *)
    let node_and_edge_tuples : (art_node * predicate) list = 
        create_node_and_edge_tuple (flatten_path) in
    (* Step2. Preprocess *)
    let new_node_and_edge_tuples = 
        List.rev (List.tl (List.rev node_and_edge_tuples)) in
    let new_fp_criteria = List.tl fp_criteria in
    (* Step3. Create list to save post_conditions & var to save preconds *)
    let pcs : (predicate list) list ref = ref [[True]] in
    let pre_conds : predicate list ref = ref [True] in
    (* Debug *)
    (*debug ("Forward analysis");*)
    (* Step4. Forward analysis. Foreach tuples & its criteria, find *)
    (* support predicates of it                                     *)
    (try
        (List.iter2
            (fun ((a_node : art_node), (an_edge : predicate)) (alabel, fp_criterian) ->
                if fp_criterian != [] then begin
                    (* Compute post_conditions *)
                    let new_post_conditions : predicate list = 
                        get_strongest_post_eq_conditions_with_criteria (!pre_conds)
                          (an_edge) (a_node.art_type) (fp_criterian)
                    in
                    (* Update information *)
                    pcs := !pcs @ [new_post_conditions];
                    pre_conds := new_post_conditions;
                end else begin
                    pcs := !pcs @ [[True]];
                    pre_conds := [True];
                end
            )
            new_node_and_edge_tuples
            new_fp_criteria
    )
    with
    | Invalid_argument _ -> assert false);
    (* Debug *)
    (*debug "Find pivot node & support predicates";*)
    (* Step5. Find pivot node & support predicates *)
    let art_pivot_label : (string * int) ref = ref ("", -1) in
    let found_pivot_flag : bool ref = ref false in
    let sps_tuples : ((string * int) * predicate) list ref = ref [] in
    let temp_art_label : (string * int) ref = ref ("", -1) in
    (try
        (List.iter2 
            (fun a_pcs anode -> 
                (* Create new suport predicates *)
                let new_sp : predicate = (
                    (match a_pcs with
                    | [] -> True
                    | pred_list -> And pred_list)
                ) in
                (* Update to sps_tuples: *)
                if !found_pivot_flag = false then begin
                    if (not (is_valid new_sp)) then begin 
                        found_pivot_flag := true;
                        art_pivot_label := !temp_art_label;
                        (*let clabel = art_label_to_cfa_label (anode.art_label) in
                        sps_tuples := !sps_tuples @ [clabel, new_sp]*)
                    end else temp_art_label := anode.art_label
                    (*art_pivot_label := anode.art_label;
                    found_pivot_flag := true;*)
                    (*let clabel = art_label_to_cfa_label (anode.art_label) in
                    sps_tuples := !sps_tuples @ [clabel, new_sp]*) 
                end; (*else begin
                    if (normalize new_sp) != True then begin
                        let clabel = art_label_to_cfa_label (anode.art_label) in
                        sps_tuples := !sps_tuples @ [clabel, new_sp]
                    end 
                end*)
                let clabel = art_label_to_cfa_label (anode.art_label) in
                sps_tuples := !sps_tuples @ [anode.art_label, new_sp]
       
            )
            !pcs
            flatten_path.art_nodes_in_path
    )
    with
    | Invalid_argument _ -> assert false);
    (* Step6. Return *)
    (!art_pivot_label, !sps_tuples)  
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(**  
 - Example: 
 - Param: {ul {li {b } []}}
 - Pseudo Code: 
{ul {li []}} 
 - Return: 
 - Since: 20090611
 *)
let analysis_function_pointers_in_path (path : art_path) 
                                       (gvars: expression list)
                                       (cutlist : Sexpr.function_entry_t list)
: (string * int) * (((string * int) * predicate) list) = 
begin
    (* debug *)
    debug "Analyzing function pointers in path: ";
    (*debug (get_string_of_art_path path);*)
    (*debug (get_string_of_labels_in_art_path path);
    debug ("\n" ^ string_of_predicate (And path.art_edges_in_path));
    debug "--------------------------------\n";*)
    if (!la_debug) then pause ()
    else ();
    (* debug *)
    (*let temp_not_flatten_path : art_path =
        change_index_of_variables_in_path (path) (gvars) (false) (false) in
    debug ("to debug: not flatten path: ");
    debug (get_string_of_art_path temp_not_flatten_path);*)
    (* Step1. Change index of vars in path with flatten = false *)
    let not_flatten_path : art_path = 
        change_index_of_variables_in_path (path) (gvars) (false) (false) (false) 
          (ref []) in
    (* Step2. Create based_not_flatten_path *)
    let based_not_flatten_path = create_based_vars_in_path (not_flatten_path) in
    (* Encode based_not_flatten_path: *)
    (*let (encoded_edges, encoded_hash) = 
        encode_vars_in_path (based_not_flatten_path.art_edges_in_path) in
    let encoded_path : art_path = {
        art_nodes_in_path = based_not_flatten_path.art_nodes_in_path;
        art_edges_in_path = encoded_edges;
    } in*)
    (* Step3. Slice path, reduces all edges irrelevant to function pointers *)
    let (((*encoded_*)sliced_flatten_path : art_path), 
        (fp_criteria : ((string * int) * (expression list)) list)) = 
        slice_path_with_function_pointers (*encoded_path*) 
          (based_not_flatten_path) (cutlist)
    in
    let (encoded_edges, encoded_hash) = 
        encode_vars_in_path (sliced_flatten_path.art_edges_in_path) in
    let encoded_sliced_flatten_path : art_path = {
        art_nodes_in_path = sliced_flatten_path.art_nodes_in_path;
        art_edges_in_path = encoded_edges;
    } in
    let encoded_fp_criteria = ref [] in
    List.iter
        (fun (label, lcriteria) -> 
            let encoded_lcriteria : expression list = 
                encode_vars_in_expressions lcriteria (Hashtbl.create 0) in
            encoded_fp_criteria := !encoded_fp_criteria @ [label, encoded_lcriteria]
        )
        fp_criteria;
    (* Step4. Compute support predicates with fp_criteria *)
    let (pivot_label, labels_and_interpolants) =
        find_pivot_label_and_sps_with_fp_criteria (*encoded_sliced_path*) 
          (encoded_sliced_flatten_path)(*encoded_path*) (*fp_criteria*)
          (!encoded_fp_criteria)
    in
    (* Decode labels_and_interpolants *)
    let decoded_labels_and_interpolants : ((string * int) * predicate) list ref = ref [] in
    List.iter
        (fun (label, interpolant) -> 
            let decoded_interpolant : predicate = 
                decode_vars_in_predicate (interpolant) (encoded_hash) in
            decoded_labels_and_interpolants := 
                !decoded_labels_and_interpolants @ [label, decoded_interpolant];
        )
        labels_and_interpolants;
    (* debug *)
    let string_of_pivot_label = string_of_art_label pivot_label in 
    debug ("---> Label_of_pivot_node: " ^ string_of_pivot_label);
    debug ("---> Labels_and_interpolants:");
    (*let rev_labels_and_interpolants = List.rev labels_and_interpolants in*)
    List.iter
        (fun ((a_label, _), an_itp) ->
            if (not (is_valid an_itp)) & (is_sat an_itp) then begin
                let string_of_an_itp = string_of_predicate an_itp in
                debug (a_label ^ " : " ^ string_of_an_itp)
            end
        )
        !decoded_labels_and_interpolants;
    (*if (!la_debug) then let _ = read_int() in ()
    else ();*)
     
    (* Step4. Return *)
    (pivot_label, !decoded_labels_and_interpolants)
end      

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rr_in_epath (epath : art_path) (gvars: expression list) : art_path = 
begin
    (* First, change index of variables in path *)
    let null_pointer_exceptions = ref [] in
    let changed_index_path : art_path = 
        change_index_of_variables_in_path (epath) (gvars) (false) (false) (true) 
          (null_pointer_exceptions) in
    (* encode variables of path before slicing path *)
    let (encoded_edges_in_path, encoded_hash) =
        encode_vars_in_path (changed_index_path.art_edges_in_path) in
    let encoded_path : art_path = {
        art_nodes_in_path = changed_index_path.art_nodes_in_path;
        art_edges_in_path = encoded_edges_in_path;
    } in
    (* Slice path *)
    let sliced_path : art_path = slice_path (encoded_path) in
    (* Compute weakest - pre - conditions *)
    let wpconds : predicate list ref = ref [] in
    let current_cond : predicate ref = ref True in
    (try
        (List.iter2
            (fun a_node an_edge ->
                match a_node.art_type with
                | ART_COND -> begin
                    current_cond := And [!current_cond; an_edge];
                    wpconds := !current_cond :: !wpconds;
                  end
                | ART_ASSUME -> begin
                    current_cond := And [!current_cond; an_edge];
                    wpconds := !current_cond :: !wpconds;
                  end
                | _ (* maybe predicates of an assignment *) -> begin
                    current_cond := get_weakest_precondition !current_cond an_edge;
                    wpconds := !current_cond :: !wpconds;          
                  end
            )
            (List.rev sliced_path.art_nodes_in_path)
            (List.rev sliced_path.art_edges_in_path)
        )
    with
    | Invalid_argument _ -> assert false);
    (* Update current_cond to path *)
    let newnodes = ref [] in
    (try
        (List.iter2
            (fun wpcond anode -> 
                let decoded_wpcond = decode_vars_in_predicate wpcond encoded_hash in
                anode.art_weakest_precond <- 
                    create_based_vars_in_predicate (decoded_wpcond);
                newnodes := !newnodes @ [anode])
            !wpconds
            epath.art_nodes_in_path
        )
        with
    | Invalid_argument _ -> assert false);
    (* create art_path to return *)
    let newpath = {
        art_nodes_in_path = !newnodes;
        art_edges_in_path = epath.art_edges_in_path;
    } in
    newpath
end 

