(* la/preprocess.ml                                              *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

open La_common

let print_id_list (id_list : (string * int) list) = (
    List.iter (
        fun (str, i) ->
            debug (str ^ "     " ^ (string_of_int i));
    ) id_list;
)

(* find all functions that called in a function *)
let find_funcs_called_in_a_func (a_cfa : Cfa.cfa) : string list = (
    let funcs_called = ref [] in
    (match a_cfa with
    | Cfa.CFA_VARDEF a_vardef -> ()
    | Cfa.CFA_FUNDEF a_fundef -> (
       Hashtbl.iter (
          fun str a_node -> 
              match a_node.Cfa.cfa_node_stmt with
              | Cfa.CFA_CALL (exprs, _) -> (
                let a_expr = try (List.hd exprs) with Not_found -> assert false in
                match a_expr with
                | LogicUtil.Logic.Variable func_name -> 
                  let func_name = String.sub func_name (0) (String.rindex func_name '%') in
                  if (List.mem func_name !funcs_called) = false then
                  funcs_called := !funcs_called @ [func_name];
                | _ ->
                  La_common.debug 
                      ("Skipped a call to function \"" 
                       ^ (LogicUtil.Logic.string_of_expression a_expr) 
                       ^ "\""))
              | _ -> ()
       ) a_fundef.Cfa.cfa_node_hashtbl)
    | _ -> ());
    !funcs_called
)

(* find all functions that reachable from a specific function *)
let rec find_funcs_reachable_from_curr_func (ctx : Checker.check_context_t)
                                            (current_cfa : Cfa.cfa)
                                            : Cfa.cfa list = (
    let result = ref [current_cfa] in
    let funcs_in_current_cfa = find_funcs_called_in_a_func current_cfa in
    List.iter
        (fun callee_name ->
            try
              let callee = 
                  Checker.find_cfa_in_context ctx
                                              ~caller:(Some current_cfa)
                                              callee_name
                                              
              in
              let new_funcs =
                   if List.mem callee !result = false then
                       (find_funcs_reachable_from_curr_func ctx callee)
                   else []
              in
              List.iter (fun a_func ->
                             if (List.mem a_func !result) = false then
                                 result := !result @ [a_func])
                        new_funcs
            with Not_found -> ())
        funcs_in_current_cfa;
    !result
)

let check_existence_of_error_node_in_a_cfa (a_cfa : Cfa.cfa) : bool = (
    match a_cfa with
    | Cfa.CFA_VARDEF _ -> false
    | Cfa.CFA_FUNDEF a_func -> (
      let result = ref false in
      Hashtbl.iter (
          fun str a_node -> 
              match a_node.Cfa.cfa_node_stmt with
              | Cfa.CFA_ERROR -> result := true
              | _ -> () 
      ) a_func.Cfa.cfa_node_hashtbl;
      !result
      )
    | Cfa.CFA_OTHERDEF -> false
)

let check_reachability_of_error_node (ctx : Checker.check_context_t) = (
    let main_cfa =
        try Some (Checker.find_cfa_of_function_entry
                    ctx
                    ctx.Checker.chk_afile.Cfa.afile_start)
        with Not_found -> (
            La_common.warning ("Cannot find the entry function `"
                               ^ (Sexpr.string_of_function_entry
                                      ctx.Checker.chk_afile.Cfa.afile_start)
                               ^ "'");
            None)
    in
    match main_cfa with
    | None -> ()
    | Some main_cfa -> (
        let cfas = find_funcs_reachable_from_curr_func ctx main_cfa in
        let result = ref false in
        List.iter (fun a_cfa ->
                       if (!result = false) then
                           result := check_existence_of_error_node_in_a_cfa a_cfa)
                  cfas;        
        if !result = false then (
            La_common.warning ("No assertions reachable from \""
                               ^ (Sexpr.string_of_function_entry
                                     ctx.Checker.chk_afile.Cfa.afile_start)
                               ^ "\""))
    )
)
