(* parser/clex.ml 2009-01-07 *)
(* Copyright (C) 2007-2009 Yonezawa Lab, The University of Tokyo *)

(* C Token Scanner *)

(* NOTES: (1) {getc} returns null character for EOF. (2) Distinction
   among traditional/C89/C99/GCC is unstrict.  GCC extension refers to
   GCC4.3.  (X) The '$' in identifiers is controlled by
   {target_no_dollars_in_identifier_p}.  (4) The '@' appearing in
   first taken is a single token.  The '@' in the middle of identifier
   is as a constituent.  (5) Trigraphs are not supported.  ('<:'='['
   ':>'=']' '<%'='{' '%>'='}' '%:'='#' '%:%:'='##') (6) Floating
   constants are held as double for single and double. Care be taken
   if precision matters.  Range checks to class integers are not
   performed for long long.  (7) Character constants are calculated as
   in a big-endian value.  (8) "__func__" is replaced with a string
   constant in a lexer. *)

open Num

let string_equal = Misc.string_equal

let yyTRACE = ref false

let newline_warning_p = ref false

(* BEHAVIOR CONTROLLING VARIABLES. *)

type c_features_t = C89 | C99 | GCC | NONSTANDARD

let no_dollars_in_identifier_p = false
let parse_atmark_comments_p = true
(*let ignore_escape_p = ref false*)

let yylval = ref ""

(* Reference to the body of YYLEX, used to call YYLEX inside. *)

let yylexref = ref (fun (lexbuf : Lexing.lexbuf) -> Cg.Q_NONE)

let syntaxerror = Cexpr.syntaxerror

(* ANLAB:
   - current state of the specification reader, either doing NOTHING, reading a
     MULTIPLE_LINES specs, or reading a SINGLE_LINE specs.
 *)
type spec_reader_state_t = NOTHING | MULTIPLE_LINES | SINGLE_LINE
let spec_reader_state = ref NOTHING
let spec_buffer_size = 100
(* ================================================================ *)

(*AHO*) (* REMOVED LATER (FOR DEBUGGING RULES) *)

let constant_to_string (e : Cexpr.expr_t) : string = (
  match e with
    (Cexpr.CONST_I (tn, v, s)) -> s
  | (Cexpr.CONST_F (tn, v, s)) -> s
  | (Cexpr.STRING (b, s, _)) -> s
  | _ -> (assert false))

let token_to_string (token : Cg.token) : string = (
  match token with
    Cg.Q_NONE -> "NONE"
  | Cg.Q_SPEC str -> ("SPEC:"^str)
  | (Cg.Q_IDENTIFIER s) -> "IDENTIFIER(" ^ s ^ ")"
  | (Cg.Q_TYPEDEFNAME v) -> "TYPEDEFNAME(" ^ (fst v) ^ ")"
  | (Cg.Q_TYPESPEC s) -> "TYPESPEC(" ^ s ^ ")"
  | (Cg.Q_TYPEQUAL s) -> "TYPEQUAL(" ^ s ^ ")"
  | (Cg.Q_CONSTANTI e) -> "CONSTANT/I(" ^ (constant_to_string e) ^ ")"
  | (Cg.Q_CONSTANTF e) -> "CONSTANT/F(" ^ (constant_to_string e) ^ ")"
  | (Cg.Q_STRING e) -> "STRING(" ^ (constant_to_string e) ^ ")"
  | Cg.Q_TYPEDEF -> "TYPEDEF"
  | Cg.Q_EXTERN -> "EXTERN"
  | Cg.Q_STATIC -> "STATIC"
  | Cg.Q_AUTO -> "AUTO"
  | Cg.Q_REGISTER -> "REGISTER"
  | Cg.Q_INLINE -> "INLINE"
  | Cg.Q_CONST -> "CONST"
  | Cg.Q_VOLATILE -> "VOLATILE"
  | Cg.Q_RESTRICT -> "RESTRICT"
  | Cg.Q_SIZEOF -> "SIZEOF"
  | Cg.Q_ENUM -> "ENUM"
  | Cg.Q_STRUCT -> "STRUCT"
  | Cg.Q_UNION -> "UNION"
  | Cg.Q_IF -> "IF"
  | Cg.Q_ELSE -> "ELSE"
  | Cg.Q_WHILE -> "WHILE"
  | Cg.Q_DO -> "DO"
  | Cg.Q_FOR -> "FOR"
  | Cg.Q_SWITCH -> "SWITCH"
  | Cg.Q_CASE -> "CASE"
  | Cg.Q_DEFAULT -> "DEFAULT"
  | Cg.Q_BREAK -> "BREAK"
  | Cg.Q_CONTINUE -> "CONTINUE"
  | Cg.Q_RETURN -> "RETURN"
  | Cg.Q_GOTO -> "GOTO"
  | Cg.Q_OFFSETOF -> "OFFSETOF"
  | Cg.Q_ASM -> "ASM"
  | Cg.Q_TYPEOF -> "TYPEOF"
  | Cg.Q_ALIGNOF -> "ALIGNOF"
  | Cg.Q_ATTRIBUTE -> "ATTRIBUTE"
  | Cg.Q_LABEL -> "LABEL"
  | Cg.Q_TYPES_COMPATIBLE_P -> "TYPES_COMPATIBLE_P"
  (*| Cg.Q_EXTENSION -> "EXTENSION"*)
  | Cg.Q_ELLIPSIS -> "ELLIPSIS"
  | Cg.Q_LPA -> "'('"
  | Cg.Q_RPA -> "')'"
  | Cg.Q_LBC -> "'{'"
  | Cg.Q_RBC -> "'}'"
  | Cg.Q_LBK -> "'['"
  | Cg.Q_RBK -> "']'"
  | Cg.Q_DOT -> "'.'"
  | Cg.Q_ARW -> "'->'"
  | Cg.Q_AMP -> "'&'"
  | Cg.Q_STR -> "'*'"
  | Cg.Q_EQL -> "'='"
  | Cg.Q_QES -> "'?'"
  | Cg.Q_COL -> "':'"
  | Cg.Q_COM -> "','"
  | Cg.Q_SEM -> "';'"
  | Cg.Q_ATM -> "'@'"
  | Cg.Q_DOL -> "'$'"
  | Cg.Q_PLS -> "'+'"
  | Cg.Q_MNS -> "'-'"
  | Cg.Q_DIV -> "'/'"
  | Cg.Q_MOD -> "'%'"
  | Cg.Q_BAR -> "'|'"
  | Cg.Q_HAT -> "'^'"
  | Cg.Q_TLD -> "'~'"
  | Cg.Q_BAN -> "'!'"
  | Cg.Q_PLSPLS -> "'++'"
  | Cg.Q_MNSMNS -> "'--'"
  | Cg.Q_BARBAR -> "'||'"
  | Cg.Q_AMPAMP -> "'&&'"
  | Cg.Q_LSH -> "'<<'"
  | Cg.Q_RSH -> "'>>'"
  | (Cg.Q_ASSIGN o) ->
      (begin
	match o with
	  Cexpr.SET -> "'='"
	| Cexpr.ADD -> "'+='"
	| Cexpr.SUB -> "'-='"
	| Cexpr.MUL -> "'*='"
	| Cexpr.DIV -> "'/='"
	| Cexpr.MOD -> "'%='"
	| Cexpr.LSH -> "'<<='"
	| Cexpr.RSH -> "'>>='"
	| Cexpr.AND -> "'&='"
	| Cexpr.IOR -> "'|='"
	| Cexpr.XOR -> "'^='"
	| _ -> (assert false)
      end)
  | (Cg.Q_EQUAL o) ->
      (begin
	match o with
	  Cexpr.EQ -> "'=='"
	| Cexpr.NE -> "'!='"
	| _ -> (assert false)
      end)
  | (Cg.Q_COMPARE o) ->
      (begin
	match o with
	  Cexpr.GE -> "'>='"
	| Cexpr.GT -> "'>'"
	| Cexpr.LE -> "'<='"
	| Cexpr.LT -> "'<'"
	| _ -> (assert false)
      end)
  | Cg.Q_Error -> "Error"
  | Cg.Q_Ignore -> "Ignore"
      (*NONSTANDARD*)
  | Cg.Q_ASSERT -> "$ASSERT"
  | Cg.Q_ASSUME -> "$ASSUME"
  | Cg.Q_REQUIRES -> "$REQUIRES"
  | Cg.Q_ENSURES -> "$ENSURES"
  | Cg.Q_INVARIANT -> "$INVARIANT"
  | Cg.Q_CHECK -> "$CHECK"
  | Cg.Q_IMP -> "'==>'"
  | Cg.Q_IFF -> "'<=>'"
  | Cg.Q_FORALL -> "$FORALL"
  | Cg.Q_EXISTS -> "$EXISTS")

(* ================================================================ *)

(* SCANNER *)

(* Here defines variables and small utilities.	It uses {lexbuf}, but
   only fields for refilling. *)

exception GETC_EOF_EXCEPTION

let getc_eof_mark : char = '\000'
let getc_ungetcbuf0 = ref getc_eof_mark
let getc_ungetcbuf1 = ref getc_eof_mark

(* Table of keywords. *)

let keyword_table = (Hashtbl.create 50)

(* Initializes keyword table. *)

let init_keywords () : unit = (
  (List.iter
      (fun ((n, t, f) as e) ->
	match e with
	  (n, t, f) ->
	    (Hashtbl.add keyword_table n e))
      (* name, syntax_class, features *)
      [("void", (Cg.Q_TYPESPEC "void"), C89) ;
       ("char", (Cg.Q_TYPESPEC "char"), C89) ;
       ("short", (Cg.Q_TYPESPEC "short"), C89) ;
       ("int", (Cg.Q_TYPESPEC "int"), C89) ;
       ("long", (Cg.Q_TYPESPEC "long"), C89) ;
       ("float", (Cg.Q_TYPESPEC "float"), C89) ;
       ("double", (Cg.Q_TYPESPEC "double"), C89) ;
       ("signed", (Cg.Q_TYPESPEC "signed"), C89) ;
       ("unsigned", (Cg.Q_TYPESPEC "unsigned"), C89) ;
       ("_Bool", (Cg.Q_TYPESPEC "_Bool"), C99) ;
       ("_Complex", (Cg.Q_TYPESPEC "_Complex"), C99) ;
       ("_Imaginary", (Cg.Q_TYPESPEC "_Imaginary"), C99) ;

       ("typedef", Cg.Q_TYPEDEF, C89) ;
       ("extern", Cg.Q_EXTERN, C89) ;
       ("static", Cg.Q_STATIC, C89) ;
       ("auto", Cg.Q_AUTO, C89) ;
       ("register", Cg.Q_REGISTER, C89) ;
       ("inline", Cg.Q_INLINE, C99) ;

       ("const", Cg.Q_CONST, C89) ;
       ("volatile", Cg.Q_VOLATILE, C89) ;
       ("restrict", Cg.Q_RESTRICT, C99) ;

       ("enum", Cg.Q_ENUM, C89) ;
       ("struct", Cg.Q_STRUCT, C89) ;
       ("union", Cg.Q_UNION, C89) ;

       ("break", Cg.Q_BREAK, C89) ;
       ("case", Cg.Q_CASE, C89) ;
       ("continue", Cg.Q_CONTINUE, C89) ;
       ("default", Cg.Q_DEFAULT, C89) ;
       ("do", Cg.Q_DO, C89) ;
       ("else", Cg.Q_ELSE, C89) ;
       ("for", Cg.Q_FOR, C89) ;
       ("goto", Cg.Q_GOTO, C89) ;
       ("if", Cg.Q_IF, C89) ;
       ("return", Cg.Q_RETURN, C89) ;
       ("sizeof", Cg.Q_SIZEOF, C89) ;
       ("switch", Cg.Q_SWITCH, C89) ;
       ("while", Cg.Q_WHILE, C89) ;

       ("offsetof", Cg.Q_OFFSETOF, C99) ; (* macro in C99 *)
       ("asm", Cg.Q_ASM, GCC) ;
       ("__asm__", Cg.Q_ASM, GCC) ;
       ("__asm", Cg.Q_ASM, GCC) ;

       ("__Pragma", Cg.Q_Ignore, C99) ;
       ("__func__", Cg.Q_Ignore, C99) ; (* macro in C99 *)
       ("__FUNCTION__", Cg.Q_Ignore, GCC) ;
       ("__PRETTY_FUNCTION__", Cg.Q_Ignore, GCC) ;

       ("fortran", Cg.Q_Error, GCC) ;

       ("typeof", Cg.Q_TYPEOF, GCC) ;
       ("__typeof__", Cg.Q_TYPEOF, GCC) ;
       ("__typeof", Cg.Q_TYPEOF, GCC) ;

       ("__alignof__", Cg.Q_ALIGNOF, GCC) ;
       ("__alignof", Cg.Q_ALIGNOF, GCC) ;

       (* GCC ADDED (See /usr/include/sys/cdefs.h). *)

       ("__extension__", Cg.Q_Ignore, GCC) ;

       ("__const__", Cg.Q_CONST,  GCC) ;
       ("__const", Cg.Q_CONST, GCC) ;
       ("__inline__", Cg.Q_INLINE, GCC) ;
       ("__inline", Cg.Q_INLINE, GCC) ;
       ("_inline", Cg.Q_INLINE, GCC) ;
       ("__restrict__", Cg.Q_RESTRICT, GCC) ;
       ("__restrict", Cg.Q_RESTRICT, GCC) ;
       ("__signed__", (Cg.Q_TYPESPEC "signed"), GCC) ;
       ("__signed", (Cg.Q_TYPESPEC "signed"), GCC) ;
       ("__volatile__", Cg.Q_VOLATILE, GCC) ;
       ("__volatile", Cg.Q_VOLATILE, GCC) ;

       (*("__builtin_object_size", Cg.Q_Error, GCC)*)
       ("__builtin_va_list", (Cg.Q_TYPESPEC "__builtin_va_list"), GCC) ;
       ("__builtin_offsetof", Cg.Q_OFFSETOF, GCC) ;
       ("__builtin_choose_expr", Cg.Q_Error, GCC) ;
       ("__builtin_types_compatible_p", Cg.Q_TYPES_COMPATIBLE_P, GCC) ;
       ("__builtin_va_arg", (Cg.Q_TYPESPEC "__builtin_va_arg"), GCC) ;

       ("__attribute__", Cg.Q_ATTRIBUTE, GCC) ;
       ("__attribute", Cg.Q_ATTRIBUTE, GCC) ;

       ("__deprecated__", (Cg.Q_IDENTIFIER "__deprecated__"), GCC) ;
       ("__format_arg__", (Cg.Q_IDENTIFIER "__format_arg__"), GCC) ;
       ("__noinline__", (Cg.Q_IDENTIFIER "__noinline__"), GCC) ;
       ("__nonnull__", (Cg.Q_IDENTIFIER "__nonnull__"), GCC) ;
       ("__nothrow__", (Cg.Q_IDENTIFIER "__nothrow__"), GCC) ;
       ("__pure__", (Cg.Q_IDENTIFIER "__pure__"), GCC) ;
       ("__strfmon__", (Cg.Q_IDENTIFIER "__strfmon__"), GCC) ;
       ("__unused__", (Cg.Q_IDENTIFIER "__unused__"), GCC) ;
       ("__used__", (Cg.Q_IDENTIFIER "__used__"), GCC) ;

       ("__blockattribute__", Cg.Q_Error, GCC) ;
       ("__blockattribute", Cg.Q_Error, GCC) ;

       ("__label__", Cg.Q_LABEL, GCC) ;

       ("__thread", Cg.Q_Error, GCC) ;

       ("_Decimal32", Cg.Q_Error, GCC) ;
       ("_Decimal64", Cg.Q_Error, GCC) ;
       ("_Decimal128", Cg.Q_Error, GCC) ;
       ("_Fract", Cg.Q_Error, GCC) ;
       ("_Accum", Cg.Q_Error, GCC) ;
       ("_Sat", Cg.Q_Error, GCC) ;
       ("__imag", Cg.Q_Error, GCC) ;
       ("__imag__", Cg.Q_Error, GCC) ;
       ("__real", Cg.Q_Error, GCC) ;
       ("__real__", Cg.Q_Error, GCC) ;

       (*("__bounded", Cg.Q_Error, GCC)*)
       (*("__unbounded", Cg.Q_Error, GCC)*)
       (*("__ptrvalue", Cg.Q_Error, GCC)*)

       (*NONSTANDARD*)

       ("$assert", Cg.Q_ASSERT, NONSTANDARD) ;
       ("$assume", Cg.Q_ASSUME, NONSTANDARD) ;
       ("$requires", Cg.Q_REQUIRES, NONSTANDARD) ;
       ("$ensures", Cg.Q_ENSURES, NONSTANDARD) ;
       ("$invariant", Cg.Q_INVARIANT, NONSTANDARD) ;
       ("$check", Cg.Q_CHECK, NONSTANDARD) ;
       ("$forall", Cg.Q_FORALL, NONSTANDARD) ;
       ("$exists", Cg.Q_EXISTS, NONSTANDARD)]))

let init_scanner (lexbuf : Lexing.lexbuf) : unit = (
  (*make_c_keywords_reserved () ;*)
  (Cexpr.init_basic_type_mapping ()) ;
  (Cexpr.init_builtin_identifiers ()) ;
  (Cattr.init_attribute_null_handler ()) ;
  (init_keywords ()) ;
  Cexpr.current_line := 1 ;
  Cexpr.current_trueline := 1 ;
  (* Reset positions in lexbuf. *)
  lexbuf.Lexing.lex_mem <- (Array.create (4 * 1024) (-1)) ;
  lexbuf.Lexing.lex_start_pos <- lexbuf.Lexing.lex_curr_pos ;
  lexbuf.Lexing.lex_last_pos <- lexbuf.Lexing.lex_curr_pos ;
  lexbuf.Lexing.lex_last_action <- (-1))

(* Gets and returns a character. *)

let rec getc_raw (lexbuf : Lexing.lexbuf) (raise_on_eof_p : bool) : char = (
  if (!getc_ungetcbuf0 != getc_eof_mark) then
    (begin
      let c = !getc_ungetcbuf0 in
      getc_ungetcbuf0 := !getc_ungetcbuf1 ;
      getc_ungetcbuf1 := getc_eof_mark ;
      c
    end)
  else if (lexbuf.Lexing.lex_curr_pos >= lexbuf.Lexing.lex_buffer_len) then
    if (lexbuf.Lexing.lex_eof_reached) then
      if (raise_on_eof_p) then
	raise GETC_EOF_EXCEPTION
      else
	getc_eof_mark
    else
      (begin
	(lexbuf.Lexing.refill_buff lexbuf) ;
	(getc_raw lexbuf raise_on_eof_p)
      end)
  else
    (begin
      let c = lexbuf.Lexing.lex_buffer.[lexbuf.Lexing.lex_curr_pos] in
      lexbuf.Lexing.lex_curr_pos <- (lexbuf.Lexing.lex_curr_pos + 1) ;
      if (c == '\n') then
	(begin
	  (incr Cexpr.current_line) ;
	  (incr Cexpr.current_trueline) ;
	  (Cexpr.current_cnum_from_bol := 0) ;
	end)
      else
        (incr Cexpr.current_cnum_from_bol) ;
      c
    end))

(* Reads a char.  It may return a null character '\000' for EOF when
   {raise_on_eof_p} is false.  It handles continuation lines (by
   concatenating lines). *)

let rec getc (lexbuf : Lexing.lexbuf) (raise_on_eof_p : bool) : char = (
  let c0 = (getc_raw lexbuf raise_on_eof_p) in
  if (c0 == '\\') then
    (begin
      let c1 = (getc_raw lexbuf raise_on_eof_p) in
      if (c1 == '\n') then
	(getc lexbuf raise_on_eof_p)
      else
	(begin
	  (ungetc c1 lexbuf) ;
	  c0
	end)
    end)
  else
    c0)

(* NOTE: {ungetc} should put char in lexbuf, but not so now. *)

and ungetc (c : char) (lexbuf : Lexing.lexbuf) : unit = (
  assert (!getc_ungetcbuf1 == getc_eof_mark) ;
  assert (c != getc_eof_mark) ;
  getc_ungetcbuf1 := !getc_ungetcbuf0 ;
  getc_ungetcbuf0 := c)

(* LINE_NOTE is a record of a file name and a line number. *)

type linenote_t = {
  linenote_file : string ;
  linenote_line : int
}

(*NOTUSED*)
let print_linenote (ln : linenote_t) (stream : out_channel) level : unit = (
  (Printf.fprintf stream "#<%s:%d>"
      (ln.linenote_file) (ln.linenote_line)))

(* ================================================================ *)

let make_integer_constant tn (value : num) (token : string) : Cg.token = (
  (Cg.Q_CONSTANTI (Cexpr.CONST_I (tn, value, token))))

let make_floating_constant tn (value : float) (token : string) : Cg.token = (
  (Cg.Q_CONSTANTF (Cexpr.CONST_F (tn, value, token))))

let make_string_constant (wide_character_p : bool) (s : string) (content : int array) : Cg.token = (
  (Cg.Q_STRING (Cexpr.STRING (wide_character_p, s, content))))

let alpha_char_p (c : char) : bool = (
  ('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z'))

let alphanumericp (c : char) : bool = (
  ('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z') || ('0' <= c && c <= '9'))

let digit_char_p (c : char) (base : int) : bool = (
  assert (base == 8 || base == 10 || base == 16) ;
  if (base == 8) then
    ('0' <= c && c <= '7')
  else if (base == 10) then
    ('0' <= c && c <= '9')
  else
    ('0' <= c && c <= '9') || ('a' <= c && c <= 'f') || ('A' <= c && c <= 'F'))

let digit_char_value (c : char) (base : int) : int = (
  if ('0' <= c && c <= '9') then
    ((int_of_char c) - (int_of_char '0'))
  else if ('a' <= c && c <= 'f') then
    ((int_of_char c) - (int_of_char 'a'))
  else
    ((int_of_char c) - (int_of_char 'A')))

(* Miscellaneouses for Scanner *)

(* Skip all characters in the stream until the first occurrence of
   the specified key, meanwhile each skipped character is copied into
   a temporary buffer for later reference. *)

let skip_until_key_appears (stream) (key : string) (spec_buffer) : bool = (
  let rec skip_until_key_appears1 (stream)
                                  (key : string)
				  (pos : int)
				  (spec_buffer) : bool = (
    let c = (getc stream false) in
    if c = String.get key pos then (
      (Buffer.add_char spec_buffer c);
      if pos = ((String.length key) - 1) then
        true
      else
        (skip_until_key_appears1 (stream) (key) (pos + 1) spec_buffer) )
    else
      match c with
        '\000' (*EOF*) ->
          (failwith "EOF while skipping a Doxygen-style comment")
      | '*' ->
          (begin
	     let c1 = (getc stream false) in
	     match c1 with
	       '\000' (*EOF*) ->
	          (failwith "EOF while skipping a Doxygen-style comment")
	     | '/' ->
	          false;
             | _ -> (
                  (Buffer.add_char spec_buffer c) ;
	          (ungetc c1 stream) ;
	          (skip_until_key_appears1 stream key 0 spec_buffer))
           end)
      | _ ->
          ((Buffer.add_char spec_buffer c);
           (skip_until_key_appears1 stream key 0 spec_buffer))
  )
  in skip_until_key_appears1 stream key 0 spec_buffer
)

(* Skips whitespaces and comments, and returns a next character.  This
   returns a lastly read character, to avoid backing up pre-read
   characters. *)

let rec skip_whitespaces_and_comments stream (return_at_newline : bool) = (
  let c = (getc stream false) in
  match c with
    '\000' (*EOF*) ->
      c
  | '\n' ->
      if (return_at_newline) then
	c
      else
	(skip_whitespaces_and_comments stream return_at_newline)
  | ' ' | '\t' | '\b' | '\x12' | '\x17' ->
      (* 0x12='\f', 0x17='\v'. *)
      (skip_whitespaces_and_comments stream return_at_newline)
  | '#' ->
      ignore (scan_sharp_directive stream);
      (skip_whitespaces_and_comments stream return_at_newline)
  | '\r' ->
      if (!Cexpr.pedantic_warn_p && (not !newline_warning_p)) then
	begin
	  newline_warning_p := true
	end ;
      (skip_whitespaces_and_comments stream return_at_newline)
  | '\\' ->
      let c1 = (getc stream false) in
      assert (c1 <> '\n') ;
      (failwith "Stray '\\' in program")
  | '/' ->
      (* Check if a comment starts. *)
      (begin
	let c1 = (getc stream false) in
	match c1 with
	  '\000' (*EOF*) ->
	    c
	| '*' -> (
            let c2 = (getc stream false) in
            match c2 with
	    | '!' ->
	        (* The `real` specification (for checking) does not start here,
		   but right after the keyword '@code' *)
	        spec_reader_state := MULTIPLE_LINES;
                c2
            | '@' ->
                spec_reader_state := MULTIPLE_LINES;
		Cexpr.spec_startpos_ln := !Cexpr.current_line;
		Cexpr.spec_startpos_cnum_from_bol :=
			!Cexpr.current_cnum_from_bol;
                c2
            | _ ->
                (ungetc c2 stream;
	        skip_comment_to_starslash stream ))
	| '/' -> (
            let c2 = (getc stream false) in
            match c2 with
            | '@' -> (
	    	Cexpr.spec_startpos_ln := !Cexpr.current_line;
		Cexpr.spec_startpos_cnum_from_bol :=
			!Cexpr.current_cnum_from_bol;
                spec_reader_state := SINGLE_LINE;
                c2
            )
            | _ ->  (skip_comment_to_eol stream)
        )
	| _ ->
	    (ungetc c1 stream) ;
	    c
      end)
  | _ ->
      c)

and skip_comment_to_starslash stream = (
  let c = (getc stream false) in
  match c with
    '\000' (*EOF*) ->
      (failwith "EOF while skipping a comment")
  | '*' ->
      (begin
	      let c1 = (getc stream false) in
	      match c1 with
	        '\000' (*EOF*) ->
	        (failwith "EOF while skipping a comment")
	      | '/' ->
	        (skip_whitespaces_and_comments stream false)
              | _ ->
	        (ungetc c1 stream) ;
	        (skip_comment_to_starslash stream)
      end)
  | _ ->
      (skip_comment_to_starslash stream))

(* Skips to end-of-line, and then skips whitespaces again. *)

and skip_comment_to_eol stream = (
  let c = (getc stream false) in
  match c with
    '\000' (*EOF*) ->
      (failwith "EOF while skipping a comment")
  | '\n' ->
      (skip_whitespaces_and_comments stream false)
(*  | '@' ->
        spec_reader_state := SINGLE_LINE;
        c		*)
  | _ ->
      (skip_comment_to_eol stream))


(*
let rec skip_whitespaces_and_comments (stream : Lexing.lexbuf) (return_at_newline_p : bool) : char = (

  let c = (getc stream false) in

  match c with
    '\000' (*EOF*) ->
      c
  | '\n' ->
      if (return_at_newline_p) then
	c
      else
	(skip_whitespaces_and_comments stream return_at_newline_p)
  | ' ' | '\t' | '\b' | '\x12' | '\x17' ->
      (* 0x12='\f', 0x17='\v'. *)
      (skip_whitespaces_and_comments stream return_at_newline_p)
  | '#' ->
      ignore (scan_sharp_directive stream) ;
      (skip_whitespaces_and_comments stream return_at_newline_p)
  | '\r' ->
      if (!Cexpr.pedantic_warn_p && (not !newline_warning_p)) then
	(begin
	  (Cexpr.warn_in_parse "Carriage return in the source file") ;
	  newline_warning_p := true
	end) ;
      (skip_whitespaces_and_comments stream return_at_newline_p)
  | '\\' ->
      (* Extended identifier. *)
      c
  | '/' ->
      (* Check if a comment starts. *)
      (begin
	let c1 = (getc stream false) in
	match c1 with
	  '\000' (*EOF*) ->
	    c
	| '*' ->
	    if (not parse_atmark_comments_p) then
	      (skip_comment_to_starslash stream)
	    else
	      (begin
		let c2 = (getc stream false) in
		match c2 with
		  '\000' (*EOF*) ->
		    (skip_comment_to_starslash stream)
		| '@' ->
		    (skip_whitespaces_and_comments stream return_at_newline_p)
		| _ ->
		    (ungetc c2 stream) ;
		    (skip_comment_to_starslash stream)
	      end)
	| '/' ->
	    if (not parse_atmark_comments_p) then
	      (skip_comment_to_eol stream)
	    else
	      (begin
		let c2 = (getc stream false) in
		match c2 with
		  '\000' (*EOF*) ->
		    (skip_comment_to_eol stream)
		| '@' ->
		    (skip_whitespaces_and_comments stream return_at_newline_p)
		| _ ->
		    (ungetc c2 stream) ;
		    (skip_comment_to_eol stream)
	      end)
	| _ ->
	    (ungetc c1 stream) ;
	    c
      end)
  | '@' ->
      if (not parse_atmark_comments_p) then
	c
      else
	(begin
	  let c1 = (getc stream false) in
	  match c1 with
	    '\000' (*EOF*) ->
	      c
	  | '*' ->
	      (begin
		let c2 = (getc stream false) in
		match c2 with
		  '\000' (*EOF*) ->
		    (ungetc c1 stream) ;
		    c
		| '/' ->
		    (skip_whitespaces_and_comments stream return_at_newline_p)
		| _ ->
		    (ungetc c1 stream) ; (ungetc c2 stream) ;
		    c
	      end)
	  | _ ->
	      (ungetc c1 stream) ;
	      c
	end)
  | _ ->
      c)

and skip_comment_to_starslash (stream : Lexing.lexbuf) : char = (
  let c = (getc stream false) in
  match c with
    '\000' (*EOF*) ->
      (syntaxerror ("EOF while skipping a comment"))
  | '*' ->
      (begin
	let c1 = (getc stream false) in
	match c1 with
	  '\000' (*EOF*) ->
	    (syntaxerror "EOF while skipping a comment")
	| '/' ->
	    (skip_whitespaces_and_comments stream false)
	| _ ->
	    (ungetc c1 stream) ;
	    (skip_comment_to_starslash stream)
      end)
  | _ ->
      (skip_comment_to_starslash stream))

(* Skips to end-of-line, and then skips whitespaces again. *)

and skip_comment_to_eol (stream : Lexing.lexbuf) : char = (
  let c = (getc stream false) in
  match c with
    '\000' (*EOF*) ->
      (syntaxerror "EOF while skipping a comment")
  | '\n' ->
      (skip_whitespaces_and_comments stream false)
  | _ ->
      (skip_comment_to_eol stream))
*)
and scan_sharp_directive (stream : Lexing.lexbuf) : string list = (
  (* Skip spaces and tabs. *)
  let c = (skip_spaces_and_tabs stream) in
  if (digit_char_p c 10) then
    (begin
      (* Line number directive. *)
      (ungetc c stream) ;
      let tokens = (scan_sharp_directive_tokens ["line"] stream) in
      (lookat_line_directive tokens) ;
      []
    end)
  else
    (begin
      (* Get sharp directive keyword. *)
      let token : string = (scan_pp_token c stream) in
      if (string_equal token "") then
	[]
      else
	(begin
	  let c2 = (skip_whitespaces_and_comments stream true) in
	  (ungetc c2 stream) ;
	  match token with
	    "line" ->
	      let tokens = (scan_sharp_directive_tokens ["line"] stream) in
	      (lookat_line_directive tokens) ;
	      []
	  | "pragma" ->
	      let tokens = (scan_sharp_directive_tokens ["pragma"] stream) in
	      (Vector.add (List.hd !Cexpr.pragma_scope_stack) tokens) ;
	      []
	  | "error" ->
	      (Cexpr.warn_pedantic_in_parse "Directive error ignored") ;
	      (scan_sharp_directive_tokens ["error"] stream)
	  | _ ->
	      (* Ignore other directives. *)
	      (Cexpr.warn_in_parse ("Ignore directive '" ^ token ^ "'")) ;
	      (begin ignore (skip_to_eol stream) ; [] end)
	end)
    end))

and lookat_line_directive (tokens : string list) : unit = (
  match tokens with
    (directive :: line :: file :: _) ->
      let n = (int_of_string line) in
      Cexpr.current_line := n ;
      Cexpr.current_file := file
  | _ ->
      (Cexpr.warn_in_parse ("Strange line directive (ignored)")))

and scan_sharp_directive_tokens (tokens : string list) (stream : Lexing.lexbuf) : string list = (
  let c = (getc stream false) in
  match c with
    '\000' (*EOF*) ->
      tokens
  | '\n' ->
      tokens
  | '\\' ->
      let c1 = (getc stream false) in
      if (c1 == getc_eof_mark) then
	(syntaxerror "EOF after backslash")
      else
	(syntaxerror "Stray backslash")
  | _ ->
      (ungetc c stream) ;
      ignore (!yylexref stream) ;
      let c2 = (skip_whitespaces_and_comments stream true) in
      (ungetc c2 stream) ;
      (scan_sharp_directive_tokens (tokens @ [!yylval]) stream))

(* Skips white spaces, and returns the next character or EOF (the
   returned char is not a white space). *)

and skip_spaces_and_tabs (stream : Lexing.lexbuf) : char = (
  let c = (getc stream false) in
  match c with
    '\000' (*EOF*) ->
      c
  | ' ' | '\t' ->
      (skip_spaces_and_tabs stream)
  | _ ->
      c)

(* Skips to EOF or EOF, and returns the last char (nl or null). *)

and skip_to_eol (stream : Lexing.lexbuf) : char = (
  let c = (getc stream false) in
  match c with
    '\000' (*EOF*) ->
      c
  | '\n' ->
      c
  | _ ->
      (skip_to_eol stream))

and scan_pp_token (c : char) (stream : Lexing.lexbuf) : string = (
  let token = (Buffer.create 10) in
  (Buffer.add_char token c) ;
  (scan_pp_token_loop token stream))

and scan_pp_token_loop (token : Buffer.t) (stream : Lexing.lexbuf) : string = (
  let c = (getc stream false) in
  if (c == getc_eof_mark || (alpha_char_p c)) then
    (begin
      (Buffer.add_char token c) ;
      (scan_pp_token_loop token stream)
    end)
  else
    (begin
      (ungetc c stream) ;
      (Buffer.contents token)
    end))

(* ================================================================ *)

(* Character/String Constant Scanner *)

(* Scans character constant. *)

let rec scan_character_constant (stream : Lexing.lexbuf) (wide_character_p : bool) : (Cg.token * string) = (
  let token = (Buffer.create 10) in
  (Buffer.add_char token '\'') ;
  let v = (num_of_int (scan_character_constant_loop stream token 0 0)) in
  let s = (Buffer.contents token) in
  ((make_integer_constant Cexpr.INTEGER v s), s))

and scan_character_constant_loop (stream : Lexing.lexbuf) (token : Buffer.t) (value : int) (count : int) : int = (
  let c = (getc stream false) in
  match c with
    '\000' (*EOF*) ->
      (syntaxerror "EOF in a character constant")
  | '\'' ->
      if (count == 0) then
	(syntaxerror "Empty character constant")
      else if (count > Cexpr.maximum_characters_in_character_constant) then
	(syntaxerror "Character constant too long")
      else
	(begin
	  if ((count >= 2) && (not !Cexpr.traditional_p)) then
	    (Cexpr.warn_in_parse "Multi-character character constant") ;
	  (Buffer.add_char token c) ;
	  value
	end)
  | '\\' ->
      (Buffer.add_char token c) ;
      let v = (scan_escaped_character stream token) in
      (scan_character_constant_loop stream token
	  (value * 256 + v) (count + 1))
  | '\n' ->
      (Cexpr.warn_pedantic_in_parse "C forbids newline in character constant.\n") ;
      (Buffer.add_char token c) ;
      (scan_character_constant_loop stream token
	  (value * 256 + (int_of_char c)) (count + 1))
  | _ ->
      (Buffer.add_char token c) ;
      (scan_character_constant_loop stream token
	  (value * 256 + (int_of_char c)) (count + 1)))

(* Scans a string constant. *)

and scan_string_constant (c : char) (stream : Lexing.lexbuf) (wide_character_p : bool) : (Cg.token * string) = (
  let token = (Buffer.create 10) in
  let content = (Vector.create 10 0) in
  (scan_string_constant_loop stream token content) ;
  let s = (Buffer.contents token) in
  ((make_string_constant wide_character_p s (Vector.to_array content)), s))

(* Collects characters in a string buffer {token}. *)

and scan_string_constant_loop (stream : Lexing.lexbuf) (token : Buffer.t) (content: int Vector.t) : unit = (
  let c = (getc stream false) in
  match c with
    '\000' (*EOF*) ->
      (syntaxerror "EOF in a string constant")
  | '\"' ->
      (Vector.add content 0) ;
      ()
  | '\\' ->
      (Buffer.add_char token c) ;
      let v = (scan_escaped_character stream token) in
      (Vector.add content v) ;
      (scan_string_constant_loop stream token content)
  | '\n' ->
      (Cexpr.warn_pedantic_in_parse "A newline in a string constant.\n") ;
      (Buffer.add_char token c) ;
      let v = 10 in
      (Vector.add content v) ;
      (scan_string_constant_loop stream token content)
  | _ ->
      (Buffer.add_char token c) ;
      let v = (Char.code c) in
      (Vector.add content v) ;
      (scan_string_constant_loop stream token content))

(* Scans backslash-escaped character.  It returns an integer value of
   the character. *)

and scan_escaped_character (stream : Lexing.lexbuf) (token : Buffer.t) : int = (
  (* warn as "escape sequence out of range for character" *)
  let c = (getc stream false) in
  match c with
    '\000' (*EOF*) ->
      (syntaxerror "EOF in escape sequence")
  | 'x' ->
      (* Scan hexadecimal. *)
      (Buffer.add_char token c) ;
      (scan_escaped_character_digits stream token 16 (-1) 0 0)
  | '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' ->
      (* Scan octal. *)
      (Buffer.add_char token c) ;
      let v = (digit_char_value c 8) in
      (scan_escaped_character_digits stream token 8 3 0 v)
  | 'a' | 'b' | 'f' | 'n' | 'r' | 't' | 'v' | '\\' | '\'' | '\"' | '?' ->
      (* '\a' is a bell, '\v' is a vertical tab. *)
      (begin
	(Buffer.add_char token c) ;
	match c with
	  'a' -> 0x07 (*BEL/alert*)
	| 'b' -> 0x08 (*BS*)
	| 'f' -> 0x0C (*FF*)
	| 'n' -> (int_of_char '\n') (*NL*)
	| 'r' -> (int_of_char '\r') (*CR*)
	| 't' -> 0x09 (*HT*)
	| 'v' -> 0x0B (*VT*)
	| '\\' -> (int_of_char '\\')
	| '\'' -> (int_of_char '\'')
	| '\"' -> (int_of_char '\"')
	| '?' -> 0xFF (*DEL/???*)
	| _ -> (assert false)
      end)
  | '\n' ->
      (assert false)
	(*(Buffer.add_char token c) ;*)
	(*(int_of_char c)*)
  | 'e' | 'E' ->
      (Cexpr.warn_pedantic_in_parse ("Non-standard escape sequence '\\e'")) ;
      (Buffer.add_char token c) ;
      0x1B (*ESC*)
  | 'u' | 'U' ->
      (* Universal character. *)
      (Buffer.add_char token c) ;
      let digits = (if (c == 'u') then 4 else 8) in
      (scan_escaped_character_digits stream token 16 digits 0 0)
  | '(' | '{' | '[' | ')' | '}' | ']' ->
      (*AHO*) (* (I forgot why these are here). *)
      (Cexpr.warn_pedantic_in_parse ("Non-standard escape sequence, '\\"
				      ^ (Char.escaped c) ^ "'")) ;
      (Buffer.add_char token c) ;
      (int_of_char c)
  | _ ->
      (Cexpr.warn_in_parse ("Unknown escape sequence '\\"
			     ^ (Char.escaped c) ^ "'")) ;
      (Buffer.add_char token c) ;
      (int_of_char c))

(* Scans digits in {base} upto {digits} count (count is unlimited when
   {digits} is -1). *)

and scan_escaped_character_digits (stream : Lexing.lexbuf) (token : Buffer.t) (base : int) (digits : int) (count : int) (value : int) : int = (
  if (count == digits) then
    value
  else
    let c = (getc stream false) in
    if (c == getc_eof_mark) then
      (begin
	if (count == 0) then
	  (syntaxerror "\\x or \\0 followed by no digits")
	else
	  value
      end)
    else if (digit_char_p c base) then
      (begin
	(Buffer.add_char token c) ;
	(scan_escaped_character_digits stream token base digits
	    (count + 1) (value * base + (digit_char_value c base)))
      end)
    else
      (begin
	(ungetc c stream) ;
	if (count == 0) then
	  if (base == 8) then
	    (* Allowed '\0' only. *)
	    value
	  else
	    (syntaxerror "\\x or \\0 followed by no digits")
	else
	  value
      end))


(* ================================================================ *)
(* RAW specifications - string *)

(* BEHAVIOUR: keeps on reading the specs until
       - EOL in case of SINGLE_LINE specs
       - or closing mark of specs *)

let rec scan_spec_string_loop stream spec_buffer = (
  assert (!spec_reader_state = MULTIPLE_LINES
                || !spec_reader_state = SINGLE_LINE);
  let c = (getc stream false) in
  match c with
  | '\000' (*EOF*) ->
     (failwith "EOF within a specification string.")
  | '*' ->
     (begin
     let c1 = (getc stream false) in
         match c1 with
         | '\000' (*EOF*) ->
            (failwith "EOF within a specification string.")
         | '/' ->
            () (* at the end of a spec string -> do nothing*)
         | _ ->
            (Buffer.add_char spec_buffer c);
            (Buffer.add_char spec_buffer c1);
            (scan_spec_string_loop stream spec_buffer)
     end)
  | '@' ->
     (scan_spec_string_loop stream spec_buffer)
  | '\n' ->
     (* - continue reading specifications in following lines in case of
          multiple-line specification (/*@*/)
        - stop reading in case of single-line specs (//@) *)
     if (!spec_reader_state = MULTIPLE_LINES) then
        (begin
        (Buffer.add_char spec_buffer c);
        (scan_spec_string_loop stream spec_buffer)
        end)
     else (* SINGLE_LINE *)
        (Buffer.add_char spec_buffer c)
  | _ ->
     (Buffer.add_char spec_buffer c);
     (scan_spec_string_loop stream spec_buffer))

let scan_spec_string stream : string = (
  let spec_buffer = (Buffer.create spec_buffer_size) in
  (scan_spec_string_loop stream spec_buffer) ;
  (Buffer.contents spec_buffer)
  )

(* ================================================================ *)

(* IDENTIFIER SCANNER *)

(* Scans an identifier. *)

let rec scan_identifier (c : char) (stream : Lexing.lexbuf) : (Cg.token * string) = (
  let token = (Buffer.create 10) in
  (Buffer.add_char token c) ;
  let id =
    if (c != '\\') then
      (scan_identifier_loop stream token)
    else
      (scan_identifier_loop_for_backslash stream token) in
  (* Try recognize a keyword. *)
  if (Hashtbl.mem keyword_table id) then
    (begin
      let (n, t, f) = (Hashtbl.find keyword_table id) in
      if (t == Cg.Q_Error) then
	(syntaxerror ("Bad (unsupported) keyword: " ^ id)) ;
      if (f != C89 && f != C99) then
	(Cexpr.warn_pedantic_in_parse ("Non-standard keyword '" ^ id ^ "'")) ;
      match n with
	"__func__" | "__FUNCTION__" | "__PRETTY_FUNCTION__" ->
	  let f = !Cexpr.current_function in
	  if (f == Cexpr.nil_declaration) then
	    (syntaxerror ("__func__ outside of a function")) ;
	  let s = (fst f.Cexpr.decl_id) in
	  let content = (Misc.string_to_int_array n) in
	  ((make_string_constant false s content), s)
      | "__Pragma" ->
	  let tokens = (scan_sharp_directive_tokens ["pragma"] stream) in
	  (Vector.add (List.hd !Cexpr.pragma_scope_stack) tokens) ;
	  (Cg.Q_Ignore, id)
      | _ ->
	  (t, id)
    end)
  else
    (check_typedef_name id))

(* Collects identifier constituent characters. *)

and scan_identifier_loop (stream : Lexing.lexbuf) (token : Buffer.t) : string = (
  let c = (getc stream false) in
  if (c == getc_eof_mark) then
    (Buffer.contents token)
  else if (alphanumericp c) then
    (begin
      (Buffer.add_char token c) ;
      (scan_identifier_loop stream token)
    end)
  else
    (begin
      match c with
	'_' | '@' ->
	  (Buffer.add_char token c) ;
	  (scan_identifier_loop stream token)
      | '$' ->
	  if (not no_dollars_in_identifier_p) then
	    (begin
	      (ungetc c stream) ;
	      (Buffer.contents token)
	    end)
	  else
	    (begin
	      (Buffer.add_char token c) ;
	      (scan_identifier_loop stream token)
	    end)
      | '\\' ->
	  (* Extended identifier (universal characters). *)
	  (*(syntaxerror "Backslash appears in the middle of an identifier")*)
	  (Buffer.add_char token c) ;
	  (scan_identifier_loop_for_backslash stream token)
      | _ ->
	  (ungetc c stream) ;
	  (Buffer.contents token)
    end))

(* Scans an extended identifier after seeing a backslash. *)

and scan_identifier_loop_for_backslash (stream : Lexing.lexbuf) (token : Buffer.t) : string = (
  let c = (getc stream false) in
  match c with
    '\000' (*EOF*) ->
      (syntaxerror "EOF in escape sequence")
  | 'u' | 'U' ->
      (Buffer.add_char token c) ;
      let digits = (if (c == 'u') then 4 else 8) in
      ignore (scan_escaped_character_digits stream token 16 digits 0 0) ;
      (scan_identifier_loop stream token)
  | _ ->
      (syntaxerror ("Unknown escape sequence '\\" ^ (Char.escaped c) ^ "'")))

(* Makes an IDENTIFIER/TYPEDEFNAME token, by checking if the given id
   is a typedef name with regard to the current scope. *)

and check_typedef_name (id : string) : (Cg.token * string) = (
  let b = (Cexpr.get_binding id !Cexpr.name_scope_stack false) in
  if (not (Cexpr.binding_exists_p b)) then
    ((Cg.Q_IDENTIFIER id), id)
  else if (b.Cexpr.binding_scls != Cexpr.TYPEDEF) then
    ((Cg.Q_IDENTIFIER id), id)
  else
    ((Cg.Q_TYPEDEFNAME (id, b.Cexpr.binding_tn)), id))

(* ================================================================ *)

(* Number Scanner *)

type scan_digits_state_t = SCAN_INTEGER | SCAN_FRACTION | SCAN_EXPONENT

(* Scans all digits including all alphanumerics which may be garbage
   characters.	It is called with {c} a dot or a digit.	 It scans
   either a number, a dot (structure reference) or dots (ellipse).
   Scan finishes if seen one of them.  It returns a token: one of
   ELLIPSIS, DOT, CONSTANTI, and CONSTANTF. *)

let rec scan_digits (c : char) (stream : Lexing.lexbuf) : (Cg.token * string) = (
  assert (c != getc_eof_mark);
  let token = (Buffer.create 10) in
  match c with
    '.' ->
      (* Check for a single dot, an ellipsis, or a fraction. *)
      (begin
	(Buffer.add_char token c) ;
	let c1 = (getc stream false) in
	match c1 with
	  '\000' (*EOF*) ->
	    (* A single dot. *)
	    (Cg.Q_DOT, ".")
	| '.' ->
	    (begin
	      Buffer.add_char token c1 ;
	      let c2 = (getc stream false) in
	      match c2 with
		'\000' (*EOF*) ->
		  (syntaxerror "Parse error at '..'") ;
	      | '.' ->
		  (Cg.Q_ELLIPSIS, "...")
	      | _ ->
		  (syntaxerror "Parse error at '..'")
	    end)
	| '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9' ->
	    (Buffer.add_char token c1) ;
	    (scan_digits_loop stream token 10 SCAN_FRACTION)
	| _ ->
	    (* A single dot. *)
	    (ungetc c1 stream) ;
	    (Cg.Q_DOT, ".")
      end)
  | '0' ->
      (begin
	(Buffer.add_char token c) ;
	let c1 = (getc stream false) in
	match c1 with
	  '\000' (*EOF*) ->
	    (scan_digits_loop stream token 10 SCAN_INTEGER)
	| 'x' | 'X' ->
	    (Buffer.add_char token c1) ;
	    (scan_digits_loop stream token 16 SCAN_INTEGER)
	| '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9' ->
	    (Buffer.add_char token c1) ;
	    (scan_digits_loop stream token 8 SCAN_INTEGER)
	| _ ->
	    (ungetc c1 stream) ;
	    (scan_digits_loop stream token 10 SCAN_INTEGER)
      end)
  | _ ->
      assert (digit_char_p c 10) ;
      (Buffer.add_char token c) ;
      (scan_digits_loop stream token 10 SCAN_INTEGER))

(* Scans digits, and returns a constant value token. *)

and scan_digits_loop (stream : Lexing.lexbuf) (token : Buffer.t) (base : int) (state : scan_digits_state_t) : (Cg.token * string) = (
  let c = (getc stream false) in
  if (c == getc_eof_mark) then
    (begin
      let s = (Buffer.contents token) in
      if (state == SCAN_INTEGER) then
	((convert_constant_number_integer s), s)
      else if (base == 16 && state == SCAN_FRACTION) then
	(syntaxerror ("Malformed floating constant"
		       ^ " (hexadecimal need exponent)"))
      else
	((convert_constant_number_float s), s)
    end)
  else if (c == '.') then
    (begin
      if (state == SCAN_FRACTION || state == SCAN_EXPONENT) then
	(syntaxerror ("Malformed floating constant (excess dots)"))
      else if (base == 16) then
	(begin
	  (Buffer.add_char token c) ;
	  (scan_digits_loop stream token 16 SCAN_FRACTION)
	end)
      else
	(begin
	  (Buffer.add_char token c) ;
	  (scan_digits_loop stream token 10 SCAN_FRACTION)
	end)
    end)
  else if (digit_char_p c 10) then
    (begin
      (Buffer.add_char token c) ;
      (scan_digits_loop stream token base state)
    end)
  else if ((base == 16) && (digit_char_p c 16)) then
    (begin
      (Buffer.add_char token c) ;
      (scan_digits_loop stream token base state)
    end)
  else if (c == 'e' || c == 'E') then
    (begin
      if (base != 10) then
	(syntaxerror ("Malformed floating constant (e/E for non-decimal)"))
      else if (state == SCAN_EXPONENT) then
	(syntaxerror ("Malformed floating constant (e/E in exponent)"))
      else
	(begin
	  (Buffer.add_char token c) ;
	  (scan_digits_loop_exponent stream token base state)
	end)
    end)
  else if (c == 'p' || c == 'P') then
    (begin
      if (base != 16) then
	(syntaxerror ("Malformed floating constant (p/P for non-hexadecimal)"))
      else if (state == SCAN_EXPONENT) then
	(syntaxerror ("Malformed floating constant (p/P in exponent)"))
      else
	(begin
	  (Buffer.add_char token c) ;
	  (scan_digits_loop_exponent stream token base state)
	end)
    end)
  else if (c == 'f' || c == 'F') then
    (begin
      assert (base <= 10) ;
      (Buffer.add_char token c) ;
      ignore (scan_digits_check_trailings stream) ;
      let s = (Buffer.contents token) in
      if (state == SCAN_INTEGER) then
	((convert_constant_number_float s), s)
      else if (base == 16 && state == SCAN_FRACTION) then
	(syntaxerror ("Malformed floating constant"
		       ^ " (hexadecimal need exponent)"))
      else
	((convert_constant_number_float s), s)
    end)
  else if ((c == 'l' || c == 'L') && (state != SCAN_INTEGER)) then
    (begin
      (Buffer.add_char token c) ;
      ignore (scan_digits_check_trailings stream) ;
      let s = (Buffer.contents token) in
      if (base == 16 && state == SCAN_FRACTION) then
	(syntaxerror ("Malformed floating constant"
		       ^ " (hexadecimal need exponent)"))
      else
	((convert_constant_number_float s), s)
    end)
  else if ((c == 'l' || c == 'L') && (state == SCAN_INTEGER)) then
    (begin
      (ungetc c stream) ;
      (scan_digits_spec_integer stream token 0 0)
    end)
  else if ((c == 'u' || c == 'U') && (state == SCAN_INTEGER)) then
    (begin
      (ungetc c stream) ;
      (scan_digits_spec_integer stream token 0 0)
    end)
  else if (alphanumericp c) then
    (begin
      (syntaxerror ("Non-digits in number"))
    end)
  else
    (begin
      (ungetc c stream) ;
      ignore (scan_digits_check_trailings stream) ;
      let s = (Buffer.contents token) in
      if (state == SCAN_INTEGER) then
	((convert_constant_number_integer s), s)
      else if (base == 16 && state == SCAN_FRACTION) then
	(syntaxerror ("Malformed floating constant"
		       ^ " (hexadecimal need exponent)"))
      else
	((convert_constant_number_float s), s)
    end))

and scan_digits_loop_exponent (stream : Lexing.lexbuf) (token : Buffer.t) (base : int) (state : scan_digits_state_t) : (Cg.token * string) = (
  let c1 = (getc stream false) in
  if (c1 == getc_eof_mark) then
    (syntaxerror ("Malformed floating constant (EOF in exponent)"))
  else if ((c1 == '+') || (c1 == '-')) then
    (begin
      (Buffer.add_char token c1) ;
      let c2 = (getc stream false) in
      if (c2 == getc_eof_mark) then
	(syntaxerror ("Malformed floating constant (EOF in exponent)"))
      else if (not (digit_char_p c2 10)) then
	(syntaxerror ("Floating constant exponent has no digits"))
      else
	(begin
	  (Buffer.add_char token c2) ;
	  (scan_digits_loop stream token 10 SCAN_EXPONENT)
	end)
    end)
  else if (not (digit_char_p c1 10)) then
    (syntaxerror ("Floating constant exponent has no digits"))
  else
    (begin
      (Buffer.add_char token c1) ;
      (scan_digits_loop stream token 10 SCAN_EXPONENT)
    end))

(* Checks if trailing chars are not a garbage. *)

and scan_digits_check_trailings (stream : Lexing.lexbuf) : unit = (
  let c = (getc stream false) in
  if (c == getc_eof_mark) then
    ()
  else if (alphanumericp c) then
    (syntaxerror ("Garbage at the end of a constant '" ^ (Char.escaped c) ^ "'"))
  else
    (ungetc c stream))

(* Scans a 'u' and 'l' sequence for integer. *)

and scan_digits_spec_integer (stream : Lexing.lexbuf) (token : Buffer.t) (ucount : int) (lcount : int) : (Cg.token * string) = (
  let c = (getc stream false) in
  match (c) with
    '\000' (*EOF*) ->
      let s = (Buffer.contents token) in
      ((convert_constant_number_integer s), s)
  | 'u' | 'U' ->
      (Buffer.add_char token c) ;
      if (ucount >= 1) then
	(syntaxerror ("More than one 'u' in an integer constant"))
      else
	(scan_digits_spec_integer stream token (ucount + 1) lcount)
  | 'l' | 'L' ->
      (Buffer.add_char token c) ;
      if (lcount >= 2) then
	(syntaxerror "More than two 'l' in an integer constant")
      else
	(begin
	  (scan_digits_spec_integer stream token ucount (lcount + 1))
	end)
  | _ ->
      (ungetc c stream) ;
      (scan_digits_check_trailings stream) ;
      let s = (Buffer.contents token) in
      ((convert_constant_number_integer s), s))

(* Converts a string into a constant value token.  Here constant type
   is figured out by the C89 rule.  Integers are: {int, long int, long
   long int; with each having unsigned variant}. *)

and convert_constant_number_integer (token0 : string) : Cg.token = (
  let token = (String.uppercase token0) in
  (* Count suffixes 'U' and 'L'. *)
  let ucount = ref 0 in
  let lcount = ref 0 in
  for i = 0 to ((String.length token) - 1) do (
    let c = token.[i] in
    if (c == 'U') then
      incr ucount
    else if (c == 'L') then
      incr lcount
  ) done ;
  if (!lcount > 2 || !ucount > 1) then
    (syntaxerror "Malformed integer constant") ;
  let suffix = (!ucount + !lcount) in
  (* Check base prefix '0x' or '0'. *)
  let (base, prefix) =
    if ((String.length token) >= 2 && token.[0] == '0' && token.[1] == 'X') then
      (16, 2)
    else if (token.[0] == '0') then
      (8, 1)
    else
      (10, 0)
  in
  (* Calculate a value. *)
  let value = ref (num_of_int 0) in
  for i = prefix to ((String.length token) - suffix - 1) do (
    let c = token.[i] in
    let v =
      if ('0' <= c && c <= '9') then
	((int_of_char c) - (int_of_char '0'))
      else
	((int_of_char c) - (int_of_char 'A') + 10)
    in
    value := (add_num (mult_num !value (num_of_int base)) (num_of_int v))
  ) done ;
  (* Determine its type. *)
  let tn =
    if ((!lcount <= 0)
	 && (compare_num !value Cexpr.most_positive_int) <= 0) then
      (begin
	if ((!ucount == 1) || (base != 10)) then
	  Cexpr.UNSIGNED_INTEGER
	else
	  Cexpr.INTEGER
      end)
    else if ((!lcount <= 0)
	      && (compare_num !value Cexpr.most_positive_unsigned_int) <= 0) then
      (begin
	assert ((compare_num !value Cexpr.most_positive_int) > 0) ;
	Cexpr.UNSIGNED_INTEGER
      end)
    else if ((!lcount <= 1)
	      && (compare_num !value Cexpr.most_positive_long) <= 0) then
      (begin
	if ((!ucount == 1) || (base != 10)) then
	  Cexpr.UNSIGNED_LONG_INTEGER
	else
	  Cexpr.LONG_INTEGER
      end)
    else if ((!lcount <= 1)
	      && (compare_num !value Cexpr.most_positive_unsigned_long) <= 0) then
      (begin
	assert ((compare_num !value Cexpr.most_positive_long) > 0) ;
	Cexpr.UNSIGNED_LONG_INTEGER
      end)
    else
      (begin
	(* Range check cannot be performed. *)
	if ((!ucount == 1) || (base != 10)) then
	  Cexpr.UNSIGNED_LONG_LONG_INTEGER
	else
	  Cexpr.LONG_LONG_INTEGER
      end) in
  (make_integer_constant tn !value token))

(* Converts a string into a constant value token. *)

and convert_constant_number_float (token0 : string) : Cg.token = (
  let token = (String.uppercase token0) in
  (* Count suffixes 'F' and 'L'. *)
  let fcount = ref 0 in
  let lcount = ref 0 in
  for i = 0 to ((String.length token) - 1) do (
    let c = token.[i] in
    if (c == 'F') then
      incr fcount
    else if (c == 'L') then
      incr lcount
  ) done ;
  if (!fcount + !lcount > 1) then
    (syntaxerror "Malformed floating constant (conflicting suffixes)") ;
  let suffix = (!fcount + !lcount) in
  (* Delete trailing type specifiers. *)
  let token1 = (String.sub token 0 ((String.length token) - suffix)) in
  (* Add '0' after 'E' for Ocaml reader. *)
  let token2 =
    if (token1.[(String.length token1) - 1] == 'E') then
      (token1 ^ "0")
    else
      token1
  in
  (* Add '0' before '.' for Ocaml reader. *)
  let token3 =
    if (token2.[0] == '.') then
      ("0" ^ token2)
    else
      token2
  in
  try
    (make_floating_constant
	(if (!fcount == 1) then
	  Cexpr.SINGLE_FLOAT
	else if (!lcount == 1) then
	  Cexpr.LONG_DOUBLE_FLOAT
	else
	  Cexpr.DOUBLE_FLOAT)
	(float_of_string token3) token)
  with
    Failure "float_of_string" ->
      (syntaxerror "Malformed floating constant (conversion failure)"))

let scan_number (c : char) (stream : Lexing.lexbuf) : (Cg.token * string) = (
  (scan_digits c stream))

(* ================================================================ *)

(* Operator Scanner *)

(* OPERATORS: "~" "![=]" "=[=]" "*[=]" "/[=]" "%[=]" "^[=]" "+[+,=]"
   "&[&,=]" "|[|,=]" "-[-,=,>]" "<[<,=,<=]" ">[>,=,>=]".  Note that a
   dot is handled in scan_number. *)

let scan_operator (c0 : char) (stream : Lexing.lexbuf) : (Cg.token * string) = (
  match c0 with
    '~' ->
      (Cg.Q_TLD, "~")
  | '!' | '=' | '*' | '/' | '%' | '^' ->
      (begin
	let c1 = (getc stream false) in
	if (c1 = getc_eof_mark) then
	  (begin
	    match c0 with
	      '!' -> (Cg.Q_BAN, "!")
	    | '=' -> (Cg.Q_EQL, "=")
	    | '*' -> (Cg.Q_STR, "*")
	    | '/' -> (Cg.Q_DIV, "/")
	    | '%' -> (Cg.Q_MOD, "%")
	    | '^' -> (Cg.Q_HAT, "^")
	    | _ -> (assert false)
	  end)
	else if (c1 == '=') then
	  (begin
	    if (c0 == '!') then
	      ((Cg.Q_EQUAL Cexpr.NE), "!=")
	    else if (c0 == '=') then
	      (begin
		(*NONSTANDARD*)
		let c2 = (getc stream false) in
		if (c2 == getc_eof_mark) then
		  ((Cg.Q_EQUAL Cexpr.EQ), "==")
		else if (c2 == '>') then
		  (Cg.Q_IMP, "==>")
		else
		  (begin
		    (ungetc c2 stream) ;
		    ((Cg.Q_EQUAL Cexpr.EQ), "==")
		  end)
	      end)
	    else
	      match c0 with
		'!' -> (assert false)
	      | '=' -> (assert false)
	      | '*' -> ((Cg.Q_ASSIGN Cexpr.MUL), "*=")
	      | '/' -> ((Cg.Q_ASSIGN Cexpr.DIV), "/=")
	      | '%' -> ((Cg.Q_ASSIGN Cexpr.MOD), "%=")
	      | '^' -> ((Cg.Q_ASSIGN Cexpr.XOR), "^=")
	      | _ -> (assert false)
	  end)
	else
	  (begin
	    (*NONSTANDARD*)
	    if (c0 == '=' && c1 == '>') then
	      (* This case is just "=>", but taken it as "==>". *)
	      (Cg.Q_IMP, "==>")
	    else
	      (begin
		(ungetc c1 stream) ;
		match c0 with
		  '!' -> (Cg.Q_BAN, "!")
		| '=' -> (Cg.Q_EQL, "=")
		| '*' -> (Cg.Q_STR, "*")
		| '/' -> (Cg.Q_DIV, "/")
		| '%' -> (Cg.Q_MOD, "%")
		| '^' -> (Cg.Q_HAT, "^")
		| _ -> (assert false)
	      end)
	  end)
      end)
  | '+' | '&' | '|' | '-' ->
      (begin
	let c1 = (getc stream false) in
	if (c1 = getc_eof_mark) then
	  (begin
	    match c0 with
	      '+' -> (Cg.Q_PLS, "+")
	    | '&' -> (Cg.Q_AMP, "&")
	    | '|' -> (Cg.Q_BAR, "|")
	    | '-' -> (Cg.Q_MNS, "-")
	    | _ -> (assert false)
	  end)
	else if (c0 == c1) then
	  (begin
	    match c0 with
	      '+' -> (Cg.Q_PLSPLS, "++")
	    | '&' -> (Cg.Q_AMPAMP, "&&")
	    | '|' -> (Cg.Q_BARBAR, "||")
	    | '-' -> (Cg.Q_MNSMNS, "--")
	    | _ -> (assert false)
	  end)
	else if (c1 == '=') then
	  (begin
	    match c0 with
	      '+' -> ((Cg.Q_ASSIGN Cexpr.ADD), "+=")
	    | '&' -> ((Cg.Q_ASSIGN Cexpr.AND), "&=")
	    | '|' -> ((Cg.Q_ASSIGN Cexpr.IOR), "|=")
	    | '-' -> ((Cg.Q_ASSIGN Cexpr.SUB), "-=")
	    | _ -> (assert false)
	  end)
	else if ((c0 == '-') && (c1 == '>')) then
	  (Cg.Q_ARW, "->")
	else
	  (begin
	    (ungetc c1 stream) ;
	    match c0 with
	      '+' -> (Cg.Q_PLS, "+")
	    | '&' -> (Cg.Q_AMP, "&")
	    | '|' -> (Cg.Q_BAR, "|")
	    | '-' -> (Cg.Q_MNS, "-")
	    | _ -> (assert false)
	  end)
      end)
  | '<' | '>' ->
      (begin
	let c1 = (getc stream false) in
	if (c1 = getc_eof_mark) then
	  (begin
	    if (c0 == '<') then
	      ((Cg.Q_COMPARE Cexpr.LT), "<")
	    else
	      ((Cg.Q_COMPARE Cexpr.GT), ">")
	  end)
	else if (c1 == '=') then
	  (begin
	    if (c0 == '<') then
	      (begin
		(*NONSTANDARD*)
		let c2 = (getc stream false) in
		if (c2 == getc_eof_mark) then
		  ((Cg.Q_COMPARE Cexpr.LE), "<=")
		else if (c2 == '>') then
		  (Cg.Q_IFF, "<=>")
		else
		  (begin
		    (ungetc c2 stream) ;
		    ((Cg.Q_COMPARE Cexpr.LE), "<=")
		  end)
	      end)
	    else
	      ((Cg.Q_COMPARE Cexpr.GE), ">=")
	  end)
	else if (c0 == c1) then
	  (begin
	    let c2 = (getc stream false) in
	    if (c2 == getc_eof_mark) then
	      if (c0 == '<') then
		(Cg.Q_LSH, "<<")
	      else
		(Cg.Q_RSH, ">>")
	    else if (c2 == '=') then
	      if (c0 == '<') then
		((Cg.Q_ASSIGN Cexpr.LSH), "<<=")
	      else
		((Cg.Q_ASSIGN Cexpr.RSH), ">>=")
	    else
	      (begin
		(ungetc c2 stream) ;
		if (c0 == '<') then
		  (Cg.Q_LSH, "<<")
		else
		  (Cg.Q_RSH, ">>")
	      end)
	  end)
	else
	  (begin
	    (ungetc c1 stream) ;
	    if (c0 == '<') then
	      ((Cg.Q_COMPARE Cexpr.LT), "<")
	    else
	      ((Cg.Q_COMPARE Cexpr.GT), ">")
	  end)
      end)
  | _ ->
      (assert false))

(* ================================================================ *)

(* SCANNER TOP *)

let yylex1 (lexbuf : Lexing.lexbuf) : (Cg.token * string) = (
  let stream = lexbuf in
  (*ignore_escape_p := false ;*)
  let c = (skip_whitespaces_and_comments stream false) in
  if (c == getc_eof_mark) then
    raise GETC_EOF_EXCEPTION
  else
    (begin
      match c, !spec_reader_state with
      | '!', MULTIPLE_LINES
      | '!', SINGLE_LINE ->
        (begin
            let spec_buffer = (Buffer.create spec_buffer_size) in
            (** Skip all chars in the stream until '@code' or '*/' appear,
	        the skipped chars are added into spec_buffer.

                If '@code' appears, find '@endcode' in the stream until '*/' appear.
                If @endcode appears, skip all chars in the stream until '*/' appear.
            *)
            let exist_start_code = skip_until_key_appears stream "@code" spec_buffer in
            match exist_start_code with
            | true -> (
              (* In this case, the specs starts right
	         after the keyword '@code' *)
              Cexpr.spec_startpos_ln := !Cexpr.current_line;
              Cexpr.spec_startpos_cnum_from_bol :=
	        !Cexpr.current_cnum_from_bol;
              let spec_buffer1 = (Buffer.create spec_buffer_size) in
              let exist_end_code = skip_until_key_appears stream "@endcode" spec_buffer1 in
              match exist_end_code with
              | false -> failwith "Error: @code without @endcode"
              | true -> (
                let buffer_contents = (Buffer.contents spec_buffer1) in
                let spec_string =
		    try String.sub buffer_contents 0
                                   ((String.length buffer_contents)
                                    - (String.length "@endcode") - 1)
                    with Invalid_argument _ -> assert false
                in
                skip_until_key_appears stream "*/" spec_buffer ;
                spec_reader_state := NOTHING;
                (Cg.Q_SPEC spec_string, "@") )
            )
            | false -> (
                spec_reader_state := NOTHING;
              let doxygen_comments = Buffer.contents spec_buffer in
              try let _ = ExtString.String.find doxygen_comments "@endcode" in
                  failwith "Error: @endcode without @code"
  	      with ExtString.Invalid_string -> (
	         spec_reader_state := NOTHING;
		 (Cg.Q_Ignore, ""))
            )
        end)
    | _, _ ->
        (match c with
        |'$' ->
	    if (no_dollars_in_identifier_p) then
	       (Cg.Q_DOL, "$")
	    else
	       (scan_identifier c stream)
        | '@' ->
            if not (!spec_reader_state = NOTHING) then
              (begin
                (*(Cexpr.save_current_trueline := !Cexpr.current_trueline);*)
                Cexpr.note_of_curr_spec := (Cexpr.make_linenote ());
                let spec_data = (scan_spec_string stream) in
                spec_reader_state := NOTHING;
               (Cg.Q_SPEC spec_data, "@")
              end)
            else
              (Cg.Q_ATM, "@")
        | '?' -> (Cg.Q_QES, "?")
        | ':' -> (Cg.Q_COL, ":")
        | ';' -> (Cg.Q_SEM, ";")
        | ',' -> (Cg.Q_COM, ",")
        | '(' -> (Cg.Q_LPA, "(")
        | ')' -> (Cg.Q_RPA, ")")
        | '{' -> (Cg.Q_LBC, "{")
        | '}' -> (Cg.Q_RBC, "}")
        | '[' -> (Cg.Q_LBK, "[")
        | ']' -> (Cg.Q_RBK, "]")
        | 'A' | 'B' | 'C' | 'D' | 'E' | 'F' | 'G' | 'H'
        | 'I' | 'J' | 'K' | 'L' | 'M' | 'N' | 'O' | 'P'
        | 'Q' | 'R' | 'S' | 'T' | 'U' | 'V' | 'W' | 'X'
        | 'Y' | 'Z'
        | 'a' | 'b' | 'c' | 'd' | 'e' | 'f' | 'g' | 'h'
        | 'i' | 'j' | 'k' | 'l' | 'm' | 'n' | 'o' | 'p'
        | 'q' | 'r' | 's' | 't' | 'u' | 'v' | 'w' | 'x'
        | 'y' | 'z' | '_' ->
	      if (c == 'L') then
	      (begin
	          (* Check if it is a char/string constant. *)
	          let c1 = (getc stream false) in
	          if (c1 == getc_eof_mark) then
		            raise GETC_EOF_EXCEPTION
	          else
		            match c1 with
		             '\'' ->
		              (scan_character_constant stream true)
		            | '\"' ->
		              (scan_string_constant c stream true)
		            | _ ->
			      (ungetc c1 stream) ;
		              (scan_identifier 'L' stream)
	      end)
	      else
	          (scan_identifier c stream)
        | '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9' | '.' ->
	  (scan_number c stream)
        | '\'' ->
	  (scan_character_constant stream false)
        | '\"' ->
	  (scan_string_constant c stream false)
        | '+' | '-' | '&' | '|' | '<' | '>' | '*' | '/'
        | '%' | '^' | '!' | '=' | '~' ->
	  (scan_operator c stream)
        | '\\' ->
 	  (scan_identifier c stream)
        | _ ->
	  (syntaxerror ("Illegal character: " ^ (Char.escaped c)))
        )
    end))

let rec yylex (lexbuf : Lexing.lexbuf) : Cg.token = (
  let (x, s) = (yylex1 lexbuf) in
  if (x == Cg.Q_Ignore) then
    (yylex lexbuf)
  else
    (begin
      yylval := s ;
      if (!yyTRACE) then
	(prerr_endline (";; yylex: " ^ (token_to_string x)
			 ^ " (" ^ !Cexpr.current_file ^ ":"
			 ^ (string_of_int !Cexpr.current_line)^ "/"
			 ^ (string_of_int !Cexpr.current_trueline)
			 ^ ").")) ;
      x
    end))

let _ = yylexref := yylex

(* ================================================================ *)

(* PARSER EXAMPLE *)

let cparse (f : string) : unit = (
  try
    let s = (open_in f) in
    let lexbuf = (Lexing.from_channel s (*stdin*)) in
    (init_scanner lexbuf) ;
    Cexpr.is_in_cparser := true;
    while (true) do
      let result = (Cg.program yylex lexbuf) in
      (print_string result) ;
      (print_newline ()) ;
      (flush stdout)
    done
  with
    GETC_EOF_EXCEPTION ->
      Cexpr.is_in_cparser := false;)

(*let _ = (cparse ())*)
