(* parser/csyntax.ml 2009-01-07 *)
(* Copyright (C) 2007-2009 Yonezawa Lab, The University of Tokyo *)

(* Simple Syntaxer (Converter to a goto program) *)

(* MEMO: An assignment through a reference at calls is only handled
   when address-of is taken at an argument.  Thus, only "f(&v)" is
   handled. *)

let bbTRACE_dump_parsed = ref false
let bbTRACE_dump_flatten = ref false
let bbTRACE_dump_breakup = ref false
let bbTRACE_dump_collapse = ref false
let bbTRACE_dump_variables = ref false
let bbTRACE_dump_ssa_binding = ref false
let bbTRACE_dump_ssa_flow = ref false
let bbTRACE_dump_ssa_equalset = ref false

let debug = Cexpr.debug
let info = Cexpr.info
let warning = Cexpr.warning

let xTRACE (s : string) = (if (true) then (debug (";; BBT " ^ s)))
let xDEBUG (s : string) = (if (true) then (debug (";; BBT " ^ s)))

open Num
open Cexpr

(* Disallows equality. *)

let (=) = (Misc.(=))
let (<>) = (Misc.(<>))

let warning_message (message : string) (memo : string) : unit = (
  (warning (message ^ " " ^ memo)))

let string_equal = Misc.string_equal
let string_list_equal = Misc.string_list_equal

module IntSet = Set.Make (
  struct
  type t = int
  let compare x y = (compare x y)
  end)

module StringSet = Form.StringSet

module String2Set = Set.Make (
  struct
  type t = (string * string)
  let compare x y = (compare x y)
  end)

module StringInt2Set = Set.Make (
  struct
  type t = (string * int * int)
  let compare x y = (compare x y)
  end)

type vartag_set_t = StringInt2Set.t

type expr_t = Cexpr.expr_t
type stmt_t = Cexpr.stmt_t
type typenode_t = Cexpr.typenode_t
type varid_t = Cexpr.varid_t

(* Basic Block.  {blk_phis} are phi-functions of SSA (in {BBJUNC}
   nodes).  They can be null when there is no variables.  {blk_phifix}
   are fixing assignments of phi-functions, where outputs of
   phi-functions are once assigned to temporary variables to avoid
   recurrence of the same variables.  {blk_inputs} are input
   variables, and they are imprecise in statement blocks ({BBSTMT}).
   {blk_outputs} are output variables.  Fields {blk_rdvars},
   {blk_livevars} and {blk_phivars} are temporarily used during
   variable version binding and their information is not correct at
   the end.  {blk_rdvars} holds live variables referenced by this and
   following blocks.  {blk_livevars} holds live (versioned) variables
   before removing redundant variables.  NOTE: In early stages,
   {blk_incoming_labels} are empty. *)

type block_type_t = BBHEAD | BBTAIL | BBSTMT | BBCALL | BBJUNC | BBCOND | BBASME | BBASSIGN

type stmt_block_t = {
  blk_type : block_type_t ;
  blk_label : string ;
(*  blk_memo : string ;*)
  blk_memo : stmt_t list ;
  blk_outgoing_labels : string list ;
  blk_incoming_labels : string list ;
  blk_stmts : stmt_t list ;
  blk_phis : stmt_t list ;
  blk_phifix : stmt_t list ;
  blk_callfix : stmt_t list ;
  blk_inputs : StringInt2Set.t ;
  blk_outputs : StringInt2Set.t ;
  (* Below are temporary. *)
  mutable blk_rdvars : StringInt2Set.t ;
  mutable blk_wrvars : StringInt2Set.t ;
  blk_livevars : (varid_t, expr_t) Hashtbl.t ;
  blk_phivars : (varid_t, expr_t) Hashtbl.t ;
  mutable blk_info_rdvars : (string * int * int) list ;
  mutable blk_info_wrvars : (string * int * int) list
}

type cdefinition_t =
    CFUNDEF of cfundef_t
    | CVARDEF of cvardef_t
    | COTHERDEF of stmt_t

and cfundef_t = {
  cfdef_declaration : Cexpr.declaration_t ;
  cfdef_blocks : (string, stmt_block_t) Hashtbl.t
}

and cvardef_t = {
  cvdef_declaration : Cexpr.declaration_t
}

(* Pseudo labels for entry/return blocks, and an outgoing label from a
   return block. *)

let pseudo_entry_label = "#Lentry"
let pseudo_return_label = "#Lreturn"
let pseudo_elsewhere_label = "#Lelsewhere"

(* Flag to let insert a goto after a call to such as "exit". *)

let insert_goto_after_noreturn_call_p = true

(* Name mapping for static variables in file scope. *)

let file_scope_static_variable_name_map : (varid_t, varid_t) Hashtbl.t ref =
  ref (Hashtbl.create 10)

let variable_list_to_set (l : (string * int * int) list) : StringInt2Set.t = (
  (List.fold_left
      (fun s v -> (StringInt2Set.add v s))
      (StringInt2Set.empty) l))

let variable_set_to_list (s : StringInt2Set.t) : (string * int * int) list = (
  (StringInt2Set.elements s))

let dump_memo (memo : Cexpr.stmt_t list): string =
  match memo with
  | [] -> ""
  | h::tail -> (begin
  	(Cexpr.stmt_to_string h)
      ^ (List.fold_left (fun s m -> s ^ "; " ^ (Cexpr.stmt_to_string m)) "" tail)
    end)

let dump_stmt_block (b : stmt_block_t) : unit = (
  let block_type_name (bb : block_type_t) : string = (
    match bb with
      BBHEAD -> "BB.HEAD" | BBTAIL -> "BB.TAIL" | BBSTMT -> "BB.STMT" | BBASME -> "BB.ASSUME"
    | BBCALL -> "BB.CALL" | BBJUNC -> "BB.JUNC" | BBCOND -> "BB.COND" | BBASSIGN -> "BB.ASSIGN") in
  let variable_set_to_string (s : StringInt2Set.t) : string = (
    (String.concat ";"
	(List.map Cexpr.vartag_to_string (variable_set_to_list s)))) in
  let live_variables_to_string (tbl : ((string * int), expr_t) Hashtbl.t) : string = (
    (String.concat ";"
	(List.map Cexpr.expr_to_string (Misc.hashtbl_values tbl)))) in
  match b with
    {
      blk_type = ((BBHEAD|BBTAIL|BBSTMT|BBASME|BBASSIGN) as bbtype) ;
      blk_label = label ;
      blk_memo = memo ;
      blk_outgoing_labels = [outgoing] ;
      blk_incoming_labels = incomings ;
      blk_stmts = stmts ;
      blk_phis = phis ;
      blk_phifix = phifix ;
      blk_callfix = callfix ;
      blk_inputs = ivars ;
      blk_outputs = ovars ;
      blk_rdvars = rvars ;
      blk_wrvars = wvars ;
      blk_livevars = livevars
    } ->
      assert (phis == []) ;
      assert ((List.length incomings) <= 1) ;
      let bbname = (block_type_name bbtype) in
      (print_endline (bbname ^ " {" ^ "(label=" ^ label
		       ^ "|outgoing=" ^ outgoing
		       ^ "|incoming=" ^ (String.concat ";" incomings) ^ ")")) ;
      (print_endline ("|memo=" ^ (dump_memo memo))) ;
      (List.iter
	  (fun s ->
	    (print_endline ("|stmt=" ^ (Cexpr.stmt_to_string_body 1 s))))
	  stmts) ;
      (print_endline ("|ivars=" ^ (variable_set_to_string ivars))) ;
      (print_endline ("|ovars=" ^ (variable_set_to_string ovars))) ;
      xDEBUG ("rdvars=" ^ (variable_set_to_string rvars)) ;
      xDEBUG ("wrvars=" ^ (variable_set_to_string wvars)) ;
      xDEBUG ("livevars=" ^ (live_variables_to_string livevars)) ;
      (print_endline ("}"))
  | {
      blk_type = BBCALL ;
      blk_label = label ;
      blk_memo = memo ;
      blk_outgoing_labels = [outgoings] ;
      blk_incoming_labels = incomings ;
      blk_stmts = stmts ;
      blk_phis = phis ;
      blk_phifix = phifix ;
      blk_callfix = callfix ;
      blk_inputs = ivars ;
      blk_outputs = ovars ;
      blk_rdvars = rvars ;
      blk_wrvars = wvars ;
      blk_livevars = livevars
    } ->
      assert (phis == []) ;
      assert ((List.length incomings) <= 1) ;
      (print_endline ("BB.CALL{" ^ "(label=" ^ label
		       ^ "|outgoing=" ^ outgoings
		       ^ "|incoming=" ^ (String.concat ";" incomings)
		       ^ ")")) ;
      (print_endline ("|memo=" ^ (dump_memo memo))) ;
      (List.iter
	  (fun s ->
	    (print_endline ("|stmt=" ^ (Cexpr.stmt_to_string_body 1 s))))
	  stmts) ;
      (List.iter
	  (fun s ->
	    (print_endline ("|callfix=" ^ (Cexpr.stmt_to_string_body 1 s))))
	  callfix) ;
      (print_endline ("|ivars=" ^ (variable_set_to_string ivars))) ;
      (print_endline ("|ovars=" ^ (variable_set_to_string ovars))) ;
      xDEBUG ("rdvars=" ^ (variable_set_to_string rvars)) ;
      xDEBUG ("wrvars=" ^ (variable_set_to_string wvars)) ;
      xDEBUG ("livevars=" ^ (live_variables_to_string livevars)) ;
      (print_endline ("}"))
  | {
      blk_type = BBJUNC ;
      blk_label = label ;
      blk_memo = memo ;
      blk_outgoing_labels = [outgoings] ;
      blk_incoming_labels = incomings ;
      blk_stmts = stmts ;
      blk_phis = phis ;
      blk_phifix = phifix ;
      blk_callfix = callfix ;
      blk_inputs = ivars ;
      blk_outputs = ovars ;
      blk_rdvars = rvars ;
      blk_wrvars = wvars ;
      blk_livevars = livevars
    } ->
      assert (stmts == []) ;
      assert ((List.length incomings) > 1) ;
      (print_endline ("BB.JUNC{" ^ "(label=" ^ label
		       ^ "|outgoing=" ^ outgoings
		       ^ "|incoming=" ^ (String.concat ";" incomings)
		       ^ ")")) ;
      (print_endline ("|memo=" ^ (dump_memo memo))) ;
      (List.iter
	  (fun s ->
	    (print_endline ("|phi=" ^ (Cexpr.stmt_to_string_body 1 s))))
	  phis) ;
      (List.iter
	  (fun s ->
	    (print_endline ("|phifix=" ^ (Cexpr.stmt_to_string_body 1 s))))
	  phifix) ;
      (print_endline ("|ivars=" ^ (variable_set_to_string ivars))) ;
      (print_endline ("|ovars=" ^ (variable_set_to_string ovars))) ;
      xDEBUG ("rdvars=" ^ (variable_set_to_string rvars)) ;
      xDEBUG ("wrvars=" ^ (variable_set_to_string wvars)) ;
      xDEBUG ("livevars=" ^ (live_variables_to_string livevars)) ;
      (print_endline ("}"))
  | {
      blk_type = BBCOND ;
      blk_label = label ;
      blk_memo = memo ;
      blk_outgoing_labels = [then_out_label ; else_out_label] ;
      blk_incoming_labels = incomings ;
      blk_stmts = [stmt] ;
      blk_phis = phis ;
      blk_phifix = phifix ;
      blk_callfix = callfix ;
      blk_inputs = ivars ;
      blk_outputs = ovars ;
      blk_rdvars = rvars ;
      blk_wrvars = wvars ;
      blk_livevars = livevars
    } ->
      assert (phis == []) ;
      assert ((List.length incomings) <= 1) ;
      (print_endline ("BB.COND{" ^ "(label=" ^ label
		       ^ "|then=" ^ then_out_label
		       ^ "|else=" ^ else_out_label
		       ^ "|incoming=" ^ (String.concat ";" incomings)
		       ^ ")")) ;
      (print_endline ("|memo=" ^ (dump_memo memo))) ;
      (print_endline ("|" ^ (Cexpr.stmt_to_string_body 1 stmt))) ;
      (print_endline ("|ivars=" ^ (variable_set_to_string ivars))) ;
      (print_endline ("|ovars=" ^ (variable_set_to_string ovars))) ;
      xDEBUG ("rdvars=" ^ (variable_set_to_string rvars)) ;
      xDEBUG ("wrvars=" ^ (variable_set_to_string wvars)) ;
      xDEBUG ("livevars=" ^ (live_variables_to_string livevars)) ;
      (print_endline ("}"))
  | _ -> (assert false))

let rec dump_stmt_blocks (ctx : cdefinition_t) : unit = (
  match ctx with
    (CFUNDEF d) ->
      let bbset = d.cfdef_blocks in
      (print_endline ("FUNDEF="
		       ^ (Cexpr.decl_to_string_body 1 d.cfdef_declaration))) ;
      (print_endline ("BLOCKS=" ^ (string_of_int (Hashtbl.length bbset)))) ;
      (print_endline "BLOCKDUMP=...") ;
      (dump_stmt_blocks_loop bbset [pseudo_entry_label] StringSet.empty)
  | (CVARDEF d) ->
      (print_endline ("VARDEF="
		       ^ (Cexpr.decl_to_string_body 1 d.cvdef_declaration)))
  | (COTHERDEF s) ->
      (print_endline ("OTHERDEF=" ^ (Cexpr.stmt_to_string s))))

and dump_stmt_blocks_loop (bbset : (string, stmt_block_t) Hashtbl.t) (labels : string list) (visited : StringSet.t) : unit = (
  match labels with
    [] ->
      ()
  | (label :: tl) ->
      if ((Hashtbl.length bbset) == 0) then
	()
      else if (label == pseudo_elsewhere_label) then
	(dump_stmt_blocks_loop bbset tl visited)
      else if (StringSet.mem label visited) then
	(dump_stmt_blocks_loop bbset tl visited)
      else if (not (Hashtbl.mem bbset label)) then
	(assert false)
      else
	(begin
	  let b = (Hashtbl.find bbset label) in
	  (dump_stmt_block b) ;
	  let newvisited = (StringSet.add label visited) in
	  (dump_stmt_blocks_loop bbset
	      (b.blk_outgoing_labels @ tl) newvisited)
	end))

let variable_expr_to_name (v : expr_t) : (string * int * int) = (
  match v with
    (VREF (tn, id, level, sequence)) ->
      (id, level, sequence)
  | _ -> (assert false))

let variable_set_of_assigned (l : stmt_t list) : StringInt2Set.t = (
  (variable_list_to_set
      (Misc.list_filtering_map
	  (fun s ->
	    match s with
	      (ASGN (v, _)) ->
		(Some (variable_expr_to_name v))
	    | _ ->
		None)
	  l)))

let vartag_set_to_string (s : StringInt2Set.t) : string = (
  ("{"
    ^ (String.concat ","
	  (List.map Cexpr.vartag_to_string (StringInt2Set.elements s)))
    ^ "}"))

(* ================================================================ *)

(* FLATTENING STATEMENTS *)

(* MEMO: Scopes of variables (appearing in a program) are resolved in
   the parser.  And, thus no need to check scopes in the syntaxer. *)

(* NOTE: Temporaries introduced for binary conditionals (?:) are NOT
   singly assigned, and they are assigned with a scope level for
   parameters to let them involved in SSA. *)

(* Pseudo variable names. *)

let pseudo_result_variable = Cexpr.pseudo_result_variable
let pseudo_varargs = Cexpr.pseudo_varargs

let result_value_reference (tn : typenode_t) : expr_t = (
  (VREF (tn, pseudo_result_variable, 1, 0)))

let pseudo_nondet_value (tn : typenode_t) : expr_t = (
  (PSEUDO_NONDET_VALUE (tn, "integer")))

let result_value_reference_p (e : expr_t) : bool = (
  match e with
    (VREF (_, id, 1, _)) when (string_equal id pseudo_result_variable) ->
      true
  | _ ->
      false)

let assignment_stmt (tn : typenode_t) (lval : expr_t) (rval : expr_t) : stmt_t = (
  (ASGN (lval, rval)))

(* Required value in flattening expressions: no-value {NOVAL}, lvalue
   {LVAL}, arithmetic-value {AVAL}, and conditional boolean-value
   {BVAL}.  Conditional expressions are treated separately to make
   predicates properly (not zero/one integer). *)

type need_value_t = NOVAL | LVAL | AVAL | BVAL

(* Syntaxing Context: {ctx_file_hash_name} and {ctx_function_name} are
   used to replace static variable names.  {ctx_referred_labels} holds
   the labels whose value is taken, along with variable names which
   are created to represent their values. *)

type flatten_context_t = {
  ctx_file_hash_name : string ;
  ctx_function_name : string ;
  ctx_return_type : typenode_t ;
  ctx_return_label : string ;
  ctx_break_label : string option ;
  ctx_continue_label : string option ;
  ctx_static_variable_name_map : (varid_t, varid_t) Hashtbl.t ;
  mutable ctx_case_label_map : ((expr_t * expr_t) * string) list ;
  mutable ctx_static_variables : stmt_t list ;
  mutable ctx_referred_labels : (string * string) list
}

(* Recognizes {exit} and generates a goto for the return label. *)

let extra_goto_after_noreturn_call (ctx : flatten_context_t) (e0 : expr_t) : stmt_t list = (
  if (insert_goto_after_noreturn_call_p) then
    (begin
      match (Cexpr.ignore_cast e0) with
	(VREF (vtn, id, level, _)) ->
	  if (level == global_scope_level
	       && ((Cattr.attribute_check_noreturn vtn)
		    || (string_equal id "exit"))) then
	    (begin
	      let dummylabel = (Cexpr.make_new_label ()) in
	      let tn = ctx.ctx_return_type in
	      if (tn != VOID_TYPE) then
		[(assignment_stmt tn (result_value_reference tn)
		     (pseudo_nondet_value tn)) ;
		 (GOTO (ctx.ctx_return_label)) ;
		 (LABEL (NAME_LABEL dummylabel))]
	      else
		[(GOTO (ctx.ctx_return_label)) ;
		 (LABEL (NAME_LABEL dummylabel))]
	    end)
	  else
	    []
      | _ ->
	  []
    end)
  else
    [])

(* Generates (pseudoly) assignments as an asm statement (for flow
   analysis).  gccasm := asm(string [ : operand... [ : operand... [ :
   string... ]]]).  operand := [ id ] string (expr). *)

let rec handle_gcc_asm_statement (s : stmt_t) : stmt_t list = (
  match s with
    (ASM e) ->
      (begin
	match e with
	  (STRING (_, asm, _)) ->
	    (* No operands; ignored *)
	    []
	| (ASMSEQUENCE [asm ; opr0]) ->
	    (handle_gcc_asm_operand_list opr0)
	| (ASMSEQUENCE [asm ; opr0 ; opr1]) ->
	    ((handle_gcc_asm_operand_list opr0)
	      @ (handle_gcc_asm_operand_list opr1))
	| (ASMSEQUENCE [asm ; opr0 ; opr1 ; strings]) ->
	    ((handle_gcc_asm_operand_list opr0)
	      @ (handle_gcc_asm_operand_list opr1))
	| _ -> (syntaxerror "Unsupported asm syntax")
      end)
  | _ -> (assert false))

and handle_gcc_asm_operand_list (opr : expr_t) : stmt_t list = (
  match (opr) with
    (ASMSEQUENCE l) ->
      (List.flatten (List.map handle_gcc_asm_operand l))
  | _ -> (syntaxerror "Unsupported asm syntax"))

and handle_gcc_asm_operand (opr : expr_t) : stmt_t list = (
  match opr with
    (ASMSEQUENCE [registerconstraint ; expr])
  | (ASMSEQUENCE [_ ; registerconstraint ; expr]) ->
      (begin
	match registerconstraint with
	  (STRING (b, s, _)) ->
	    if (String.contains s '=') then
	      (begin
		let (lvaluep, tn) = (check_lvalue_p expr true) in
		if (lvaluep) then
		  [(assignment_stmt tn expr (pseudo_nondet_value tn))]
		else
		  (syntaxerror "Non-lvalue in asm statement")
	      end)
	    else
	      []
	| _ -> (syntaxerror "Unsupported asm syntax")
      end)
  | _ -> (syntaxerror "Unsupported asm syntax"))

let make_logical_value (e : expr_t) : expr_t = (
  if (Cexpr.logical_value_p e) then
    e
  else
    (OPER (INTEGER, NE, e, Cexpr.constant_zero, NIL)))

let make_zero_one_value_of_conditional (e : expr_t) : expr_t = (
  if (Cexpr.logical_value_p e) then
    (OPER (INTEGER, COND, e, Cexpr.constant_one, Cexpr.constant_zero))
  else
    e)

let simple_transform_expression (e : expr_t) : expr_t = (
  match e with
    (OPER (tn, ((OROR|ANDAND) as op), e0, e1, _)) ->
      (begin
	let zero = Cexpr.constant_zero in
	let one = Cexpr.constant_one in
	if (op == ANDAND) then
	  (OPER (tn, COND, e0, (OPER (tn, COND, e1, one, zero)), zero))
	else
	  (OPER (tn, COND, e0, one, (OPER (tn, COND, e1, one, zero))))
      end)
  | _ -> (assert false))

(* BinhLT - 20091002 *)
let null_note = NOTE(LINE(true, "", 0, 0))
let current_note_while_flatten = ref null_note
let insert_checking_array_index = ref false

(* Flattens a toplevel declaration.  Flattening converts statements
   into forms: (ASGN e e), (ASGN e (CALL e)), (STMT e), (STMT (CALL
   e)), (IF (e, (GOTO l), (GOTO l))), (GOTO l), (LABEL l), where
   expressions are effect-free.  It also moves static local variables
   outside.  It returns a list of statements and a list of referred
   labels.  Referred labels are used not to remove a label whose value
   is taken. *)

let rec flatten_syntax (hashname : string) (s : stmt_t) : ((stmt_t list) * ((string * string) list)) = (
  match s with
    (DEFUN (d, (BEGIN l))) ->
      (begin
	let returnlabel = (Cexpr.make_new_label ()) in
	let tn = (Cexpr.type_of_return_value d.Cexpr.decl_tn) in
	let ctx : flatten_context_t = {
	  ctx_file_hash_name = hashname ;
	  ctx_function_name = (fst d.Cexpr.decl_id) ;
	  ctx_return_type = tn ;
	  ctx_return_label = returnlabel ;
	  ctx_break_label = None ;
	  ctx_continue_label = None ;
	  ctx_static_variable_name_map =
	    (Hashtbl.copy !file_scope_static_variable_name_map) ;
	  ctx_case_label_map = [] ;
	  ctx_static_variables = [] ;
	  ctx_referred_labels = []
	} in
	let body =
	  ([(STMT PSEUDO_PROLOGUE)]
	    @ (List.flatten (List.map (flatten_stmt_body ctx) l))
	    @ [(LABEL (NAME_LABEL returnlabel))]
	    @ [(STMT PSEUDO_EPILOGUE)]) in
        let d1 = (rename_file_scope_statics hashname d) in
	let defs = (ctx.ctx_static_variables @ [(DEFUN (d1, (BEGIN body)))]) in
	(defs, ctx.ctx_referred_labels)
      end)
  | (DEFUN (d, _)) -> (assert false)
  | (SPEC _) | (ASSUME _) | (RESET _) -> (assert false)
(*  | (DEFVAR d) ->
      let d1 = (rename_file_scope_statics hashname d) in
      [(DEFVAR d1)]*)
  (* 20090710. Binh deals with global arrays --START *)
  | (DEFVAR d) ->(begin
      let d1 = (rename_file_scope_statics hashname d) in
      let inits = d1.decl_inits in
      let tn = d1.decl_tn in
      let name, level = d1.decl_id in
      let params = d1.decl_params in
      let s_define = d1.decl_struct_defined_p in
      let scls = d1.decl_scls in
      let var = VREF(tn, name, level, 0) in
	     let ctx : flatten_context_t = {
	         ctx_file_hash_name = hashname ;
	         ctx_function_name = (fst d.Cexpr.decl_id) ;
	         ctx_return_type = tn ;
	         ctx_return_label = "";
	         ctx_break_label = None ;
	         ctx_continue_label = None ;
	         ctx_static_variable_name_map =
	         (Hashtbl.copy !file_scope_static_variable_name_map) ;
	         ctx_case_label_map = [] ;
	         ctx_static_variables = [] ;
	         ctx_referred_labels = []
	     } in
      let alloc =(
          if (Cexpr.array_type_p tn) then
            (begin
               [(assignment_stmt tn var (PSEUDO_MEMORY (INTEGER, tn)))]
            end)
          else
            []
      )in
      let new_field_inits = Ccommon.init_fields_of_declaration d in
      let new_flatten_field_inits = (
        List.flatten (
            List.map (fun e ->
                (snd (flatten_expr_body ctx NOVAL false e))
            ) new_field_inits)
      ) in
      let new_inits = (
          match inits with
          | [PROGN(t, is)] -> [PROGN(t, (alloc @ is @ (new_flatten_field_inits)))]
                              
          | _ -> assert false
      )in
      let d1 = {
          decl_id = (name, level);
          decl_tn = tn;
          decl_scls= scls;
          decl_inits = new_inits;
          decl_params= params;
          decl_struct_defined_p= s_define;
          decl_specs = d1.decl_specs;
          decl_location = d.Cexpr.decl_location;
      }in
      ([(DEFVAR d1)], [])
   end) (* 20090710. Binh deals with global arrays -- END *)
  | (STMT e) -> (assert false)
  | (ASGN (e0, e1)) -> (assert false)
  | (CALLASGN (e0, l, _)) -> (assert false)
  | (IF (e, l0, l1)) -> (assert false)
  | (BEGIN l) -> (assert false)
  | (SWITCH (e, l)) -> (assert false)
  | (DOWHILE (b, e, l)) -> (assert false)
  | (FOR (i, e1, e2, l)) -> (assert false)
  | (BREAK) -> (assert false)
  | (CONTINUE) -> (assert false)
  | (LABEL l) -> (assert false)
  | (GOTO s) -> (assert false)
  | (RETURN None) -> (assert false)
  | (RETURN (Some e)) -> (assert false)
  | (ANNOTATION _) ->
      ([s], [])
  | (ASM e) ->
      (warning_message
	  ("IGNORE ASM: '" ^ (expr_to_string e) ^ "'") "") ;
      ([(NOTE (LINE(false, "/*ASM IGNORED*/", 0, 0)))], [])
  | (NOTE _) -> (assert false))

and rename_file_scope_statics (hashname : string) (decl : Cexpr.declaration_t) : Cexpr.declaration_t = (
  match decl with
    {
      decl_id = ((id, level) : varid_t) ;
      decl_tn = (_ : typenode_t) ;
      decl_scls = (scls : storage_class_t) ;
      decl_inits = (inits : expr_t list) ;
      decl_params = (_ : string list) ;
      decl_struct_defined_p = (_ : bool)
    } ->
      (begin
	(* Make a dummy context. *)
	let ctx0 : flatten_context_t = {
	  ctx_file_hash_name = hashname ;
	  ctx_function_name = "(noname)" ;
	  ctx_return_type = Cexpr.NOTYPE ;
	  ctx_return_label = pseudo_elsewhere_label ;
	  ctx_break_label = None ;
	  ctx_continue_label = None ;
	  ctx_static_variable_name_map =
	    (Hashtbl.copy !file_scope_static_variable_name_map) ;
	  ctx_case_label_map = [] ;
	  ctx_static_variables = [] ;
	  ctx_referred_labels = []
	} in
	match scls with
	  NILSCLS | AUTO | REGISTER | EXTERN | TYPEDEF ->
	    let ctx = ctx0 in
      let old_check = !insert_checking_array_index in
      insert_checking_array_index := false;
	    let newinits =
	      (PROGN
		  (VOID_TYPE,
		  (List.flatten
		      (List.map
			  (fun e ->
			    (*if (not (Cexpr.constant_expression_p e)) then
			      (syntaxerror "Non-constants in initializer");*)
			    (* Discard {fst} of flattening as for effect. *)
			    (snd (flatten_expr_body ctx NOVAL false e)))
			  inits)))) in
      insert_checking_array_index := old_check;
	    {
	      decl with
		decl_inits = [newinits]
	    }
	| STATIC ->
	    let newid = (hashname ^ "%" ^ id) in
	    (Hashtbl.replace !file_scope_static_variable_name_map
		(id, level) (newid, Cexpr.global_scope_level)) ;
	    (* Make a dummy context. *)
	    let ctx = {
	      ctx0 with
		ctx_static_variable_name_map =
		(Hashtbl.copy !file_scope_static_variable_name_map)
	    } in
      let old_check = !insert_checking_array_index in
      insert_checking_array_index := false;
	    let newinits =
	      (PROGN
		  (VOID_TYPE,
		  (List.flatten
		      (List.map
			  (fun e ->
			    (*if (not (Cexpr.constant_expression_p e)) then
			      (syntaxerror "Non-constants in initializer");*)
			    (* Discard {fst} of flattening as for effect. *)
			    (snd (flatten_expr_body ctx NOVAL false e)))
			  inits)))) in
      insert_checking_array_index := old_check;
	    {
	      decl with
		decl_id = (newid, Cexpr.global_scope_level) ;
		decl_inits = [newinits]
	    }
      end))

(* Flattens a statement, replacing {BEGIN}, {SWITCH}, {DOWHILE},
   {FOR}, {BREAK}, {RETURN}, and {CONTINUE} with {IF} and {GOTO}. *)

and flatten_stmt_body (ctx : flatten_context_t) (s : stmt_t) : stmt_t list = (
  match s with
    (DEFUN (d, l)) -> (assert false)
  | (SPEC _) -> (assert false)
  | (DEFVAR ({
      decl_id = ((id, level) : varid_t) ;
      decl_tn = (tn : typenode_t) ;
      decl_scls = (scls : storage_class_t) ;
      decl_inits = (inits : expr_t list) ;
      decl_params = (_ : string list) ;
      decl_struct_defined_p = (_ : bool)
    } as d)) ->
      (begin
	match scls with
	  NILSCLS | AUTO | REGISTER | EXTERN | TYPEDEF ->
	    let v = (VREF (tn, id, level, 0)) in
            (flatten_auto_decl ctx tn v inits)
            @ (create_and_flatten_field_inits ctx d)
	| STATIC ->
	    let newid = (ctx.ctx_file_hash_name ^ "%"
			  ^ ctx.ctx_function_name ^ "%"
			  ^ (string_of_int level) ^ "%" ^ id) in
	    (Hashtbl.replace ctx.ctx_static_variable_name_map
		(id, level) (newid, Cexpr.global_scope_level)) ;
      let old_check = !insert_checking_array_index in
      insert_checking_array_index := false;
	    let newinits =
	      (PROGN
		  (VOID_TYPE,
		  (List.flatten
		      (List.map
			  (fun e ->
			    (* Discard {fst} of flattening as for effect. *)
			    (snd (flatten_expr_body ctx NOVAL false e)))
			  inits)))) in
      insert_checking_array_index := old_check;
	    let defvar = (DEFVAR {
	      d with
		decl_id = (newid, Cexpr.global_scope_level) ;
		decl_inits = [newinits]
	    }) in
	    ctx.ctx_static_variables <- (ctx.ctx_static_variables @ [defvar]) ;
	    (* Remove a static variable declaration from here. *)
	    []
      end)
  | (STMT e) ->
      (* Discard {fst} of flattening as for effect. *)
(*      Gprint.eprint (s, !current_note_while_flatten); Scommon.pause();*)
        (snd (flatten_expr_body ctx NOVAL false e))
  | (ASGN _) -> (assert false)
  | (CALLASGN _) -> (assert false)
  | (IF (e0, l1, [])) ->
      (begin
	let thenlabel = (Cexpr.make_new_label ()) in
	let nextlabel = (Cexpr.make_new_label ()) in
	let (_, es) =
	  (flatten_expr_condition ctx (thenlabel, nextlabel) false e0) in
	(es
	  @ [(LABEL (NAME_LABEL thenlabel))]
	  @ (List.flatten (List.map (flatten_stmt_body ctx) l1))
	  @ [(GOTO nextlabel)]
	  @ [(LABEL (NAME_LABEL nextlabel))])
      end)
  | (IF (e0, l1, l2)) ->
      (begin
	let thenlabel = (Cexpr.make_new_label ()) in
	let elselabel = (Cexpr.make_new_label ()) in
	let nextlabel = (Cexpr.make_new_label ()) in
	let (_, es) =
	  (flatten_expr_condition ctx (thenlabel, elselabel) false e0) in
	(es
	  @ [(LABEL (NAME_LABEL thenlabel))]
	  @ (List.flatten (List.map (flatten_stmt_body ctx) l1))
	  @ [(GOTO nextlabel)]
	  @ [(LABEL (NAME_LABEL elselabel))]
	  @ (List.flatten (List.map (flatten_stmt_body ctx) l2))
	  @ [(GOTO nextlabel)]
	  @ [(LABEL (NAME_LABEL nextlabel))])
      end)
  | (BEGIN l) ->
      (List.flatten (List.map (flatten_stmt_body ctx) l))
  | (SWITCH (e1, l1)) ->
      (begin
	let leavelabel = (Cexpr.make_new_label ()) in
	let newctx = {
	  ctx with
	    ctx_break_label = Some leavelabel ;
	    ctx_case_label_map = []
	} in
	let (casevar, s1) = (flatten_expr_body ctx AVAL true e1) in
	let body = (List.flatten (List.map (flatten_stmt_body newctx) l1)) in
	let gotodefault =
	  if (List.mem_assoc (NIL, NIL) newctx.ctx_case_label_map) then
	    [(GOTO (List.assoc (NIL, NIL) newctx.ctx_case_label_map))]
	  else
	    [(GOTO leavelabel)] in
	([(NOTE (LINE(false, "/*memo(switch)*/", 0, 0)))]
	  @ s1
	  @ (List.flatten
		(List.map
		    (fun ((v0, v1), l) ->
		      if (v0 == NIL && v1 == NIL) then
			[]
		      else if (v1 == NIL) then
			(begin
			  let nlabel = (Cexpr.make_new_label ()) in
			  [(IF ((OPER (INTEGER, EQ, casevar, v0, NIL)),
			       [(GOTO l)], [(GOTO nlabel)])) ;
			   (LABEL (NAME_LABEL nlabel))]
			end)
		      else
			(begin
			  let label0 = (Cexpr.make_new_label ()) in
			  let label1 = (Cexpr.make_new_label ()) in
			  [(IF ((OPER (INTEGER, GE, casevar, v0, NIL)),
			       [(GOTO label0)], [(GOTO label1)])) ;
			   (LABEL (NAME_LABEL label0)) ;
			   (IF ((OPER (INTEGER, LE, casevar, v1, NIL)),
			       [(GOTO l)], [(GOTO label1)])) ;
			   (LABEL (NAME_LABEL label1))]
			end))
		    newctx.ctx_case_label_map))
	  @ gotodefault
	  @ body
	  @ [(LABEL (NAME_LABEL leavelabel))])
      end)
  | (DOWHILE (true, e0, l1)) ->
      (begin
	(*while-loop*)
	let looplabel = (Cexpr.make_new_label ()) in
	let leavelabel = (Cexpr.make_new_label ()) in
	let newctx = {
	  ctx with
	    ctx_continue_label = Some looplabel ;
	    ctx_break_label = Some leavelabel
	} in
	let bodylabel = (Cexpr.make_new_label ()) in
	let (_, s1) =
	  (flatten_expr_condition ctx (bodylabel, leavelabel) false e0) in
	let v0 = (Cexpr.fold_constant_integer e0) in
	let constantp = (Cexpr.constant_integer_p v0) in
	let v1 = (Cexpr.constant_integer_value true v0) in
	if (constantp && (eq_num v1 (num_of_int 0))) then
	  ([(NOTE (LINE(false, "/*memo(while(false))*/", 0, 0)))])
	else if (constantp && (not (eq_num v1 (num_of_int 0)))) then
	  ([(NOTE (LINE(false, "/*memo(while(true))*/", 0, 0)))]
	    @ [(LABEL (NAME_LABEL looplabel))]
	    @ (List.flatten (List.map (flatten_stmt_body newctx) l1))
	    @ [(GOTO looplabel)]
	    @ [(LABEL (NAME_LABEL leavelabel))])
	else
	  ([(NOTE (LINE(false, "/*memo(while)*/", 0, 0)))]
	    @ [(LABEL (NAME_LABEL looplabel))]
	    @ s1
	    @ [(LABEL (NAME_LABEL bodylabel))]
	    @ (List.flatten (List.map (flatten_stmt_body newctx) l1))
	    @ [(GOTO looplabel)]
	    @ [(LABEL (NAME_LABEL leavelabel))])
      end)
  | (DOWHILE (false, e0, l1)) ->
      (begin
	(*do-loop*)
	let looplabel = (Cexpr.make_new_label ()) in
	let leavelabel = (Cexpr.make_new_label ()) in
	let newctx = {
	  ctx with
	    ctx_continue_label = Some looplabel ;
	    ctx_break_label = Some leavelabel
	} in
	let (_, s1) =
	  (flatten_expr_condition ctx (looplabel, leavelabel) false e0) in
	let v0 = (Cexpr.fold_constant_integer e0) in
	let constantp = (Cexpr.constant_integer_p v0) in
	let v1 = (Cexpr.constant_integer_value true v0) in
	if (constantp && (eq_num v1 (num_of_int 0))) then
	  ([(NOTE (LINE(false, "/*memo(do(false))*/", 0, 0)))]
	    @ [(LABEL (NAME_LABEL looplabel))]
	    @ (List.flatten (List.map (flatten_stmt_body newctx) l1))
	    @ [(LABEL (NAME_LABEL leavelabel))])
	else if (constantp && (not (eq_num v1 (num_of_int 0)))) then
	  ([(NOTE (LINE(false, "/*memo(do(true))*/", 0, 0)))]
	    @ [(LABEL (NAME_LABEL looplabel))]
	    @ (List.flatten (List.map (flatten_stmt_body newctx) l1))
	    @ [(GOTO looplabel)]
	    @ [(LABEL (NAME_LABEL leavelabel))])
	else
	  ([(NOTE (LINE(false, "/*memo(do)*/", 0, 0)))]
	    @ [(LABEL (NAME_LABEL looplabel))]
	    @ (List.flatten (List.map (flatten_stmt_body newctx) l1))
	    @ s1
	    @ [(LABEL (NAME_LABEL leavelabel))])
      end)
  | (FOR (i0, e1, e2, l3)) ->
      (begin
	let looplabel = (Cexpr.make_new_label ()) in
	let entrylabel = (Cexpr.make_new_label ()) in
	let bodylabel = (Cexpr.make_new_label ()) in
	let leavelabel = (Cexpr.make_new_label ()) in
	let newctx = {
	  ctx with
	    ctx_continue_label = Some looplabel ;
	    ctx_break_label = Some leavelabel
	} in
	let s0 = (List.flatten (List.map (flatten_stmt_body ctx) i0)) in
	let (_, s1) =
	  (flatten_expr_condition ctx (bodylabel, leavelabel) false e1) in
	let (x2, s2) = (flatten_expr_body ctx NOVAL false e2) in
	if ((e1 == NIL) && (e2 == NIL)) then
	  ([(NOTE (LINE(false, "/*memo(for(i;;))*/", 0, 0)))]
	    @ s0
	    @ [(LABEL (NAME_LABEL looplabel))]
	    @ (List.flatten (List.map (flatten_stmt_body newctx) l3))
	    @ [(GOTO looplabel)]
	    @ [(LABEL (NAME_LABEL leavelabel))])
	else if (e1 == NIL) then
	  ([(NOTE (LINE(false, "/*memo(for(i;;n))*/", 0, 0)))]
	    @ s0
	    @ [(GOTO entrylabel)]
	    @ [(LABEL (NAME_LABEL looplabel))]
	    @ s2
	    @ [(LABEL (NAME_LABEL entrylabel))]
	    @ (List.flatten (List.map (flatten_stmt_body newctx) l3))
	    @ [(GOTO looplabel)]
	    @ [(LABEL (NAME_LABEL leavelabel))])
	else if (e2 == NIL) then
	  ([(NOTE (LINE(false, "/*memo(for(i;c;)*/", 0, 0)))]
	    @ s0
	    @ [(LABEL (NAME_LABEL looplabel))]
	    @ s1
	    @ [(LABEL (NAME_LABEL bodylabel))]
	    @ (List.flatten (List.map (flatten_stmt_body newctx) l3))
	    @ [(GOTO looplabel)]
	    @ [(LABEL (NAME_LABEL leavelabel))])
	else
	  ([(NOTE (LINE(false, "/*memo(for(i;c;n)*/", 0, 0)))]
	    @ s0
	    @ [(GOTO entrylabel)]
	    @ [(LABEL (NAME_LABEL looplabel))]
	    @ s2
	    @ [(LABEL (NAME_LABEL entrylabel))]
	    @ s1
	    @ [(LABEL (NAME_LABEL bodylabel))]
	    @ (List.flatten (List.map (flatten_stmt_body newctx) l3))
	    @ [(GOTO looplabel)]
	    @ [(LABEL (NAME_LABEL leavelabel))])
      end)
  | (BREAK) ->
      assert (not (ctx.ctx_break_label == None)) ;
      [(GOTO (Misc.some_of (ctx.ctx_break_label)))]
  | (CONTINUE) ->
      assert (not (ctx.ctx_continue_label == None)) ;
      [(GOTO (Misc.some_of (ctx.ctx_continue_label)))]
  | (LABEL l) ->
      (begin
	match l with
	  (NAME_LABEL n) ->
	    [s]
	| (CASE_LABEL (e0, e1)) ->
	    let v0 = (Cexpr.fold_constant_integer e0) in
	    let v1 = (if (e1 == NIL) then NIL 
	      else (Cexpr.fold_constant_integer e1)) in
	    if (not (Cexpr.constant_integer_p v0)) then
	      (syntaxerror "Non-constant case") ;
	    if (v1 != NIL && (not (Cexpr.constant_integer_p v1))) then
	      (syntaxerror "Non-constant case") ;
	    let n = (Cexpr.make_new_label ()) in
	    (ctx.ctx_case_label_map
	      <- (ctx.ctx_case_label_map @ [((v0, v1), n)])) ;
	    [(LABEL (NAME_LABEL n))]
	| (DFLT_LABEL) ->
	    let n = (Cexpr.make_new_label ()) in
	    (ctx.ctx_case_label_map
	      <- (ctx.ctx_case_label_map @ [((NIL, NIL), n)])) ;
	    [(LABEL (NAME_LABEL n))]
      end)
  | (GOTO n) ->
      [s]
  | (RETURN None) ->
      [(GOTO (ctx.ctx_return_label))]
  | (RETURN (Some e)) ->
      (* Already be casted to a proper type in the parser. *)
      assert (same_type_p None ctx.ctx_return_type (type_of_expr e)) ;
      let (x0, s0) = (flatten_expr_body ctx AVAL false e) in
      let tn = ctx.ctx_return_type in
      (s0
	@ [(assignment_stmt tn (result_value_reference tn) x0)]
	@ [(GOTO (ctx.ctx_return_label))])
  | (ASSUME _) | (RESET _)
  | (ANNOTATION _) ->
      [s]
  | (ASM e) ->
      (warning_message
	  ("IGNORE ASM: '" ^ (expr_to_string e) ^ "'") "") ;
      ([(NOTE (LINE(false, "/*ASM IGNORED*/", 0, 0)))]
	@ (handle_gcc_asm_statement s))
  | (NOTE _) -> (
(*      current_note_while_flatten := s;*)
      [s]
  )
)

(* Flattens an expression.  It returns a pair of an expression and a
   sequence of statements, where an expression is effect-free.  Calls
   assign a new variable for a return value (it is for nesting of
   calls).  It returns a variable for an expression, when {flat} and
   {RVAL} is specified.  It removes uses of operators {COND}, {OROR},
   {ANDAND}, {INCR_PRE}, {DECR_PRE}, {INCR_POST}, {DECR_POST}, and
   {COMMA}. *)

and flatten_expr_body (ctx : flatten_context_t) (needvalue : need_value_t) (flatp : bool) ?(check_array_index = false) (e : expr_t) : (expr_t * (stmt_t list)) = (
  match e with
    NIL ->
      (e, [])
  | (PSEUDO_NONDET_VALUE (tn, _)) ->
      (e, [])
  | (OPER (tn, SET, _, _, _)) -> (assert false)
  | (OPER (tn, (INCR_PRE|DECR_PRE), _, _, _)) -> (assert false)
  | (OPER (tn, (INCR_POST|DECR_POST), _, _, _)) -> (assert false)
  | (OPER (tn, COND, e0, NIL, e2)) ->
      (*GCC*)
      (begin
	assert (not (needvalue == LVAL)) ;
	if (needvalue == AVAL) then assert (not (tn == VOID_TYPE)) ;
	if (needvalue == BVAL) then assert (not (tn == VOID_TYPE)) ;
	let thenlabel = (Cexpr.make_new_label ()) in
	let elselabel = (Cexpr.make_new_label ()) in
	let nextlabel = (Cexpr.make_new_label ()) in
	let (x0, s0) = (flatten_expr_body ctx needvalue false e0) in
	let x1 = (make_logical_value x0) in
	let (x2, s2) = (flatten_expr_body ctx needvalue false e2) in
	(* Check constantness of the condition. *)
	let cx = (Cexpr.fold_constant_integer e0) in
	if (Cexpr.constant_integer_p cx) then
	  (begin
	    let cv = (constant_integer_value false cx) in
	    match needvalue with
	      LVAL -> (assert false)
	    | NOVAL ->
		(begin
		  if (not (eq_num cv num_ZERO)) then
		    (NIL, [])
		  else
		    (NIL, s2)
		end)
	    | AVAL ->
		(begin
		  if (not (eq_num cv num_ZERO)) then
		    (x0, [])
		  else
		    (x2, s2)
		end)
	    | BVAL -> (assert false)
	  end)
	else
	  (begin
	    match needvalue with
	      LVAL -> (assert false)
	    | NOVAL ->
		(begin
		  let ss =
		    (s0
		      @ [(IF (x1, [(GOTO nextlabel)], [(GOTO elselabel)]))]
		      @ [(LABEL (NAME_LABEL elselabel))]
		      @ s2
		      @ [(LABEL (NAME_LABEL nextlabel))]) in
		  (NIL, ss)
		end)
	    | AVAL ->
		(begin
		  (* THIS TEMPORARY DEPENDS ON FLOW (it needs SSA). *)
		  let result = (Cexpr.make_new_variable tn) in
		  let ss =
		    (s0
		      @ [(IF (x1, [(GOTO thenlabel)], [(GOTO elselabel)]))]
		      @ [(LABEL (NAME_LABEL thenlabel))]
(*                  @ [!current_note_while_flatten] (* Binh - 20091002 *) *)
		      @ [(assignment_stmt tn result x0)]
		      @ [(GOTO nextlabel)]
		      @ [(LABEL (NAME_LABEL elselabel))]
		      @ s2
(*                  @ [!current_note_while_flatten] (* Binh - 20091002 *) *)
		      @ [(assignment_stmt tn result x2)]
		      @ [(LABEL (NAME_LABEL nextlabel))]) in
		  (result, ss)
		end)
	    | BVAL -> (assert false)
	  end)
      end)
  | (OPER (tn, COND, e0, e1, e2)) ->
      (begin
	assert (not (needvalue == LVAL)) ;
	if (needvalue == AVAL) then assert (not (tn == VOID_TYPE)) ;
	if (needvalue == BVAL) then assert (not (tn == VOID_TYPE)) ;
	let thenlabel = (Cexpr.make_new_label ()) in
	let elselabel = (Cexpr.make_new_label ()) in
	let nextlabel = (Cexpr.make_new_label ()) in
	let (_, s0) =
	  (flatten_expr_condition ctx (thenlabel, elselabel) false e0) in
	let (x1, s1) = (flatten_expr_body ctx needvalue false e1) in
	let (x2, s2) = (flatten_expr_body ctx needvalue false e2) in
	(* Check constantness of the condition. *)
	let cx = (Cexpr.fold_constant_integer e0) in
	if (Cexpr.constant_integer_p cx) then
	  (begin
	    let cv = (constant_integer_value false cx) in
	    match needvalue with
	      LVAL -> (assert false)
	    | NOVAL ->
		(begin
		  if (not (eq_num cv num_ZERO)) then
		    (NIL, s1)
		  else
		    (NIL, s2)
		end)
	    | AVAL ->
		(begin
		  if (not (eq_num cv num_ZERO)) then
		    (x1, s1)
		  else
		    (x2, s2)
		end)
	    | BVAL -> (assert false)
	  end)
	else
	  (begin
	    match needvalue with
	      LVAL -> (assert false)
	    | NOVAL ->
		(begin
		  let ss =
		    (s0
		      @ [(LABEL (NAME_LABEL thenlabel))]
		      @ s1
		      @ [(GOTO nextlabel)]
		      @ [(LABEL (NAME_LABEL elselabel))]
		      @ s2
		      @ [(LABEL (NAME_LABEL nextlabel))]) in
		  (NIL, ss)
		end)
	    | AVAL ->
		(begin
		  (* THIS TEMPORARY DEPENDS ON FLOW (it needs SSA). *)
		  let result = (Cexpr.make_new_variable tn) in
		  let ss =
		    (s0
		      @ [(LABEL (NAME_LABEL thenlabel))]
		      @ s1
(*                  @ [!current_note_while_flatten] (* Binh - 20091002 *) *)
		      @ [(assignment_stmt tn result x1)]
		      @ [(GOTO nextlabel)]
		      @ [(LABEL (NAME_LABEL elselabel))]
		      @ s2
(*                  @ [!current_note_while_flatten] (* Binh - 20091002 *) *)
		      @ [(assignment_stmt tn result x2)]
		      @ [(LABEL (NAME_LABEL nextlabel))]) in
		  (result, ss)
		end)
	    | BVAL -> (assert false)
	  end)
      end)
  | (OPER (tn, (OROR|ANDAND), e0, e1, _)) ->
      (begin
	assert (not (needvalue == LVAL)) ;
	if (needvalue == AVAL) then assert (not (tn == VOID_TYPE)) ;
	if (needvalue == BVAL) then assert (not (tn == VOID_TYPE)) ;
	let ee = (simple_transform_expression e) in
	(flatten_expr_body ctx needvalue flatp ee)
      end)
  | (OPER (tn, COMMA, e0, e1, _)) ->
      (begin
	if (needvalue == LVAL) then (syntaxerror "Comma for lvalue") ;
	let (_, s0) = (flatten_expr_body ctx NOVAL false e0) in
	let (x1, s1) = (flatten_expr_body ctx needvalue false e1) in
	(x1, (s0 @ s1))
      end)
  | (OPER (tn, ((ADD|SUB|MUL|DIV) as op), e0, e1, _))
  | (OPER (tn, ((MOD|LSH|RSH|AND|IOR|XOR) as op), e0, e1, _))
  | (OPER (tn, (((CAST _)|NEG|BITNOT) as op), e0, e1, _)) ->
      (begin
	assert (not (needvalue == LVAL)) ;
	if (needvalue == NOVAL) then
	  (begin
	    let (x0, s0) = (flatten_expr_body ctx NOVAL false e0) in
	    let (x1, s1) = (flatten_expr_body ctx NOVAL false e1) in
	    let ss = (s0 @ s1) in
	    (NIL, ss)
	  end)
	else
	  (begin
	    let (x0, s0) = (flatten_expr_body ctx AVAL false e0) in
	    let (x1, s1) = (flatten_expr_body ctx AVAL false e1) in
	    let xx =
	      match needvalue with
		NOVAL | LVAL -> (assert false)
	      | AVAL ->
		  (OPER (tn, op, x0, x1, NIL))
	      | BVAL ->
		  (make_logical_value (OPER (tn, op, x0, x1, NIL))) in
	    if (not flatp) then
	      (begin
		let ss = (s0 @ s1) in
		(xx, ss)
	      end)
	    else
	      (begin
		let result = (Cexpr.make_new_temporary_variable tn) in
(*                let curr_note =  [!current_note_while_flatten] in(* Binh - 20091002 *) *)
		let ss =
                  if ((not !insert_checking_array_index)
                      || check_array_index) then
                    (s0 @ s1 @ [(assignment_stmt tn result xx)])
                  else (
                    (* Check accessing array.
                       Eg. a [i] if ((i >= 0) and (i < sizeof (a))), SAFE
                                 else UNSAFE *)
                    let assert_lst =
                      if Ccommon.is_array_oper e then (
                        assert (Cexpr.integral_type_p (Cexpr.type_of_expr e1));
                        let conditions =
                          let lowerbound_cond =
                            Cexpr.make_binary_operation Cexpr.GE
                                                        e1
                                                        Cexpr.constant_zero
                          in
                          let ex_op = Ccommon.get_array_size_oper e in
                          match ex_op with
                          | None -> [lowerbound_cond]
                          | Some ex -> (
                            let upperbound_cond =
                                Cexpr.make_binary_operation Cexpr.LT
                                                          e1
                                                          ex
                            in
                            [lowerbound_cond; upperbound_cond])
                        in
                        (Stranslate.make_list_assert conditions))
                      else [] in
                    let flatten_assert_lst =
                      List.flatten (List.map (flatten_stmt_body ctx) assert_lst) in
		        (s0 @ s1 @ flatten_assert_lst @ [(assignment_stmt tn result xx)]))
    in
		(result, ss)
	      end)
	  end)
      end)
  | (OPER (tn, ((EQ|NE|GE|GT|LE|LT) as op), e0, e1, _)) ->
      (begin
	assert (not (needvalue == LVAL)) ;
	match needvalue with
	  NOVAL ->
	    (begin
	      let (x0, s0) = (flatten_expr_body ctx NOVAL false e0) in
	      let (x1, s1) = (flatten_expr_body ctx NOVAL false e1) in
	      let ss = (s0 @ s1) in
	      (NIL, ss)
	    end)
	| LVAL -> (assert false)
	| AVAL ->
	    (begin
	      let ee = (make_zero_one_value_of_conditional e) in
	      (flatten_expr_body ctx needvalue flatp ee)
	    end)
	| BVAL ->
	    (begin
	      let (x0, s0) = (flatten_expr_body ctx AVAL false e0) in
	      let (x1, s1) = (flatten_expr_body ctx AVAL false e1) in
	      let xx = (OPER (tn, op, x0, x1, NIL)) in
	      if (not flatp) then
		(begin
		  let ss = (s0 @ s1) in
		  (xx, ss)
		end)
	      else
		(begin
		  let result = (Cexpr.make_new_temporary_variable tn) in
(*                  let curr_note =  [!current_note_while_flatten] in(* Binh - 20091002 *)*)
		  let ss = (s0 @ s1 @ [(assignment_stmt tn result xx)]) in
		  (result, ss)
		end)
	    end)
      end)
  | (OPER (tn, LOGNOT, e0, _, _)) ->
      (begin
	assert (not (needvalue == LVAL)) ;
        match e0 with
        | (OPER (tn, LOGNOT, e1, _, _)) -> (
            flatten_expr_body ctx needvalue flatp e1)
        | (PSEUDO_ACSL_VALID (tn, e1)) -> (
            let pseudo_func_tn =
                make_function_type INTEGER [(make_pointer_type VOID_TYPE)]
            in
            let pseudo_func_expr =
                VREF (pseudo_func_tn, "null_addr", global_scope_level, 0)
            in
            let pseudo_func_call_expr =
                make_function_call pseudo_func_expr [e1]
            in
            let new_e =
                make_conditional_expr COND pseudo_func_call_expr
                                           constant_one
                                           constant_zero
            in 
            (flatten_expr_body ctx needvalue false new_e))
        | _ ->
          (begin
            match needvalue with
              NOVAL ->
	        (begin
	          (flatten_expr_body ctx NOVAL false e0)
	        end)
	    | LVAL -> (assert false)
	    | AVAL ->
	      (begin
	        let ee = (make_zero_one_value_of_conditional e) in
	        (flatten_expr_body ctx needvalue flatp ee)
	      end)
	    | BVAL -> (assert false)
          end)
      end)
  | (OPER (tn, (IMP|IFF), _, _, _)) -> (assert false)
  | (OPER (tn, (PSEUDO_AMP|PSEUDO_STR|PSEUDO_PLS), _, _, _)) -> (assert false)
  | (OPER (tn, (PSEUDO_MNS|PSEUDO_TLD|PSEUDO_BAN), _, _, _)) -> (assert false)
  | (SETF (tn, SET, e0, e1)) ->
      (begin
	(* Always generates a flat expression. *)
	assert (not (needvalue == LVAL)) ;
	let (x0, s0) = (flatten_expr_body ctx LVAL true e0) in
	let (x1, s1) = (flatten_expr_body ctx AVAL false e1) in
(*        let curr_note =  [!current_note_while_flatten] in(* Binh - 20091002 *) *)
	let ss = (s0 @ s1 @ [(assignment_stmt tn x0 x1)]) in
	let xx =
	  match needvalue with
	    NOVAL ->
	      NIL
	  | LVAL -> (assert false)
	  | AVAL ->
	      x1
	  | BVAL ->
	      (make_logical_value x1) in
	(xx, ss)
      end)
  | (SETF (tn, ((INCR_PRE|DECR_PRE) as op), e0, _)) ->
      (begin
	assert (not (needvalue == LVAL)) ;
	let incop = (if (op == INCR_PRE) then ADD else SUB) in
	let (x0, s0) = (flatten_expr_body ctx LVAL true e0) in
	let one = Cexpr.constant_one in
	let stmt = (assignment_stmt tn x0 (OPER (tn, incop, x0, one, NIL))) in
(*        let curr_note =  [!current_note_while_flatten] in(* Binh - 20091002 *) *)
	let ss = (s0 @ [stmt]) in
	let xx =
	  match needvalue with
	    NOVAL ->
	      NIL
	  | LVAL -> (assert false)
	  | AVAL ->
	      x0
	  | BVAL ->
	      (make_logical_value x0) in
	(xx, ss)
      end)
  | (SETF (tn, ((INCR_POST|DECR_POST) as op), e0, _)) ->
      (begin
	(* Always generates a flat expression. *)
	assert (not (needvalue == LVAL)) ;
	let incop = (if (op == INCR_POST) then ADD else SUB) in
	let (x0, s0) = (flatten_expr_body ctx LVAL true e0) in
	let one = Cexpr.constant_one in
	let stmt = (assignment_stmt tn x0 (OPER (tn, incop, x0, one, NIL))) in
(*        let curr_note =  !current_note_while_flatten in(* Binh - 20091002 *) *)
	match needvalue with
	  NOVAL ->
	    let ss = (s0 @ [stmt]) in
	    (NIL, ss)
	| LVAL -> (assert false)
	| AVAL ->
	    let result = (Cexpr.make_new_temporary_variable tn) in
	    let ss = (s0 @ [(assignment_stmt tn result x0) ; stmt]) in
	    (result, ss)
	| BVAL ->
	    let result = (Cexpr.make_new_temporary_variable tn) in
	    let ss = (s0 @ [(assignment_stmt tn result x0) ; stmt]) in
	    let xx = (make_logical_value result) in
	    (xx, ss)
      end)
  | (SETF (tn, ((ADD|SUB|MUL|DIV) as op), e0, e1))
  | (SETF (tn, ((MOD|LSH|RSH|AND|IOR|XOR) as op), e0, e1)) ->
      (begin
	(* Always generates a flat expression. *)
	assert (not (needvalue == LVAL)) ;
	let (x0, s0) = (flatten_expr_body ctx LVAL true e0) in
	let (x1, s1) = (flatten_expr_body ctx AVAL false e1) in
(*        let curr_note =  !current_note_while_flatten in(* Binh - 20091002 *) *)
	if (needvalue == NOVAL) then
	  (begin
	    let stmt = (assignment_stmt tn x0 (OPER (tn, op, x0, x1, NIL))) in
	    let ss = (s0 @ s1 @ [stmt]) in
	    (NIL, ss)
	  end)
	else
	  (begin
	    let result = (Cexpr.make_new_temporary_variable tn) in
	    let stmt =
	      [(assignment_stmt tn result (OPER (tn, op, x0, x1, NIL))) ;
	       (assignment_stmt tn x0 result)] in
	    let ss = (s0 @ s1 @ stmt) in
	    match needvalue with
	      NOVAL | LVAL -> (assert false)
	    | AVAL ->
		(result, ss)
	    | BVAL ->
		((make_logical_value result), ss)
	  end)
      end)
  | (SETF (tn, (EQ|NE|GE|GT|LE|LT), e0, e1)) -> (assert false)
  | (SETF (tn, (LOGNOT), e0, e1)) -> (assert false)
  | (SETF (tn, ((CAST _)|NEG|BITNOT), e0, e1)) -> (assert false)
  | (SETF (tn, (COND|OROR|ANDAND|COMMA), e0, e1)) -> (assert false)
  | (SETF (tn, (IMP|IFF), e0, e1)) -> (assert false)
  | (SETF (tn, (PSEUDO_AMP|PSEUDO_STR|PSEUDO_PLS), e0, e1)) -> (assert false)
  | (SETF (tn, (PSEUDO_MNS|PSEUDO_TLD|PSEUDO_BAN), e0, e1)) -> (assert false)
  | (VREF (tn, id, level, _)) ->
      (begin
	match needvalue with
	  NOVAL ->
	    (NIL, [])
	| LVAL | AVAL ->
	    ((map_static_variable_name ctx e), [])
	| BVAL ->
	    (flatten_expr_body ctx needvalue flatp (make_logical_value e))
      end)
  | (PREF (tn, deref, e1, e2)) ->
      (begin
(*        let curr_note =  !current_note_while_flatten in(* Binh - 20091002 *) *)
	match needvalue with
	  NOVAL ->
	    (flatten_expr_body ctx NOVAL flatp e1)
	| LVAL ->
	    (begin
	      if (not deref) then
		(syntaxerror "Array reference for lvalue") ;
	      let (x1, s1) = (flatten_expr_body ctx AVAL true e1)
	      and (x2, s2) = (flatten_expr_body ctx AVAL true ~check_array_index:true e2) in
              let s = s1 @ s2 in
	      let xx = (PREF (tn, deref, x1, x2)) in
	      (xx, s)
	    end)
	| AVAL ->
	    (begin
	      let (x1, s1) = (flatten_expr_body ctx AVAL true e1)
	      and (x2, s2) = (flatten_expr_body ctx AVAL true ~check_array_index:true e2) in
	      let xx = (PREF (tn, deref, x1, x2)) in
	      if (flatp || deref) then
		(begin
		  let result = (Cexpr.make_new_temporary_variable tn) in
		  let stmt = [(assignment_stmt tn result xx)] in
		  (result, (s1 @ s2 @ stmt))
		end)
	      else
		(begin
		  (xx, s1)
		end)
	    end)
	| BVAL ->
	    (flatten_expr_body ctx needvalue flatp (make_logical_value e))
      end)
  | (SREF (tn, deref, e0, s)) ->
      if (deref) then
	(begin
	  match needvalue with
	    NOVAL ->
	      (flatten_expr_body ctx NOVAL flatp e0)
	  | LVAL | AVAL ->
	      let (x0, s0) = (flatten_expr_body ctx AVAL true e0) in
	      ((SREF (tn, deref, x0, s)), s0)
	  | BVAL ->
	      (flatten_expr_body ctx needvalue flatp (make_logical_value e))
	end)
      else
	(begin
	  match needvalue with
	    NOVAL ->
	      (flatten_expr_body ctx NOVAL flatp e0)
	  | LVAL | AVAL ->
	      let (x0, s0) = (flatten_expr_body ctx needvalue true e0) in
	      ((SREF (tn, deref, x0, s)), s0)
	  | BVAL ->
	      (flatten_expr_body ctx needvalue flatp (make_logical_value e))
	end)
  | (DSGN_AREF (tn, e0, e1)) -> (assert false)
  | (CALL (_, _, _, _)) ->
      (begin
	assert (not (needvalue == LVAL)) ;
	let x = (convert_builtin_function_call ctx needvalue flatp e) in
	match x with
	  (CALL (tn, e0, ee, spec_op)) ->
	    (begin
	      (* Always generates a flat expression. *)
	      let extra = (extra_goto_after_noreturn_call ctx e0) in
	      match needvalue with
		NOVAL ->
		  (begin
		    let (x0, s0) = (flatten_expr_body ctx AVAL false e0) in
		    let l = (List.map (flatten_expr_body ctx AVAL false) ee) in
		    let xx = (List.map fst l) in
		    let ss = (List.flatten (List.map snd l)) in
		    let tn = (Cexpr.type_of_expr x) in
		    let aa =
		      ((STMT NIL)
			:: (list_variables_addrof_taken_in_call ctx xx)) in
		    let sx = [(CALLASGN ((CALL (tn, x0, xx, spec_op)), aa, []))] in
		    (NIL, (ss @ sx @ extra))
		  end)
	      | LVAL -> (assert false)
	      | AVAL ->
		  (begin
		    let (x0, s0) = (flatten_expr_body ctx AVAL false e0) in
		    let l = (List.map (flatten_expr_body ctx AVAL false) ee) in
		    let xx = (List.map fst l) in
		    let ss = (List.flatten (List.map snd l)) in
		    let tn = (Cexpr.type_of_expr x) in
		    let vt = (Cexpr.make_new_temporary_variable tn) in
		    let aa =
		      ((assignment_stmt tn vt (pseudo_nondet_value tn))
			:: (list_variables_addrof_taken_in_call ctx xx)) in
		    let sx = [(CALLASGN ((CALL (tn, x0, xx, spec_op)), aa, []))] in
		    (vt, (ss @ sx @ extra))
		  end)
	      | BVAL ->
		  (begin
		    let (x0, s0) =
		      (flatten_expr_body ctx BVAL flatp
			  (make_logical_value e)) in
		    (x0, (s0 @ extra))
		  end)
	    end)
	| _ ->
	    (* Builtin function is rewritten.*)
	    (flatten_expr_body ctx needvalue flatp x)
      end)
  | (ADDROF (tn, e0)) ->
      (begin
	if ((Cexpr.type_of_expr e0) != NOTYPE) then
	  (begin
	    match needvalue with
	      NOVAL ->
		(flatten_expr_body ctx NOVAL flatp e0)
	    | LVAL -> (assert false)
	    | AVAL ->
		let (x0, s0) = (flatten_expr_body ctx LVAL false e0) in
		((ADDROF (tn, x0)), s0)
	    | BVAL ->
		let one = Cexpr.constant_one in
		(one, [])
	  end)
	else
	  (begin
	    (*GCC*) (* Label values. *)
	    match needvalue with
	      NOVAL ->
		(NIL, [])
	    | LVAL -> (assert false)
	    | AVAL ->
		let x0 = (make_value_for_label ctx e0) in
		(x0, [])
	    | BVAL ->
		let one = Cexpr.constant_one in
		(one, [])
	  end)
      end)
  | (SIZEOF (tn, tt)) ->
      (e, [])
  | (PROGN (tn, l)) ->
      (*GCC*)
      (begin
	assert (not (needvalue == LVAL)) ;
	(flatten_stmt_return_last ctx needvalue flatp l)
      end)
  | (PSEUDO_MEMORY (tn, tt)) ->
      (e, [])
  | (INIT_FILL (tt, x)) ->
      (e, [])
  | (CONST_I (tn, v, s)) ->
      (begin
	match needvalue with
	  NOVAL ->
	    (NIL, [])
	| LVAL -> (assert false)
	| AVAL ->
	    (e, [])
	| BVAL ->
	    let zero = Cexpr.constant_zero in
	    let one = Cexpr.constant_one in
	    if (eq_num v (num_of_int 0)) then
	      (zero, [])
	    else
	      (one, [])
      end)
  | (CONST_F (tn, v, s)) ->
      (begin
	match needvalue with
	  NOVAL ->
	    (NIL, [])
	| LVAL -> (assert false)
	| AVAL ->
	    (e, [])
	| BVAL ->
	    let one = Cexpr.constant_one in
	    (one, [])
      end)
  | (STRING (b, s, _)) ->
      (begin
	match needvalue with
	  NOVAL ->
	    (NIL, [])
	| LVAL -> (assert false)
	| AVAL ->
	    (e, [])
	| BVAL ->
	    let one = Cexpr.constant_one in
	    (one, [])
      end)
  | (CLITERAL (tn, l)) ->
      (begin
	match needvalue with
	  NOVAL ->
	    (NIL, [])
	| LVAL -> (assert false)
	| AVAL ->
	    let v = (make_new_temporary_variable tn) in
	    let (initexpr, maxindex) = (make_initializers v tn l) in
	    let s = (flatten_auto_decl ctx tn v initexpr) in
	    (v, s)
	| BVAL ->
	    let one = Cexpr.constant_one in
	    (one, [])
      end)
  | (ASMSEQUENCE l) -> (*AHO*) (assert false)
  | (PHI _) -> (assert false)
  | (PRAGMA s) ->
      (e, [])
  | (QUANTIFIER (s, vs, e0)) ->
      (*AHO*)
      (e, [])
  | PSEUDO_PROLOGUE | PSEUDO_EPILOGUE -> (assert false)
  | PSEUDO_ACSL_RESULT (tn, e0) ->
      (begin
        let x, ss = flatten_expr_body ctx needvalue flatp e0 in
        let xx = match needvalue with
                 | NOVAL -> NIL
                 | LVAL -> (assert false)
                 | AVAL -> (PSEUDO_ACSL_RESULT (tn, x))
                 | BVAL -> (make_logical_value (PSEUDO_ACSL_RESULT (tn, x)))
        in
        (xx, ss)
      end)
  | PSEUDO_ACSL_OLD (tn, e0, e1) ->
      (begin
        let x0, ss0 = flatten_expr_body ctx needvalue flatp e0 in
        let x1, ss1 = flatten_expr_body ctx needvalue flatp e1 in
        let xx = match needvalue with
                 | NOVAL -> NIL
                 | LVAL -> (assert false)
                 | AVAL -> (PSEUDO_ACSL_OLD (tn, x0, x1))
                 | BVAL -> (make_logical_value (PSEUDO_ACSL_OLD (tn, x0, x1)))
        in
        (xx, ss0 @ ss1)
      end)
  | ALL_FIELD _ -> assert false
  | PSEUDO_ACSL_UNIQUE (tn, e0) ->
      (begin
        let x, ss = flatten_expr_body ctx AVAL flatp e0 in
        let xx = (PSEUDO_ACSL_UNIQUE (tn, x))
        in
        (xx, ss)
      end)
  | PSEUDO_ACSL_UNIQUEBLOCK (tn, e0, e1) ->
      (begin
        let x, ss = flatten_expr_body ctx AVAL flatp e0 in
        let xx = (PSEUDO_ACSL_UNIQUEBLOCK (tn, x, e1))
        in
        (xx, ss)
      end)
  | PSEUDO_ACSL_VALID (tn, e0) ->
      (begin
        let new_e0 =
            Cexpr.make_binary_operation Cexpr.NE e0 Cexpr.constant_zero
        in
        flatten_expr_body ctx needvalue flatp new_e0
      end))

(* Makes a new variable representing the label value, and returns it
   after recording it in {file_label_table}.  The variable name for
   the referred label is given 1 for the SSA version, because it is
   assigned that value in later stages. *)

and make_value_for_label (ctx : flatten_context_t) (e : expr_t) : expr_t = (
  match e with
    (VREF (tn, id, level, _)) ->
      (begin
	assert (tn == Cexpr.NOTYPE) ;
	let label =
	  if (level == Cexpr.parameter_scope_level) then
	    id
	  else
	    (id ^ "%" ^ (string_of_int level)) in
	let name = (ctx.ctx_file_hash_name ^ "%"
		    ^ ctx.ctx_function_name ^ "%"
		    ^ (string_of_int level) ^ "%" ^ id) in
	let var = (Cexpr.vartag_to_string (name, global_scope_level, 1)) in
	ctx.ctx_referred_labels <- (ctx.ctx_referred_labels @ [(label, var)]) ;
	let pos = (ctx.ctx_file_hash_name, ctx.ctx_function_name, label) in
	Cexpr.file_label_table := (!Cexpr.file_label_table @ [(var, pos)]) ;
	(VREF (Cexpr.void_pointer_type, name, Cexpr.global_scope_level, 0))
      end)
  | _ -> (assert false))

(* Flattens a conditional expression.  It generates a conditional goto
   on a value of an expression to the labels of then/else parts.  It
   is to avoid assignment of a value of a conditional expression to a
   variable, and to handle booleans as predicates.  It handles only
   logical-to-logical expressions, such like {COND} and {LOGNOT}.
   MEMO: {NIL} can appear as a condition of {for} and is taken as
   true.  {COND} with constants may be introduced by {OROR} and
   {ANDAND}, and they are handled separately just for simpler
   output. *)

and flatten_expr_condition (ctx : flatten_context_t) ((thenlabel, elselabel) : string * string) (flatp : bool) (e : expr_t) : (expr_t * (stmt_t list)) = (
  let labelpair = (thenlabel, elselabel) in
  match e with
    NIL ->
      (NIL, [(GOTO thenlabel)])
  | (PSEUDO_NONDET_VALUE (tn, _)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (OPER (tn, SET, _, _, _)) -> (assert false)
  | (OPER (tn, (INCR_PRE|DECR_PRE), _, _, _)) -> (assert false)
  | (OPER (tn, (INCR_POST|DECR_POST), _, _, _)) -> (assert false)
  | (OPER (tn, COND, e0, NIL, e2)) ->
      (*GCC*)
      (begin
	let xelselabel = (Cexpr.make_new_label ()) in
	let (_, s0) =
	  (flatten_expr_condition ctx (thenlabel, xelselabel) false e0) in
	let (_, s2) =
	  (flatten_expr_condition ctx labelpair false e2) in
	let ss =
	  (s0
	    @ [(LABEL (NAME_LABEL xelselabel))]
	    @ s2) in
	(NIL, ss)
      end)
  | (OPER (tn, COND, e0, ((CONST_I _) as e1), ((CONST_I _) as e2))) ->
      (begin
	let xthenlabel =
	  if (not (constant_zero_p e1)) then thenlabel else elselabel in
	let xelselabel =
	  if (not (constant_zero_p e2)) then thenlabel else elselabel in
	(flatten_expr_condition ctx (xthenlabel, xelselabel) false e0)
      end)
  | (OPER (tn, COND, e0, e1, ((CONST_I _) as e2))) ->
      (begin
	let xthenlabel = (Cexpr.make_new_label ()) in
	let xelselabel =
	  if (not (constant_zero_p e2)) then thenlabel else elselabel in
	let (_, s0) =
	  (flatten_expr_condition ctx (xthenlabel, xelselabel) false e0) in
	let (x1, s1) =
	  (flatten_expr_condition ctx labelpair false e1) in
	let ss =
	  (s0
	    @ [(LABEL (NAME_LABEL xthenlabel))]
	    @ s1) in
	(NIL, ss)
      end)
  | (OPER (tn, COND, e0, ((CONST_I _) as e1), e2)) ->
      (begin
	let xthenlabel =
	  if (not (constant_zero_p e1)) then thenlabel else elselabel in
	let xelselabel = (Cexpr.make_new_label ()) in
	let (_, s0) =
	  (flatten_expr_condition ctx (xthenlabel, xelselabel) false e0) in
	let (x2, s2) =
	  (flatten_expr_condition ctx labelpair false e2) in
	let ss =
	  (s0
	    @ [(LABEL (NAME_LABEL xelselabel))]
	    @ s2) in
	(NIL, ss)
      end)
  | (OPER (tn, COND, e0, e1, e2)) ->
      (begin
	let xthenlabel = (Cexpr.make_new_label ()) in
	let xelselabel = (Cexpr.make_new_label ()) in
	let (_, s0) =
	  (flatten_expr_condition ctx (xthenlabel, xelselabel) false e0) in
	let (x1, s1) =
	  (flatten_expr_condition ctx labelpair false e1) in
	let (x2, s2) =
	  (flatten_expr_condition ctx labelpair false e2) in
	let ss =
	  (s0
	    @ [(LABEL (NAME_LABEL xthenlabel))]
	    @ s1
	    @ [(LABEL (NAME_LABEL xelselabel))]
	    @ s2) in
	(NIL, ss)
      end)
  | (OPER (tn, (OROR|ANDAND), e0, e1, _)) ->
      (begin
	let ee = (simple_transform_expression e) in
	(flatten_expr_condition ctx labelpair flatp ee)
      end)
  | (OPER (tn, COMMA, e0, e1, _)) ->
      (begin
	let (_, s0) = (flatten_expr_body ctx NOVAL false e0) in
	let (_, s1) = (flatten_expr_condition ctx labelpair flatp e1) in
	(NIL, (s0 @ s1))
      end)
  | (OPER (tn, LOGNOT, e0, _, _)) ->
      (flatten_expr_condition ctx (elselabel, thenlabel) false e0)
  | (OPER (tn, (ADD|SUB|MUL|DIV), e0, e1, _))
  | (OPER (tn, (MOD|LSH|RSH|AND|IOR|XOR), e0, e1, _))
  | (OPER (tn, ((CAST _)|NEG|BITNOT), e0, e1, _)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (OPER (tn, (EQ|NE|GE|GT|LE|LT), e0, e1, _)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (OPER (tn, (IMP|IFF), _, _, _)) -> (assert false)
  | (OPER (tn, (PSEUDO_AMP|PSEUDO_STR|PSEUDO_PLS), _, _, _)) -> (assert false)
  | (OPER (tn, (PSEUDO_MNS|PSEUDO_TLD|PSEUDO_BAN), _, _, _)) -> (assert false)
  | (SETF (tn, SET, e0, e1)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (SETF (tn, (INCR_PRE|DECR_PRE), e0, _)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (SETF (tn, (INCR_POST|DECR_POST), e0, _)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (SETF (tn, (ADD|SUB|MUL|DIV), e0, e1))
  | (SETF (tn, (MOD|LSH|RSH|AND|IOR|XOR), e0, e1)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (SETF (tn, (EQ|NE|GE|GT|LE|LT), e0, e1)) -> (assert false)
  | (SETF (tn, (LOGNOT), e0, e1)) -> (assert false)
  | (SETF (tn, ((CAST _)|NEG|BITNOT), e0, e1)) -> (assert false)
  | (SETF (tn, (COND|OROR|ANDAND|COMMA), e0, e1)) -> (assert false)
  | (SETF (tn, (IMP|IFF), e0, e1)) -> (assert false)
  | (SETF (tn, (PSEUDO_AMP|PSEUDO_STR|PSEUDO_PLS), e0, e1)) -> (assert false)
  | (SETF (tn, (PSEUDO_MNS|PSEUDO_TLD|PSEUDO_BAN), e0, e1)) -> (assert false)
  | (VREF (tn, id, level, _)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (PREF (tn, deref, e1, _)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (SREF (tn, _, e0, s)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (DSGN_AREF (tn, e0, e1)) -> (assert false)
  | (CALL (_, _, _, _)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (ADDROF (tn, e0)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (SIZEOF (tn, tt)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (PROGN (tn, l)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (PSEUDO_MEMORY (tn, tt)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (INIT_FILL (tt, x)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (CONST_I (tn, v, s)) ->
      if (not (constant_zero_p e)) then
	(NIL, [(GOTO thenlabel)])
      else
	(NIL, [(GOTO elselabel)])
  | (CONST_F (tn, v, s)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (STRING (b, s, _)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (CLITERAL (tn, l)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (ASMSEQUENCE l) -> (*AHO*) (assert false)
  | (PHI _) -> (assert false)
  | (PRAGMA s) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | (QUANTIFIER (s, vs, e0)) ->
      (flatten_expr_condition_other ctx labelpair flatp e)
  | PSEUDO_PROLOGUE | PSEUDO_EPILOGUE -> (assert false)
  | PSEUDO_ACSL_RESULT _ ->  (e, [])
  | PSEUDO_ACSL_OLD _ -> (e, [])
  | ALL_FIELD _ -> assert false
  | PSEUDO_ACSL_UNIQUEBLOCK _
  | PSEUDO_ACSL_UNIQUE _ -> assert false
  | PSEUDO_ACSL_VALID _ ->  (e, []))

and flatten_expr_condition_other (ctx : flatten_context_t) ((thenlabel, elselabel) : string * string) (flatp : bool) (e : expr_t) : (expr_t * (stmt_t list)) = (
  if (thenlabel == elselabel) then
    (begin
      (* Degenerate case. *)
      let (_, s0) = (flatten_expr_body ctx BVAL flatp e) in
      (NIL, (s0 @ [(GOTO thenlabel)]))
    end)
  else
    (begin
      let (x0, s0) = (flatten_expr_body ctx BVAL flatp e) in
      let ss =
	(s0 @ [(IF (x0, [(GOTO thenlabel)], [(GOTO elselabel)]))]) in
      (NIL, ss)
    end))

(* initiate fields of struct that have not initiated yet *)
and create_and_flatten_field_inits (ctx : flatten_context_t)
                                   (decl : declaration_t)
: (stmt_t list) = (
    let (id, level) = decl.decl_id
    and tn = decl.decl_tn in
    let v = (VREF (tn, id, level, 0)) in
    let new_field_inits = Ccommon.init_fields_of_declaration decl in
    let new_flatten_field_inits = (
        List.flatten (
            List.map (fun e ->
                (snd (flatten_expr_body ctx NOVAL false e))
            ) new_field_inits)
    ) in
    match new_flatten_field_inits with
    | [] -> []
    | _ -> (
        let zero = Cexpr.CONST_I (Cexpr.INTEGER, Num.Int(0), "0") in
        let new_flatten_field_inits = (
            match tn with
            | Cexpr.POINTER_TO (Cexpr.STRUCTURE_TYPE _) -> (
                let cond = Cexpr.make_binary_operation Cexpr.NE v zero in
                let thenlabel = (Cexpr.make_new_label ()) in
                let elselabel = (Cexpr.make_new_label ()) in
                let nextlabel = (Cexpr.make_new_label ()) in
                let (_, es) = flatten_expr_condition ctx
                                                     (thenlabel, elselabel)
                                                     false
                                                     cond
                in
                (es
                 @ [(LABEL (NAME_LABEL thenlabel))]
                 @ (new_flatten_field_inits)
                 @ [(GOTO nextlabel)]
                 @ [(LABEL (NAME_LABEL elselabel))]
                 @ ([])
                 @ [(GOTO nextlabel)]
                 @ [(LABEL (NAME_LABEL nextlabel))])
              )
            | _ -> new_flatten_field_inits)
        in
        new_flatten_field_inits)
)

and flatten_auto_decl (ctx : flatten_context_t) (tn : typenode_t) (var : expr_t) (inits : expr_t list) : stmt_t list = (
  let alloc =
    (begin
      if (Cexpr.array_type_p tn) then
	(begin
	  [(assignment_stmt tn var (PSEUDO_MEMORY (INTEGER, tn)))]
	end)
      else if (Cexpr.structure_type_p tn) then
	(begin
	  [(assignment_stmt tn var (PSEUDO_MEMORY (INTEGER, tn)))]
	end)
      else
	[]
    end) in
(*  let curr_note =  !current_note_while_flatten in(* Binh - 20091002 *) *)
  let old_check = !insert_checking_array_index in
  insert_checking_array_index := false;
  let newinits =
    (List.flatten
	(List.map
	    (fun e ->
	      (* Discard {fst} of flattening as for effect. *)
	      (snd (flatten_expr_body ctx NOVAL false e)))
	    inits)) in
  insert_checking_array_index := old_check;
  (alloc @ newinits))

(* Lists lvalues address-of is taken in call arguments.  It is called
   for flattened argument expressions. *)

and list_variables_addrof_taken_in_call (ctx : flatten_context_t) (l : expr_t list) : stmt_t list = (
  (Misc.list_filtering_map
      (fun (ea : expr_t) ->
	let e = (ignore_cast ea) in
	match e with
	  NIL ->
	    None
	| (PSEUDO_NONDET_VALUE (tn, _)) ->
	    None
	| (OPER (tn, op, e0, e1, e2)) ->
	    None
	| (CALL (tn, e0, ee, _)) -> (assert false)
	| (SETF _) -> (assert false)
	| (VREF (tn, id, level, 0)) ->
	    None
	| (VREF _) -> (assert false)
	| (PREF (tn, _, e0, _)) ->
	    None
	| (SREF (tn, _, e0, s)) ->
	    None
	| (DSGN_AREF (tn, e0, e1)) -> (assert false)
	| (ADDROF (tn, e0)) ->
	    (check_variable_addrof_taken ctx (ignore_cast e0))
	| (SIZEOF (tn, tt)) | (PSEUDO_MEMORY (tn, tt)) ->
	    None
	| (INIT_FILL _) -> (assert false)
	| (PROGN (tn, l)) -> (assert false)
	| (CONST_I _) | (CONST_F _) | (STRING _) ->
	    None
	| (CLITERAL (tn, l)) -> (assert false)
	| (QUANTIFIER (s, vs, e0)) ->
	    None
	| (PSEUDO_PROLOGUE) ->
	    None
	| (PSEUDO_EPILOGUE) ->
	    None
	| (ASMSEQUENCE ee) -> (assert false)
	| (PHI _) -> (assert false)
	| (PRAGMA s) -> (assert false)
	| (ALL_FIELD _) -> (assert false)
        | PSEUDO_ACSL_RESULT _ | PSEUDO_ACSL_OLD _ | PSEUDO_ACSL_UNIQUE _
        | PSEUDO_ACSL_VALID _ | PSEUDO_ACSL_UNIQUEBLOCK _ ->
            None)
      l))

and check_variable_addrof_taken (ctx : flatten_context_t) (e : expr_t) : stmt_t option = (
  match e with
    (VREF (tn, id, level, 0)) ->
      (begin
	if (temporary_variable_p e) then
	  (assert false)
	else if (function_reference_p e) then
	  None
	else
	  (Some (assignment_stmt tn e (pseudo_nondet_value tn)))
      end)
  | (SREF (tn, deref, e2, s)) ->
      if (deref) then
	None
      else
	let e3 = (ignore_cast e2) in
	(check_variable_addrof_taken ctx e3)
  | _ ->
      None)

and map_static_variable_name (ctx : flatten_context_t) (e : expr_t) : expr_t = (
  match e with
    (VREF (tn, id, level, _)) ->
      if (Hashtbl.mem ctx.ctx_static_variable_name_map (id, level)) then
	let (id1, level1) =
	  (Hashtbl.find ctx.ctx_static_variable_name_map (id, level)) in
	(VREF (tn, id1, level1, 0))
      else
	e
  | _ -> (assert false))

(* Note it has to skip a NOTE at the last entry. *)

and flatten_stmt_return_last (ctx : flatten_context_t) (needvalue : need_value_t) (flatp : bool) (l : stmt_t list) : (expr_t * (stmt_t list)) = (
  let nothing = (STMT NIL) in
  let xe = (last_effective_expression_loop l nothing) in
  let laststmt = (if (needvalue == NOVAL) then nothing else xe) in
  if (needvalue != NOVAL && xe == nothing) then
    (syntaxerror "No value in statement expression") ;
  (List.fold_left
      (fun ((e0, sl) : expr_t * (stmt_t list)) (s : stmt_t) ->
	if (needvalue != NOVAL && s == laststmt) then
	  (begin
	    match s with
	      (STMT e) ->
		let (xe, xs) = (flatten_expr_body ctx needvalue false e) in
		(xe, (sl @ xs))
	    | _ ->
		(syntaxerror "No value in statement expression")
	  end)
	else
	  (begin
	    let xs = (flatten_stmt_body ctx s) in
	    (e0, (sl @ xs))
	  end))
      (NIL, [])
      l))

(* Converts a builtin function call.  It is empty.  Some builtin
   functions are handled in {Cexpr.make_builtin_function_call}. *)

and convert_builtin_function_call (ctx : flatten_context_t) (needvalue : need_value_t) (flatp : bool) (e : expr_t) : expr_t = (
  e)

(* ================================================================ *)

(* MAKING BASIC BLOCKS *)

(* Lays out a statement list as basic blocks.  It removes {NOTE}
   statements and places line number information in memo. *)

(* Binh 20091003*)
let notes = ref []
let clean_notes () : unit = (
    notes := []
)
let append_notes (note : stmt_t) : unit = (
    notes := (!notes) @ [note]
)

let rec breakup_blocks (s : stmt_t) : cdefinition_t = (
  match s with
    (DEFUN (d, (BEGIN body))) ->
      (begin
        let f_name = (
            match (Ccommon.get_function_name d) with
            | Some str -> str
            | _ -> assert false
        ) in
	let note = (
            try
                (Hashtbl.find Ccommon.note_of_entry_function f_name)
            with
            | _ -> (NOTE (LINE(true, "", 0, 0)))
        ) in
        (append_notes note); (* Binh - 2091003*)
	let bbset = (Hashtbl.create 10) in
	let newlabel = (Cexpr.make_new_label ()) in
	let label = pseudo_entry_label in
	match body with
	  (((STMT PSEUDO_PROLOGUE) as hd) :: truebody) ->
            (breakup_emit_block bbset BBHEAD label !notes [hd] (GOTO newlabel));
	    (breakup_stmt_body bbset truebody newlabel note note []) ;
	    let ctx : cfundef_t = {
	      cfdef_declaration = d ;
	      cfdef_blocks = bbset
	    } in
	    (CFUNDEF ctx)
	| _ -> (assert false)
      end)
  | (DEFUN (d, _)) -> (assert false)
  | (SPEC _) | (ASSUME _) | (RESET _) -> (assert false)
  | (DEFVAR d) ->
      let ctx : cvardef_t = {
	cvdef_declaration = d
      } in
      (CVARDEF ctx)
  | (STMT _) -> (assert false)
  | (ASGN _) -> (assert false)
  | (CALLASGN _) -> (assert false)
  | (IF _) -> (assert false)
  | (BEGIN _) -> (assert false)
  | (SWITCH _) -> (assert false)
  | (DOWHILE _) -> (assert false)
  | (FOR _) -> (assert false)
  | (BREAK) -> (assert false)
  | (CONTINUE) -> (assert false)
  | (LABEL _) -> (assert false)
  | (GOTO n) -> (assert false)
  | (RETURN _) -> (assert false)
  | (ANNOTATION _) -> (assert false)
  | (ASM e) ->
      (COTHERDEF s)
  | (NOTE _) -> (assert false))

and breakup_stmt_body (ctx : (string, stmt_block_t) Hashtbl.t) (l : stmt_t list) (label : string) (note : stmt_t) (lastnote : stmt_t) (stmts : stmt_t list) : unit = (
(*Gprint.eprint ("list_stmt", l); Scommon.pause();*)
  match l with
    [] ->
      (breakup_emit_block ctx BBSTMT label !notes stmts (STMT NIL))
  | (s :: tl) ->
      (begin
	match s with
	  (DEFUN _) | (DEFVAR _) | (SPEC _) -> (assert false)
	| (STMT (CALL _)) -> (assert false)
	| (ASGN (_, (CALL _))) -> (assert false)
	| (STMT _) | (ASGN _) ->(
            (append_notes note);
	    (breakup_stmt_body ctx tl label note lastnote (stmts @ [s]))
        )
  (* Memo : RESET - ANLAB *)
    | (RESET _) -> (
        (begin
          let rec find_block_reset (stmts : Cexpr.stmt_t list)
                                     (result : Cexpr.stmt_t list) :
                                     (Cexpr.stmt_t list * Cexpr.stmt_t list) = (
              match stmts with
              | ((RESET _) as s) :: tl ->
                  (find_block_reset tl (result@[s]))
              | _ -> result, stmts

          )in
          let reset,new_tl = find_block_reset l [] in
          let newlabel0 = (Cexpr.make_new_label ()) in
          let newlabel1 = (Cexpr.make_new_label ()) in
          (breakup_emit_block ctx BBSTMT label !notes stmts (GOTO newlabel0)) ;
          (breakup_emit_block ctx BBASSIGN newlabel0 [lastnote] (reset) (GOTO newlabel1));
          (breakup_stmt_body ctx new_tl newlabel1 lastnote lastnote [])
(*          (breakup_emit_block ctx BBASSIGN label lastnote ([s]@reset)
          (GOTO newlabel)) ;
          (breakup_stmt_body ctx new_tl newlabel lastnote lastnote [])*)
        end)
    )
  (* Memo : ASSUME - ANLAB *)
  | (ASSUME _) ->
        (begin
          let rec find_block_assumes (stmts : Cexpr.stmt_t list)
                                     (result : Cexpr.stmt_t list) :
                                     (Cexpr.stmt_t list * Cexpr.stmt_t list) = (
              match stmts with
              | ((ASSUME _) as s) :: tl ->
                  (find_block_assumes tl (result@[s]))
              | _ -> result, stmts

          )in
          let ass,new_tl = find_block_assumes l [] in
          let newlabel0 = (Cexpr.make_new_label ()) in
          let newlabel1 = (Cexpr.make_new_label ()) in
          (breakup_emit_block ctx BBSTMT label !notes stmts (GOTO newlabel0)) ;
          (breakup_emit_block ctx BBASME newlabel0 [lastnote] (ass) (GOTO newlabel1));
          (breakup_stmt_body ctx new_tl newlabel1 lastnote lastnote [])

(*          (breakup_emit_block ctx BBASME label lastnote ([s]@ass)
          (GOTO newlabel)) ;
          (breakup_stmt_body ctx new_tl newlabel lastnote lastnote [])*)
        end)
	| (CALLASGN ((CALL _), _, rs)) ->
	    (* MEMO: It emits an empty block when stmts is empty.
	       Skipping emitting an empty block makes it work badly,
	       making some used variables become NIL in redundant
	       variable removal.  Reason unknown. *) (*AHO*)
        if (List.length rs) > 0 then
	    (begin
	      let newlabel0 = (Cexpr.make_new_label ()) in
	      let newlabel1 = (Cexpr.make_new_label ()) in
	      let newlabel2 = (Cexpr.make_new_label ()) in
	      (*(breakup_emit_block ctx BBSTMT label note stmts
		  (GOTO newlabel0)) ;
	      (breakup_emit_block ctx BBCALL newlabel0 lastnote [s]
		  (GOTO newlabel1)) ;*)
          (breakup_emit_block ctx BBSTMT label !notes stmts
		  (GOTO newlabel0)) ;
	      (breakup_emit_block ctx BBCALL newlabel0 [lastnote] [s]
		  (GOTO newlabel1)) ;
	      (breakup_emit_block ctx BBASSIGN newlabel1 [lastnote] [s]
		  (GOTO newlabel2)) ;
          (breakup_stmt_body ctx tl newlabel2 lastnote lastnote [])
	      (*(breakup_stmt_body ctx tl newlabel1 lastnote lastnote [])*)
	    end)
      else (begin (* rs is empty -- ANLAB *)
          let newlabel0 = (Cexpr.make_new_label ()) in
          let newlabel1 = (Cexpr.make_new_label ()) in
          (breakup_emit_block ctx BBSTMT label !notes stmts
                              (GOTO newlabel0)) ;
          (breakup_emit_block ctx BBCALL newlabel0 [lastnote] [s]
                              (GOTO newlabel1)) ;
          (breakup_stmt_body ctx tl newlabel1 lastnote lastnote [])
      end)
	| (CALLASGN _) -> (assert false)
	| (IF (_, [(GOTO _)], [(GOTO _)])) ->
	    if (not (stmts == [])) then
	      (begin
		let newlabel = (Cexpr.make_new_label ()) in
		(breakup_emit_block ctx BBSTMT label !notes stmts
		    (GOTO newlabel)) ;
		(breakup_emit_block ctx BBCOND newlabel [note] [] s) ;
		(breakup_stmt_body_newblock ctx lastnote tl)
	      end)
	    else
	      (begin
		(breakup_emit_block ctx BBCOND label [note] [] s) ;
		(breakup_stmt_body_newblock ctx lastnote tl)
	      end)
	| (IF _) -> (assert false)
	| (BEGIN _) -> (assert false)
	| (SWITCH _) -> (assert false)
	| (DOWHILE _) -> (assert false)
	| (FOR _) -> (assert false)
	| (BREAK) -> (assert false)
	| (CONTINUE) -> (assert false)
	| (LABEL (NAME_LABEL nextlabel)) ->
	    (breakup_emit_block ctx BBSTMT label !notes stmts
		(GOTO nextlabel)) ;
	    (breakup_stmt_body ctx tl nextlabel lastnote lastnote [])
	| (LABEL _) -> (assert false)
	| (GOTO n) ->
	    (breakup_emit_block ctx BBSTMT label !notes stmts s) ;
	    (breakup_stmt_body_newblock ctx lastnote tl)
	| (RETURN _) -> (assert false)
	| (ANNOTATION _) ->
	    (breakup_stmt_body ctx tl label note lastnote (stmts @ [s]))
	| (ASM e) ->
	    (breakup_stmt_body ctx tl label note lastnote (stmts @ [s]))
(*	| ((NOTE (_, _, _, _)) as newnote) ->
	    (breakup_stmt_body ctx tl label note newnote stmts)*)
	| ((NOTE (LINE(true, _, _, _))) as newnote)
        | ((NOTE (RANGE _)) as newnote) ->
            (* FIXME: may need to consider more *)
	    (breakup_stmt_body ctx tl label newnote newnote stmts)
	| (NOTE _) ->
	    (breakup_stmt_body ctx tl label note lastnote stmts)
      end))

(* Skips statements (possibly notes) after goto. *)

and breakup_stmt_body_newblock (ctx : (string, stmt_block_t) Hashtbl.t) (note : stmt_t) (l : stmt_t list) : unit = (
  let pmsg (skipped : stmt_t) : unit = (
    (warning ("Some statement ignored ("
		     ^ (Cexpr.stmt_to_string skipped) ^ ")."))) in
  match l with
    [] ->
      (* Never should come here. *)
      (assert false)
  | (s :: tl) ->
      (begin
	match s with
	  (LABEL (NAME_LABEL nextlabel)) ->
	    (breakup_stmt_body ctx tl nextlabel note note [])
	| ((NOTE (LINE(true, _, _, _))) as newnote)
	| ((NOTE (RANGE _ )) as newnote) ->	       (* FIXME: Dzungdk uhmm *)
	    (breakup_stmt_body_newblock ctx newnote tl)
	| (NOTE _) ->
	    (breakup_stmt_body_newblock ctx note tl)
	| (GOTO _) ->
	    (*(pmsg s);*)
	    (breakup_stmt_body_newblock ctx note tl)
	| _ ->
	    if (true) then
	      (begin
		(pmsg s) ;
		(breakup_stmt_body_newblock ctx note tl)
	      end)
	    else
	      (failwith ("Unexecuted statement after goto: "
			  ^ (stmt_to_string s)))
      end))

and note_string (note : stmt_t) = (
  match note with
    (NOTE (LINE(true, s, l, k))) ->
      ("\"" ^ s ^ " " ^ (string_of_int l) ^ " (" ^ (string_of_int k) ^ ")\"")
  | _ -> (assert false))

and breakup_emit_block (bbset : (string, stmt_block_t) Hashtbl.t) (bb : block_type_t) (label : string) (notes : stmt_t list) (stmts : stmt_t list) (jump : stmt_t) : unit = (
(*  let memo = (note_string note) in (* Binh *)*)
  let memo = notes in (* Binh *)
  (clean_notes ());

(*  Gprint.eprint (note, stmts); Scommon.pause();*)
  (* First element should be a label, which names this BB. *)
  match jump with
    (GOTO nextlabel) ->
      (begin
	assert (bb == BBHEAD || bb == BBTAIL || bb == BBSTMT || bb == BBCALL || bb == BBASME || bb == BBASSIGN) ;
	let block = {
	  blk_type = bb ;
	  blk_label = label ;
	  blk_memo = memo ;
	  blk_outgoing_labels = [nextlabel] ;
	  blk_incoming_labels = [] ;
	  blk_stmts = stmts ;
	  blk_phis = [] ;
	  blk_phifix = [] ;
	  blk_callfix = [] ;
	  blk_inputs = (StringInt2Set.empty) ;
	  blk_outputs = (StringInt2Set.empty) ;
	  blk_rdvars = (StringInt2Set.empty) ;
	  blk_wrvars = (StringInt2Set.empty) ;
	  blk_livevars = (Hashtbl.create 10) ;
	  blk_phivars = (Hashtbl.create 10) ;
	  blk_info_rdvars = [] ;
	  blk_info_wrvars = []
	  (*blk_info_varrds = (StringInt2Set.empty)*)
	} in
	assert (not (Hashtbl.mem bbset label)) ;
	(Hashtbl.replace bbset label block)
      end)
  | (IF (e1, [(GOTO thenlabel)], [(GOTO elselabel)])) ->
      (begin
	assert (bb == BBCOND) ;
	assert (stmts == []) ;
	let block = {
	  blk_type = BBCOND ;
	  blk_label = label ;
	  blk_memo = memo ;
	  blk_outgoing_labels = [thenlabel ; elselabel] ;
	  blk_incoming_labels = [] ;
	  blk_stmts = [jump] ;
	  blk_phis = [] ;
	  blk_phifix = [] ;
	  blk_callfix = [] ;
	  blk_inputs = (StringInt2Set.empty) ;
	  blk_outputs = (StringInt2Set.empty) ;
	  blk_rdvars = (StringInt2Set.empty) ;
	  blk_wrvars = (StringInt2Set.empty) ;
	  blk_livevars = (Hashtbl.create 10) ;
	  blk_phivars = (Hashtbl.create 10) ;
	  blk_info_rdvars = [] ;
	  blk_info_wrvars = []
	  (*blk_info_varrds = (StringInt2Set.empty)*)
	} in
	assert (not (Hashtbl.mem bbset label)) ;
	(Hashtbl.replace bbset label block) ;
      end)
  | (STMT NIL) ->
      (begin
	(* Procedure end (stmts includes EPILOGUE and nonempty). *)
	assert (bb == BBSTMT) ;
	assert (epilogue_statement_p stmts) ;
	let block = {
	  blk_type = BBTAIL ;
	  blk_label = label ;
	  blk_memo = memo ;
	  blk_outgoing_labels = [pseudo_elsewhere_label] ;
	  blk_incoming_labels = [] ;
	  blk_stmts = stmts ;
	  blk_phis = [] ;
	  blk_phifix = [] ;
	  blk_callfix = [] ;
	  blk_inputs = (StringInt2Set.empty) ;
	  blk_outputs = (StringInt2Set.empty) ;
	  blk_rdvars = (StringInt2Set.empty) ;
	  blk_wrvars = (StringInt2Set.empty) ;
	  blk_livevars = (Hashtbl.create 10) ;
	  blk_phivars = (Hashtbl.create 10) ;
	  blk_info_rdvars = [] ;
	  blk_info_wrvars = []
	  (*blk_info_varrds = (StringInt2Set.empty)*)
	} in
	assert (not (Hashtbl.mem bbset label)) ;
	(Hashtbl.replace bbset label block) ;
      end)
  | _ -> (assert false))

(* ================================================================ *)

(* EMPTY BLOCK REMOVAL *)

(* Checks if a block is empty. *)

let block_empty_p (b : stmt_block_t) : bool = (
  match b with
    {
      blk_type = BBSTMT ;
      blk_label = _ ;
      blk_memo = _ ;
      blk_outgoing_labels = _ ;
      blk_incoming_labels = _ ;
      blk_stmts = stmts ;
      blk_phis = [] ;
      blk_phifix = []
    } ->
      (stmts == [])
  | _ ->
      false)

(* Follows a mapping to a transitive limit.  It is used to skip empty
   blocks.  It returns (_, true) at limit of the follow count
   (indicating an infinite loop).  *)

let rec follow_map (map : ('a, 'a) Hashtbl.t) (k : 'a) (limit : int) : ('a * bool) = (
  if (limit == 0) then
    (k, true)
  else if (Hashtbl.mem map k) then
    (follow_map map (Hashtbl.find map k) (limit - 1))
  else
    (k, false))

(* Removes empty blocks, and redirects gotoes. *)

let rec collapse_blocks (bbset : (string, stmt_block_t) Hashtbl.t) (referredlabels : (string * string) list) : unit = (
  let incomemap : (string, (string list)) Hashtbl.t = (Hashtbl.create 10) in
  let forwarding : (string, string) Hashtbl.t =
    (find_empty_blocks bbset referredlabels) in
  (* Relabel gotos. *)
  (begin
    (* Do update only on non-empty blocks. *)
    let nonforwardinglabels =
      (List.filter
	  (fun s -> (not (Hashtbl.mem forwarding s)))
	  (Misc.hashtbl_keys bbset)) in
    (List.iter
	(fun label ->
	  let b = (Hashtbl.find bbset label) in
	  (* MEMO: No junction block yet. *)
	  assert (not (b.blk_type == BBJUNC)) ;
	  match b with
	    {
	      blk_type = (BBHEAD|BBTAIL|BBSTMT|BBCALL|BBASME|BBASSIGN) ;
	      blk_outgoing_labels = [outgoing]
	    } ->
	      (begin
		let newgoing = (check_forwarding forwarding outgoing) in
		if (not (string_equal newgoing outgoing)) then
		  (begin
		    let newblk = {
		      b with blk_outgoing_labels = [newgoing]
		    } in
		    (Hashtbl.replace bbset label newblk)
		  end) ;
		(add_incoming_to_follower incomemap label newgoing) ;
	      end)
	  | {
	      blk_type = BBCOND ;
	      blk_outgoing_labels = [thenout ; elseout] ;
	      blk_stmts = [(IF (condexpr, _, _))]
	    } ->
	      (begin
		let thennew = (check_forwarding forwarding thenout) in
		let elsenew = (check_forwarding forwarding elseout) in
		if (not ((string_equal thennew thenout)
			  && (string_equal elsenew elseout))) then
		  (begin
		    let jump = (IF (condexpr,
				   [(GOTO thennew)],
				   [(GOTO elsenew)])) in
		    let newblk = {
		      b with
			blk_outgoing_labels = [thennew ; elsenew] ;
			blk_stmts = [jump]
		    } in
		    (Hashtbl.replace bbset label newblk) ;
		  end) ;
		(add_incoming_to_follower incomemap label thennew) ;
		(add_incoming_to_follower incomemap label elsenew)
	      end)
	  | _ -> (assert false))
	nonforwardinglabels) ;
    let removelabels = (Misc.hashtbl_keys forwarding) in
    (List.iter (Hashtbl.remove bbset) removelabels) ;
  end) ;
  (* Remove blocks with no incomings. *)
  (remove_blocks_no_incomings bbset incomemap) ;
  (* Replace or insert a return block. *)
  (relabel_return_block bbset incomemap) ;
  (* Split junction (phi) nodes. *)
  (split_junction_blocks bbset incomemap) ;
  (* Update incoming labels. *)
  (List.iter
      (fun label ->
	if (not (string_equal label pseudo_entry_label)) then
	  (begin
	    let b = (Hashtbl.find bbset label) in
	    let incomings = (Hashtbl.find incomemap label) in
	    let newblk = {
	      b with blk_incoming_labels = incomings
	    } in
	    (Hashtbl.replace bbset label newblk)
	  end))
      (Misc.hashtbl_keys bbset)))

(* Removes blocks with no incomings, and repeats while some are
   removed.  Unreachable return is fatal.  MEMO: The entry node
   {pseudo_entry_label} has no incomings. *)

and remove_blocks_no_incomings (bbset : (string, stmt_block_t) Hashtbl.t) (incomemap : (string, (string list)) Hashtbl.t) : unit = (
  let noincomings = (List.filter
			(fun label ->
			  ((not (string_equal label pseudo_entry_label))
			    && (not (Hashtbl.mem incomemap label))))
			(Misc.hashtbl_keys bbset)) in
  if (not (noincomings == [])) then
    (begin
      (List.iter
	  (fun label ->
	    assert (not (string_equal label pseudo_entry_label)) ;
	    (* No pseudo_return_label yet. *)
	    let b = (Hashtbl.find bbset label) in
	    if (b.blk_type == BBTAIL) then
	      (begin
		(* Unreachable return block; keep it. *)
		(warning_message
		    (";; Unreachable return")
		    ""(*b.blk_memo*)) ;
		(Hashtbl.replace incomemap label [pseudo_elsewhere_label])
	      end)
	    else
	      (begin
		(warning_message
		    (";; Remove unreachable block: '" ^ label ^ "'")
		    ""(* b.blk_memo*)) ;
		(Hashtbl.remove bbset label) ;
		(List.iter
		    (fun follower ->
		      (remove_incoming_from_follower incomemap
			  label follower))
		    b.blk_outgoing_labels)
	      end))
	  noincomings) ;
      (* Do it again if non-empty. *)
      (remove_blocks_no_incomings bbset incomemap)
    end))

and relabel_return_block (bbset : (string, stmt_block_t) Hashtbl.t) (incomemap : (string, (string list)) Hashtbl.t) : unit = (
  assert (not (Hashtbl.mem incomemap pseudo_return_label)) ;
  (List.iter
      (fun label ->
	assert (not (string_equal label pseudo_return_label)) ;
	let b = (Hashtbl.find bbset label) in
	if (b.blk_type == BBTAIL) then
	  (begin
	    assert (b.blk_type == BBTAIL) ;
	    assert (epilogue_statement_p b.blk_stmts) ;
	    assert (string_list_equal b.blk_outgoing_labels
		       [pseudo_elsewhere_label]) ;
	    let incomings = (Hashtbl.find incomemap label) in
	    if ((List.length incomings) == 1) then
	      (begin
		(* Rename this block as return block. *)
		let modblk = {
		  b with
		    blk_type = BBTAIL ;
		    blk_label = pseudo_return_label ;
		} in
		(Hashtbl.remove bbset label) ;
		(Hashtbl.replace bbset pseudo_return_label modblk) ;
		let prevlabel = (List.hd incomings) in
		if (string_equal prevlabel pseudo_elsewhere_label) then
		  (* Do nothing for an unreachable return block. *)
		  ()
		else
		  (begin
		    let prev = (Hashtbl.find bbset prevlabel) in
		    assert (List.mem label prev.blk_outgoing_labels) ;
		    let outlabels =
		      (List.map
			  (fun s ->
			    if (string_equal s label) then
			      pseudo_return_label
			    else
			      s)
			  prev.blk_outgoing_labels) in
		    let modprev = {
		      prev with
			blk_outgoing_labels = outlabels ;
		    } in
		    (Hashtbl.remove bbset prev.blk_label) ;
		    (Hashtbl.replace bbset prev.blk_label modprev)
		  end) ;
		assert (not (Hashtbl.mem incomemap pseudo_return_label)) ;
		(Hashtbl.remove incomemap label) ;
		(Hashtbl.replace incomemap pseudo_return_label [prevlabel]) ;
	      end)
	    else
	      (begin
		(* Add a return block; this block will be a junction. *)
		let modblk = {
		  b with
		    blk_type = BBSTMT ;
		    blk_outgoing_labels = [pseudo_return_label] ;
		    blk_stmts = []
		} in
		(Hashtbl.remove bbset label) ;
		(Hashtbl.replace bbset label modblk) ;
		let newblk = {
		  b with
		    blk_type = BBTAIL ;
		    blk_label = pseudo_return_label ;
		    blk_stmts = [(STMT PSEUDO_EPILOGUE)] ;
		    blk_livevars = (Hashtbl.create 10) ;
		    blk_phivars = (Hashtbl.create 10)
		} in
		(Hashtbl.remove bbset pseudo_return_label) ;
		(Hashtbl.replace bbset pseudo_return_label newblk) ;
		(add_incoming_to_follower incomemap label pseudo_return_label)
	      end)
	  end))
      (Misc.hashtbl_keys bbset)))

(* Splits junction (phi) blocks. *)

and split_junction_blocks (bbset : (string, stmt_block_t) Hashtbl.t) (incomemap : (string, (string list)) Hashtbl.t) : unit = (
  (List.iter
      (fun label ->
	if (not (string_equal label pseudo_entry_label)) then
	  (begin
	    let b = (Hashtbl.find bbset label) in
	    let incomings = (Hashtbl.find incomemap label) in
	    if ((List.length incomings) > 1) then
	      (begin
		if (b.blk_stmts == []) then
		  (begin
		    (* Special case for immediately before return block. *)
		    assert (string_list_equal b.blk_outgoing_labels
			       [pseudo_return_label]) ;
		    let phiblk = {
		      b with
			blk_type = BBJUNC ;
			blk_outgoing_labels = b.blk_outgoing_labels ;
			blk_stmts = []
		    } in
		    (Hashtbl.remove bbset label) ;
		    (Hashtbl.replace bbset label phiblk) ;
		  end)
		else
		  (begin
		    let newlabel = (Cexpr.make_new_label ()) in
		    let phiblk = {
		      b with
			blk_type = BBJUNC ;
			blk_outgoing_labels = [newlabel] ;
			blk_stmts = []
		    } in
		    (Hashtbl.remove bbset label) ;
		    (Hashtbl.replace bbset label phiblk) ;
		    let newblk = {
		      b with
			blk_label = newlabel ;
			blk_livevars = (Hashtbl.create 10) ;
			blk_phivars = (Hashtbl.create 10)
		    } in
		    (Hashtbl.replace bbset newlabel newblk) ;
		    (add_incoming_to_follower incomemap label newlabel) ;
		    (List.iter
			(fun going ->
			  (remove_incoming_from_follower incomemap label going) ;
			  (add_incoming_to_follower incomemap newlabel going))
			b.blk_outgoing_labels)
		  end)
	      end)
	  end))
      (Misc.hashtbl_keys bbset)))

(* Replaces a label to a forwarding one, when it exists. *)

and check_forwarding (forwarding : (string, string) Hashtbl.t) (label : string) : string = (
    if (Hashtbl.mem forwarding label) then
      (Hashtbl.find forwarding label)
    else
      label)

(* Add a label to the follower's incoming list. *)

and add_incoming_to_follower (incomemap : (string, (string list)) Hashtbl.t) (label : string) (follower : string) : unit = (
  if (not (string_equal follower pseudo_elsewhere_label)) then
    (begin
      let incomings =
	if (Hashtbl.mem incomemap follower) then
	  (Hashtbl.find incomemap follower)
	else
	  [] in
      if (not (List.mem label incomings)) then
	(Hashtbl.replace incomemap follower (label :: incomings))
    end))

and remove_incoming_from_follower (incomemap : (string, (string list)) Hashtbl.t) (label : string) (follower : string) : unit = (
  if (not (string_equal follower pseudo_elsewhere_label)) then
    (begin
      let incomings =
	if (Hashtbl.mem incomemap follower) then
	  (Hashtbl.find incomemap follower)
	else
	  [] in
      let newincomings = (List.filter
			     (fun l -> (not (string_equal l label)))
			     incomings) in
      if (newincomings == []) then
	(Hashtbl.remove incomemap follower)
      else
	(Hashtbl.replace incomemap follower newincomings)
    end))

(* Finds empty blocks, makes a forwarding mapping for them, and
   returns the mapping (as a hashtable).  It does not remove blocks
   with a label whose value is taken. *)

and find_empty_blocks (bbset : (string, stmt_block_t) Hashtbl.t) (referredlabels : (string * string) list) : (string, string) Hashtbl.t = (
  let labelpairs0 : String2Set.t =
    (Hashtbl.fold
	(fun l b s ->
	  if ((block_empty_p b)
	       && (not (List.mem_assoc l referredlabels))) then
	    (begin
	      match b with
		{
		  blk_type = BBSTMT ;
		  blk_label = label ;
		  blk_outgoing_labels = [outgoing]
		} ->
		  if (not (Cexpr.temporary_label_p label)) then
		    (warning_message
			("IGNORE LABEL: '" ^ label ^ "'") ""(*b.blk_memo*)) ;
		  (String2Set.add (label, outgoing) s)
	      | _ -> (assert false)
	    end)
	  else
	    s)
	bbset (String2Set.empty)) in
  let labelpairs = labelpairs0 in
  (* Make a forwarding map. *)
  let map1 = (Hashtbl.create 10) in
  (List.iter
      (fun (l0, l1) ->
	assert (not (Hashtbl.mem map1 l0)) ;
	(Hashtbl.replace map1 l0 l1))
      (String2Set.elements labelpairs)) ;
  let map2 = (Hashtbl.create 10) in
  (List.iter
      (fun (l0, l1) ->
	let (follower, looping) = (follow_map map1 l1 20) in
	if (not looping) then
	  (begin
	    assert (not (Hashtbl.mem map2 l0)) ;
	    (Hashtbl.replace map2 l0 follower)
	  end)
	else if (looping) then
	  (begin
	    (* Possible looping of empty blocks, keep it. *)
	    (*AHO*) (* Replace stmt with assert(0) for a failure mark. *)
	    let b = (Hashtbl.find bbset l0) in
	    (warning_message
		(";; Possible looping of empty blocks: (" ^ l0 ^ ")")
		""(*b.blk_memo*)) ;
	    assert (b.blk_stmts == []) ;
	    let newblk = {
	      b with blk_stmts = [(ANNOTATION
				      ("$assert", Cexpr.constant_zero))] ;
	    } in
	    (Hashtbl.replace bbset l0 newblk)
	  end))
      (String2Set.elements labelpairs)) ;
  map2)

and find_forwarding_source (tolabel : string) (labelpairs : (string * string) list) : string = (
  match labelpairs with
    [] ->
      ""
  | ((l0, l1) :: tl) ->
      if (string_equal l1 tolabel) then
	l0
      else
	(find_forwarding_source tolabel tl))

(* ================================================================ *)

(* LISTING VARIABLE USAGE (FOR SSA-LIKE TRANSFORMATION) *)

(* NOTE: Globals are assigned sequence numbers (versions) for each
   function.  They are zero at entry. *)

(* Marks reads on variables in {blk_varrds}.  They are live variables.
   They do not include temporaries and write-then-read variables.
   {propagate_variable_usage} propagates usage to include uses of the
   following blocks.  In addition, it fills {vartypes}, which maps
   variables to types.  It is called before SSA yet. *)

let rec find_variable_usage (bbset : (string, stmt_block_t) Hashtbl.t) (vartypes : (varid_t, typenode_t) Hashtbl.t) : unit = (
  let alllabels = (Misc.hashtbl_keys bbset) in
  (List.iter
      (fun label ->
	let b = (Hashtbl.find bbset label) in
	assert (string_equal b.blk_label label) ;
	let (rds, wrs) = (find_variable_usage_stmts vartypes b.blk_stmts
			     ((StringInt2Set.empty), (StringInt2Set.empty))) in
	b.blk_rdvars <- rds ;
	b.blk_wrvars <- wrs ;
	b.blk_info_rdvars <- (StringInt2Set.elements b.blk_rdvars) ;
	b.blk_info_wrvars <- (StringInt2Set.elements b.blk_wrvars))
      alllabels))

(* NOTE: Assignment to a structure member (for a variable) is treated
   as writes to a whole variable. *)

and find_variable_usage_stmts (vartypes : (varid_t, typenode_t) Hashtbl.t) (l : stmt_t list) ((rds, wrs) : (StringInt2Set.t * StringInt2Set.t)) : (StringInt2Set.t * StringInt2Set.t) = (
  match l with
    [] ->
      (rds, wrs)
  | (s :: tl) ->
      (begin
	match s with
	  (DEFVAR _) | (DEFUN _) | (SPEC _) -> (assert false)
	| (STMT (CALL _)) -> (assert false)
        (* ANLAB memo : RESET *)
    | (RESET es) ->(
                List.hd(List.map(
                    fun e0 ->
                        let (rds1, wrs1) =
                           (variable_usage_expr vartypes e0 (rds, wrs)) in
                           (find_variable_usage_stmts vartypes tl (rds1, wrs1))
                )es)
         )
        (* ANLAB memo : ASSUME *)
    | (ASSUME e0) -> (
	    let (rds1, wrs1) =
	      (variable_usage_expr vartypes e0 (rds, wrs)) in
	    (find_variable_usage_stmts vartypes tl (rds1, wrs1))
         )
	| (STMT e0) ->
	    let (rds1, wrs1) =
	      (variable_usage_expr vartypes e0 (rds, wrs)) in
	    (find_variable_usage_stmts vartypes tl (rds1, wrs1))
	| (ASGN (_, (CALL _))) -> (assert false)
	| (ASGN (((VREF (tn, id, level, 0)) as lval), rval)) ->
	    (find_variable_usage_assign_variable vartypes lval rval
		tl (rds, wrs))
	| (ASGN ((PREF (tn, _, e0, e1)), rval)) ->
	    let (rds1, wrs1) =
	      (Misc.list_fold
		  (variable_usage_expr vartypes)
		  [e0 ; e1 ; rval] (rds, wrs)) in
	    (find_variable_usage_stmts vartypes tl (rds1, wrs1))
	| (ASGN ((SREF (tn, deref, e0, n)), rval)) ->
	    (begin
	      let e1 = (ignore_cast e0) in
	      match e1 with
		(VREF (vtn, id, level, 0)) ->
		  (find_variable_usage_assign_variable vartypes e1 rval
		      tl (rds, wrs))
	      | _ ->
		  let (rds1, wrs1) =
		    (Misc.list_fold
			(variable_usage_expr vartypes)
			[e0 ; rval] (rds, wrs)) in
		  (find_variable_usage_stmts vartypes tl (rds1, wrs1))
	    end)
	| (ASGN (_, rval)) -> (assert false)
	| (CALLASGN ((CALL (tn, e0, ee, spec_op)), assignments, _)) ->
	    let (rds1, wrs1) =
	      (Misc.list_fold
		  (variable_usage_expr vartypes)
		  (e0 :: ee) (rds, wrs)) in
	    let (rds2, wrs2) =
	      (find_variable_usage_stmts vartypes assignments (rds1, wrs1)) in
	    (find_variable_usage_stmts vartypes tl (rds2, wrs2))
	| (CALLASGN _) -> (assert false)
	| (IF (e0, [(GOTO _)], [(GOTO _)])) ->
	    let (rds1, wrs1) =
	      (variable_usage_expr vartypes e0 (rds, wrs)) in
	    (find_variable_usage_stmts vartypes tl (rds1, wrs1))
	| IF _ -> (assert false)
	| BEGIN _ | SWITCH _ | DOWHILE _ | FOR _ -> (assert false)
	| BREAK | CONTINUE | LABEL _ | GOTO _ | RETURN _ -> (assert false)
	| (ANNOTATION (s, e0)) ->
	    let (rds1, wrs1) =
	      (variable_usage_expr vartypes e0 (rds, wrs)) in
	    (find_variable_usage_stmts vartypes tl (rds1, wrs1))
	| (ASM e0) ->
	    (failwith "NO ASM YET")
	| NOTE _ -> (assert false)
      end))

and find_variable_usage_assign_variable (vartypes : (varid_t, typenode_t) Hashtbl.t) (lval : expr_t) (rval : expr_t) (l : stmt_t list) ((rds, wrs) : (StringInt2Set.t * StringInt2Set.t)) : (StringInt2Set.t * StringInt2Set.t) = (
  match lval with
    (VREF (tn, id, level, 0)) ->
      (Hashtbl.replace vartypes (id, level) tn) ;
      let (rds1, wrs1) = (variable_usage_expr vartypes rval (rds, wrs)) in
      if (temporary_variable_p lval) then
	(find_variable_usage_stmts vartypes l (rds1, wrs1))
      else if (function_reference_p lval) then
	(find_variable_usage_stmts vartypes l (rds1, wrs1))
      else
	let wrs2 = (StringInt2Set.add (id, level, 0) wrs1) in
	(find_variable_usage_stmts vartypes l (rds1, wrs2))
  | _ -> (assert false))

(* NOTE: Variables read after written are not marked as read, to make
   flow analyzer ignore it in read-set. *)

and variable_usage_expr (vartypes : (varid_t, typenode_t) Hashtbl.t) (e : expr_t) ((rds, wrs) : (StringInt2Set.t * StringInt2Set.t)) : (StringInt2Set.t * StringInt2Set.t) = (
  match e with
    NIL ->
      (rds, wrs)
  | (PSEUDO_NONDET_VALUE (tn, _)) ->
      (rds, wrs)
  | (OPER (tn, op, e0, e1, e2)) ->
      (Misc.list_fold
	  (variable_usage_expr vartypes)
	  [e0 ; e1 ; e2] (rds, wrs))
  | (SETF _) -> (assert false)
  | (VREF (tn, id, level, 0)) ->
      (Hashtbl.replace vartypes (id, level) tn) ;
      if (temporary_variable_p e) then
	(rds, wrs)
      else if ((function_reference_p e) && (global_variable_p e)) then
	(* Ignore functions except function pointer variables. *)
	(rds, wrs)
      else if (enumerator_p tn id) then
        (* Exclude enumerator from variable set. *)
        (rds, wrs)
      else
	(begin
	  if (not (StringInt2Set.mem (id, level, 0) wrs)) then
	    ((StringInt2Set.add (id, level, 0) rds), wrs)
	  else
	    (rds, wrs)
	end)
  | (VREF _) -> (assert false)
  | (PREF (tn, _, e0, _)) ->
      (variable_usage_expr vartypes e0 (rds, wrs))
  | (SREF (tn, _, e0, s)) ->
      (variable_usage_expr vartypes e0 (rds, wrs))
  | (CALL (tn, e0, ee, _)) -> (assert false)
  | (ADDROF (tn, e0)) ->
      (variable_usage_addrof vartypes e0 (rds, wrs))
  | (SIZEOF (tn, tt)) | (PSEUDO_MEMORY (tn, tt)) ->
      (variable_usage_type vartypes tt (rds, wrs))
  | (INIT_FILL _) -> (assert false)
  | (PROGN (tn, l)) -> (assert false)
  | (CONST_I _) | (CONST_F _) | (STRING _) ->
      (rds, wrs)
  | (CLITERAL (tn, l)) -> (assert false)
  | (QUANTIFIER (s, vs, e0)) ->
      (variable_usage_expr vartypes e0 (rds, wrs))
  | (PSEUDO_PROLOGUE) ->
      (rds, wrs)
  | (PSEUDO_EPILOGUE) ->
      (rds, wrs)
  | (ASMSEQUENCE ee) -> (assert false)
  | (PHI _) -> (assert false)
  | (DSGN_AREF (tn, e0, e1)) -> (assert false)
  | (PRAGMA s) -> (assert false)
  | PSEUDO_ACSL_RESULT (tn, e0) ->
      (variable_usage_addrof vartypes e0 (rds, wrs))
  | PSEUDO_ACSL_OLD (tn, e0, e1) ->
      (Misc.list_fold (variable_usage_expr vartypes) [e0; e1] (rds, wrs))
  | ALL_FIELD _ -> assert false
  | PSEUDO_ACSL_UNIQUE (tn, e0) ->
      (variable_usage_addrof vartypes e0 (rds, wrs))
  | PSEUDO_ACSL_UNIQUEBLOCK (tn, e0, e1) ->
      (Misc.list_fold (variable_usage_expr vartypes) [e0; e1] (rds, wrs))
  | PSEUDO_ACSL_VALID (tn, e0) ->
      (variable_usage_addrof vartypes e0 (rds, wrs)))

and variable_usage_addrof (vartypes : (varid_t, typenode_t) Hashtbl.t) (ea : expr_t) ((rds, wrs) : (StringInt2Set.t * StringInt2Set.t)) : (StringInt2Set.t * StringInt2Set.t) = (
  let e = (ignore_cast ea) in
  match e with
    (VREF (tn, id, level, 0)) ->
      (rds, wrs)
  | (VREF _) -> (assert false)
  | (SREF (tn, deref, e0, s)) ->
      if (deref) then
	(variable_usage_expr vartypes e0 (rds, wrs))
      else
	(variable_usage_addrof vartypes e0 (rds, wrs))
  | (PREF (tn, deref, e0, _)) ->
      if (deref) then
	(variable_usage_expr vartypes e0 (rds, wrs))
      else
	(variable_usage_addrof vartypes e0 (rds, wrs))
  | _ ->
      (* NOT CERTAIN ABOUT OTHER CASES. *)
      (variable_usage_expr vartypes e (rds, wrs)))

and variable_usage_type (vartypes : (varid_t, typenode_t) Hashtbl.t) (tn : typenode_t) ((rds, wrs) : (StringInt2Set.t * StringInt2Set.t)) : (StringInt2Set.t * StringInt2Set.t) = (
  match (Cexpr.strip_off_pseudo_type tn true) with
    (ARRAY_OF (t0, Some e0)) ->
      let (rds1, wrs1) =
	(variable_usage_type vartypes t0 (rds, wrs)) in
      (variable_usage_expr vartypes e0 (rds1, wrs1))
  | (ARRAY_OF (t0, None)) ->
      (variable_usage_type vartypes t0 (rds, wrs))
  | _ ->
      (rds, wrs))

(* Collects read usage of variables in {blk_varrds} of all the
   following blocks, by taking an union of reads filtered by writes.
   It starts with the return label in favor of working backwards.
   NOTE: It collects assigned globals and forces them as read in the
   return block, which is later used to reflect updates to globals. *)

let rec propagate_variable_usage (bbset : (string, stmt_block_t) Hashtbl.t) : unit = (
  (collect_assigned_globals bbset) ;
  let alllabels = (Misc.hashtbl_keys bbset) in
  (propagate_variable_usage_loop bbset
      ([pseudo_return_label] @ alllabels)))

and collect_assigned_globals (bbset : (string, stmt_block_t) Hashtbl.t) : unit = (
  let globals : StringInt2Set.t =
    (Hashtbl.fold
	(fun label b vs ->
	  let wrs = b.blk_wrvars in
	  let wrs1 = (StringInt2Set.filter varid_global_p wrs) in
	  (*let rds = blk_rdvars in*)
	  (*let rds1 = (StringInt2Set.filter varid_global_p rds) in*)
	  (*(StringInt2Set.union wrs1 (StringInt2Set.union rds1 vs))*)
	  (StringInt2Set.union wrs1 vs))
	bbset (StringInt2Set.empty)) in
  let b = (Hashtbl.find bbset pseudo_return_label) in
  (b.blk_rdvars <- (StringInt2Set.union globals b.blk_rdvars)) ;
  (b.blk_info_rdvars <- (StringInt2Set.elements b.blk_rdvars)))

and propagate_variable_usage_loop (bbset : (string, stmt_block_t) Hashtbl.t) (labels : string list) : unit = (
  match labels with
    [] ->
      ()
  | (label :: tl) ->
      let b = (Hashtbl.find bbset label) in
      (*let rds = (StringInt2Set.diff b.blk_rdvars b.blk_wrvars) in*)
      let rds = b.blk_rdvars in
      let visits =
	(Misc.list_filtering_map
	    (fun prevlabel ->
	      if (string_equal prevlabel pseudo_elsewhere_label) then
		(begin
		  (* Do nothing for an unreachable return block. *)
		  assert (b.blk_type == BBTAIL) ;
		  None
		end)
	      else
		(begin
		  let prev = (Hashtbl.find bbset prevlabel) in
		  let rds1 = (StringInt2Set.diff rds prev.blk_wrvars) in
		  let newrds = (StringInt2Set.union prev.blk_rdvars rds1) in
		  if ((StringInt2Set.cardinal newrds)
			 != (StringInt2Set.cardinal prev.blk_rdvars)) then
		    (begin
		      (prev.blk_rdvars <- newrds) ;
		      (b.blk_info_rdvars <- (StringInt2Set.elements
						b.blk_rdvars)) ;
		      (Some prevlabel)
		    end)
		  else
		    None
		end))
	    b.blk_incoming_labels) in
      (propagate_variable_usage_loop bbset (tl @ visits)))

(* ================================================================ *)

(* TRANSFORMATION INTO SSA-LIKE FORMS (TEMPORARY BINDING ASSIGNMENT) *)

(* Binder of fresh variables. *)

type binder_t = {
  binder_varseq : (varid_t, int ref) Hashtbl.t
}

(* Returns a fresh sequence number for a variable. *)

let var_seq (ctx : binder_t) (id : string) (level : int) : int = (
  if (not (Hashtbl.mem ctx.binder_varseq (id, level))) then
    (Hashtbl.replace ctx.binder_varseq (id, level) (ref 1)) ;
  let seqref : int ref = (Hashtbl.find ctx.binder_varseq (id, level)) in
  let seq = (!seqref) in
  (incr seqref) ;
  seq)

(* Pseudoly inserts bindings for {$result}. *)

let pseudo_bind_result (livevars : (varid_t, expr_t) Hashtbl.t) (tn : typenode_t) : unit = (
  if (not (tn == VOID_TYPE)) then
    let id = pseudo_result_variable in
    let level = 2 in
    let seq = 0 in
    (Hashtbl.replace livevars (id, level) (VREF (tn, id, level, seq))))

(* Pseudoly inserts bindings for parameters. *)

let pseudo_bind_parameters (ctx : binder_t) (bbset : (string, stmt_block_t) Hashtbl.t) (decl : Cexpr.declaration_t) : unit = (
  let b = (Hashtbl.find bbset pseudo_entry_label) in
  let livevars = b.blk_livevars in
  let tn = (strip_off_pseudo_type (decl.Cexpr.decl_tn) true) in
  match (tn) with
    (FUNCTION_OF (tn0, None)) ->
      (pseudo_bind_result livevars tn0) ;
      ()
  | (FUNCTION_OF (tn0, (Some tl))) ->
      (pseudo_bind_result livevars tn0) ;
      (List.iter2
	  (fun (tn : typenode_t) (id : string) ->
	    if (tn != ELLIPSIS_TYPE) then
	      (begin
		let level = parameter_scope_level in
		let newseq = (var_seq ctx id level) in
		let newvar = (VREF (tn, id, level, newseq)) in
		(Hashtbl.replace livevars (id, level) newvar)
	      end)
	    else
	      (begin
		(* Skip ellipsis. *)
		assert (Cexpr.ellipsis_parameter_p id)
	      end))
	  tl
	  decl.Cexpr.decl_params)
  | _ -> (assert false))

let pseudo_bind_globals (ctx : binder_t) (bbset : (string, stmt_block_t) Hashtbl.t) (vartypes : (varid_t, typenode_t) Hashtbl.t) : unit = (
  let b = (Hashtbl.find bbset pseudo_entry_label) in
  let vars = (StringInt2Set.elements b.blk_rdvars) in
  let livevars = b.blk_livevars in
  (List.iter
      (fun (id, level, _) ->
	if (level == 1) then
	  (begin
	    let tn = (Hashtbl.find vartypes (id, level)) in
	    let newseq = (var_seq ctx id level) in
	    let newvar = (VREF (tn, id, level, newseq)) in
	    (Hashtbl.replace livevars (id, level) newvar)
	  end))
      vars))

let make_binder (bbset : (string, stmt_block_t) Hashtbl.t) : binder_t = (
  let ctx : binder_t = {
    binder_varseq = (Hashtbl.create 10) ;
  } in
  ctx)

(* Runs over non-junction blocks and assigns new bindings. *)

let rec bind_variables_forward (ctx : binder_t) (bbset : (string, stmt_block_t) Hashtbl.t) (vartypes : (varid_t, typenode_t) Hashtbl.t) (labels : string list) (visited : StringSet.t) : unit = (
  match labels with
    [] ->
      ()
  | (label :: tl) ->
      if (string_equal label pseudo_elsewhere_label) then
	(bind_variables_forward ctx bbset vartypes tl visited)
      else if (StringSet.mem label visited) then
	(bind_variables_forward ctx bbset vartypes tl visited)
      else
	(begin
	  let b = (Hashtbl.find bbset label) in
	  (bind_variables_straight ctx bbset b vartypes) ;
	  (bind_variables_forward ctx bbset vartypes
	      (b.blk_outgoing_labels @ tl) (StringSet.add label visited))
	end))

(* Assigns new bindings to variables for non-junction blocks. *)

and bind_variables_straight (ctx : binder_t) (bbset : (string, stmt_block_t) Hashtbl.t) (b : stmt_block_t) (vartypes : (varid_t, typenode_t) Hashtbl.t) : unit = (
  match (b.blk_type) with
    BBJUNC ->
      (* Ignore junction block. *)
      ()
  | BBHEAD ->
      assert (b.blk_incoming_labels == []) ;
      let newstmts = (bind_variables_stmts ctx b.blk_livevars b.blk_stmts) in
      let livevars = b.blk_livevars in
      let newblk = {
	b with
	  blk_stmts = newstmts ;
	  blk_outputs = (variable_list_to_set
			    (List.map variable_expr_to_name
				(Misc.hashtbl_values livevars)))
      } in
      (Hashtbl.replace bbset b.blk_label newblk)
  | BBTAIL | BBSTMT | BBCALL | BBCOND | BBASME | BBASSIGN->
      assert (List.length (b.blk_incoming_labels) == 1) ;
      let prevlabel = (List.hd b.blk_incoming_labels) in
      let prev = (Hashtbl.find bbset prevlabel) in
      let vars = (StringInt2Set.elements b.blk_rdvars) in
      (List.iter
	  (fun (id, level, _) ->
	    if (Hashtbl.mem prev.blk_livevars (id, level)) then
	      (Hashtbl.replace b.blk_livevars (id, level)
		  (Hashtbl.find prev.blk_livevars (id, level))))
	  vars) ;
      let newstmts = (bind_variables_stmts ctx b.blk_livevars b.blk_stmts) in
      (* Pseudoly assign globals for updates in the call. *)
      let callfix =
	if (not (b.blk_type == BBCALL)) then
	  []
	else
	  (Misc.list_filtering_map
	      (fun (id, level, sequence) ->
		if (level == Cexpr.global_scope_level) then
		  (begin
		    let oldvar =
		      (Hashtbl.find b.blk_livevars (id, level)) in
		    let tn = (Hashtbl.find vartypes (id, level)) in
		    let newseq = (var_seq ctx id level) in
		    let newvar = (VREF (tn, id, level, newseq)) in
		    (Hashtbl.replace b.blk_livevars (id, level) newvar) ;
		    (Some (ASGN (newvar, oldvar)))
		  end)
		else
		  None)
	      vars) in
      let newblk = {
	b with
	  blk_stmts = newstmts ;
	  blk_callfix = callfix ;
	  blk_outputs = (variable_list_to_set
			    (List.map variable_expr_to_name
				(Misc.hashtbl_values b.blk_livevars)))
      } in
      (Hashtbl.replace bbset b.blk_label newblk))

(* Assigns new bindings to variables for junction blocks.  It creates a
   temporarily variables for outputs of phi-functions for following
   references. *)

and bind_variables_arbitrary_junction (ctx : binder_t) (bbset : (string, stmt_block_t) Hashtbl.t) (vartypes : (varid_t, typenode_t) Hashtbl.t) : unit = (
  let alllabels = (Misc.hashtbl_keys bbset) in
  (List.iter
      (fun label ->
	let b = (Hashtbl.find bbset label) in
	if ((List.length b.blk_incoming_labels) > 1) then
	  (bind_variables_arbitrary_phi ctx bbset b vartypes))
      alllabels))

and bind_variables_arbitrary_phi (ctx : binder_t) (bbset : (string, stmt_block_t) Hashtbl.t) (b : stmt_block_t) (vartypes : (varid_t, typenode_t) Hashtbl.t) : unit = (
  let vars = (StringInt2Set.elements b.blk_rdvars) in
  (List.iter
      (fun (id, level, _) ->
	let newsequence = (var_seq ctx id level) in
	let tn = (Hashtbl.find vartypes (id, level)) in
	let v = (VREF (tn, id, level, newsequence)) in
	(Hashtbl.replace b.blk_livevars (id, level) v) ;
	(Hashtbl.replace b.blk_phivars (id, level) v))
      vars) ;
  let newstmts = (bind_variables_stmts ctx b.blk_livevars b.blk_stmts) in
  let newblk = { b with blk_stmts = newstmts } in
  (Hashtbl.replace bbset b.blk_label newblk))

and bind_variables_junction (ctx : binder_t) (bbset : (string, stmt_block_t) Hashtbl.t) (vartypes : (varid_t, typenode_t) Hashtbl.t) : unit = (
  let alllabels = (Misc.hashtbl_keys bbset) in
  (List.iter
      (fun label ->
	let b = (Hashtbl.find bbset label) in
	if ((List.length b.blk_incoming_labels) > 1) then
	  (bind_variables_phi ctx bbset b vartypes))
      alllabels))

and bind_variables_phi (ctx : binder_t) (bbset : (string, stmt_block_t) Hashtbl.t) (b : stmt_block_t) (vartypes : (varid_t, typenode_t) Hashtbl.t) : unit = (
  let vars = (Misc.hashtbl_keys b.blk_phivars) in
  let phis =
    (Misc.list_filtering_map
	(fun (id, level) ->
	  let phivar = (Hashtbl.find b.blk_phivars (id, level)) in
	  let tn = (Hashtbl.find vartypes (id, level)) in
	  let phimap =
	    (List.map
		(fun prevlabel ->
		  let prev = (Hashtbl.find bbset prevlabel) in
		  if (Hashtbl.mem prev.blk_livevars (id, level)) then
		    (begin
		      let v = (Hashtbl.find prev.blk_livevars (id, level)) in
		      (prevlabel, v)
		    end)
		  else
		    (begin
		      xDEBUG ("prev=(" ^ prevlabel ^ ") livevars="
			       ^ (String.concat ";"
				     (List.map
					 Cexpr.expr_to_string
					 (Misc.hashtbl_values
					     prev.blk_livevars)))) ;
		      xDEBUG ("curr=(" ^ b.blk_label ^ ") phivar="
			       ^ (Cexpr.expr_to_string phivar)) ;
		      (syntaxerror ("Reference to no-live variable: "
				     ^ (Cexpr.expr_to_string phivar)))
		    end))
		b.blk_incoming_labels) in
	  let phi = (bind_variables_uniqify_phi tn phimap) in
	  if (phi == NIL) then
	    None
	  else
	    (Some (ASGN (phivar, phi))))
	vars) in
  let outputs = (variable_list_to_set
		    (List.map variable_expr_to_name
			(Misc.hashtbl_values b.blk_livevars))) in
  let newblk = {
    b with
      blk_phis = phis ;
      blk_outputs = outputs
  } in
  (Hashtbl.replace bbset b.blk_label newblk))

(* Assigns temporary variables for phi-function output.  It is to
   avoid recurrence of the same variables in calculating state
   transitions. *)

and bind_temporaries_junction (ctx : binder_t) (bbset : (string, stmt_block_t) Hashtbl.t) : unit = (
  let alllabels = (Misc.hashtbl_keys bbset) in
  (List.iter
      (fun label ->
	let b = (Hashtbl.find bbset label) in
	if ((List.length b.blk_incoming_labels) > 1) then
	  (bind_temporaries_phi ctx bbset b))
      alllabels))

and bind_temporaries_phi (ctx : binder_t) (bbset : (string, stmt_block_t) Hashtbl.t) (b : stmt_block_t) : unit = (
  let modphis =
    (List.map
	(fun (s : stmt_t) ->
	  match s with
	    (ASGN (phivar, phi)) ->
	      (begin
		match phivar with
		  (VREF (tn, id, level, sequence)) ->
		    let newsequence = (var_seq ctx id level) in
		    let newvar = (VREF (tn, id, level, newsequence)) in
		    ((ASGN (newvar, phi)), (ASGN (phivar, newvar)))
		| _ -> (assert false)
	      end)
	  | _ -> (assert false))
	b.blk_phis) in
  let phis = (List.map fst modphis) in
  let phifix = (List.map snd modphis) in
  let newblk = {
    b with
      blk_phis = phis ;
      blk_phifix = phifix
  } in
  (Hashtbl.replace bbset b.blk_label newblk))

and bind_variables_uniqify_phi (tn : typenode_t) (phimap : (string * expr_t) list) : expr_t = (
  let uniqseqs =
    (List.fold_left
	(fun ss (label, v) ->
	  match v with
	    (VREF (tn, id, level, sequence)) ->
	      (IntSet.add sequence ss)
	  | NIL ->
	      ss
	  | _ -> (assert false))
	(IntSet.empty) phimap) in
  match (IntSet.cardinal uniqseqs) with
    0 ->
      NIL
  | 1 ->
      let (label, v) = (List.hd phimap) in
      v
  | _ ->
      (PHI (tn, phimap)))

(* Replaces variables references with live bindings and updates a
   mapping of current live variables.  NOTE: It does not bind
   variables in invariants. *)

and bind_variables_stmts (ctx : binder_t) (livevars : (varid_t, expr_t) Hashtbl.t) (stmts : stmt_t list) : stmt_t list = (
  (List.map
      (fun s ->
	match s with
	  DEFVAR _ | DEFUN _ | SPEC _ -> (assert false)
	| (STMT (CALL _)) -> (assert false)
    | RESET es -> (
        RESET(
            List.map(
                fun e0 ->
                    (bind_new_variable_on_vref_new ctx livevars e0)
            )es
        )
    )
    | ASSUME e0 -> (
            let x0 = (bind_new_variable_on_vref_new ctx livevars e0) in
            (ASSUME x0)

    )
	| (STMT e0) ->
	    let x0 = (bind_variables_exprs livevars e0) in
	    (STMT x0)
	| (ASGN (_, (CALL _))) -> (assert false)
	| (ASGN (lval, rval)) ->
	    (begin
	      match lval with
		(VREF (tn, id, level, 0)) ->
		  let newrval = (bind_variables_exprs livevars rval) in
		  let newlval = (bind_new_variable_on_vref ctx livevars lval) in
		  (ASGN (newlval, newrval))
	      | (PREF (tn, deref, e1, e2)) ->
		  let x1 = (bind_variables_exprs livevars e1)
		  and x2 = (bind_variables_exprs livevars e2) in
		  let newrval = (bind_variables_exprs livevars rval) in
		  (ASGN ((PREF (tn, deref, x1, x2)), newrval))
	      | (SREF (tn, deref, e1, n)) ->
		  if (variable_reference_p e1) then
		    (begin
		      let newrval = (bind_variables_exprs livevars rval) in
		      let newlval = (bind_new_variable_on_vref ctx livevars e1) in
		      (ASGN ((SREF (tn, deref, newlval, n)), newrval))
		    end)
		  else
		    (begin
		      let x1 = (bind_variables_exprs livevars e1) in
		      let newrval = (bind_variables_exprs livevars rval) in
		      (ASGN ((SREF (tn, deref, x1, n)), newrval))
		    end)
	      | _ -> (assert false)
	    end)
	| (CALLASGN ((CALL (tn, e0, ee, spec_op)), assignments, rs)) ->
	    (*assert (Cexpr.temporary_variable_p lval) ;*)
	    let x0 = (bind_variables_exprs livevars e0) in
	    let xx = (List.map (bind_variables_exprs livevars) ee) in
	    let ss = (bind_variables_stmts ctx livevars assignments) in
	    (CALLASGN ((CALL (tn, x0, xx, spec_op)), ss, rs))
	| (CALLASGN _) -> (assert false)
	| (IF (e0, go0, go1)) ->
	    let x0 = (bind_variables_exprs livevars e0) in
	    (IF (x0, go0, go1))
	| BEGIN _ | SWITCH _ | DOWHILE _ | FOR _ -> (assert false)
	| BREAK | CONTINUE | LABEL _ | GOTO _ | RETURN _ -> (assert false)
	| (ANNOTATION ("$invariant", e0)) ->
	    s
	| (ANNOTATION ("$check", e0)) ->
	    s
	| (ANNOTATION (s, e0)) ->
	    let x0 = (bind_variables_exprs livevars e0) in
	    (ANNOTATION (s, x0))
	| (ASM e0) ->
	    (failwith "NO ASM YET")
	| (NOTE _) -> (assert false))
      stmts))

and bind_new_variable_on_vref_new (ctx : binder_t)
                        (livevars : (varid_t, expr_t) Hashtbl.t)
                        (e : expr_t) : expr_t = (
  match e with
    (OPER (tn, c, e1, e2, e3)) ->
      let x1 = (bind_new_variable_on_vref_new ctx livevars e1) in
      (OPER (tn, c, x1, e2, e3))
  | (VREF (tn, id, level, 0)) ->
      if (temporary_variable_p e) then
        e
      else
        (begin
          let newseq = (var_seq ctx id level) in
          let newvar = (VREF (tn, id, level, newseq)) in
          (Hashtbl.replace livevars (id, level) newvar) ;
          newvar
        end)
  | (PREF (tn, deref, e0, e1)) ->
      let x0 = (bind_variables_exprs livevars e0)
      and x1 = (bind_variables_exprs livevars e1) in
      (PREF (tn, deref, x0, x1))
  | (SREF (tn, deref, e0, s)) ->(
      if (variable_reference_p e0) then (
          let x0 = (bind_new_variable_on_vref_new ctx livevars e0) in
          (SREF (tn, deref, x0, s))
      )
      else (
          let x0 = (bind_variables_exprs livevars e0) in
          (SREF (tn, deref, x0, s))
      )
  )
  | (ADDROF (tn, e0)) ->
      (*let x0 = (bind_variables_exprs livevars e0) in*)
      (ADDROF (tn, e0))
  | _ -> (assert false))

and bind_new_variable_on_vref (ctx : binder_t) (livevars : (varid_t, expr_t) Hashtbl.t) (e : expr_t) : expr_t = (
  match e with
    (OPER (tn, (CAST b), e1, _, _)) ->
      let x1 = (bind_new_variable_on_vref ctx livevars e1) in
      (OPER (tn, (CAST b), x1, NIL, NIL))
  | (VREF (tn, id, level, 0)) ->
      if (temporary_variable_p e) then
	e
      else
	(begin
	  let newseq = (var_seq ctx id level) in
	  let newvar = (VREF (tn, id, level, newseq)) in
	  (Hashtbl.replace livevars (id, level) newvar) ;
	  newvar
	end)
  | _ -> (assert false))

and bind_variables_exprs (livevars : (varid_t, expr_t) Hashtbl.t) (e : Cexpr.expr_t) = (
  match e with
  | NIL ->
      e
  | (PSEUDO_NONDET_VALUE (tn, _)) ->
      e
  | (OPER (tn, op, e0, e1, e2)) ->
      let x0 = (bind_variables_exprs livevars e0) in
      let x1 = (bind_variables_exprs livevars e1) in
      let x2 = (bind_variables_exprs livevars e2) in
      (OPER (tn, op, x0, x1, x2))
  | (SETF _) -> (assert false)
  | (VREF (tn, id, level, _)) ->
      if (Cexpr.temporary_variable_p e) then
	e
      else if (level == global_scope_level) then
	(begin
	  (* Assume globals are safely referenced. *)
	  if (Hashtbl.mem livevars (id, level)) then
	    (Hashtbl.find livevars (id, level))
	  else
	    e
	end)
      else if (enumerator_p tn id) then
	e
      else if (not (Hashtbl.mem livevars (id, level))) then
	(begin
	  xDEBUG ("livevars="
		   ^ (String.concat ";"
			 (List.map
			     expr_to_string
			     (Misc.hashtbl_values livevars)))) ;
	  (syntaxerror ("Reference to no-live variable: "
			 ^ (expr_to_string e)))
	end)
      else
	(begin
	  assert (Hashtbl.mem livevars (id, level)) ;
	  (Hashtbl.find livevars (id, level))
	end)
  | (PREF (tn, deref, e0, e1)) ->
      let x0 = (bind_variables_exprs livevars e0)
      and x1 = (bind_variables_exprs livevars e1) in
      (PREF (tn, deref, x0, x1))
  | (SREF (tn, deref, e0, s)) ->
      let x0 = (bind_variables_exprs livevars e0) in
      (SREF (tn, deref, x0, s))
  | (CALL (tn, e0, ee, _)) -> (assert false)
  | (ADDROF (tn, e0)) ->
      (*let x0 = (bind_variables_exprs livevars e0) in*)
      (ADDROF (tn, e0))
  | (SIZEOF (tn, t0)) ->
      let t1 = (bind_variables_types livevars t0) in
      (SIZEOF (tn, t1))
  | (PROGN (tn, l)) -> (assert false)
  | (PSEUDO_MEMORY (tn, t0)) ->
      let t1 = (bind_variables_types livevars t0) in
      (PSEUDO_MEMORY (tn, t1))
  | (INIT_FILL (t0, x0)) ->
      let t1 = (bind_variables_types livevars t0) in
      (*let x1 = (bind_variables_exprs livevars x0) in*)
      (INIT_FILL (t1, x0))
  | (CONST_I _) | (CONST_F _) | (STRING _) ->
      e
  | (CLITERAL (tn, l)) -> (assert false)
  | (QUANTIFIER (s, vs, e0)) ->
      (*NONSTANDARD*)
      (* Add quantified vars to livevars to avoid no-live reference. *)
      (List.iter
	  (fun (v) ->
	    (match v with
	      (VREF (tn, id, level, _)) ->
		(Hashtbl.replace livevars (id, level) v)
	    | _ -> (assert false)))
	  vs) ;
      let x0 = (bind_variables_exprs livevars e0) in
      (QUANTIFIER (s, vs, x0))
  | (PSEUDO_PROLOGUE) ->
      e
  | (PSEUDO_EPILOGUE) ->
      e
  | (ASMSEQUENCE ee) -> (assert false)
  | (PHI _) -> (assert false)
  | (DSGN_AREF (tn, e0, e1)) -> (assert false)
  | (PRAGMA s) -> (assert false)
  | PSEUDO_ACSL_RESULT (tn, e0) -> (
      let x0 = (bind_variables_exprs livevars e0) in
      PSEUDO_ACSL_RESULT (tn, x0)
    )
  | PSEUDO_ACSL_OLD (tn, e0, e1) -> (
      let x0 = (bind_variables_exprs livevars e0) in
      let x1 = (bind_variables_exprs livevars e1) in
      PSEUDO_ACSL_OLD (tn, x0, x1)
    )
  | ALL_FIELD _ -> assert false
  | PSEUDO_ACSL_UNIQUE (tn, e0) -> (
      let x0 = (bind_variables_exprs livevars e0) in
      PSEUDO_ACSL_UNIQUE (tn, x0)
    )
  | PSEUDO_ACSL_UNIQUEBLOCK (tn, e0, e1) -> (
      let x0 = (bind_variables_exprs livevars e0)
      and x1 = (bind_variables_exprs livevars e1) in
      PSEUDO_ACSL_UNIQUEBLOCK (tn, x0, x1)
    )
  | PSEUDO_ACSL_VALID (tn, e0) -> (
      let x0 = (bind_variables_exprs livevars e0) in
      PSEUDO_ACSL_VALID (tn, x0)
    )
)

(* NOTE: IT DROPS QUALIFIERS. *)

and bind_variables_types (livevars : (varid_t, expr_t) Hashtbl.t) (tn : typenode_t) : typenode_t = (
  match (Cexpr.strip_off_pseudo_type tn true) with
    (ARRAY_OF (t0, Some e0)) ->
      let t1 = (bind_variables_types livevars t0) in
      let e1 = (bind_variables_exprs livevars e0) in
      (ARRAY_OF (t1, Some e1))
  | (ARRAY_OF (t0, None)) ->
      let t1 = (bind_variables_types livevars t0) in
      (ARRAY_OF (t1, None))
  | _ ->
      tn)

(* ================================================================ *)

(* REDUNDANT ASSIGNMENT REMOVAL AFTER SSA *)

(* NOTE: Removal works on the field {blk_outputs}, and the temporary
   fields {blk_livevars}, {blk_rdvars}, and {blk_wrvars} are not
   updated and their contents become obsolete. *)

(* Maps a variable name triple (id,level,sequence) to a new one by
   replacing a sequence found in a {eqlmap}.  It removes duplicate
   entries. *)

let rec variable_set_map (eqlmap : ((string * int * int), int) Hashtbl.t) (l : StringInt2Set.t) : StringInt2Set.t = (
      (StringInt2Set.fold
	  (fun (id, level, sequence) s ->
	    let newsequence =
	      if (Hashtbl.mem eqlmap (id, level, sequence)) then
		(Hashtbl.find eqlmap (id, level, sequence))
	      else
		sequence in
	    (StringInt2Set.add (id, level, newsequence) s))
	  l (StringInt2Set.empty)))

let rec variable_list_map (eqlmap : ((string * int * int), int) Hashtbl.t) (l : (string * int * int) list) : (string * int * int) list = (
  (variable_set_to_list
      (List.fold_left
	  (fun s (id, level, sequence) ->
	    let newsequence =
	      if (Hashtbl.mem eqlmap (id, level, sequence)) then
		(Hashtbl.find eqlmap (id, level, sequence))
	      else
		sequence in
	    (StringInt2Set.add (id, level, newsequence) s))
	  (StringInt2Set.empty) l)))

(* Removes trivial variable-to-variable assignments. *)

let rec remove_redundant_assignments (bbset : (string, stmt_block_t) Hashtbl.t) : unit = (
  let eqlset : ((string * int * int), IntSet.t) Hashtbl.t =
    (Hashtbl.create 10) in
  (find_equal_assignments bbset eqlset) ;
  (remove_equal_assignments bbset eqlset))

and find_equal_assignments (bbset : (string, stmt_block_t) Hashtbl.t) (eqlset : ((string * int * int), IntSet.t) Hashtbl.t) : unit = (
  let alllabels = (Misc.hashtbl_keys bbset) in
  (List.iter
      (fun label ->
	let b = (Hashtbl.find bbset label) in
	(List.iter (find_equal_assignments_stmt eqlset) b.blk_phis) ;
	(List.iter (find_equal_assignments_stmt eqlset) b.blk_stmts))
      alllabels))

and find_equal_assignments_stmt (eqlset : ((string * int * int), IntSet.t) Hashtbl.t) (s : stmt_t) : unit = (
  match s with
    DEFVAR _ | DEFUN _ | SPEC _  -> (assert false)
  | (STMT (CALL _)) -> (assert false)
  | ASSUME _ -> ()
  | RESET _ -> ()
  | (STMT e0) ->
      ()
  | (ASGN (_, (CALL _))) -> (assert false)
  | (ASGN (lval, rval)) ->
      (find_equal_variable_assignment eqlset lval rval)
  | (CALLASGN ((CALL (tn, e0, ee, _)), assignements, _)) ->
      (List.iter (find_equal_assignments_stmt eqlset) assignements)
  | (CALLASGN _) -> (assert false)
  | (IF (e0, [(GOTO _)], [(GOTO _)])) ->
      ()
  | IF _ -> (assert false)
  | BEGIN _ | SWITCH _ | DOWHILE _ | FOR _ -> (assert false)
  | BREAK | CONTINUE | LABEL _ | GOTO _ | RETURN _ -> (assert false)
  | (ANNOTATION (s, e0)) ->
      ()
  | (ASM e0) ->
      (failwith "NO ASM YET")
  | NOTE _ -> (assert false))

(* Identifies variable to variable assignment.  NOTE: It specially
   handles a case "x=phi(l0:x;l1:y)" with two inputs, which can be
   reduced to y. *)

and find_equal_variable_assignment (eqlset : ((string * int * int), IntSet.t) Hashtbl.t) (e0 : expr_t) (e1 : expr_t) : unit = (
  let insert_to_eqlset eqlset id level seqset = (
    if (IntSet.mem (-1) seqset) then
      let nilset = (IntSet.add (-1) (IntSet.empty)) in
      (IntSet.iter
	  (fun seq ->
	    if (seq != (-1)) then
	      (Hashtbl.replace eqlset (id, level, seq) nilset))
	  seqset)
    else
      (IntSet.iter
	  (fun seq ->
	    (Hashtbl.replace eqlset (id, level, seq) seqset))
	  seqset)) in
  if (not (result_value_reference_p e0)) then
    (begin
      match (e0, e1) with
	((VREF (tn0, id0, level0, sequence0)),
	(PHI (_, phimap))) ->
	  (begin
	    assert (sequence0 != (-1)) ;
	    let seqset =
	      (List.fold_left
		  (fun r (label, v) ->
		    match v with
		      (VREF (_, id1, level1, sequence1)) ->
			assert ((string_equal id0 id1) && level0 == level1) ;
			assert (sequence1 != (-1)) ;
			(IntSet.add sequence1 r)
		    | NIL ->
			(IntSet.add (-1) r)
		    | _ -> (assert false))
		  (IntSet.add sequence0 (IntSet.empty))
		  phimap) in
	    if ((IntSet.cardinal seqset) == 2) then
	      (insert_to_eqlset eqlset id0 level0 seqset)
	  end)
      | ((VREF (tn0, id0, level0, sequence0)),
	(VREF (tn1, id1, level1, sequence1)))
	  when ((string_equal id0 id1) && level0 == level1) ->
	  (begin
	    assert (sequence0 != (-1) && sequence1 != (-1)) ;
	    let seqs0 =
	      if (Hashtbl.mem eqlset (id0, level0, sequence0)) then
		(Hashtbl.find eqlset (id0, level0, sequence0))
	      else
		(IntSet.add sequence0 (IntSet.empty)) in
	    let seqs1 =
	      if (Hashtbl.mem eqlset (id1, level1, sequence1)) then
		(Hashtbl.find eqlset (id1, level1, sequence1))
	      else
		(IntSet.add sequence1 (IntSet.empty)) in
	    let seqset = (IntSet.union seqs0 seqs1) in
	    (insert_to_eqlset eqlset id0 level0 seqset)
	  end)
      | ((VREF (tn0, id0, level0, sequence0)), NIL) ->
	  assert (sequence0 != (-1)) ;
	  (Hashtbl.replace eqlset (id0, level0, sequence0)
	      (IntSet.add (-1) (IntSet.empty)))
      | _ ->
	  ()
    end))

and remove_equal_assignments (bbset : (string, stmt_block_t) Hashtbl.t) (eqlset : ((string * int * int), IntSet.t) Hashtbl.t) : unit = (
  let alllabels = (Misc.hashtbl_keys bbset) in
  let eqlmap : (((string * int * int), int) Hashtbl.t) = (Hashtbl.create 10) in
  (Hashtbl.iter
      (fun (id, level, sequence) seqset ->
	let minseq = (IntSet.min_elt seqset) in
	if (Hashtbl.mem eqlmap (id, level, sequence)) then
	  assert ((Hashtbl.find eqlmap (id, level, sequence)) == minseq)
	else
	  (Hashtbl.replace eqlmap (id, level, sequence) minseq))
      eqlset) ;
  (if (!bbTRACE_dump_ssa_equalset) then (trace_dump_equalmap eqlmap)) ;
  let neweqlset : (((string * int * int), IntSet.t) Hashtbl.t) =
    (Hashtbl.create 10) in
  (List.iter
      (fun label ->
	let b = (Hashtbl.find bbset label) in
	let phis =
	  (List.flatten
	      (List.map (remove_equals_stmt eqlmap neweqlset) b.blk_phis)) in
	let stmts =
	  (List.flatten
	      (List.map (remove_equals_stmt eqlmap neweqlset) b.blk_stmts)) in
	assert (b.blk_inputs == (StringInt2Set.empty)) ;
	let outputs = (variable_set_map eqlmap b.blk_outputs) in
	let newblk = {
	  b with
	    blk_phis = phis ;
	    blk_stmts = stmts ;
	    blk_outputs = outputs
	} in
	(Hashtbl.replace bbset label newblk))
      alllabels) ;
  if ((Hashtbl.length neweqlset) != 0) then
    (remove_equal_assignments bbset neweqlset))

and remove_equals_stmt (eqlmap : ((string * int * int), int) Hashtbl.t) (neweqlset : ((string * int * int), IntSet.t) Hashtbl.t) (s : stmt_t) : stmt_t list = (
  match s with
    DEFVAR _ | DEFUN _ | SPEC _ -> (assert false)
  | ASSUME _ -> [s]
  | RESET _ -> [s]
  | (STMT (CALL _)) -> (assert false)
  | (STMT e0) ->
      let x0 = (remove_equals_expr eqlmap neweqlset e0) in
      [(STMT x0)]
  | (ASGN (_, (CALL _))) -> (assert false)
  | (ASGN (lval, rval)) ->
      (begin
	let newlval = (remove_equals_expr eqlmap neweqlset lval) in
	let newrval = (remove_equals_expr eqlmap neweqlset rval) in
	let stmt = (ASGN (newlval, newrval)) in
	match (newlval, newrval) with
	  ((VREF (tn0, id0, level0, sequence0)),
	  (VREF (tn1, id1, level1, sequence1)))
	    when ((string_equal id0 id1)
		   && level0 == level1 && sequence0 == sequence1) ->
	      (if (!bbTRACE_dump_ssa_equalset) then (trace_dump_remove s)) ;
	      []
	| ((VREF (tn0, id0, level0, sequence0)),
	  (VREF (tn1, id1, level1, sequence1))) ->
	    (find_equal_assignments_stmt neweqlset stmt) ;
	    [stmt]
	| ((VREF (tn0, id0, level0, sequence0)), NIL) ->
	    (find_equal_assignments_stmt neweqlset stmt) ;
	    [stmt]
	| _ ->
	    (find_equal_assignments_stmt neweqlset stmt) ;
	    [stmt]
      end)
  | (CALLASGN ((CALL (tn, e0, ee, spec_op)), assignments, rs)) ->
      (begin
	let x0 = (remove_equals_expr eqlmap neweqlset e0) in
	let xx = (List.map (remove_equals_expr eqlmap neweqlset) ee) in
	let ss =
	  (List.flatten
	      (List.map (remove_equals_stmt eqlmap neweqlset) assignments)) in
	[(CALLASGN ((CALL (tn, x0, xx, spec_op)), ss, rs))]
      end)
  | (CALLASGN _) -> (assert false)
  | (IF (e0, go0, go1)) ->
      let x0 = (remove_equals_expr eqlmap neweqlset e0) in
      [(IF (x0, go0, go1))]
  | BEGIN _ | SWITCH _ | DOWHILE _ | FOR _ -> (assert false)
  | BREAK | CONTINUE | LABEL _ | GOTO _ | RETURN _ -> (assert false)
  | (ANNOTATION (s0, e0)) ->
      [s]
  | (ASM e0) ->
      (failwith "NO ASM YET")
  | NOTE _ -> (assert false))

and remove_equals_expr (eqlmap : ((string * int * int), int) Hashtbl.t) (neweqlset : ((string * int * int), IntSet.t) Hashtbl.t) (e : expr_t) : expr_t = (
  match e with
  | NIL ->
      e
  | (PSEUDO_NONDET_VALUE (tn, _)) ->
      e
  | (PHI (tn, phis)) ->
      let phimap =
	(List.map
	    (fun (label, e0) ->
	      let x0 = (remove_equals_expr eqlmap neweqlset e0) in
	      (label, x0))
	    phis) in
      (bind_variables_uniqify_phi tn phimap)
  | (OPER (tn, op, e0, e1, e2)) ->
      let x0 = (remove_equals_expr eqlmap neweqlset e0) in
      let x1 = (remove_equals_expr eqlmap neweqlset e1) in
      let x2 = (remove_equals_expr eqlmap neweqlset e2) in
      (OPER (tn, op, x0, x1, x2))
  | (VREF (tn, id, level, sequence)) ->
      if (Hashtbl.mem eqlmap (id, level, sequence)) then
	let seq = (Hashtbl.find eqlmap (id, level, sequence)) in
	if (seq != (-1)) then
	  (VREF (tn, id, level, seq))
	else
	  NIL
      else
	e
  | (PREF (tn, deref, e0, e1)) ->
      let x0 = (remove_equals_expr eqlmap neweqlset e0)
      and x1 = (remove_equals_expr eqlmap neweqlset e1) in
      (PREF (tn, deref, x0, x1))
  | (SREF (tn, deref, e0, s)) ->
      let x0 = (remove_equals_expr eqlmap neweqlset e0) in
      (SREF (tn, deref, x0, s))
  | (ADDROF (tn, e0)) ->
      let x0 = (remove_equals_expr eqlmap neweqlset e0) in
      (ADDROF (tn, x0))
  | (SIZEOF (tn, tt)) ->
      e
  | (PROGN (tn, l)) -> (assert false)
  | (PSEUDO_MEMORY (tn, tt)) ->
      e
  | (INIT_FILL _) ->
      e
  | (CONST_I _) | (CONST_F _) | (STRING _) ->
      e
  | (CLITERAL (tn, l)) -> (assert false)
  | (QUANTIFIER (s, vs, e0)) ->
      let x0 = (remove_equals_expr eqlmap neweqlset e0) in
      (QUANTIFIER (s, vs, x0))
  | (PSEUDO_PROLOGUE) ->
      e
  | (PSEUDO_EPILOGUE) ->
      e
  | (SETF _) -> (assert false)
  | (CALL (tn, e0, ee, _)) -> (assert false)
  | (ASMSEQUENCE ee) -> (assert false)
  | (DSGN_AREF (tn, e0, e1)) -> (assert false)
  | (PRAGMA s) -> (assert false)
  | PSEUDO_ACSL_RESULT (tn, e0) -> (
      let x0 = remove_equals_expr eqlmap neweqlset e0 in
      PSEUDO_ACSL_RESULT (tn, x0))
  | PSEUDO_ACSL_OLD (tn, e0, e1) -> (
      let x0 = remove_equals_expr eqlmap neweqlset e0 in
      let x1 = remove_equals_expr eqlmap neweqlset e1 in
      PSEUDO_ACSL_OLD (tn, x0, x1))
  | ALL_FIELD _ -> assert false
  | PSEUDO_ACSL_UNIQUE (tn, e0) -> (
      let x0 = remove_equals_expr eqlmap neweqlset e0 in
      PSEUDO_ACSL_UNIQUE (tn, x0))
  | PSEUDO_ACSL_UNIQUEBLOCK (tn, e0, e1) -> (
      let x0 = remove_equals_expr eqlmap neweqlset e0
      and x1 = remove_equals_expr eqlmap neweqlset e1 in
      PSEUDO_ACSL_UNIQUEBLOCK (tn, x0, x1))
  | PSEUDO_ACSL_VALID (tn, e0) -> (
      let x0 = remove_equals_expr eqlmap neweqlset e0 in
      PSEUDO_ACSL_VALID (tn, x0)))

and trace_dump_equalmap (eqlmap : ((string * int * int), int) Hashtbl.t) : unit = (
  (Hashtbl.iter
      (fun (id, level, sequence) newsequence ->
	(debug ("EQSET=" ^ id ^ "%" ^ (string_of_int level)
			 ^ ":" ^ (string_of_int sequence)
			 ^ "->" ^ (string_of_int newsequence))))
      eqlmap))

and trace_dump_remove (stmt : stmt_t) : unit = (
  (debug ("REMOVE=" ^ (Cexpr.stmt_to_string_body 1 stmt))))

(* Makes an input variable list for junction blocks.  It is called
   after removal of redundant phi-assignments, because inputs of the
   removed phis shall appear in the input list.  *)

let fixup_input_variables (bbset : (string, stmt_block_t) Hashtbl.t) : unit = (
  let alllabels = (Misc.hashtbl_keys bbset) in
  (List.iter
      (fun label ->
	let b = (Hashtbl.find bbset label) in
	match b with
	  {
	    blk_type = BBHEAD ;
	    blk_incoming_labels = [] ;
	    blk_stmts = stmts ;
	    blk_inputs = ivars ;
	    blk_outputs = ovars
	  } ->
	    (* Entry block. *)
	    assert (ivars == (StringInt2Set.empty)) ;
	    let xvars = (variable_set_of_assigned stmts) in
	    let ivars = (StringInt2Set.diff ovars xvars) in
	    let newblk = { b with blk_inputs = ivars } in
	    (Hashtbl.replace bbset label newblk)
	| {
	    blk_type = BBTAIL|BBSTMT|BBASME|BBASSIGN ;
	    blk_incoming_labels = [incoming] ;
	    blk_inputs = ivars ;
	    blk_outputs = ovars
	  } ->
	    (* NOTE: IMPRECISE. *)
	    assert (ivars == (StringInt2Set.empty)) ;
	    if (string_equal incoming pseudo_elsewhere_label) then
	      (* Do nothing for an unreachable return block. *)
	      assert (b.blk_type == BBTAIL)
	    else
	      (begin
		let prev = (Hashtbl.find bbset incoming) in
		let newblk = { b with blk_inputs = (prev.blk_outputs) } in
		(Hashtbl.replace bbset label newblk)
	      end)
	| {
	    blk_type = BBCALL ;
	    blk_incoming_labels = [incoming] ;
	    blk_inputs = ivars ;
	    blk_outputs = ovars
	  } ->
	    (* NOTE: IMPRECISE. *)
	    assert (ivars == (StringInt2Set.empty)) ;
	    let prev = (Hashtbl.find bbset incoming) in
	    let newblk = { b with blk_inputs = (prev.blk_outputs) } in
	    (Hashtbl.replace bbset label newblk)
	| {
	    blk_type = BBJUNC ;
	    blk_phis = phis ;
	    blk_inputs = ivars ;
	    blk_outputs = ovars
	  } ->
	    assert (ivars == (StringInt2Set.empty)) ;
	    let phivars = (variable_set_of_assigned phis) in
	    let ivars = (StringInt2Set.diff ovars phivars) in
	    let newblk = { b with blk_inputs = ivars } in
	    (Hashtbl.replace bbset label newblk)
	| {
	    blk_type = BBCOND ;
	    blk_incoming_labels = [incoming] ;
	    blk_inputs = ivars
	  } ->
	    (* NOTE: IMPRECISE. *)
	    assert (ivars == (StringInt2Set.empty)) ;
	    let prev = (Hashtbl.find bbset incoming) in
	    let newblk = { b with blk_inputs = (prev.blk_outputs) } in
	    (Hashtbl.replace bbset label newblk)
	| _ -> (assert false))
      alllabels))

(* ================================================================ *)

(* Assigns new SSA bindings to variables.  It first assigns
   temporarily variables for outputs of phi-functions in junction
   blocks.  Then, it follows blocks (in a topologically sorted order)
   from the entry label. *)

let rec bind_new_variables (bbset : (string, stmt_block_t) Hashtbl.t) (decl : Cexpr.declaration_t) (vartypes : (varid_t, typenode_t) Hashtbl.t) = (
  let ctx : binder_t = (make_binder bbset) in
  (pseudo_bind_parameters ctx bbset decl) ;
  (pseudo_bind_globals ctx bbset vartypes) ;
  (bind_variables_arbitrary_junction ctx bbset vartypes) ;
  (bind_variables_forward ctx bbset vartypes
      [pseudo_entry_label] (StringSet.empty)) ;
  if (!bbTRACE_dump_ssa_flow) then
    (dump_stmt_blocks (CFUNDEF { cfdef_declaration = decl ;
				 cfdef_blocks = bbset })) ;
  (bind_variables_junction ctx bbset vartypes) ;
  (remove_redundant_assignments bbset) ;
  (fixup_input_variables bbset) ;
  (bind_temporaries_junction ctx bbset))

let rec fixate_minor_on_blocks (bbset : (string, stmt_block_t) Hashtbl.t) : unit = (
  let b = (Hashtbl.find bbset pseudo_entry_label) in
  assert (b.blk_incoming_labels == []) ;
  let newblk = {
    b with
      blk_incoming_labels = [pseudo_elsewhere_label] ;
  } in
  (Hashtbl.replace bbset pseudo_entry_label newblk))
