(* parser/misc.ml 2009-01-07 *)
(* Copyright (C) 2007-2009 Yonezawa Lab, The University of Tokyo *)

(* Misc Functions *)

open Num

(* Tests equalness, to make visibly distinguishable from '='. *)

let string_equal (x : string) (y : string) : bool = (x = y)
let string_list_equal (l0 : string list) (l1 : string list) : bool = (
  (List.for_all2 string_equal l0 l1))

exception Found

(* ================================================================ *)

(* Disallows equality. *)

module None_ = (struct type none = NONE end)
let (=) (x : None_.none) (y : None_.none) : bool =
  (ignore (failwith "Equality disallowed"); true)
let (<>) (x : None_.none) (y : None_.none) : bool =
  (ignore (failwith "Equality disallowed"); true)

(* Takes the body of {Some}. *)

let some_of (v : 'a option) : 'a = (
  match v with
    Some e -> e
  | None -> raise (Invalid_argument "None is given to some_of."))

let some_of_safe (dflt : 'a) (v : 'a option) : 'a = (
  match v with
    Some e -> e
  | None -> dflt)

(* ================================================================ *)

(* Maps on a string (calling f with a char and an index). *)

let string_mapi (f : char -> int -> 'a) (s : string) : 'a list = (
  let rec string_mapi_loop_ f (s : string) i n = (
    if (i == n) then
      []
    else
      ((f s.[i] i) :: (string_mapi_loop_ f s (i + 1) n))) in
  (string_mapi_loop_ f s 0 (String.length s)))

let string_iteri (f : char -> int -> unit) (s : string) : unit = (
  let rec string_iteri_loop_ f (s : string) i n = (
    if (i == n) then
      ()
    else
      ((f s.[i] i) ; (string_iteri_loop_ f s (i + 1) n))) in
  (string_iteri_loop_ f s 0 (String.length s)))

let string_to_int_array (s : string) : int array = (
  let a = (Array.make ((String.length s) + 1) 0) in
  (string_iteri (fun c i -> (a.(i) <- (Char.code c))) s) ;
  a)

(* Checks if a string is with a given prefix. *)

let string_starts_with (prefix : string) (s : string) : bool = (
  ((String.length s) >= (String.length prefix)
    && (string_equal prefix (String.sub s 0 (String.length prefix)))))

(* Same as {String.index} but returns -1 if not found. *)

let string_index_safe (s : string) (c : char) : int = (
  try
    (String.index s c)
  with
    Not_found ->
      -1)

let string_replace (oldchar : char) (newchar : char) (s : string) : string = (
  let n = (String.copy s) in
  for i = 0 to ((String.length s) - 1) do (
    if (s.[i] == oldchar) then (String.set n i newchar)
  ) done ;
  n)

let rec string_index_safe_ (delim : char list) (s : string) (i : int) (l : int) : int = (
  if (i == l) then
    (-1)
  else if (List.mem (String.get s i) delim) then
    i
  else
    (string_index_safe_ delim s (i + 1) l))

let rec string_tokens (delim : char list) (s : string) : string list = (
  let i = (string_index_safe_ delim s 0 (String.length s)) in
  if (i == -1) then
    [s]
  else
    let s1 = (String.sub s 0 i) in
    let s2 = (String.sub s (i + 1) ((String.length s) - (i + 1))) in
    (s1 :: (string_tokens delim s2)))

(* ================================================================ *)

(* Adds a new entry after checking non-existence. *)

let hashtbl_unique_add (t : ('a, 'b) Hashtbl.t) (k : 'a) (v : 'b) : unit = (
  assert (not (Hashtbl.mem t k)) ;
  (Hashtbl.replace t k v))

(* Finds an entry without exception. *)

let hashtbl_find_safe (dflt : 'b) (t : ('a, 'b) Hashtbl.t) (k : 'a) : 'b = (
  if (Hashtbl.mem t k) then
    (Hashtbl.find t k)
  else
    dflt)

(* Checks if an element exists that satisfies the predicate p. *)

let hashtbl_mem_if (p : 'a -> 'b -> bool) (t : ('a, 'b) Hashtbl.t) : bool = (
  try
    (Hashtbl.iter (fun k b -> if (p k b) then raise Found else ()) t) ;
    false
  with
    Found -> true)

let hashtbl_keys (t : ('a, 'b) Hashtbl.t) : 'a list = (
  (Hashtbl.fold (fun k v e -> (k :: e)) t []))

let hashtbl_values (t : ('a, 'b) Hashtbl.t) : 'b list = (
  (Hashtbl.fold (fun k v e -> (v :: e)) t []))

(* Iterates over a hashtable, and collects results. *)

let hashtbl_map (f : 'a -> 'b -> 'c) (t : ('a, 'b) Hashtbl.t) : 'c list = (
  (Hashtbl.fold (fun k v r -> ((f k v) :: r)) t []))

(* ================================================================ *)

(* Folds left, taking arguments in the same order as {Set} or {Map},
   (intended for use when commutative). *)

let list_fold (f : 'b -> 'a -> 'a) (l : 'b list) (v : 'a) : 'a = (
  (List.fold_left (fun v x -> (f x v)) v l))

let list_fold2 (f : 'b -> 'c -> 'a -> 'a) (xl : 'b list) (yl : 'c list) (v : 'a) : 'a = (
  (List.fold_left2 (fun v x y -> (f x y v)) v xl yl))

(* Conses a new element if it is new (tested with equal). *)

let list_add_new (e : 'a) (l : 'a list) : 'a list = (
  (if (not (List.mem e l)) then (e :: l) else l))

(* Takes a last of a list. *)

let rec list_last (l : 'a list) : 'a = (
  match l with
    [] -> (failwith "Empty list to list_last.")
  | [e] -> e
  | (hd :: tl) -> list_last tl)

(* Drops a tail of a list. *)

let rec list_droptail (l : 'a list) : 'a list = (
  match l with
    [] -> (failwith "Empty list to list_droptail.")
  | [e] -> []
  | (hd :: tl) -> (hd :: list_droptail tl))

(* Maps like List.map but with index. *)

let list_mapi (f : ('a -> int -> 'b)) (l : 'a list) : 'b list = (
  let rec list_mapi_loop_ f (l : 'a list) (i : int) : 'b list = (
    match l with
      [] -> []
    | (hd :: tl) -> ((f hd i) :: (list_mapi_loop_ f tl (i + 1)))) in
  (list_mapi_loop_ f l 0))

(* Splits a list into two by a predicate. *) (*SLOW*)

let list_split (f : ('a -> bool)) (l : 'a list) : ('a list * 'a list) = (
  (List.fold_left
      (fun (tparts, fparts) e ->
	if (f e) then
	  ((tparts @ [e]), fparts)
	else
	  (tparts, (fparts @ [e])))
      ([], []) l))

(* Splits a list into two with specified length. *) (*SLOW*)

let list_cutat (position : int) (l : 'a list) : ('a list * 'a list) = (
  let rec list_cutat_loop_ (position : int) (l : 'a list) (r : 'a list) (i : int) : ('a list * 'a list) = (
    if (i == position) then
      (l, r)
    else
      match r with
	[] ->
	  (l, r)
      | (hd :: tl) ->
	  (list_cutat_loop_ position (l @ [hd]) tl (i + 1))) in
  (list_cutat_loop_ position [] l 0))

(* Maps a function on a list collecting values with {Some}. *)

let rec list_filtering_map (f : ('a -> 'b option)) (l : 'a list) : 'b list = (
  match l with
    [] ->
      []
  | (hd :: tl) ->
      (begin
	match (f hd) with
	  None -> (list_filtering_map f tl)
	| Some v -> (v :: (list_filtering_map f tl))
      end))

(*
let list_filtering_map (f : ('a -> 'b option)) (l : 'a list) : 'b list = (
  (List.flatten
      (List.map
	  (fun (e : 'a) ->
	    let v = (f e) in
	    if (v == None) then
	      []
	    else
	      [(some_of v)])
	  l)))
*)

(* Takes a singleton element in a list. *)

let list_singleton (l : 'a list) : 'a = (
  match l with
    [a] ->
      a
  | _ ->
      (failwith "Non-singleton list passed to list_singleton."))

(* Uniqifies elements in the list.  Equality is taken as hashtable
   entries.  (unstable; does not keep the original order). *)
(*SLOW*)

let pseudo_uniqify_list (l : 'a list) : 'a list = (
  let tbl : ('a, unit) Hashtbl.t = (Hashtbl.create 10) in
  (List.iter (fun e -> (Hashtbl.replace tbl e ())) l) ;
  (hashtbl_keys tbl))

let rec list_find_safe (dflt : 'a) (f : 'a -> bool) (l : 'a list) : 'a = (
  match l with
    [] ->
      dflt
  | (e :: tl) ->
      if (f e) then
	e
      else
	(list_find_safe dflt f tl))

(* Finds an element with which {f} returns {Some}.  It returns a value
   of {f}. *)

let rec list_find_some (f : ('a -> 'a option)) (l : 'a list) : 'a option = (
  match l with
    [] ->
      None
  | (hd :: tl) ->
      (begin
	match (f hd) with
	  None -> (list_find_some f tl)
	| Some v -> Some v
      end))

let rec list_map_compare (f : 'a -> 'a -> int) (xl : 'a list) (yl : 'a list) : int = (
  match (xl, yl) with
    ([], []) ->
      0
  | ([], _) ->
      (1)
  | (_, []) ->
      (-1)
  | ((x :: xt), (y :: yt)) ->
      let v = (f x y) in
      if (v != 0) then
	v
      else
	(list_map_compare f xt yt))

(* Generates a list of zero to (s - 1) ints. *)

let rec list_int_sequence (s : int) : int list = (
  (list_int_sequence_loop_ s []))

and list_int_sequence_loop_ (s : int) (l : int list) : int list = (
  if (s <= 0) then
    l
  else
    (list_int_sequence_loop_ (s - 1) ((s - 1) :: l)))

(* ================================================================ *)

(* Calculates ceiling of {x} to a multiple of{m}. *)

let ceiling_to (x : num) (m : num) : num = (
  (mult_num (ceiling_num (div_num x m)) m))

let pseudo_bitnot_num (x : num) : num = (
  let x0 = (int_of_num x) in
  (num_of_int (lnot x0)))

let pseudo_and_num (x : num) (y : num) : num = (
  let x0 = (int_of_num x) in
  let y0 = (int_of_num y) in
  (num_of_int (x0 land y0)))

let pseudo_ior_num (x : num) (y : num) : num = (
  let x0 = (int_of_num x) in
  let y0 = (int_of_num y) in
  (num_of_int (x0 lor y0)))

let pseudo_xor_num (x : num) (y : num) : num = (
  let x0 = (int_of_num x) in
  let y0 = (int_of_num y) in
  (num_of_int (x0 lxor y0)))

(* ================================================================ *)

(* Mutable store of a Set module (not a multi-set). *)

module Bag = functor (BaseSet : Set.S) ->
  (struct
  type elt = BaseSet.elt
  type t = {
    mutable s : BaseSet.t
  }
  let compare (g0 : t) (g1 : t) : int = (raise (Invalid_argument "Bag"))
  let equal (g0 : t) (g1 : t) : bool = (raise (Invalid_argument "Bag"))
  let create () : t = { s = BaseSet.empty }
  let clear (g : t) : unit = (g.s <- BaseSet.empty)
  let add (g : t) (e : elt) : unit = (g.s <- (BaseSet.add e g.s))
  let remove (g : t) (e : elt) : unit = (g.s <- (BaseSet.remove e g.s))
  let mem (g : t) (e : elt) : bool = (BaseSet.mem e g.s)
  let iter (f : (elt -> unit)) (g : t) : unit = (BaseSet.iter f g.s)
  let fold (f: (elt -> 'a -> 'a)) (g : t) (v : 'a) = (BaseSet.fold f g.s v)
  let cardinal (g : t) : int = (BaseSet.cardinal g.s)
  let length (g : t) : int = (cardinal g)
  let size (g : t) : int = (cardinal g)
  let elements (g : t) : elt list = (BaseSet.elements g.s)
  let contents (g : t) : elt list = (elements g)
  let as_set (g : t) : BaseSet.t = (g.s)
  end)

(* Set with effects implemented as hashtable (pseudo and unoptimized).
   It takes Set module to return a set of contents. *)

module HBag = functor (Xset : Set.S) ->
  (struct
  type elt = Xset.elt
  type t = (elt, unit) Hashtbl.t
  let create () = ((Hashtbl.create 40) :> t)
  let clear (g : t) = (Hashtbl.clear g)
  let copy (g : t) = ((Hashtbl.copy g) :> t)
  let add_multiple (g : t) (e : elt) = (Hashtbl.add g e ())
  let add (g : t) (e : elt) = (
    if (Hashtbl.mem g e) then () else (Hashtbl.add g e ()))
  let remove (g : t) (e : elt) = (Hashtbl.remove g e)
  let mem (g : t) (e : elt) = (Hashtbl.mem g e)
  let iter (f : (elt -> unit)) (g : t) = (
    (Hashtbl.iter (fun k v -> (f k)) g))
  let fold (f: (elt -> 'b -> 'b)) (g : t) (e0 : 'b) = (
    (Hashtbl.fold (fun k v e -> (f k e)) g e0))
  let length (g : t) = (Hashtbl.length g)
  let contents (g : t) : elt list = (
    (fold (fun e r -> (e :: r)) g []))

  let as_set (g : t) = (
    (fold (fun e r -> (Xset.add e r)) g Xset.empty))

  let find_ (g : t) (e : elt) = (Hashtbl.find g e)
  let find_all_ (g : t) (e : elt) = (Hashtbl.find_all g e)
  let replace_ (g : t) (e1 : elt) (e2 : elt) = (
    (Hashtbl.remove g e1) ; (Hashtbl.add g e2))
  end)

(*
module HBag = functor (Xset : Set.S) ->
  (struct
  type 'a t = ('a, unit) Hashtbl.t
  let create () = ((Hashtbl.create 40) :> 'a t)
  let clear (g : 'a t) = (Hashtbl.clear g)
  let copy (g : 'a t) = ((Hashtbl.copy g) :> 'a t)
  let add_multiple (g : 'a t) (e : 'a) = (Hashtbl.add g e ())
  let add (g : 'a t) (e : 'a) = (
    if (Hashtbl.mem g e) then () else (Hashtbl.add g e ()))
  let remove (g : 'a t) (e : 'a) = (Hashtbl.remove g e)
  let mem (g : 'a t) (e : 'a) = (Hashtbl.mem g e)
  let iter (f : ('a -> unit)) (g : 'a t) = (
    (Hashtbl.iter (fun k v -> (f k)) g))
  let fold (f: ('a -> 'b -> 'b)) (g : 'a t) (e0 : 'b) = (
    (Hashtbl.fold (fun k v e -> (f k e)) g e0))
  let length (g : 'a t) = (Hashtbl.length g)
  let contents (g : 'a t) : 'a list = (
    (fold (fun e r -> (e :: r)) g []))

  let as_set (g : 'a t) = (
    (fold (fun e r -> (Xset.add e r)) g Xset.empty))

  let find_ (g : 'a t) (e : 'a) = (Hashtbl.find g e)
  let find_all_ (g : 'a t) (e : 'a) = (Hashtbl.find_all g e)
  let replace_ (g : 'a t) (e1 : 'a) (e2 : 'a) = (
    (Hashtbl.remove g e1) ; (Hashtbl.add g e2))
  end)
*)

(* Same as above, without facility to return contents as a set. *)

(*
module XBag =
  (struct
  type 'a t = ('a, unit) Hashtbl.t
  let create () = ((Hashtbl.create 40) :> 'a t)
  let clear (g : 'a t) = (Hashtbl.clear g)
  let copy (g : 'a t) = ((Hashtbl.copy g) :> 'a t)
  let add_multiple (g : 'a t) (e : 'a) = (Hashtbl.add g e ())
  let add (g : 'a t) (e : 'a) = (
    if (Hashtbl.mem g e) then () else (Hashtbl.add g e ()))
  let remove (g : 'a t) (e : 'a) = (Hashtbl.remove g e)
  let mem (g : 'a t) (e : 'a) = (Hashtbl.mem g e)
  let iter (f : ('a -> unit)) (g : 'a t) = (
    (Hashtbl.iter (fun k v -> (f k)) g))
  let fold (f: ('a -> 'b -> 'b)) (g : 'a t) (e0 : 'b) = (
    (Hashtbl.fold (fun k v e -> (f k e)) g e0))
  let length (g : 'a t) = (Hashtbl.length g)
  let contents (g : 'a t) : 'a list = (
    (fold (fun e r -> (e :: r)) g []))

  let find_ (g : 'a t) (e : 'a) = (Hashtbl.find g e)
  let find_all_ (g : 'a t) (e : 'a) = (Hashtbl.find_all g e)
  let replace_ (g : 'a t) (e1 : 'a) (e2 : 'a) = (
    (Hashtbl.remove g e1) ; (Hashtbl.add g e2))
  end)
*)
