(* parser/specs/scommon.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

(* Some auxiliary logging functions *)
let warning (s : string) : unit =
    LogUtil.log LogUtil.SPECS LogUtil.WARNING s
   
let debug (s : string) : unit =
    LogUtil.log LogUtil.SPECS LogUtil.DEBUG s

let error (s : string) : unit =
    LogUtil.log LogUtil.SPECS LogUtil.ERROR s
    
let info (s : string) : unit =
    LogUtil.log LogUtil.SPECS LogUtil.INFO s

let emerg (s : string) : unit =
    LogUtil.log LogUtil.SPECS LogUtil.EMERG s

(** {2 Parsing utilities} *)

(** Hack to allow typedefs whose names are ACSL keywords: the state of the
    lexer depends on the parser rule. See logic_lexer.mll and
    logic_parser.mly for more details.
*)

(** *)
let doxygen_mode = ref false

let enter_doxygen_mode () = doxygen_mode := true

let exit_doxygen_mode () = doxygen_mode := false

(**
  - false => keywords are all ACSL keywords
  - true => only C keywords are recognized as such.
   (other remains plain identifiers/typenames)
*)
let acsl_term_mode = ref false

let enter_acsl_term_mode () = acsl_term_mode := true

let exit_acsl_term_mode () = acsl_term_mode := false

let is_acsl_term_mode () = !acsl_term_mode

let kw_c_mode = ref false

let enter_kw_c_mode () = kw_c_mode := true

let exit_kw_c_mode () = kw_c_mode := false

let is_kw_c_mode () = !kw_c_mode

(**********************)

let typenames : (string, unit) Hashtbl.t = Hashtbl.create 13

let add_typename t = Hashtbl.add typenames t ()

let remove_typename t = Hashtbl.remove typenames t

let reset_typenames () = Hashtbl.clear typenames

let typename_status t = Hashtbl.mem typenames t

let add_typenames ts = List.iter add_typename ts

(**********************)

(* process_spec indicates whether specs reader is enabled or not *)
let process_specs = ref false

(* check pre-condition for initializer field or not *)
let no_check_initializer = ref false

(*----------------------------*)

let string_of_strings (strs: string list) : string = (
    String.concat "," strs 
)

let rec find_substring (source : string) (substring : string) : int = (
    let len_source = String.length source in
    let len_sub = String.length substring in
    if len_source < len_sub then -1
    else (
        let c = String.get substring 0 in
        try 
            let index = String.index source c in
            let sub_str = String.sub source index len_sub in
            if Misc.string_equal sub_str substring then index
            else
                let new_source = String.sub source (index+1) 
                                            (len_source - index - 1) in
                find_substring new_source substring
        with 
        | Not_found -> -1
    )
)
(* *****************************************************************
   Function :
           to split the input string into a list of substrings
   Params : 
           source : input string, to be splitted into strings
           sep    : separator
   Ex     :
          source = "hello_world_hola"
          sep    = "_" 
    ==>   output = ["hello"; "world"; "hola" ]  
****************************************************************)
let rec split_string source sep =
    match (find_substring source sep) with
    | -1 -> 
        [source]
    | n -> 
        let len_source = String.length source in
        let len_sep = String.length sep in
        let new_source = String.sub source 
                                    (n + len_sep) 
                                    (len_source - n - len_sep) in
        (String.sub source 0 n) :: (split_string new_source sep)

(*           
   Ex     :
           str_source   = "hello_hello_hola"
           str_replaced = "_"
           str          = "+"
    ==>
          output        = "hello+hello+hola"
*)
let replace_string (source : string) 
                   (old_sep : string) 
                   (new_sep : string) : string = (
    let strs = (split_string source old_sep) in
    List.fold_left (fun s e -> 
                       if (String.length s) == 0 then 
                          e
                       else
                          s ^ new_sep ^ e
                   ) 
                   "" strs
)

let rec get_mem_strings_of_strings (strs1 : string list) 
                                   (strs2 : string list) = (
    match strs2 with
     | [] -> []
     | head::tail -> (if List.mem head strs1 then [head] 
                      else []) @
                     (get_mem_strings_of_strings strs1 tail ) 
)

(*
   params : 
           strs1 : string list 
           strs2 : string list
   output :           
*)
let check_include_list_string ( strs1 : string list ) 
                              ( strs2 : string list ) = (
   let res1 = get_mem_strings_of_strings strs1 strs2 in
   let res2 = get_mem_strings_of_strings strs2 strs1 in
   if (List.length res1) + (List.length res2) > 0 then true
   else false
)

(*
   params :
            strs : string list 
   output :
   Ex     :
           [d;a;c;d;a;c;b] --> [d;a;c;b]
*)
let rec check_redundant_string_list (strs : string list) 
                                    (strs_res : string list) = (
   match strs with
    | [] -> strs_res
    | head::tail -> if List.mem head strs_res then 
                       check_redundant_string_list tail 
                                                   strs_res
                    else 
                       check_redundant_string_list tail 
                                                   (head::strs_res)
)

