(* parser/specs/smain.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

(* stop_at indicates the step after which the program ends.
   By default, the program runs through all steps.
   NOTE: not in use at the moment!*)
open Translate
open Scommon

type stop_at_t = 
     TRANSLATION
   | PARSE
   | ANYWHERE
let stop_at = ref ANYWHERE

(* + parsed_stmts: consists of all parsed declarations including 
     "raw" specifications.
   + stmt_merger: do the merging whenever a declaration is parsed *)
let parsed_stmts = ref [] 

let stmt_merger (decl : Cexpr.stmt_t) = (
    parsed_stmts := (!parsed_stmts) @ [decl]
)

let get_and_replace_params_conds
    (func_contracts_args : (('pred, 'locs) Sexpr.function_contract_t * Cexpr.expr_t list) list)
    (d_params : Cexpr.expr_t list)
: ('pred list (* precondition *)
   * 'pred list (* postcondition *)) = (
    List.fold_left (fun (pre_cond, post_cond) (func_contract, s_params) ->
        let conds = func_contract.Sexpr.fspec_conditions in
        let conds_replaced_params =
            Ccommon.replace_exprs_in_conds conds
                                           s_params
                                           d_params in
        let new_pre_cond = Sexpr.pre_conds_of_func_conds conds_replaced_params
        in
        let new_post_cond = Sexpr.post_conds_of_func_conds conds_replaced_params
        in
        (pre_cond @ new_pre_cond, post_cond @ new_post_cond)
    ) ([], []) func_contracts_args;
)

let memo_of_note (note : Cexpr.note_t)
: string * string * string * string * string = (
    match note with
    | Cexpr.LINE (_, fn, ln, cn) -> (
        let string_of_ln = string_of_int ln in
        let string_of_cn = string_of_int cn in
        (fn, string_of_ln, string_of_cn, string_of_ln, string_of_cn)
      )
    | Cexpr.RANGE (fn, start_ln, start_cn, end_ln, end_cn) -> (
        (fn, string_of_int start_ln, string_of_int start_cn,
        string_of_int end_ln, string_of_int end_cn)
      )
)

exception Unsafe_subtyping of (string * (string * string * string * string *
string));;
let type_checking (decls : Cexpr.stmt_t list) : unit = (
    Hashtbl.iter (fun key value ->
        let specs_before_pointer, pointer_name =
        (match fst key with
        | None -> (
            (* find specs before pointer declaration *)
            let pointer_var_name = snd key in
            (Hashtbl.find_all Ccommon.contract_of_fun_pointer
                                pointer_var_name,
             pointer_var_name))
        | Some struct_var_name -> (
            let pointer_field_name = snd key in
            let struct_name =
                try Hashtbl.find Sinstrument.f_pointer_field_hashtbl
                                 (struct_var_name, pointer_field_name)
                with Not_found -> assert false
            in
            (Hashtbl.find_all Ccommon.contract_of_field_hashtbl
                              (struct_name, pointer_field_name),
             struct_name ^ "." ^ pointer_field_name))
          ) in
        (* find function contract of real function relevent to pointer *)
        let expr = fst value
        and note = snd value
        in
        let fun_name, _ =
            try ExtString.String.split (Ccommon.get_name_of_expr expr) "%"
            with _ -> (Ccommon.get_name_of_expr expr), ""
        in
        List.iter (fun decl ->
            match decl with
            | Cexpr.DEFVAR (header)
            | Cexpr.DEFUN (header, _) -> (
                if (fst header.Cexpr.decl_id) = fun_name then
                let specs_before_real_function = header.Cexpr.decl_specs in
                let d_param_names = header.Cexpr.decl_params in
                let d_param_types =
                    let tn =
                        Cexpr.strip_off_pseudo_type (header.Cexpr.decl_tn) true
                    in
                    match tn with
                    | Cexpr.FUNCTION_OF (typenode, typenodes_op) -> (
                        match typenodes_op with
                        | None -> []
                        | Some typenodes -> typenodes)
                    | _ -> []
                in
                let d_params =
                    List.map2 (fun param_name param_type ->
                        Cexpr.VREF (param_type, param_name,
                                    Cexpr.parameter_scope_level, 0)
                    ) d_param_names d_param_types in
                let f_fun_conds =
                    match specs_before_real_function with
                    | None -> []
                    | Some (Sexpr.SPEC_FUN func_cond) ->
                        let full_expr = func_cond.Sexpr.fspec_conditions in
                        full_expr
                    | _ -> assert false
                in
                let f_pre_exprs = Sexpr.pre_conds_of_func_conds f_fun_conds
                and f_post_exprs = Sexpr.post_conds_of_func_conds f_fun_conds
                in
                let f_pre_form = Stranslate.cexprs_to_form f_pre_exprs
                and f_post_form = Stranslate.cexprs_to_form f_post_exprs
                in
                let pre_exprs, post_exprs =
                    get_and_replace_params_conds specs_before_pointer
                                                 d_params in
                let loc = memo_of_note note in
                let pre_form = Stranslate.cexprs_to_form pre_exprs
                and post_form = Stranslate.cexprs_to_form post_exprs
                in
                let form1 = Form.IMP(pre_form, f_pre_form) in
                let form2 = Form.IMP(f_post_form, post_form) in
                let pred1 = Form.form_to_predicate form1
                and pred2 = Form.form_to_predicate form2
                in
                let result = (LogicUtil.Logic.is_valid pred1)
                             & (LogicUtil.Logic.is_valid pred2) in
                let msg =
                        ("The specification of function '" ^ fun_name ^ "'"
                        ^ " does not match the specification of pointer "
                        ^ "'" ^ pointer_name ^ "'")
                in
                if (result = false) then
                    raise (Unsafe_subtyping (msg, loc))
              )
            | _ -> ()
        ) decls;
    ) Sinstrument.fp_to_f_hashtbl
)

(******************************************************************************)
let clone_stmt_block (bound : int)
                         (source : Cexpr.stmt_t list)
                         : Cexpr.stmt_t list = (
    let rec append_stmt_block (bound : int)
                          (source : Cexpr.stmt_t list)
                          (target : Cexpr.stmt_t list)
                          : Cexpr.stmt_t list =
        match bound with
        | x when (x < 0) -> assert false
        | 0 -> target
        | _ -> (append_stmt_block (bound - 1) source (source @ target))
    in
    append_stmt_block bound source [])

let rec transform_loops (bound : int) (stmt : Cexpr.stmt_t) : Cexpr.stmt_t = (
    match stmt with
    | Cexpr.DEFUN (header, body) -> (
        let new_body = transform_loops bound body in
        Cexpr.DEFUN(header, new_body))
    | Cexpr.DEFVAR _ -> stmt
    | Cexpr.BEGIN stmts -> (
        let new_stmts = List.map (transform_loops bound) stmts in
        Cexpr.BEGIN new_stmts)
    | Cexpr.IF (cond, then_stmts, else_stmts) -> (
        let new_then_stmts = List.map (transform_loops bound) then_stmts
        and new_else_stmts = List.map (transform_loops bound) else_stmts
        in
        Cexpr.IF (cond, new_then_stmts, new_else_stmts))
    | Cexpr.SWITCH (e, stmts) -> (
        let new_stmts = List.map (transform_loops bound) stmts in
        Cexpr.SWITCH (e, new_stmts))
    | Cexpr.DOWHILE (b, e, stmts) -> (
        let escape_label = Cexpr.make_new_label () in
        let escape_stmt = Cexpr.GOTO escape_label in
        let not_e =
            let tn = Cexpr.type_of_expr e in 
            Cexpr.OPER(tn, Cexpr.LOGNOT, e, Cexpr.NIL, Cexpr.NIL)
        in
        let one_block =
            if b then
                (Cexpr.IF (not_e, [escape_stmt], [])) :: stmts
            else
                stmts @ [Cexpr.IF (not_e, [escape_stmt], [])]
        in
        let bounded_blocks = clone_stmt_block bound one_block in
        Cexpr.BEGIN (bounded_blocks
                     @ [(Cexpr.LABEL (Cexpr.NAME_LABEL escape_label))])
        )
    | Cexpr.FOR (i, e1, e2, l) -> (
        let escape_label = Cexpr.make_new_label () in
        let escape_stmt = Cexpr.GOTO escape_label in
        let new_body = (Cexpr.STMT e2) :: l in
        let not_e1 =
            let tn = Cexpr.type_of_expr e1 in 
            Cexpr.OPER(tn, Cexpr.LOGNOT, e1, Cexpr.NIL, Cexpr.NIL)
        in
        let one_block =
            (Cexpr.IF (not_e1, [escape_stmt], [])) :: new_body
        in
        let bounded_blocks = clone_stmt_block bound one_block in
        Cexpr.BEGIN (i
                     @ bounded_blocks
                     @ [(Cexpr.LABEL (Cexpr.NAME_LABEL escape_label))]))
    | _ -> stmt)

let last_note : Cexpr.note_t ref = ref (Cexpr.LINE (false, "", -1, -1))
let position_stack : Cexpr.note_t Stack.t = Stack.create ()

let rec augment_bound_into_loops (bound : int)
                                 (stmt : Cexpr.stmt_t)
                                 : Cexpr.stmt_t list = (
    let make_itermax_check (counter : Cexpr.expr_t) : Cexpr.stmt_t list =
        let cond_expr = Cexpr.make_binary_operation
                            Cexpr.GE
                            counter
                            (Cexpr.constant_integer (Num.num_of_int bound))
        in
        let func_tn = Cexpr.make_function_type Cexpr.VOID_TYPE [] in
        let func_expr = Cexpr.VREF (func_tn, Ccommon.raise_itermax_err_fn,
                                    Cexpr.global_scope_level, 0)
        in
        let func_call = Cexpr.STMT (Cexpr.make_function_call func_expr []) in
        let note = Stack.pop position_stack in
        [(Cexpr.NOTE note); (Cexpr.IF (cond_expr, [func_call], []))]
    in
    let make_augmentation_args (pre_check : bool)
                               (cond : Cexpr.expr_t)
                               : Cexpr.stmt_t
                                 * Cexpr.expr_t
                                 * Cexpr.stmt_t
                                 * (Cexpr.stmt_t list) = (
        let counter =
            Cexpr.make_new_temporary_variable Cexpr.UNSIGNED_INTEGER
        in
        let init_counter =
            if pre_check then
                Cexpr.STMT (Cexpr.make_assignment Cexpr.SET
                                                  counter
                                                  Cexpr.constant_one)
            else
                Cexpr.STMT (Cexpr.make_assignment Cexpr.SET
                                                  counter
                                                  Cexpr.constant_zero)
        in
        let bound_cond = Cexpr.make_binary_operation
                             Cexpr.LT
                             counter
                             (Cexpr.constant_integer (Num.num_of_int bound))
        in
        let new_cond =
            Cexpr.make_conditional_expr Cexpr.ANDAND cond bound_cond Cexpr.NIL
        in
        let incr_counter_stmt =
            Cexpr.STMT (Cexpr.make_unary_operation Cexpr.INCR_POST counter)
        in
        let check_itermax_stmt =
            make_itermax_check counter
        in
        (init_counter, new_cond, incr_counter_stmt, check_itermax_stmt))
    in
    match stmt with
    | Cexpr.DEFUN (header, body) -> (
        let new_body = List.hd (augment_bound_into_loops bound body) in
        [Cexpr.DEFUN(header, new_body)])
    | Cexpr.DEFVAR _ -> [stmt]
    | Cexpr.BEGIN stmts -> (
        let new_stmts =
            List.flatten (List.map (augment_bound_into_loops bound) stmts)
        in
        [Cexpr.BEGIN new_stmts])
    | Cexpr.IF (cond, then_stmts, else_stmts) -> (
        let new_then_stmts =
            List.flatten (List.map (augment_bound_into_loops bound) then_stmts)
        and new_else_stmts =
            List.flatten (List.map (augment_bound_into_loops bound) else_stmts)
        in
        [Cexpr.IF (cond, new_then_stmts, new_else_stmts)])
    | Cexpr.SWITCH (e, stmts) -> (
        let new_stmts =
            List.flatten (List.map (augment_bound_into_loops bound) stmts)
        in
        [Cexpr.SWITCH (e, new_stmts)])
    | Cexpr.DOWHILE (b, e, stmts) -> (
        (Stack.push !last_note position_stack);
        let init_counter, new_e, incr_counter, itermax_check =
            make_augmentation_args b e
        in
        let new_stmts =
            List.flatten (List.map (augment_bound_into_loops bound) stmts)
        in
        ([init_counter; Cexpr.DOWHILE (b, new_e, (new_stmts @ [incr_counter]))]
            @ itermax_check))
    | Cexpr.FOR (i, e1, e2, l) -> (
        (Stack.push !last_note position_stack);
        let init_counter, new_e1, incr_counter, itermax_check =
            make_augmentation_args true e1
        in
        let new_l =
            List.flatten (List.map (augment_bound_into_loops bound) l)
        in
        ([init_counter; Cexpr.FOR (i, new_e1, e2, (new_l @ [incr_counter]))]
            @ itermax_check))
    | Cexpr.NOTE note -> (
        last_note := note;
        [stmt])
    | _ -> [stmt])

(* main entry of the specification reader. *)

let parse_source_file (filename : string) 
                      (file_path : string) 
                      (loop_bound_opt : int option)
                      (subtyping_check : bool)
                      : (Cexpr.ofile_t) = (
  let name = (Filename.chop_extension (Filename.basename file_path)) in
  assert ((String.length name) > 0) ;
  
  Scommon.process_specs := true;
  Cexpr.process_declaration := stmt_merger ;
  parsed_stmts := [] ;
  
  let hash = (Printf.sprintf "%04x" ((Random.bits ()) land 0xFFFF)) in
  Interp.file_hash_name := (name ^ "%" ^ hash) ;
  Cexpr.file_typebody_table := (Some (Hashtbl.create 40)) ;
  Csyntax.file_scope_static_variable_name_map := (Hashtbl.create 10) ;
  Cexpr.defined_size_t_type := None ;

  (* MEMO: file_label_table below is intentionally not reset to
     collect label value references from all source files. *)

  (*Cexpr.file_label_table := [] ;*)

  if (!Interp.aiTRACE_linking) then
    (debug (";; READING FILE '" ^ file_path ^ "'...")) ;
  (Clex.cparse filename) ;
  
  let implicit_invariants_hashtbl = 
        Hashtbl.copy (Ccommon.implicit_type_invariants_hashtbl) in

  if (!Interp.aiTRACE_linking) then
    (debug (";; READING FILE '" ^ file_path ^ "' DONE")) ;

  if !stop_at = PARSE then (
      print_string "\nStop after Parsing as requested!\n" ;
      exit 0
  );
  (*
   * Now enters the specification instrumentation
   *)
  let decls =
      Sinstrument.main !parsed_stmts
                       implicit_invariants_hashtbl
  in
  (* check sub-typing *)
  if subtyping_check then (
      debug (";; CHECKING SUBTYPING");
      (type_checking decls))
  else
      debug (";; SKIPPING SUBTYPING CHECK");

  let decls =
      match loop_bound_opt with
      | Some bound -> (
          (Stack.clear position_stack);
          (last_note := (Cexpr.LINE (false, "", -1, -1)));
          List.flatten (List.map (augment_bound_into_loops bound) decls))
      | None -> decls
  in
  (assert (0 = (Stack.length position_stack)));
  
  let decls = (Ccommon.init_pointer_stmts decls) in
  let decls = [Ccommon.artificial_assert] @ decls in
  let decls = Cexpr.preprocess_assume_in_stmts decls in
  let stmt_ofile : Cexpr.ofile_t = 
      Cexpr.create_stmt_ofile decls 
                              file_path
                              !Interp.file_hash_name
                              Sinstrument.f_pointer_field_hashtbl
                              Ccommon.contract_of_field_hashtbl
  in stmt_ofile
)

let translate_ofile (stmt_ofile : Cexpr.ofile_t) : Translate.ofile_t =
begin
    Interp.file_hash_name := stmt_ofile.Cexpr.ofile_hash_name;
    let defs : (string, Translate.definition_t) Hashtbl.t = (Hashtbl.create 10) in

    Interp.parsed_definitions := (Some defs) ;
    List.iter 
        (fun (_, st) ->
            Interp.translate_declaration st
        )
        !(stmt_ofile.Cexpr.ofile_defs);

    let globalpreds =
        (Form.uniqify_form_list
	          (Hashtbl.fold
	              (fun n def tl ->
	                  match def with
		                (Translate.FUNDEF d) ->
		                  (d.Translate.fundef_global_preds @ tl)
	                  | (Translate.VARDEF d) ->
                          tl
	                  | (Translate.OTHERDEF s) ->
                          tl)
	          defs [])) in
    let o : Translate.ofile_t = {
        Translate.ofile_hash_name = stmt_ofile.Cexpr.ofile_hash_name ;
        Translate.ofile_file_name = stmt_ofile.Cexpr.ofile_file_name ;
        Translate.ofile_defs = defs ;
        Translate.ofile_global_predicates = globalpreds
    } in
    (o)     
end
