(* parser/specs/stranslate.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

open Num
open Lexing

(* {scan_string_constant_loop} is intentionally to holds
   a reference to function Clex.scan_string_constant_loop.
   The main purpose of this is to resolve mutual module dependency
   between modules Clex, Cg, and Sinstrument.
   This body function is dummy. *)

let scan_string_constant_loop
    : (Lexing.lexbuf -> Buffer.t -> int Vector.t -> unit) ref 
    = ref
    (fun stream token content -> 
         (Printf.fprintf stderr "Unscanned string constant"))

(*********** type ************)

type position_of_insertion =
     | PRE_FUN
     | PRE_FUN_COND
     | POST_FUN
     | PRE_FUN_BEHAV
     | POST_FUN_BEHAV
     | POST_RESULT
     | PRE_LOOP
     | POS_BREAK
     | POS_CONTINUE
     | POS_RETURN
     | FUNCTION_INVARIANT
     | HERE

type insert_triple = ((string list) * 
                      position_of_insertion * 
                      (Cexpr.stmt_t list))


(*****************make note about location ***********************
   FIXME: to be considered more cases
 ***********************************************)
let make_range_note (loc : Sexpr.range_t) = 
    let fn, start_pos, end_pos = loc in
    Cexpr.NOTE (Cexpr.RANGE(fn, 
                            (fst start_pos), (snd start_pos), 
                            (fst end_pos), (snd end_pos)))

(*****************make assert statement ***********************)
(*let make_assert (ex : Cexpr.expr_t) = (
        Cexpr.ANNOTATION ("$assert", ex)
)*)
let make_assert (ex : Cexpr.expr_t) = (
    let cond_expr = Cexpr.OPER(
                          Cexpr.INTEGER, 
                          Cexpr.EQ, 
                          ex, 
                          Cexpr.CONST_I ((Cexpr.INTEGER, Num.Int (0), "0")),
                          Cexpr.NIL) in
    Cexpr.STMT (
            Cexpr.OPER (
                        Cexpr.INTEGER,
                        Cexpr.COND,
                        cond_expr,
                        Cexpr.CALL (Cexpr.INTEGER,
                                    Cexpr.OPER (
                                                Cexpr.POINTER_TO (
                                                    Cexpr.FUNCTION_OF (
                                                        Cexpr.INTEGER,None)),
                                                Cexpr.CAST (false),
                                                Cexpr.VREF (
                                                    Cexpr.FUNCTION_OF (
                                                        Cexpr.INTEGER,None),
                                                "__assert_fail",1,0),
                                                Cexpr.NIL,
                                                Cexpr.NIL
                                     ),
                                     [ Cexpr.STRING (
                                           false,
                                           (Sdump.string_of_expr ex false), 
                                           [|0|]);
                                       Cexpr.STRING (
                                           false," unknown file name", [|0|]);
                                       Cexpr.CONST_I (
                                           Cexpr.INTEGER, Num.Int (0), "0");
                                       Cexpr.STRING (
                                           false,"__PRETTY_FUNCTION__", [|0|]) 
                                     ],
                                     None
                        ),
                         Cexpr.OPER (Cexpr.VOID_TYPE,
                                    Cexpr.CAST (true),
                                    Cexpr.CONST_I (Cexpr.UNSIGNED_INTEGER, Num.Int (0),"0"),
                                    Cexpr.NIL,Cexpr.NIL
                        )
             )
    )

)


(* make list of assert statement *)
let rec make_list_assert (exs : Cexpr.expr_t list) = 
   List.map make_assert exs

(**************** make assumes **************)

let make_assume (ex : Cexpr.expr_t): Cexpr.stmt_t = (
    Cexpr.ANNOTATION("$assume", ex)
)
let make_assumes (exs : Cexpr.expr_t list): Cexpr.stmt_t list= (
    List.map make_assume exs
)


(********** make if statement ***********************)

(* make if statement *)
let make_if_statement ( condition : Cexpr.expr_t )
                      ( assert_stmts : Cexpr.stmt_t list ) = 
   let begin_stmt = Cexpr.BEGIN(assert_stmts) in
   let cond = Cexpr.OPER (Cexpr.INTEGER,
                          Cexpr.NE,
                          condition,
                          Cexpr.CONST_I(
                                Cexpr.INTEGER, 
                                Num.Int (0), 
                                "0"),
                          Cexpr.NIL
              )
   in
   [Cexpr.IF(cond,[begin_stmt],[])]

(**************)
let make_assumes_and_assert (condition : Cexpr.expr_t option)
                            (exs : Cexpr.expr_t list) 
                            (default_behav : bool): Cexpr.stmt_t list = (
    let if_stmts = (
        if default_behav then (make_list_assert exs)
        else (
            match condition with
            | Some c -> (make_if_statement c (make_list_assert exs))
            | None -> assert false)
    ) in
    if_stmts)

(**************)
(* 
   assert ( a => b ) 
--> 
   assert( (!a) || b ) 
*)
let make_assert_behavior (behav_cond : Cexpr.expr_t) 
                         (cond : Cexpr.expr_t)= (
    let new_behav_cond =
        Cexpr.OPER(
               Cexpr.INTEGER,
               Cexpr.NEG, 
               Cexpr.OPER (Cexpr.INTEGER,
                           Cexpr.NE,
                           behav_cond,
                           Cexpr.CONST_I(
                                 Cexpr.INTEGER, 
                                 Num.Int (0), 
                                 "0"),
                           Cexpr.NIL),
               Cexpr.NIL,
               Cexpr.NIL)
    in
    let new_cond = 
        Cexpr.OPER (Cexpr.INTEGER,
                    Cexpr.OROR,
                    new_behav_cond,
                    cond,
                    Cexpr.NIL)
    in
    Cexpr.ANNOTATION ("$assert", new_cond)
)
(*
   struct test {
       int a;
       int b;
   };
   id = a 
          =>  int
   id = c 
          => fail
*)
let find_struct_component_type (struct_name : string)
                               (comps : ( string * Cexpr.typenode_t ) list)
                               (id : string) : Cexpr.typenode_t = (
   let ls_t = 
       List.filter (
        fun com ->
            let str, typ = com in
            if (Misc.string_equal str id) then true
            else false            
      ) comps 
   in
   match ls_t with
    | [] ->
        raise (Sexpr.Specs_error ("Struct '" ^ struct_name 
                                  ^ "' does not have field '" ^ id ^ "'", 
                                  None))
    | [(_,ty)] -> ty
    | _ ->
        raise (Sexpr.Specs_error ("Duplicate field '" ^ id 
                                  ^ "' in struct '" ^ struct_name ^ "'",
                                  None))
)

let find_enum_component_type 
    (enum_name : string)
    (comps : ( string * Cexpr.expr_t ) list) 
    (id : string) = (
    let ls_t = 
        List.filter (
            fun com ->
                let str, _ = com in
                if str = id then true
                else false            
            ) comps 
    in
    match ls_t with
    | [] -> 
        raise (Sexpr.Specs_error ("Enum '" ^ enum_name 
                                  ^ "' does not have field '" ^ id ^ "'",
                                  None))
    | [(_,_)] -> Cexpr.INTEGER
    | _ -> 
        raise (Sexpr.Specs_error ("Duplicate field '" ^ id 
                                  ^ "' in enum '" ^ enum_name ^ "'",
                                  None))
)

let rec get_type_struct (var_name : string ) ( id : string ) = (
    let tyn = Ccommon.get_struct_type_from_binding var_name in
    match tyn with
    | Cexpr.STRUCTURE_TYPE (struct_type) -> (
        let unique_id = struct_type.Cexpr.strc_unique_id in
        let table = struct_type.Cexpr.strc_typebody_table in
        let type_body = (
            try
                Hashtbl.find table unique_id
            with
            | Not_found -> assert false
        ) in
        let components = (
            match type_body with
            | Cexpr.STRC_BODY (str_type) -> (
                str_type.Cexpr.strc_components            
              )
            | _ -> assert false
       ) in
       let struct_name = struct_type.Cexpr.strc_tag in
       (find_struct_component_type struct_name components id)
     )
    | Cexpr.ENUMERATION_TYPE (struct_type) -> (
        let unique_id = struct_type.Cexpr.enum_unique_id in
        let table = struct_type.Cexpr.enum_typebody_table in
        let type_body = (
            try
                Hashtbl.find table unique_id
            with
            | Not_found -> assert false
        ) in
        let components = (
            match type_body with
            | Cexpr.ENUM_BODY (str_type) -> (
                str_type.Cexpr.enum_components
              )
            | _ -> assert false
       ) in
       let enum_name = struct_type.Cexpr.enum_tag in
       (find_enum_component_type enum_name components id)
     )
    | _ -> assert false
)
let rec get_name (tyn : Cexpr.typenode_t ) = (
    match tyn with 
    | Cexpr.STRUCTURE_TYPE stype -> 
         Some stype.Cexpr.strc_tag
    | Cexpr.STRCDEF_TYPE (ty,_) -> get_name ty
    | Cexpr.POINTER_TO ty -> get_name ty
    | _ -> None
)

(******************************************)
let keep_acsl_constructs = ref false

let translate_ikind (ik : Sexpr.ikind) : Cexpr.typenode_t = (
    match ik with
    | Sexpr.IBool -> Cexpr.BOOLEAN
    | Sexpr.IChar -> Cexpr.CHARACTER
    | Sexpr.ISChar -> Cexpr.SIGNED_CHARACTER
    | Sexpr.IUChar -> Cexpr.UNSIGNED_CHARACTER
    | Sexpr.IInt -> Cexpr.INTEGER
    | Sexpr.IUInt -> Cexpr.UNSIGNED_INTEGER
    | Sexpr.IShort -> Cexpr.SHORT_INTEGER
    | Sexpr.IUShort -> Cexpr.UNSIGNED_SHORT_INTEGER
    | Sexpr.ILong -> Cexpr.LONG_INTEGER
    | Sexpr.IULong -> Cexpr.UNSIGNED_LONG_INTEGER
    | Sexpr.ILongLong -> Cexpr.LONG_LONG_INTEGER
    | Sexpr.IULongLong -> Cexpr.UNSIGNED_LONG_LONG_INTEGER
)

let translate_fkind (fk : Sexpr.fkind) : Cexpr.typenode_t = (
    match fk with
    | Sexpr.FFloat -> Cexpr.SINGLE_FLOAT
    | Sexpr.FDouble -> Cexpr.DOUBLE_FLOAT
    | Sexpr.FLongDouble -> Cexpr.LONG_DOUBLE_FLOAT
)

let rec translate_logictype (logictype : Sexpr.logic_type) : Cexpr.typenode_t = (
    match logictype with
    | Sexpr.LTvoid ->
        Cexpr.VOID_TYPE
    | Sexpr.LTinteger ->
        Cexpr.INTEGER
    | Sexpr.LTint ik ->
        (translate_ikind ik)
    | Sexpr.LTfloat fk ->
        (translate_fkind fk)
    | Sexpr.LTarray logictype ->
        Cexpr.ARRAY_OF ((translate_logictype logictype), None)
    | Sexpr.LTpointer logictype ->
        Cexpr.POINTER_TO (translate_logictype logictype)
    | Sexpr.LTstruct id
    | Sexpr.LTunion id ->
        Ccommon.find_struct_type id
    | Sexpr.LTenum str ->
        (assert false)
    | Sexpr.LTreal ->
        (assert false)
    | Sexpr.LTnamed str ->
        (Ccommon.find_typedef_type str)
)

exception Translate_failed;;

let translate_int_constant (s : string) : (Num.num * Cexpr.typenode_t) =
begin
    (* extract suffix *)
    let len = String.length s in
    let s_upcase = String.uppercase s in
    let begin_index = ref len in
    let is_done = ref false in
    while not !is_done do
       if ((!begin_index > 1) && ((s_upcase.[!begin_index - 1] = 'U')
                               || (s_upcase.[!begin_index - 1] = 'L'))) then
           begin_index := !begin_index - 1
       else
           is_done := true
    done;
    let end_index = len - 1 in
    let suffix = (try String.sub s_upcase !begin_index (end_index - !begin_index + 1)
                  with _ -> "") in
    (* extract prefix about base number & data string of number *)
    let str = String.sub s_upcase 0 !begin_index in
    let (base, data) = (
        let len = String.length str in
        if (len > 2) then
        begin
            if ((String.sub str 0 2) = "0X") then
                (16, String.sub str 2 (len - 2))
            else if ((String.sub str 0 1) = "0") then
                (8, String.sub str 1 (len - 1))
            else
                (10, str)
        end
        else if (len >= 1) then
        begin
            if ((String.sub str 0 1) = "0") then
                (8, String.sub str 1 (len - 1))
            else
                (10, str)
        end
        else
             (10, str)
    ) in
    (* get value of the number stored in input string*)
    let (value : Num.num ref) = ref (Num.num_of_int 0) in
    String.iter (
        fun ch ->
        begin
            let v = (
                if (base == 8) && ((ch >= '0') && (ch <= '7')) then
                    (int_of_char ch) - (int_of_char '0')
                else
                if (base == 10) && ((ch >= '0') && (ch <= '9')) then
                    (int_of_char ch) - (int_of_char '0')
                else
                if (base == 16) && ((ch >= '0') && (ch <= '9')) then
                    (int_of_char ch) - (int_of_char '0')
                else
                if (base == 16) && ((ch >= 'A') && (ch <= 'F')) then
                    (int_of_char ch) - (int_of_char 'A') + 10
                else
                    raise Translate_failed;
            ) in
            value := add_num (mult_num !value (num_of_int base)) (num_of_int v);
        end
    ) data;
    (* get typenode of number *)
    let (ucount, lcount) = (
        match suffix with
        | ""            -> (0, 0)
        | "U"           -> (1, 0)
        | "L"           -> (0, 1)
        | "LL"          -> (0, 2)
        | "UL" | "LU"   -> (1, 1)
        | "ULL" | "LLU" -> (1, 2)
        | _             -> raise Translate_failed
    ) in
    let typenode =
    if ((lcount <= 0) && (compare_num !value Cexpr.most_positive_int) <= 0) then
    begin
        if ((ucount == 1) || (base != 10)) then
            Cexpr.UNSIGNED_INTEGER
        else
            Cexpr.INTEGER
    end
    else if ((lcount <= 0)
          && (compare_num !value Cexpr.most_positive_unsigned_int) <= 0) then
    begin
        assert ((compare_num !value Cexpr.most_positive_int) > 0);
        Cexpr.UNSIGNED_INTEGER
    end
    else if ((lcount <= 1)
          && (compare_num !value Cexpr.most_positive_long) <= 0) then
    begin
        if ((ucount == 1) || (base != 10)) then
            Cexpr.UNSIGNED_LONG_INTEGER
        else
            Cexpr.LONG_INTEGER
    end
    else if ((lcount <= 1)
          && (compare_num !value Cexpr.most_positive_unsigned_long) <= 0) then
    begin
        assert ((compare_num !value Cexpr.most_positive_long) > 0) ;
        Cexpr.UNSIGNED_LONG_INTEGER
    end
    else
    begin
        (* Range check cannot be performed. *)
        if ((ucount == 1) || (base != 10)) then
            Cexpr.UNSIGNED_LONG_LONG_INTEGER
        else
            Cexpr.LONG_LONG_INTEGER
    end in
    (* return *)
    (!value, typenode);
end

let translate_string_constant (s : string) (wide : bool) : Cexpr.expr_t = (
    (* manually add an enclosing mark to avoid the syntax_error exception *)
    let s = s ^ "\"" in
    let stream = Lexing.from_string s in
    let token = (Buffer.create 10) in
    let content = (Vector.create 10 0) in
    (!scan_string_constant_loop stream token content) ;
    let s = (Buffer.contents token) in
    (Cexpr.STRING (wide, s, (Vector.to_array content))))

let translate_constant (const : Sexpr.constant) (loc : Sexpr.location)
                       : Cexpr.expr_t = (
    match const with
    | Sexpr.IntConstant str ->
        let (num_int, typenode) = (
            try translate_int_constant str
            with Translate_failed -> (
                let error_msg = "Invalid integer constant: '" ^ str ^ "'" in
                let range = Sexpr.make_range (fst loc)
                                             (fst loc)
                in
                let (filename, begin_loc, end_loc) = range in
                let (begin_line, begin_pos) = begin_loc in
                let global_pos = (filename, begin_line, begin_pos) in
                raise (Sexpr.Specs_error (error_msg, Some global_pos))
            )
        ) in
        Cexpr.CONST_I (typenode, num_int, str)
    | Sexpr.FloatConstant str ->
        let num_float =
            (try float_of_string str
            with Failure _ -> assert false) in
        Cexpr.CONST_F (Cexpr.DOUBLE_FLOAT, num_float, str)
    | Sexpr.StringConstant str ->
        translate_string_constant str false
    | Sexpr.WStringConstant str ->
        translate_string_constant str true
)

let rec translate_unop (un_op : Sexpr.unop) 
                   (lexpr : Sexpr.lexpr)
                   (vars : (string, Cexpr.expr_t) Hashtbl.t) =(
    match un_op with
    | Sexpr.Uminus -> (
        let expr = translate_lexpr lexpr vars in
        Cexpr.OPER (Cexpr.type_of_expr expr,
                      Cexpr.NEG,
                      expr,
                      Cexpr.NIL,
                      Cexpr.NIL)
        )
    | Sexpr.Ustar -> (
        let expr = translate_lexpr lexpr vars  in
        let new_type = (Cexpr.type_of_expr expr) in
        match new_type with
        | Cexpr.POINTER_TO(t) -> 
            Cexpr.PREF (t, true, expr, expr)
        | _ -> 
            Cexpr.PREF (new_type, true, expr, expr)
        )
    | Sexpr.Uamp -> (
        let expr = translate_lexpr lexpr vars  in 
        Cexpr.ADDROF (Cexpr.POINTER_TO (Cexpr.type_of_expr expr), expr)
        )
    | Sexpr.Ubw_not -> (
        let expr = translate_lexpr lexpr vars  in 
        Cexpr.OPER (Cexpr.type_of_expr expr, 
                   Cexpr.BITNOT, 
                   expr, 
                   Cexpr.NIL, 
                   Cexpr.NIL)
        )
)

and translate_relation (lexpr1 : Sexpr.lexpr)
                       (rel_op : Sexpr.relation)  
                       (lexpr2 : Sexpr.lexpr)
                       (vars : (string, Cexpr.expr_t) Hashtbl.t) = (
    let ex1 = translate_lexpr lexpr1 vars in
    let ex2 = translate_lexpr lexpr2 vars in
    match rel_op with
    | Sexpr.Lt ->
        Cexpr.OPER(Cexpr.INTEGER, Cexpr.LT, ex1, ex2, Cexpr.NIL)
    | Sexpr.Gt ->
        Cexpr.OPER(Cexpr.INTEGER, Cexpr.GT, ex1, ex2, Cexpr.NIL)
    | Sexpr.Le ->
        Cexpr.OPER(Cexpr.INTEGER, Cexpr.LE, ex1, ex2, Cexpr.NIL)
    | Sexpr.Ge ->
        Cexpr.OPER(Cexpr.INTEGER, Cexpr.GE, ex1, ex2, Cexpr.NIL)
    | Sexpr.Eq ->
        Cexpr.OPER(Cexpr.INTEGER, Cexpr.EQ, ex1, ex2, Cexpr.NIL)
    | Sexpr.Neq ->
        Cexpr.OPER(Cexpr.INTEGER, Cexpr.NE, ex1, ex2, Cexpr.NIL)
)

and translate_binop (lexpr1 : Sexpr.lexpr) 
                    (bin_op : Sexpr.binop) 
                    (lexpr2 : Sexpr.lexpr) 
                    (vars : (string, Cexpr.expr_t) Hashtbl.t) = (
    let expr1 = translate_lexpr lexpr1 vars  in
    let expr2 = translate_lexpr lexpr2 vars  in
    let ctyp1 = Cexpr.usual_unary_conversion (Cexpr.type_of_expr expr1) in
    let ctyp2 = Cexpr.usual_unary_conversion (Cexpr.type_of_expr expr2) in
    let ctyp = Cexpr.usual_binary_conversion ctyp1 ctyp2 in
    match bin_op with
    | Sexpr.Badd ->
        Cexpr.OPER (ctyp, Cexpr.ADD, expr1, expr2, Cexpr.NIL)
    | Sexpr.Bsub ->
        Cexpr.OPER (ctyp, Cexpr.SUB, expr1, expr2, Cexpr.NIL) 
    | Sexpr.Bmul ->
        Cexpr.OPER (ctyp, Cexpr.MUL, expr1, expr2, Cexpr.NIL)
    | Sexpr.Bdiv ->
        Cexpr.OPER (ctyp, Cexpr.DIV, expr1, expr2, Cexpr.NIL)
    | Sexpr.Bmod ->
        Cexpr.OPER (ctyp, Cexpr.MOD, expr1, expr2, Cexpr.NIL)
    | Sexpr.Bbw_and ->
        Cexpr.OPER (ctyp, Cexpr.AND, expr1, expr2, Cexpr.NIL)
    | Sexpr.Bbw_or ->
        Cexpr.OPER (ctyp, Cexpr.IOR, expr1, expr2, Cexpr.NIL)
    | Sexpr.Bbw_xor ->
        Cexpr.OPER (ctyp, Cexpr.XOR, expr1, expr2, Cexpr.NIL)
    | Sexpr.Blshift ->
        Cexpr.OPER (ctyp, Cexpr.LSH, expr1, expr2, Cexpr.NIL)
    | Sexpr.Brshift ->
        Cexpr.OPER (ctyp, Cexpr.RSH, expr1, expr2, Cexpr.NIL)
)

and translate_list_lexpr (lexprs : Sexpr.lexpr list) 
                         (vars : (string, Cexpr.expr_t) Hashtbl.t)
: Cexpr.expr_t list = (
    List.map (fun lexpr -> translate_lexpr lexpr vars) lexprs
)

and translate_lexpr (lexpr : Sexpr.lexpr) 
                    (vars : (string, Cexpr.expr_t) Hashtbl.t)
: Cexpr.expr_t = (
    match lexpr.Sexpr.lexpr_node with
    | Sexpr.PLvar name -> (
        try Hashtbl.find vars name
        with
        | Not_found -> (
            let loc = fst lexpr.Sexpr.lexpr_loc in
            let pos = (loc.Lexing.pos_fname,
                       loc.Lexing.pos_lnum,
                       loc.Lexing.pos_cnum)
            and msg = "Referencing undefined identifier: " ^ name in
            raise (Sexpr.Specs_error (msg, Some pos)))
        )
    | Sexpr.PLunop (unop, lexpr1) ->
        translate_unop unop lexpr1 vars
    | Sexpr.PLbinop (lexpr1, binop, lexpr2) ->
        translate_binop  lexpr1 binop lexpr2 vars
    | Sexpr.PLdot (lexpr1, id) -> (
        let ex = translate_lexpr lexpr1 vars in
        let tyn = Cexpr.type_of_expr ex in
        match (Cexpr.structure_or_union_type_p tyn) with
        | true -> (
            try
                let type_of_sref = Cexpr.get_structure_slot tyn id in
                Cexpr.SREF (type_of_sref, false, ex, id)
            with Failure msg -> (
                let loc = fst lexpr.Sexpr.lexpr_loc in
                let pos = (loc.Lexing.pos_fname,
                           loc.Lexing.pos_lnum,
                           loc.Lexing.pos_cnum)
                in
                raise (Sexpr.Specs_error (msg, Some pos))
            )
            )
        | false ->
            let loc = fst lexpr.Sexpr.lexpr_loc in
            let pos = (loc.Lexing.pos_fname,
                       loc.Lexing.pos_lnum,
                       loc.Lexing.pos_cnum)
            and msg = "Referencing to a non-struct/union" in
            raise (Sexpr.Specs_error (msg, Some pos))
        )
    | Sexpr.PLarrow (lexpr1, id) -> (
        let ex = translate_lexpr lexpr1 vars in
        let tyn = Cexpr.type_of_expr ex in
        try
            Cexpr.make_sref_arw ex id;
        with
        | Failure msg ->
            let loc = fst lexpr.Sexpr.lexpr_loc in
            let pos = (loc.Lexing.pos_fname,
                       loc.Lexing.pos_lnum,
                       loc.Lexing.pos_cnum)
            in
            raise (Sexpr.Specs_error (msg, Some pos))
        )
    | Sexpr.PLarrget (lexpr1, lexpr2) -> (
        let expr1 = translate_lexpr lexpr1 vars
        and expr2 = translate_lexpr lexpr2 vars in
        Cexpr.make_aref expr1 expr2)
    | Sexpr.PLold lexpr1 -> (
        let t1 = (translate_lexpr lexpr1 vars) in
        let tyn = Cexpr.type_of_expr t1 in
        let name_of_expr = 
            (Ccommon.replace_special_char_in_string (Cexpr.expr_to_string t1))
        in
        let old_expr = Cexpr.VREF (tyn, ("old_of_" ^ name_of_expr), 3, 0) in
        match !keep_acsl_constructs with
        | true -> Cexpr.PSEUDO_ACSL_OLD (tyn, t1, old_expr)
        | false -> old_expr
        )
    | Sexpr.PLresult -> (
        let result_expr =
            try
                Hashtbl.find vars Ccommon.return_expr_name (*"__RET"*)
            with
            | Not_found ->
                assert false
        in
        match !keep_acsl_constructs with
        | true ->
            let ty = Cexpr.type_of_expr result_expr in
            Cexpr.PSEUDO_ACSL_RESULT (ty, result_expr)
        | false -> result_expr
        )
    | Sexpr.PLnull -> (
        Cexpr.OPER(Cexpr.INTEGER, Cexpr.EQ, (translate_lexpr lexpr vars),
                 Cexpr.CONST_I (Cexpr.UNSIGNED_INTEGER, Int (0), 
                                     string_of_int(0)),
                 Cexpr.NIL
               ) 
        )
    | Sexpr.PLcast (logictype, lexpr1) -> (
        let ctype = translate_logictype logictype in
        let expr = translate_lexpr lexpr1 vars in
        try
            Cexpr.make_cast ctype expr
        with Failure msg -> (
            let loc = fst lexpr.Sexpr.lexpr_loc in
            let pos = (loc.Lexing.pos_fname,
                       loc.Lexing.pos_lnum,
                       loc.Lexing.pos_cnum)
            in
            raise (Sexpr.Specs_error (msg, Some pos))
        )
        )
    | Sexpr.PLsizeof logictype -> (
        let typ = translate_logictype logictype in
        try
            Cexpr.make_sizeof_operation typ
        with Failure msg -> (
            let loc = fst lexpr.Sexpr.lexpr_loc in
            let pos = (loc.Lexing.pos_fname,
                       loc.Lexing.pos_lnum,
                       loc.Lexing.pos_cnum)
            in
            raise (Sexpr.Specs_error (msg, Some pos))
        )
        )
    | Sexpr.PLsizeofE lexpr1 -> (
        let expr = translate_lexpr lexpr1 vars in
        let typ = Cexpr.type_of_expr expr in
        try
            Cexpr.make_sizeof_operation typ
        with Failure msg -> (
            let loc = fst lexpr.Sexpr.lexpr_loc in
            let pos = (loc.Lexing.pos_fname,
                       loc.Lexing.pos_lnum,
                       loc.Lexing.pos_cnum)
            in
            raise (Sexpr.Specs_error (msg, Some pos))
        )
        )
    | Sexpr.PLfalse ->
        Cexpr.CONST_I ((Cexpr.INTEGER, Num.Int (0), "0"))
    | Sexpr.PLtrue ->
        Cexpr.CONST_I ((Cexpr.INTEGER, Num.Int (1), "1"))
    | Sexpr.PLrel (lexpr1, relation, lexpr2) ->
        (translate_relation lexpr1 relation lexpr2 vars)
    | Sexpr.PLand (lexpr1, lexpr2) ->
        Cexpr.OPER(Cexpr.INTEGER, Cexpr.ANDAND,
                   (translate_lexpr lexpr1 vars),
                   (translate_lexpr lexpr2 vars),
                   Cexpr.NIL)
    | Sexpr.PLor (lexpr1, lexpr2) ->
        Cexpr.OPER(Cexpr.INTEGER, Cexpr.OROR,
                   (translate_lexpr lexpr1 vars),
                   (translate_lexpr lexpr2 vars),
                   Cexpr.NIL)
    | Sexpr.PLxor (lexpr1, lexpr2) ->
        Cexpr.OPER(Cexpr.INTEGER, Cexpr.XOR,
                   (translate_lexpr lexpr1 vars),
                   (translate_lexpr lexpr2 vars),
                   Cexpr.NIL) 
    | Sexpr.PLimplies (lexpr1, lexpr2) ->
        let lexpr1_neg = {Sexpr.lexpr_node = Sexpr.PLnot lexpr1;
                          Sexpr.lexpr_loc = lexpr1.Sexpr.lexpr_loc;} in
        Cexpr.OPER(Cexpr.INTEGER, Cexpr.OROR,
                   (translate_lexpr lexpr1_neg vars), 
                   (translate_lexpr lexpr2 vars),
                   Cexpr.NIL)
    | Sexpr.PLiff (lexpr1, lexpr2) ->
        let lexpr1_imp = {Sexpr.lexpr_node = Sexpr.PLimplies (lexpr1, lexpr2);
                          Sexpr.lexpr_loc = lexpr1.Sexpr.lexpr_loc;} in
        let lexpr2_imp = {Sexpr.lexpr_node = Sexpr.PLimplies (lexpr2, lexpr1);
                          Sexpr.lexpr_loc = lexpr2.Sexpr.lexpr_loc;} in
        Cexpr.OPER(Cexpr.INTEGER, 
                   Cexpr.ANDAND, 
                   translate_lexpr lexpr1_imp vars,
                   translate_lexpr lexpr1_imp vars,
	           Cexpr.NIL)
    | Sexpr.PLnot lexpr1 ->
        Cexpr.OPER(Cexpr.INTEGER, Cexpr.LOGNOT,
                   (translate_lexpr lexpr1 vars),
                   Cexpr.NIL,
                   Cexpr.NIL)
    | Sexpr.PLif (lexpr1, lexpr2_op, lexpr3) -> (
        let expr1 = translate_lexpr lexpr1 vars in
        let expr2 = match lexpr2_op with
                    | None -> Cexpr.NIL
                    | Some lexpr2 -> translate_lexpr lexpr2 vars in
        let expr3 = translate_lexpr lexpr3 vars in 
        Cexpr.make_conditional_expr Cexpr.COND expr1 expr2 expr3 
        )
    | Sexpr.PLvalid lexpr1 -> (
        let expr = translate_lexpr lexpr1 vars in
        let tyn = Cexpr.type_of_expr expr in
        Cexpr.PSEUDO_ACSL_VALID (tyn, expr)
        )
    | Sexpr.PLuniqueloc lexpr1 -> (
        let t1 = (translate_lexpr lexpr1 vars) in
        let tyn = Cexpr.type_of_expr t1 in
        Cexpr.make_binary_operation Cexpr.NE
                                    (Cexpr.PSEUDO_ACSL_UNIQUE (tyn, t1))
                                    Cexpr.constant_zero;
        )
    | Sexpr.PLuniqueblock (lexpr1, lexpr2) -> (
        let t1 = (translate_lexpr lexpr1 vars)
        and t2 = (translate_lexpr lexpr2 vars) in
        let tyn = Cexpr.type_of_expr t1 in
        Cexpr.make_binary_operation Cexpr.NE
                                    (Cexpr.PSEUDO_ACSL_UNIQUEBLOCK (tyn, t1, t2))
                                    Cexpr.constant_zero;
        )
    | Sexpr.PLapp (id, lexprs) -> (
        let ex = Cexpr.make_variable_reference id in
        let tyn = Cexpr.type_of_expr ex in
        let params = List.map (fun a_lexpr ->
                         translate_lexpr a_lexpr vars
                     ) lexprs in
        Cexpr.CALL (tyn, ex, params, None)
        )
    | Sexpr.PLconstant const -> (
            let loc = lexpr.Sexpr.lexpr_loc in
            translate_constant const loc
        )
    | Sexpr.PLfield lexpr1 -> (
        let ex = translate_lexpr lexpr1 vars in
        let tyn = Cexpr.type_of_expr ex in
        Cexpr.ALL_FIELD (tyn, ex)
        )
)

(******* translate function contracts to Cexpr.stmt_t ***********)

let num_temp_var = ref 0 

let rec make_oper_expr (exprs : Cexpr.expr_t list) = (
    match exprs with
     | [] -> assert false
     | [expr] -> expr
     | head::tail -> Cexpr.OPER( Cexpr.INTEGER,
                                 Cexpr.ANDAND, 
                                 head, 
                                 make_oper_expr tail,
                                 Cexpr.NIL))
let translate_zone (zone : Sexpr.lzone_t)
                   (vars : (string, Cexpr.expr_t) Hashtbl.t)
: Cexpr.expr_t * Sexpr.range_t option = (
    match zone with
    | Sexpr.Location lexpr -> (
        let range = (Sexpr.make_range  (fst lexpr.Sexpr.lexpr_loc)
                                       (snd lexpr.Sexpr.lexpr_loc))
        in
        (translate_lexpr lexpr vars, Some range))
    | Sexpr.Nothing ->
        (Cexpr.NIL, None)
)
let translate_assign (assign : Sexpr.lassign_t)
                     (vars : (string, Cexpr.expr_t) Hashtbl.t)
: Cexpr.stmt_t list = (
    let zone, zones = assign in
    let translated_zone, range_opt = translate_zone zone vars in
(*    let translated_zones = List.map (fun a_zone -> translate_zone a_zone vars)*)
(*    in                                                                        *)
    match range_opt with
    | Some range ->
        let range_note = make_range_note range in
        [range_note; Cexpr.RESET [translated_zone]]
    | None ->
        []
)
let translate_assign_in_func (assign : Sexpr.lassign_t)
                             (behav_name : string) 
                             (func_name : string) 
                             (vars : (string, Cexpr.expr_t) Hashtbl.t) = (
   let zone, zones = assign in
   let lexprs = List.fold_left (fun l a_zone ->
       match a_zone with
       | Sexpr.Location lexpr -> l @ [lexpr]
       | Sexpr.Nothing -> l
   ) [] (zone :: zones) in
   let exprs = translate_list_lexpr lexprs vars in
   (Hashtbl.add Ccommon.assign_in_func func_name exprs);
   (*let res = (  (* for future - Binhlt *)
   List.map (
        fun term -> 
            let ex = translate_term term vars  in
            let typ = Cexpr.type_of_expr ex in
            let name_var = (":#Temp_var_"^(string_of_int !num_temp_var)) in
            let temp_var = Cexpr.VREF(typ, name_var , 3, 0) in
            let expr_pre = Cexpr.SETF( Cexpr.INTEGER, 
                                              Cexpr.SET,
                                              temp_var,
                                              ex
                                        
                            ) in
            let expr_assert = Cexpr.OPER(
                                    Cexpr.INTEGER,
                                    Cexpr.EQ,
                                    ex,
                                    temp_var,
                                    Cexpr.NIL
                              ) in
            let stmts_post_temp = make_assert expr_assert in
            let stmts_post = (
                       match behav_name with
                       | "behav_default" -> 
                              [stmts_post_temp]
                       | _ -> (
                                 let name = "spec_var_"^behav_name^"_temp" in
                                 let condition = Cexpr.VREF(Cexpr.INTEGER, 
                                                            name, 3, 0) in 
			         make_if_statement condition  [stmts_post_temp]
                                 )
	    )
	    in
            let (decl : Cexpr.declaration_t)  = {
                   Cexpr.decl_id= (name_var, 3); 
                   Cexpr.decl_tn= typ;
                   Cexpr.decl_scls= Cexpr.NILSCLS;
                   Cexpr.decl_inits = [ expr_pre ];
                   Cexpr.decl_params= [];
                   Cexpr.decl_struct_defined_p= false;
           }
           in
           let def = Cexpr.DEFVAR(decl) in
           [([behav_name], PRE_FUN, [def]);
            ([behav_name], POST_FUN, stmts_post)]
           (*res := (!res) @ [([behav_name], PRE_FUN, [def])]
                         @ [([behav_name], POST_FUN, stmts_post)]*)
   ) terms) in
   (List.flatten res) *)
   []
)

let translate_assigns_in_func (assigns : Sexpr.lassign_t list)
                             (behav_name : string) 
                             (func_name : string) 
                             (vars : (string, Cexpr.expr_t) Hashtbl.t) = (
    List.fold_left (fun l assign ->
        l @ (translate_assign_in_func assign behav_name func_name vars)
    ) [] assigns;
)
(* FIXME: to handle location information *)
(* translate pre/post condition to Cexpr.stmt_t  *)
let translate_function_condition 
       (func_conds : Sexpr.lfunc_condition_t list)
       (vars : (string, Cexpr.expr_t) Hashtbl.t) 
       (name_behavior:string) (func_name : string)
       (behav_cond : Sexpr.lexpr list)
       (req_to_assume : bool) = (
   (*let res = ref ([] : insert_triple list)in    *)
   List.map (fun func_cond ->		   
		   let (fun_type, lexpr) = func_cond in
       let range = (Sexpr.make_range  (fst lexpr.Sexpr.lexpr_loc)
                                      (snd lexpr.Sexpr.lexpr_loc))
       in
       let range_node = make_range_note range in
       let is_pre_cond = (fun_type == Sexpr.PRE_CONDITION) in
		   let stmts_spec, exprs = (
           match name_behavior with
           | "behav_default" -> (
             (begin
	                  let new_expr = translate_lexpr lexpr vars in
                    let stmts = (
                        if (req_to_assume && is_pre_cond) then
                            (make_assumes_and_assert None [new_expr] true)
                        else
                            (make_list_assert [new_expr])) in
                    let stmts = (range_node :: stmts) in
                    (stmts, [new_expr])
				     end))
           | _ -> (
             let name = "spec_var_"^name_behavior^"_temp" in
             let condition = Cexpr.VREF(Cexpr.INTEGER, name, 3, 0) in 
             let exs1 = (translate_list_lexpr behav_cond vars ) in
             let cond = make_oper_expr exs1 in
             let ex = (translate_lexpr lexpr vars)in
             let exs_ret = (
                 Cexpr.OPER (Cexpr.INTEGER, 
                                     Cexpr.OROR, 
                                     Cexpr.OPER (Cexpr.INTEGER,
                                                 Cexpr.LOGNOT, cond,
                                                 Cexpr.NIL, Cexpr.NIL),
                                                 ex,
                                     Cexpr.NIL)
              )in
             let stmts = (
                 if (req_to_assume && is_pre_cond) then
                     (make_assumes_and_assert (Some condition) [ex] false)
                 else 
                     (make_if_statement condition (make_list_assert [ex]))
             )in
             let stmts = (range_node :: stmts) in
             (stmts, [exs_ret]))
        ) in
        let is_result = (Sexpr.check_result_lexpr lexpr) in 
        let temp = (
            match fun_type, name_behavior with
            | Sexpr.PRE_CONDITION , "behav_default" -> 
              ([name_behavior], PRE_FUN, stmts_spec)
            | Sexpr.POST_CONDITION _, "behav_default" -> (
                (match exprs with
                | [] -> ()
                | _ ->
                  (Hashtbl.add (Ccommon.post_condition_hashtbl) func_name exprs));
                  if is_result == true then
                      ([name_behavior], POST_RESULT, stmts_spec)
                  else
                      ([name_behavior], POST_FUN, stmts_spec) 
              )
            | Sexpr.PRE_CONDITION , _ -> (
              ([name_behavior], PRE_FUN_BEHAV, stmts_spec))
            | Sexpr.POST_CONDITION _, _ -> (
                (match exprs with
                | [] -> ()
                | _ ->
                  (Hashtbl.add (Ccommon.post_condition_hashtbl)
                               func_name exprs));
                  if is_result == true then
                     ([name_behavior], POST_RESULT, stmts_spec)
                  else 
                      ([name_behavior], POST_FUN_BEHAV, stmts_spec)
              )       
        ) in
        temp
   ) func_conds
)


(* make a temp_var for condition of behavior  *)
let make_behavior_condition (str : string)
                            (vars : (string, Cexpr.expr_t) Hashtbl.t) 
                            (behav_cond : Sexpr.lexpr list) 
                            : insert_triple = ( 
   let lexprs = behav_cond in 
        begin
           let exs = (translate_list_lexpr lexprs vars ) in
           let and_exs = make_oper_expr exs in
           let name_var = ("spec_var_"^str^"_temp") in
           let temp_bool_var = Cexpr.VREF(Cexpr.INTEGER, name_var, 3, 0) in
           let expr = Cexpr.SETF( Cexpr.INTEGER, 
                                  Cexpr.SET, temp_bool_var, and_exs ) in
           let (decl : Cexpr.declaration_t)  = {
                   Cexpr.decl_id= (name_var, 3); 
                   Cexpr.decl_tn= Cexpr.INTEGER;
                   Cexpr.decl_scls= Cexpr.NILSCLS;
                   Cexpr.decl_inits = [ expr ];
                   Cexpr.decl_params= [];
                   Cexpr.decl_struct_defined_p= false;
                   Cexpr.decl_specs = None;
                   Cexpr.decl_location = None;
           }
           in
           let def = Cexpr.DEFVAR(decl) in
           ([], PRE_FUN_COND, [def])
        end)

let translate_behavior (behavs : Sexpr.lbehavior_t list) 
                       (vars : (string, Cexpr.expr_t) Hashtbl.t)
                       (func_name : string) 
                       (req_to_assume : bool)
                        : insert_triple list = ( 
   let res = ( (*ref ([] : insert_triple list) in*)
       List.map (
           function
             {
               Sexpr.b_name = str;
               Sexpr.b_assumes = behav_cond;
               Sexpr.b_conditions = fun_conds;
               Sexpr.b_assigns = assigns
             } ->
              let stmt_condition = (make_behavior_condition str vars 
                                                            behav_cond)
              in
              let behav_conditions = (translate_function_condition 
                                          fun_conds vars str func_name 
                                          behav_cond req_to_assume)
              in
              let assgn_behavs = (translate_assigns_in_func 
                                     assigns str func_name vars)
              in
              ([stmt_condition] @ assgn_behavs @ behav_conditions)
       )behavs) in
    (List.flatten res))
let translate_local_invariant (iv : string * Sexpr.lexpr)
                              : insert_triple list = ( 
    let name, lexpr = iv in
    let name_var = "checking_"^name in
    let expr = Cexpr.VREF(Cexpr.INTEGER, name_var, 1, 0) in
    let stmt_pre = Cexpr.STMT(
                     Cexpr.SETF(
                         Cexpr.INTEGER,
                         Cexpr.INCR_POST,
                         expr,Cexpr.NIL
                         (*Cexpr.CONST_I(Cexpr.INTEGER, Num.Int (1), "1")*)))
    in
    let stmt_post = Cexpr.STMT(
                     Cexpr.SETF(
                         Cexpr.INTEGER,
                         Cexpr.DECR_POST,
                         expr, Cexpr.NIL
                         (*Cexpr.CONST_I(Cexpr.INTEGER, Num.Int (0), "0")*)))
    in
    [([], PRE_FUN, [stmt_pre]); ([], POST_FUN, [stmt_post])]
)
let translate_local_invariants ivs = (
    List.flatten (List.map translate_local_invariant ivs)
)

(** Determine whether a lexpr is in a lexpr list or not 
*)
let is_mem_of_lexpr_list (lex : Sexpr.lexpr)
                         (lexs : Sexpr.lexpr list)
                         (vars : (string, Cexpr.expr_t) Hashtbl.t)
: bool =
begin
    let is_equal (lex1 : Sexpr.lexpr)
                 (lex2 : Sexpr.lexpr)
                 (vars : (string, Cexpr.expr_t) Hashtbl.t)
                 : bool = (
        let expr1 = translate_lexpr lex1 vars in
        let expr2 = translate_lexpr lex2 vars in
        Cexpr.compare_expr expr1 expr2)
    in
    List.exists (fun a_lexpr -> is_equal lex a_lexpr vars) lexs       
end

(** Check a lexpr list and remove redundant lexprs 
*)
let rec check_redundant_lexpr_list (lexs : Sexpr.lexpr list)
                                   (lexs_res : Sexpr.lexpr list)
                                   (vars : (string, Cexpr.expr_t) Hashtbl.t)
                                   : Sexpr.lexpr list = 
begin
    match lexs with
    | [] -> lexs_res
    | lex :: lex_remains -> (
        if (is_mem_of_lexpr_list lex lexs_res vars) then
            check_redundant_lexpr_list lex_remains lexs_res vars
        else
            check_redundant_lexpr_list lex_remains (lexs_res @ [lex]) vars)
end

let translate_function (func_contracts : Sexpr.lfunction_contract_t) 
                       (vars : (string, Cexpr.expr_t) Hashtbl.t)
                       (func_name : string) 
                       (req_to_assume : bool)
                       : insert_triple list = ( 
    match func_contracts with
      {
        Sexpr.fspec_conditions = func_conds;
        Sexpr.fspec_behaviors = behavs;
        Sexpr.fspec_assigns = assigns;
        Sexpr.fspec_prohibits = _;
        Sexpr.fspec_local_invs = ivs;
        Sexpr.fspec_context = _t;
        Sexpr.fspec_is_handler = _
      } ->
    let old_lexprs_raw = 
        (Sexpr.get_old_term_from_function_contract func_conds behavs)    
    in
    (* remove old redundant lexprs *)
    let old_lexprs = check_redundant_lexpr_list old_lexprs_raw [] vars in
    (* get expr of lexprs *) 
    let old_exprs = 
        List.map (fun a_arrow -> 
                     translate_lexpr a_arrow vars)
                 old_lexprs          
    in
    (* make stmt for all of old expr *)
    let old_stmts = (Ccommon.make_old_exprs_stmts old_exprs) in
    [([], PRE_FUN, old_stmts)]
    @ (translate_function_condition func_conds vars
                                    "behav_default" 
                                    func_name [] req_to_assume)
    @ (translate_assigns_in_func assigns
                                 "behav_default" 
                                 func_name vars )
    @ (translate_behavior behavs vars func_name req_to_assume)
    @ (translate_local_invariants ivs)
)
(*********** translate ghost code  ****************)
(*
let make_expr_of_ghost (name : string) 
                       (typ : Sexpr.ctype_t) (level : int) = ( 
   let type_ghost = (Scommon.convert_Ctype_to_typenode_t typ) in
   Cexpr.VREF(type_ghost, name , level, 0))
*)
(*
let translate_define (ghost_define : Sexpr.ghost_define_t) 
                     (vars : (string, Cexpr.expr_t) Hashtbl.t)
                     (ghost_vars : (string, Cexpr.expr_t) Hashtbl.t) 
                     (level : int) 
                     : insert_triple = (
    let new_vars = Hashtbl.copy vars in
    Hashtbl.iter( fun key value -> Hashtbl.add new_vars key value) ghost_vars;
    let (typ, id, ghost_init)  = ghost_define in 
    if (Hashtbl.mem vars id) then 
        raise (Sexpr.Specs_error
	                  ("Duplicate non-ghost variable " ^ id, None));
    let ex1 = (make_expr_of_ghost id typ level) in
    Hashtbl.replace ghost_vars id ex1;
    let init_exps = (
        match ghost_init with
        | Some(init_ghost) -> 
            begin
               let ter2 = init_ghost in 
               let ex2 = (translate_term ter2 new_vars ) in
               [Cexpr.SETF(Cexpr.INTEGER, Cexpr.SET, ex1, ex2)]
            end
	| None -> []
   ) in
   let (def_var : Cexpr.declaration_t) = {
       Cexpr.decl_id = (id, level);
       Cexpr.decl_tn = Cexpr.type_of_expr  ex1;
       Cexpr.decl_params = [];
       Cexpr.decl_scls= Cexpr.NILSCLS;
       Cexpr.decl_struct_defined_p = false;
       Cexpr.decl_inits = init_exps;
   } in
 ( [], HERE, [Cexpr.DEFVAR(def_var)] )
 )
*)
let rec get_name_struct (tyn : Cexpr.typenode_t) : string = (
    match tyn with
    |Cexpr.STRUCTURE_TYPE (struct_type) -> (
         struct_type.Cexpr.strc_tag
     )
    | Cexpr.POINTER_TO (ty) -> get_name_struct ty
    | _ -> assert false
)
(*
let check_fields1 (id : string)
                 (struct_name : string)
                 (modelfields : (string , string list) Hashtbl.t) = (
    try
        let field_names = 
            Hashtbl.find modelfields struct_name in
        if (List.mem id field_names) then None
        else Some (id, struct_name)
    with 
    | Not_found -> None
)
let check_fields (ter : Sexpr.term) 
                 (vars : (string, Cexpr.expr_t) Hashtbl.t)
                 (modelfields : (string, string list) Hashtbl.t)
                 = (
    match ter with
    | Sexpr.TERM_DEF_FIELD (ter1 , id) -> (
        let ex = translate_term ter1 vars in
        let tyn = Cexpr.type_of_expr ex in
        let struct_name = get_name_struct tyn in
        (check_fields1 id struct_name modelfields)
     )
    | Sexpr.TERM_REF_FIELD (ter1, id) ->  (
        let ex = translate_term ter1 vars  in
        let tyn = Cexpr.type_of_expr ex in
        let struct_name = get_name_struct tyn in
        (check_fields1 id struct_name modelfields)
     )
    | _ -> None
)
*)
(*
let translate_assign (ghost_ass : Sexpr.ghost_assign_t) 
                     (vars : (string, Cexpr.expr_t) Hashtbl.t) 
                     (ghost_vars : (string, Cexpr.expr_t) Hashtbl.t)
                     (level : int) 
                     (modelfields : (string , string list) Hashtbl.t)
                     : insert_triple = (
    let new_vars = Hashtbl.copy vars in
    Hashtbl.iter( fun key value -> Hashtbl.add new_vars key value) ghost_vars;
    let (ter1, ter2) = ghost_ass in
    let expr1 = (
        (match (check_fields ter1 new_vars modelfields) with
        | None -> ()
        | Some (f,ns) -> 
            raise (Sexpr.Specs_error ("Field '" ^ f
                                        ^ "' of struct '" ^ ns 
                                        ^ "' is not a model field",
                                        None)));
        let id = Scommon.get_name_of_term ter1 in
        let boo = (Hashtbl.mem vars id) in
        if (boo == true) then
            raise (Sexpr.Specs_error
                       ("Forbidded assignment a non-ghost variable: " ^ id,
                        None))
        else (
            try Hashtbl.find ghost_vars id
            with 
            | Not_found -> (translate_term ter1 new_vars))               
    ) 
    in
    let expr2 = translate_term ter2 new_vars in
    let typ = Cexpr.type_of_expr expr1 in
    ([], HERE, [Cexpr.STMT (Cexpr.SETF (typ,
                                        Cexpr.SET,
                                        expr1,
                                        expr2 ))])
  )
*)
(*
let translate_ghost_code (ghost_content : Sexpr.ghost_code_t) 
                         (vars : (string, Cexpr.expr_t) Hashtbl.t) 
                         (ghost_vars : (string, Cexpr.expr_t) Hashtbl.t)                            
                         (level : int ) 
                         (modelfields : (string , string list) Hashtbl.t)
                         : insert_triple = ( 
   match ghost_content with
   | Sexpr.GHOST_DEF(ghost_def) -> 
           translate_define ghost_def vars ghost_vars level
   | Sexpr.GHOST_ASG(ghost_ass) -> 
           translate_assign ghost_ass vars ghost_vars level modelfields
)
*)
(*
let translate_ghost_codes 
              (ghost_codes : Sexpr.ghost_code_t list)
              (vars : (string, Cexpr.expr_t) Hashtbl.t) 
              (ghost_vars : (string, Cexpr.expr_t) Hashtbl.t) 
              (level : int) 
              (modelfields : (string , string list) Hashtbl.t) = ( 
   List.map (
       fun ghost_code ->
             (translate_ghost_code ghost_code 
                                   vars 
                                   ghost_vars 
                                   level 
                                   modelfields)
   ) ghost_codes
)
*)
(***** translate statement specification to Cexpr.stmt_t *********)

let translate_stmt_assert
        (stmt_assert : Sexpr.identifier_t list * Sexpr.lexpr)
        (vars : (string, Cexpr.expr_t) Hashtbl.t) 
        : insert_triple = ( 
    let (strs, lexpr) = stmt_assert in
    let range = (Sexpr.make_range  (fst lexpr.Sexpr.lexpr_loc)
                                   (snd lexpr.Sexpr.lexpr_loc))
    in
    let range_note = make_range_note range in
    (strs, HERE, range_note
                 :: (make_list_assert [(translate_lexpr lexpr vars)]))
)
   
let translate_stmt_loop (stmt_loop : Sexpr.lloop_statement_t)  
                        (vars : (string, Cexpr.expr_t) Hashtbl.t)
                        = (
   match stmt_loop with
   | Sexpr.STMT_RETURN (ids, expr) -> 
       (ids, POS_RETURN, make_list_assert [(translate_lexpr expr vars)])
   | Sexpr.STMT_BREAK (ids, expr) -> 
       (ids, POS_BREAK, make_list_assert [(translate_lexpr expr vars)])
   | Sexpr.STMT_CONTINUE (ids, expr) -> 
       (ids, POS_CONTINUE, make_list_assert [(translate_lexpr expr vars)])
   | Sexpr.STMT_LOOP_INV (ids, expr) -> 
       (ids, PRE_LOOP, make_list_assert [(translate_lexpr expr vars)])

(*    | Sexpr.STMT_RETURN(stmt_return) ->  *)
(*            (translate_stmt_return stmt_return vars) *)
(*    | Sexpr.STMT_BREAK(stmt_break) ->  *)
(*            (translate_stmt_break stmt_break vars) *)
(*    | Sexpr.STMT_CONTINUE(stmt_continue) ->  *)
(*            (translate_stmt_continue stmt_continue vars) *)
(*     | Sexpr.STMT_LOOP_INV (stmt_loop) ->  *)
(*             (translate_stmt_loop_inv stmt_loop vars) *)
)

let rec translate_stmt_loops (stmt_loops : Sexpr.lloop_statement_t list) 
                             (vars : (string, Cexpr.expr_t) Hashtbl.t) = (
   match stmt_loops with 
   | [] -> []
   | head :: tail -> 
       (translate_stmt_loop head vars ) :: 
       (translate_stmt_loops tail vars )
)


let translate_stmt_assume (assume : Sexpr.lexpr)
                          (vars : (string, Cexpr.expr_t) Hashtbl.t)
: Cexpr.stmt_t list = (
    let ex : Cexpr.expr_t = (translate_lexpr assume vars) in
    let range = (Sexpr.make_range  (fst assume.Sexpr.lexpr_loc)
                                   (snd assume.Sexpr.lexpr_loc))
    in
    let range_note = make_range_note range in
    [range_note; Cexpr.ASSUME ex]
)
let translate_stmt_reset (reset : Sexpr.lassign_t list) (vars)
: Cexpr.stmt_t list = (
    List.fold_left (fun stmts assign -> stmts @ (translate_assign assign vars))
                   [] reset
)
let translate_stmt (stmt_contracts : Sexpr.lstatement_contract_t) 
                   (vars : (string, Cexpr.expr_t) Hashtbl.t) 
                   (*(level : int) *) =(
    match stmt_contracts with
    | Sexpr.STMT_ASSERT (ids, lexpr) ->
        [translate_stmt_assert (ids, lexpr) vars]
    | Sexpr.STMT_ASSUME lexpr -> (
        let stmts = (translate_stmt_assume lexpr vars) in
        [([], HERE, stmts)]
    )
    | Sexpr.STMT_RESET assigns -> (
      let stmts = translate_stmt_reset assigns vars in
        [([], HERE, stmts)]
    )
    | _ -> assert false


)

(*********** translate data invariant to Cexpr.stmt_t*****************)

(*  make a function with : 
    name_fun, list of ( name and type ) of parameter *)
let make_function_header (name_fun : string) (name_params : string list)
                         (type_param : Cexpr.typenode_t list) = ( 
   let typ_param = (
       match type_param with
         | [] -> Cexpr.FUNCTION_OF (Cexpr.VOID_TYPE, None)
         | _ -> Cexpr.FUNCTION_OF (Cexpr.VOID_TYPE, Some (type_param))
   ) 
   in
   let (return_header_function : Cexpr.declaration_t) = {
        Cexpr.decl_id = (name_fun, 1);
        Cexpr.decl_tn = typ_param;
        Cexpr.decl_params = name_params;
        Cexpr.decl_scls = Cexpr.NILSCLS;
        Cexpr.decl_inits = [];
        Cexpr.decl_struct_defined_p = false;
        Cexpr.decl_specs = None;
        Cexpr.decl_location = None;
        (*Cexpr.decl_attributes = [];*)
   }in
   return_header_function)

(* translate global invariant *)
let translate_global_invariant 
          (global_inv : Sexpr.lglobal_invariant_t)
          (vars : (string, Cexpr.expr_t) Hashtbl.t)
          = (
   (* currently skip the invariant strength *)
   let {
         Sexpr.ginv_strength = _;
         Sexpr.ginv_name = name_fun;
         Sexpr.ginv_expr = lexpr;
         Sexpr.ginv_exceptions = except
       } = global_inv in 
   let stmt = (
(*       if (check_pred_if_predicates preds ) = true then
           let exprs, stmts_temp = (translate_pred_if preds vars) in 
           stmts_temp
       else (
*)
         let expr = (translate_lexpr lexpr vars) in
         (make_assert expr) 
   )in
   let check_var = Cexpr.VREF(Cexpr.INTEGER, "checking_" ^ name_fun,1,0) in
   let if_stmt = (Cexpr.IF(check_var, [stmt], [])) in
   let fun_header = (make_function_header name_fun [] []) in
   let note_stmt = Cexpr.NOTE(Cexpr.LINE(true,"(noname)",0,0)) in
   let begin_stmt = Cexpr.BEGIN ([note_stmt; if_stmt; note_stmt]) in
   ([], FUNCTION_INVARIANT, Cexpr.DEFUN (fun_header, begin_stmt))
)

(* translate type invariant *)    
let translate_type_invariant (type_inv : Sexpr.ltype_invariant_t) 
                             (vars : (string, Cexpr.expr_t) Hashtbl.t)
                             = (
   (* currently skip the invariant strength *)
   let {
         Sexpr.tinv_strength = _;
         Sexpr.tinv_name = name_fun;
         Sexpr.tinv_type = typ;
         Sexpr.tinv_param_name = name_var;
         Sexpr.tinv_expr = lexpr;
         Sexpr.tinv_exceptions = except
       } = type_inv in
   let ctyp = translate_logictype typ in
   let var_expr = Cexpr.VREF(ctyp, name_var, 2, 0) in
   (Hashtbl.add vars name_var var_expr);
   let stmt = (
(*       if (check_pred_if_predicates preds ) = true then
           let exprs, stmts_temp = (translate_pred_if preds vars) in 
           stmts_temp
       else (
*)
         let expr = (translate_lexpr lexpr vars) in
         (make_assert expr) 
(*       )*)
   )in
   let fun_header = (make_function_header name_fun [name_var] [ctyp]) in
   let note_stmt = Cexpr.NOTE(Cexpr.LINE(true,"(noname)",0,0)) in
   let begin_stmt = Cexpr.BEGIN([note_stmt] @ [stmt] @ [note_stmt]) in
   ([], 
                 FUNCTION_INVARIANT, 
                 Cexpr.DEFUN(fun_header, begin_stmt))
)
(************** translate common specification *****************)

let translate (content : Sexpr.lspec_t) 
              (vars : (string, Cexpr.expr_t) Hashtbl.t) 
              (ghost_vars : (string, Cexpr.expr_t) Hashtbl.t)
              (level : int) (func_name : string) 
              (modelfields : (string , string list) Hashtbl.t option) 
              (req_to_assume : bool)= (
    (*copy_hash_to_hash vars_list;*)
    let new_vars = Hashtbl.copy vars in
    Hashtbl.iter(
        fun key value -> Hashtbl.add new_vars key value
    )ghost_vars; 
    match content with
    | Sexpr.SPEC_FUN (fun_contract) ->
        translate_function fun_contract new_vars func_name req_to_assume
    | Sexpr.SPEC_STMT (stmt_contracts) -> 
        List.fold_left (fun translated_stmts stmt_contract ->
                            translated_stmts @
                                (translate_stmt stmt_contract new_vars))
                       [] stmt_contracts
    | Sexpr.STMT_LOOP (stmt_loops) -> 
        (translate_stmt_loops stmt_loops vars)
(*    | Sexpr.SPEC_GHOST (ghost_codes) -> assert false *)
(*
        (
        let m_fields = match modelfields with
                       | Some a -> a
                       | None -> assert false
        in
        translate_ghost_codes ghost_codes 
                              vars 
                              ghost_vars 
                              level 
                              m_fields)
*)
    | Sexpr.SPEC_GLOBAL_INV (global_inv) ->
        assert false
    | Sexpr.SPEC_TYPE_INV (type_inv) ->
        assert false
    | Sexpr.SPEC_MODEL_FIELD _
    | Sexpr.SPEC_ENUM _
    | Sexpr.SPEC_VAR _
    | Sexpr.SPEC_ANYTHING -> []
)

(******************************************)
(* translate types in Sexpr to term or form *)
let oper_relation_to_relop_t (r : Cexpr.op_t) : Form.relop_t = (
    match r with
    | Cexpr.EQ -> Form.EQ
    | Cexpr.NE -> Form.NE
    | Cexpr.GE -> Form.GE
    | Cexpr.GT -> Form.GT
    | Cexpr.LE -> Form.LE
    | Cexpr.LT -> Form.LT
    | _ -> assert false
)

let rec bcexpr_to_term (expr1 : Cexpr.expr_t) (op : Cexpr.op_t)
                   (expr2 : Cexpr.expr_t)
: Form.term_t = (
    let term1 = cexpr_to_term expr1 in
    let term2 = cexpr_to_term expr2 in
    match op with
    | Cexpr.ADD -> Form.ARITH (Form.ADD(term1,term2))
    | Cexpr.SUB -> Form.ARITH (Form.ADD (term1, Form.ARITH (Form.NEG(term2))))
    | Cexpr.MUL -> Form.ARITH (Form.MUL(term1, term2))
    | Cexpr.DIV -> Form.UIFUN [(Form.FUN "div") ; term1 ; term2]
    | Cexpr.MOD -> Form.UIFUN [(Form.FUN "mod") ; term1 ; term2]
    | Cexpr.LSH -> Form.UIFUN [(Form.FUN "lsh") ; term1 ; term2]
    | Cexpr.RSH -> Form.UIFUN [(Form.FUN "rsh") ; term1 ; term2]
    | Cexpr.AND -> Form.UIFUN [(Form.FUN "and") ; term1 ; term2]
    | Cexpr.IOR -> Form.UIFUN [(Form.FUN "ior") ; term1 ; term2]
    | Cexpr.XOR -> Form.UIFUN [(Form.FUN "xor") ; term1 ; term2]
    | _ -> assert false
)

and expr_oper_to_term (expr1 : Cexpr.expr_t) (op : Cexpr.op_t)
                      (expr2 : Cexpr.expr_t)
: Form.term_t = (
    let term1 = cexpr_to_term expr1 in
    match op with
    | Cexpr.NEG -> Form.ARITH (Form.NEG (term1))
    | Cexpr.CAST _ -> term1
    | Cexpr.BITNOT -> (Form.UIFUN [(Form.FUN "bitnot") ; term1])
    | Cexpr.ADD | Cexpr.SUB | Cexpr.MUL | Cexpr.DIV | Cexpr.MOD | Cexpr.LSH
    | Cexpr.RSH | Cexpr.AND | Cexpr.IOR | Cexpr.XOR ->
        bcexpr_to_term expr1 op expr2
    | _ -> assert false
)

and expr_oper_to_form (expr1 : Cexpr.expr_t) (op : Cexpr.op_t)
                      (expr2 : Cexpr.expr_t) (expr3 : Cexpr.expr_t)
: Form.form_t = (
    match op with
    | Cexpr.SET
    | Cexpr.INCR_PRE | Cexpr.DECR_PRE | Cexpr.INCR_POST | Cexpr.DECR_POST ->
        assert false
    | Cexpr.CAST _ -> assert false
    | Cexpr.NEG -> assert false
    | Cexpr.BITNOT -> assert false
    | Cexpr.LOGNOT -> assert false
    | Cexpr.EQ | Cexpr.NE | Cexpr.GE | Cexpr.GT | Cexpr.LE | Cexpr.LT ->
        let relop = oper_relation_to_relop_t op in
        let form1 = cexpr_to_term expr1 in
        let form2 = cexpr_to_term expr2 in
        Form.PRED (relop, form1, form2)
    | Cexpr.COND -> (
      let form1 = cexpr_to_form expr1 in
      let form2 = cexpr_to_form expr2 in
      let form3 = cexpr_to_form expr3 in
      let and1 = Form.AND [form1; form2] in
      let and2 = Form.AND [(Form.NOT form1); form3] in
      Form.OR [and1; and2]
      )
    | Cexpr.OROR -> (
        let form1 = cexpr_to_form expr1 in
        let form2 = cexpr_to_form expr2 in
        Form.OR [form1; form2]
      )
    | Cexpr.ANDAND -> (
        let form1 = cexpr_to_form expr1 in
        let form2 = cexpr_to_form expr2 in
        Form.AND [form1; form2]
      )
    | Cexpr.COMMA -> assert false
    | Cexpr.IMP -> (
        let form1 = cexpr_to_form expr1 in
        let form2 = cexpr_to_form expr2 in
        Form.IMP (form1, form2)
      )
    | Cexpr.IFF ->  (
        let form1 = cexpr_to_form expr1 in
        let form2 = cexpr_to_form expr2 in
        Form.IFF (form1, form2)
      )
    | Cexpr.PSEUDO_AMP | Cexpr.PSEUDO_STR | Cexpr.PSEUDO_PLS
    | Cexpr.PSEUDO_MNS | Cexpr.PSEUDO_TLD | Cexpr.PSEUDO_BAN -> assert false
    | _ -> assert false
)

and cexpr_to_term (expr : Cexpr.expr_t) : Form.term_t = (
    match expr with
    | Cexpr.NIL -> assert false
    | Cexpr.PSEUDO_NONDET_VALUE _ -> assert false
    | Cexpr.OPER (_, op, expr1, expr2, expr3) ->
        expr_oper_to_term expr1 op expr2
    | Cexpr.SETF _ -> assert false
    | Cexpr.CALL (_, expr1, exprs1, _) -> (
        let term = cexpr_to_term expr1 in
        let terms = List.map cexpr_to_term exprs1 in
        (Form.UIFUN (term :: terms))
      )
    | Cexpr.VREF (_, id, _, _) ->
        Form.VAR id
    | Cexpr.PREF (_, _, expr1, expr2)-> (
        let term1 = cexpr_to_term expr1
        and term2 = cexpr_to_term expr1 in
        Form.UIFUN ([Form.FUN "pref"; term1; term2])
      )
    | Cexpr.SREF (_, _, expr1, id) -> (
        let term = cexpr_to_term expr1 in
        (Form.UIFUN [(Form.FUN ("sref")); term; Form.VAR (id ^ "%0")])
      )
    | Cexpr.DSGN_AREF _ -> assert false
    | Cexpr.ADDROF (_, expr1) -> (
        let term = cexpr_to_term expr1 in
        Form.UIFUN ([Form.FUN "addrof"; term])
      )
    | Cexpr.SIZEOF _ -> assert false
    | Cexpr.PROGN _ -> assert false
    | Cexpr.CONST_I (_, _, str) -> (
        let value, _ = translate_int_constant str in
        Form.ARITH (Form.NUM (value))
      )
    | Cexpr.CONST_F (_, _, str) -> assert false
    | Cexpr.STRING _ -> assert false
    | Cexpr.CLITERAL _ -> assert false
    | Cexpr.ASMSEQUENCE _ -> assert false
    | Cexpr.PHI _ -> assert false
    | Cexpr.PSEUDO_MEMORY _ -> assert false
    | Cexpr.INIT_FILL _ -> assert false
    | Cexpr.PRAGMA _ -> assert false
    | Cexpr.QUANTIFIER _ -> assert false
    | Cexpr.PSEUDO_PROLOGUE
    | Cexpr.PSEUDO_EPILOGUE -> assert false
    | Cexpr.PSEUDO_ACSL_RESULT (_, expr1) ->
        cexpr_to_term expr1
    | Cexpr.PSEUDO_ACSL_OLD (_, expr1, expr2) -> (
        cexpr_to_term expr2
      )
    | Cexpr.PSEUDO_ACSL_VALID _
    | Cexpr.ALL_FIELD _ -> assert false (* have not proccessed yet *)
)

and cexpr_to_form (expr : Cexpr.expr_t) : Form.form_t = (
    match expr with
    | Cexpr.OPER (_, op, expr1, expr2, expr3) ->
        expr_oper_to_form (expr1) (op) (expr2) (expr3)
    | Cexpr.NIL -> Form.TT
    | Cexpr.PSEUDO_NONDET_VALUE _ -> assert false
    | Cexpr.SETF _ -> assert false

    | Cexpr.CALL (_, expr1, exprs1, _) ->
        assert false
    | Cexpr.VREF _ -> assert false
    | Cexpr.PREF _ -> assert false
    | Cexpr.SREF _ -> assert false
    | Cexpr.DSGN_AREF _ -> assert false
    | Cexpr.ADDROF _ -> assert false
    | Cexpr.SIZEOF _ -> assert false
    | Cexpr.PROGN _ -> assert false

    | Cexpr.CONST_I _ -> assert false
    | Cexpr.CONST_F _ -> assert false
    | Cexpr.STRING _ -> assert false
    | Cexpr.CLITERAL _ -> assert false
    | Cexpr.ASMSEQUENCE _ -> assert false

    | Cexpr.PHI _ -> assert false
    | Cexpr.PSEUDO_MEMORY _ -> assert false
    | Cexpr.INIT_FILL _ -> assert false

    | Cexpr.PRAGMA _ -> assert false
    | Cexpr.QUANTIFIER _ -> assert false

    | Cexpr.PSEUDO_PROLOGUE
    | Cexpr.PSEUDO_EPILOGUE
    | Cexpr.PSEUDO_ACSL_RESULT _ -> assert false
    | Cexpr.PSEUDO_ACSL_OLD _ -> assert false
    | Cexpr.PSEUDO_ACSL_VALID _ -> assert false
    | Cexpr.ALL_FIELD _ -> assert false (* have not processed yet *)
)
and full_expr_to_form (full_expr : Cexpr.full_expr_t)
: Form.form_t * Sexpr.range_t option = (
    let form = cexpr_to_form full_expr.Cexpr.expr_node in
    let start_pos, end_pos = full_expr.Cexpr.expr_loc in
    let range = Sexpr.make_range start_pos end_pos in
    (form, Some range)
)

and full_expr_to_term (full_expr : Cexpr.full_expr_t)
: Form.term_t * Sexpr.range_t option = (
    let term = cexpr_to_term full_expr.Cexpr.expr_node in
    let start_pos, end_pos = full_expr.Cexpr.expr_loc in
    let range = Sexpr.make_range start_pos end_pos in
    (term, Some range)
)

and cexprs_to_form (exprs : Cexpr.full_expr_t list) : Form.form_t = (
    let forms = List.map (fun expr ->
        fst (full_expr_to_form expr)
        ) exprs in
    match forms with
    | [] -> Form.TT
    | [p] -> p
    | _ -> Form.AND forms
)

(***************************************)
(* Translate to the Cspec_t type *)

let cexpr_of_lexpr (lexpr : Sexpr.lexpr) 
                   (vars : (string, Cexpr.expr_t) Hashtbl.t)
: Cexpr.full_expr_t = (
    match lexpr with
      {
        Sexpr.lexpr_node = _;
        Sexpr.lexpr_loc = location
      } ->
        let expr = translate_lexpr lexpr vars in
        {
          Cexpr.expr_node = expr;
          Cexpr.expr_loc = location
        }
)

let cfunccond_of_lfunccond (func_cond : Sexpr.lexpr Sexpr.func_condition_t)
                           (vars : (string, Cexpr.expr_t) Hashtbl.t)
: Cexpr.full_expr_t Sexpr.func_condition_t = (
    let pos, lexpr = func_cond in
    let cexpr = cexpr_of_lexpr lexpr vars in
    match pos with
    | Sexpr.PRE_CONDITION ->
        (pos, cexpr)
    | Sexpr.POST_CONDITION true ->
        (assert false)
    | Sexpr.POST_CONDITION false -> (
        if (Cexpr.full_effect_in_caller cexpr) then
            (Sexpr.POST_CONDITION true, cexpr)
        else
            (Sexpr.POST_CONDITION false, cexpr)))

let cvarcond_of_lvarcond (var_cond : Sexpr.lexpr Sexpr.var_condition_t)
                         (vars : (string, Cexpr.expr_t) Hashtbl.t)
: Cexpr.full_expr_t Sexpr.var_condition_t =
begin
    let (pos, lexpr) = var_cond in
    let cexpr = cexpr_of_lexpr lexpr vars in
    (pos, cexpr)
end

let cvarspec_of_lvarspec (var_spec : Sexpr.lexpr Sexpr.var_condition_t list)
                         (vars : (string, Cexpr.expr_t) Hashtbl.t)
: Cexpr.full_expr_t Sexpr.var_condition_t list =
begin
    List.map
        (fun a_v_cond ->
            cvarcond_of_lvarcond a_v_cond vars    
        ) var_spec
end
 
let cassign_of_lassign (assign : Sexpr.lassign_t)
                       (vars : (string, Cexpr.expr_t) Hashtbl.t)
: Cexpr.full_expr_t Sexpr.assign_t = (
    let translate_zone zone vars =
        match zone with
        | Sexpr.Location lexpr -> Sexpr.Location (cexpr_of_lexpr lexpr vars)
        | Sexpr.Nothing -> Sexpr.Nothing
    in     
    let zone, zones = assign in
    let translated_zone = translate_zone zone vars
    and translated_zones = List.map (fun z -> translate_zone z vars) zones
    in (translated_zone, translated_zones))

let cbehavior_of_lbehavior 
        (behavior : Sexpr.lbehavior_t)
        (vars : (string, Cexpr.expr_t) Hashtbl.t)
: (Cexpr.full_expr_t, Cexpr.full_expr_t) Sexpr.behavior_t = (        
    match behavior with
      {
        Sexpr.b_name = name;
        Sexpr.b_assumes = assumes;
        Sexpr.b_conditions = conds;
        Sexpr.b_assigns = assigns
      } -> (begin
        let translated_assumes = 
            List.map (fun a -> cexpr_of_lexpr a vars) assumes
        and translated_conds =
            List.map (fun cond -> cfunccond_of_lfunccond cond vars)
                     conds
        and translated_assigns =
            List.map (fun a -> cassign_of_lassign a vars) assigns
        in
        {
          Sexpr.b_name = name;
          Sexpr.b_assumes = translated_assumes;
          Sexpr.b_conditions = translated_conds;
          Sexpr.b_assigns = translated_assigns
        }
        end)
    )
         
let cfunccontract_of_lfunccontract (funcspecs : Sexpr.lfunction_contract_t)
                                   (vars : (string, Cexpr.expr_t) Hashtbl.t)
: (Cexpr.full_expr_t, Cexpr.full_expr_t) Sexpr.function_contract_t = (
    match funcspecs with
      {
        Sexpr.fspec_conditions = func_conds;
        Sexpr.fspec_behaviors = behaviors;
        Sexpr.fspec_assigns = assigns;
        Sexpr.fspec_prohibits = prohibits;
        Sexpr.fspec_local_invs = local_invs;
        Sexpr.fspec_context = context;
        Sexpr.fspec_is_handler = is_handler
      } -> (
        let translated_func_conds =
            List.map (fun cond -> cfunccond_of_lfunccond cond vars)
                     func_conds
        and translated_behaviors =
            List.map (fun behav -> cbehavior_of_lbehavior behav vars)
                     behaviors
        and translated_assigns =
            List.map (fun a -> cassign_of_lassign a vars) assigns
        and translated_local_invs =
            List.map (fun (id, lexpr) -> (id, cexpr_of_lexpr lexpr vars))
                     local_invs
        in
        {
          Sexpr.fspec_conditions = translated_func_conds;
          Sexpr.fspec_behaviors = translated_behaviors;
          Sexpr.fspec_assigns = translated_assigns;
          Sexpr.fspec_prohibits = prohibits;
          Sexpr.fspec_local_invs = translated_local_invs;
          Sexpr.fspec_context = context;
          Sexpr.fspec_is_handler = is_handler
        })
    )

let cglobalinvariant_of_lglobalinvariant
        (global_inv : Sexpr.lglobal_invariant_t)
        (vars : (string, Cexpr.expr_t) Hashtbl.t)
: Cexpr.full_expr_t Sexpr.global_invariant_t = (
    let {
          Sexpr.ginv_strength = strength;
          Sexpr.ginv_name = id;
          Sexpr.ginv_expr = expr;
          Sexpr.ginv_exceptions = exceptionsopt
        } = global_inv in
    let translated_expr = cexpr_of_lexpr expr vars in
    {
      Sexpr.ginv_strength = strength;
      Sexpr.ginv_name = id;
      Sexpr.ginv_expr = translated_expr;
      Sexpr.ginv_exceptions = exceptionsopt
    }) 
                                             
let ctypeinvariant_of_ltypeinvariant
        (type_inv : Sexpr.ltype_invariant_t)
        (vars : (string, Cexpr.expr_t) Hashtbl.t)
: (Cexpr.full_expr_t, Cexpr.typenode_t) Sexpr.type_invariant_t = (
    let {
          Sexpr.tinv_strength = strength;
          Sexpr.tinv_name = id1;
          Sexpr.tinv_type = logictype;
          Sexpr.tinv_param_name = id2;
          Sexpr.tinv_expr = expr;
          Sexpr.tinv_exceptions = exceptionsopt
        } = type_inv in
    let translated_expr = cexpr_of_lexpr expr vars
    and translated_type = translate_logictype logictype
    in  {
          Sexpr.tinv_strength = strength;
          Sexpr.tinv_name = id1;
          Sexpr.tinv_type = translated_type;
          Sexpr.tinv_param_name = id2;
          Sexpr.tinv_expr = translated_expr;
          Sexpr.tinv_exceptions = exceptionsopt
        })

let cloopstatement_of_lloopstatement
        (specs : Sexpr.lloop_statement_t)
        (vars : (string, Cexpr.expr_t) Hashtbl.t)
: Cexpr.full_expr_t Sexpr.loop_statement_t = (
    match specs with
    | Sexpr.STMT_RETURN (ids, expr) ->
        Sexpr.STMT_RETURN (ids, cexpr_of_lexpr expr vars)
    | Sexpr.STMT_CONTINUE (ids, expr) ->
        Sexpr.STMT_CONTINUE (ids, cexpr_of_lexpr expr vars)
    | Sexpr.STMT_BREAK (ids, expr) ->
        Sexpr.STMT_BREAK (ids, cexpr_of_lexpr expr vars)
    | Sexpr.STMT_LOOP_INV (ids, expr) ->
        Sexpr.STMT_LOOP_INV (ids, cexpr_of_lexpr expr vars))

let cstatementcontract_of_lstatementcontract
        (specs : Sexpr.lstatement_contract_t)
        (vars : (string, Cexpr.expr_t) Hashtbl.t)
: (Cexpr.full_expr_t, Cexpr.full_expr_t) Sexpr.statement_contract_t = (
    match specs with
    | Sexpr.STMT_ASSUME expr ->
        Sexpr.STMT_ASSUME (cexpr_of_lexpr expr vars)
    | Sexpr.STMT_RESET assigns ->
        let translated_assigns =
            List.map (fun a -> cassign_of_lassign a vars) assigns
        in Sexpr.STMT_RESET translated_assigns
    | Sexpr.STMT_ASSERT (ids, expr) ->
        Sexpr.STMT_ASSERT (ids, cexpr_of_lexpr expr vars)
    | Sexpr.STMT_INIT x ->
        Sexpr.STMT_INIT x
    | Sexpr.STMT_RELEASE x ->
        Sexpr.STMT_RELEASE x)

let cspec_of_lspec (specs : Sexpr.lspec_t)
                   (vars : (string, Cexpr.expr_t) Hashtbl.t)
= ( 
    (keep_acsl_constructs := true);
    let translated_specs =
        match specs with
        | Sexpr.SPEC_FUN func_specs ->
            Sexpr.SPEC_FUN (cfunccontract_of_lfunccontract func_specs vars)
        | Sexpr.SPEC_STMT statement_specs ->
            let c_contracts =
                List.map
                    (fun l_contract ->
                         cstatementcontract_of_lstatementcontract
                             l_contract
                             vars)
                    statement_specs
            in
            Sexpr.SPEC_STMT c_contracts
        | Sexpr.STMT_LOOP loop_stats ->
            let translated_loop_stats =
                List.map (fun loopstat ->
                              cloopstatement_of_lloopstatement loopstat vars)
                         loop_stats
            in
            Sexpr.STMT_LOOP translated_loop_stats
        | Sexpr.SPEC_GLOBAL_INV global_inv ->
            Sexpr.SPEC_GLOBAL_INV (cglobalinvariant_of_lglobalinvariant
                                       global_inv
                                       vars)
        | Sexpr.SPEC_TYPE_INV type_inv ->
            Sexpr.SPEC_TYPE_INV (ctypeinvariant_of_ltypeinvariant
                                     type_inv
                                     vars)
        | Sexpr.SPEC_VAR var_specs -> (
            Sexpr.SPEC_VAR (cvarspec_of_lvarspec var_specs vars)
        )
        | Sexpr.SPEC_MODEL_FIELD _
        | Sexpr.SPEC_ENUM _
        | Sexpr.SPEC_ANYTHING -> assert false
    in
    (keep_acsl_constructs := false);
    translated_specs)

