(* parser/vector.ml (2009-01-26) *)
(* Copyright (C) 2007-2009 Yonezawa Lab, The University of Tokyo *)

(* Trivial Vector (with Fill Pointer) *)

let minimal_length = 8

type 'a t = {
  mutable size : int ;
  mutable junk : 'a ;
  mutable data : 'a array
}

let rec doubles_ (n : int) (ub : int) : int = (
  assert ((n * 2) > n) ;
  if (n >= ub) then n else (doubles_ (n * 2) ub))

let create (initialsize : int) (junkvalue : 'a) : 'a t = (
  let nlen = (doubles_ minimal_length initialsize) in
  {
    size = 0 ;
    junk = junkvalue ;
    data = (Array.make nlen junkvalue)
  })

let clear (v : 'a t) : unit = (
  v.size <- 0 ;
  (Array.fill v.data 0 (Array.length v.data) v.junk))

let copy (v : 'a t) : 'a t = (
  {
    size = v.size ;
    junk = v.junk ;
    data = (Array.copy (v.data))
  })

let length (v : 'a t) : int = (v.size)

let resize (v : 'a t) (lastindex : int) : unit = (
  if ((Array.length v.data) <= lastindex) then
    (begin
      assert ((Array.length v.data) > 0) ;
      assert (lastindex < Sys.max_array_length) ;
      let nlen = (doubles_ (Array.length v.data) (lastindex + 1)) in
      let ndata = (Array.make nlen (v.junk)) in
      (Array.blit v.data 0 ndata 0 (Array.length v.data)) ;
      v.data <- ndata
    end))

let add (v : 'a t) (value : 'a) : unit = (
  (resize v v.size) ;
  v.data.(v.size) <- value ;
  v.size <- (succ v.size))

let add_first (v : 'a t) (value : 'a) : unit = (
  (resize v v.size) ;
  (Array.blit v.data 0 v.data 1 v.size) ;
  v.data.(0) <- value ;
  v.size <- (succ v.size))

let rec find_if_loop_ (v : 'a t) f (i : int) : 'a option = (
  if (i >= v.size) then
    None
  else if (f v.data.(i)) then
    Some v.data.(i)
  else
    (find_if_loop_ v f (i + 1)))

let find_if (f : 'a -> bool) (v : 'a t) : 'a option = (find_if_loop_ v f 0)

let rec iter_loop_ f (v : 'a t) (i : int) : unit = (
  if (i >= v.size) then
    ()
  else
    (begin
      (f v.data.(i)) ;
      (iter_loop_ f v (i + 1))
    end))

let iter f (v : 'a t) : unit = (iter_loop_ f v 0)

let rec map_loop_ f (v : 'a t) (x : 'b array) (i : int) : unit = (
  if (i >= v.size) then
    ()
  else
    (begin
      x.(i) <- (f v.data.(i)) ;
      (map_loop_ f v x (i + 1))
    end))

let map f (v : 'a t) : 'b list = (
  if (v.size = 0) then
    []
  else
    (begin
      let e = (f v.data.(0)) in
      let x = (Array.make v.size e) in
      x.(0) <- e ;
      (map_loop_ f v x 1) ;
      (Array.to_list x)
    end))

let rec mapi_loop_ f (v : 'a t) (x : 'b array) (i : int) : unit = (
  if (i >= v.size) then
    ()
  else
    (begin
      x.(i) <- (f v.data.(i) i) ;
      (mapi_loop_ f v x (i + 1))
    end))

let mapi f (v : 'a t) : 'b list = (
  if (v.size = 0) then
    []
  else
    (begin
      let e = (f v.data.(0) 0) in
      let x = (Array.make v.size e) in
      x.(0) <- e ;
      (mapi_loop_ f v x 1) ;
      (Array.to_list x)
    end))

let to_array (v : 'a t) : 'a array = (
  let a = (Array.make v.size v.junk) in
  (Array.blit v.data 0 a 0 v.size) ;
  a)

let to_list (v : 'a t) : 'a list = (
  (Array.to_list (to_array v)))

let test_ () = (
  let v = (create 10 0) in
  (add v 11) ;
  (add v 12) ;
  (add v 13) ;
  (add v 14) ;
  (iter (fun v -> (print_endline ("v=" ^ (string_of_int v)))) v) ;
  let w = (create 10 (0, "junk")) in
  (add w (1, "one")) ;
  (add w (2, "two")) ;
  (add w (3, "three")) ;
  let x = (find_if (fun (i, s) -> (i = 2)) w) in
  (match x with
    None -> (print_endline ("find=None"))
  | Some (i, s) -> (print_endline ("find=" ^ (string_of_int i) ^ "/" ^ s))) ;
  let y = (map (fun (i, s) -> ((string_of_int i) ^ "/" ^ s)) w) in
  (List.map (fun s -> (print_endline ("v=" ^ s))) y)
)

(*let _ = (test_ ())*)
