(* testsuite/testclap.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

open OUnit
open Unix

type testsuite_options_t = {
  topt_output_filename : string;
  topt_clap_file : string;
  topt_include_dir : string;
  topt_desc_file : string;
  topt_is_xml : bool;
  topt_temp_dir : string;
}

(* Get full path *)
let get_full_path (pathname : string) : string = (
    let current_path = Sys.getcwd () in
    let (basename, dirname) =
        try (
            if (Sys.is_directory pathname) then
                ("", pathname)
            else
                (Filename.basename pathname, Filename.dirname pathname))
        with Sys_error _ ->
            failwith ("'" ^ pathname ^ "' is neither a file nor a directory")
    in
    try
        Sys.chdir dirname;
        let full_name = Sys.getcwd () in
        Sys.chdir current_path;
        Filename.concat full_name basename;
    with Sys_error _ -> assert false
)

let string_of_tm_in_iso8601 tm =
    Printf.sprintf "%04d-%02d-%02dT%02d:%02d:%02d"
        (tm.tm_year + 1900)
        tm.tm_mon
        tm.tm_mday
        tm.tm_hour
        tm.tm_min
        tm.tm_sec

let read_desc_file (desc_filename : string) : (string * string) list = (
    (* Read lines of desc file *)
    let lines = ref [] in
    let chan = try open_in desc_filename
               with Sys_error _ ->
                   failwith ("Cannot open file `" ^ desc_filename ^ "'") in
    try
        while true; do
        let line = input_line chan in
        let line = ExtString.String.strip ~chars:"\t " line in
        if ((String.length line) > 0)
           && not (ExtString.String.starts_with line "#") then
            lines := !lines @ [line];
        done; []
    with End_of_file ->
    close_in chan;

    (* Get command and safe information of it *)
    let rec get_info (strs : string list) : (string * string) list = (
        match strs with
        | [] -> []
        | head :: tails -> (
            let (ops, safe_info) =
                try ExtString.String.split head "|"
                with ExtString.Invalid_string (*Not_found*) ->
                    failwith ("Cannot find the expected reult in line `"
                              ^ head ^ "'")
            in
            (* remove character ' \t\r\n ' *)
            let safe_info = ExtString.String.strip safe_info in
            let ops = ExtString.String.strip ops in
            match safe_info with
              | "safe"
              | "unsafe"
              | "parser_error" -> (ops, safe_info) :: (get_info tails)
              | _ ->
                  if ExtString.String.starts_with safe_info "exit code "
                     || ExtString.String.starts_with safe_info "signal number"
                  then (ops, safe_info) :: (get_info tails)
                  else
                     failwith ("Expected result `"
                               ^ safe_info
                               ^ "' is not allowed")
        )
    ) in
    get_info !lines;
)

(* Read input arguments and get options.
   example input command: './test_suite ps -I ../includes/ test/mctest'
   then options: ["ps";"-I";"../includes"]
*)

module O = struct
    open OptParse
    let clap = StdOpt.str_option ~default:"clap" ()
    let incd = StdOpt.str_option ~default:"" ()
    let ofil = StdOpt.str_option ~default:"" ()
    let xml = StdOpt.store_true ()
    let tempdir = StdOpt.str_option ~default:"" ()

    let ops = OptParser.make ()

    open OptParser
;;
    add ops ~long_name:"clap" ~help:"path to clap" clap;
    add ops ~short_name:'I' ~long_name:"include-dir" ~help:"path to include-dir" incd;
    add ops ~short_name:'o' ~long_name:"output" ~help:"path to output-file" ofil;
    add ops ~long_name:"xml" ~help:"specify if XML output is needed" xml;
    add ops ~long_name:"temp-dir" ~help:"path to temporary directory" tempdir;

end

let get_options () =
    let rests = OptParse.OptParser.parse_argv O.ops in

    let desc_file =
        match rests with [file] -> file
        | [] -> failwith "no desc_file was specified" | _ -> failwith "too many desc_files were specified" in

    let clap_str = OptParse.Opt.get O.clap in
    let incd_str = OptParse.Opt.get O.incd in
    let ofil_str = OptParse.Opt.get O.ofil in
    let is_xml = OptParse.Opt.get O.xml in
    let tempdir_str = OptParse.Opt.get O.tempdir in
    let test_data = read_desc_file desc_file in

    test_data, {
        topt_output_filename = ofil_str;
        topt_clap_file = get_full_path clap_str;
        topt_include_dir = get_full_path incd_str;
        topt_desc_file = get_full_path desc_file;
        topt_is_xml = is_xml;
        topt_temp_dir =
            if ((String.compare tempdir_str "") != 0) then (
                let tempdir = get_full_path tempdir_str in
                (* Check for READ and WRITE permissions *)            
                try (
                    (Unix.access tempdir [Unix.R_OK; Unix.W_OK]);
                    tempdir)
                with
                  Unix.Unix_error _ ->
                      failwith ("Cannot read and/or write to '"
                                ^ tempdir ^ "'."))
            else ""
    }

type runinfo = float * string * string
let clap_results : (string * runinfo) Queue.t = Queue.create ()
let result_flavour (result : OUnit.test_result) = (
    let separate = "\n--------------------\nConclusion: " in
    match result with
    | RError (_,msg) -> (Queue.pop clap_results; msg ^ separate ^ "Error")
    | RFailure (_, msg) -> (fst (Queue.pop clap_results)) ^ separate ^ "Failed"
    | RSuccess _ -> fst (Queue.pop clap_results) ^ separate ^ "Passed"
(*
    | RSkip (_, msg) -> "Run test skip"
    | RTodo (_, msg) -> msg
*)
)

let result_path = function
    | RSuccess path
    | RError (path, _)
    | RFailure (path, _) -> path
(*
    | RSkip (path, _)
    | RTodo (path, _) -> path
*)

(* save result of testsuite to file in format *)
let save_result_txt (results : OUnit.test_result list)
                    (options : testsuite_options_t)
: unit =
begin
    let output_filename = options.topt_output_filename
    and desc_file_name = options.topt_desc_file
    in
    let tab = "         " in
    let separator1 =
        ("|_________|_____________________________________________________________________\n")
    in
    let formatResult (num : int) (result : OUnit.test_result) : string = (
        let str = "|" ^ (string_of_int num) ^ tab in
        let str = String.sub str (0) ((String.length tab) + 1) in
        let str = str ^ "|" ^ (string_of_path (result_path result)) ^ "\n"
                   ^ "Actual result: " ^ (result_flavour result) in
        let str = ExtString.String.replace_chars (fun ch ->
                           if ch = '\n' then ("\n|" ^ tab ^ "|")
                           else ExtString.String.of_char ch
                       )str in
        let str = str ^ "\n" ^ separator1 in
        str
    ) in
    let f = try open_out output_filename
            with Sys_error _ -> (
                failwith ("Creating output file `" 
                          ^ output_filename
                          ^"' failed!")) in

    output_string f ("Description file : " ^ desc_file_name ^ "\n\n");
    output_string f ("________________________________________________________________________________\n");
    output_string f ("|No       |Test result\n");
    output_string f separator1;
    let i = ref 0 in
    List.iter (fun result ->
        output_string f (formatResult !i result);
        i := !i + 1;
    ) results;
    let is_failure = function
        | RFailure _ -> true
        | RSuccess _ | RError _  (*| RSkip _ | RTodo _ *) -> false
    in
    let is_error = function
        | RError _ -> true
        | RSuccess _ | RFailure _ (*| RSkip _ | RTodo _ *) -> false
    in
    let errors = List.filter is_error results in
    let failures = List.filter is_failure results in
    let summer ="Cases: " ^ (string_of_int (List.length results)) ^ "\n"
        ^ "Errors: " ^ (string_of_int (List.length errors)) ^ "\n"
        ^ "Failures: " ^ (string_of_int (List.length failures)) in
    output_string f (summer);
    close_out f;
    print_endline ("Test result is written to file `" ^ output_filename ^ "'");
end

exception Error_test of string;;

(* Find attribute "val" in tag "result" of a xml list *)
let rec find_attrib_of_tag (attr : string) (tag : string) (xmls : Xml.xml list)
: string =
begin
    match xmls with
    | [] ->
        raise (Error_test ("ERROR oUnit: cannot find tag `"
                           ^ tag
                           ^"' in the xml file"))
    | a_xml::remain_xmls -> (
        let tag_of_xml =
            try Xml.tag a_xml
            with Xml.Not_element _ ->
                raise (Error_test ("ERROR oUnit: cannot find tag `"
                                   ^ tag
                                   ^ "' in the xml file"))
        in
        if tag_of_xml = tag then
            try (Xml.attrib a_xml attr)
            with Xml.No_attribute _ ->
                raise (Error_test ("ERROR oUnit: cannot find attribute `"
                                   ^ attr
                                   ^ "' of tag `"
                                   ^ tag
                                   ^ "' in the xml file"))
        else (
            let remain_xmls = remain_xmls
                              @ (try Xml.children a_xml
                                 with Xml.Not_element _ -> []) in
            find_attrib_of_tag attr tag remain_xmls))
end

(* Run a test
   Call command = ./clap + options + filename.
   Read file result to find output information of safety.
   Compare this information with information of safety that read in file .c.
 *)
let test_a_file (cmd : string)
                (clap_options : string)
                (info : string) _ =
    let options = ExtString.String.nsplit clap_options " " in
    (* find output file name *)
    let rec find_xml_file_name (list_of_options : string list)
        : (string * string list) = (
        match list_of_options with
        | [] -> ("result.xml",
                  ["--output-xml";"result.xml"])
        | a_option :: remain_options -> (
          if a_option = "--output-xml" then
              ((List.hd remain_options), [])
          else
              find_xml_file_name remain_options
          )
    ) in

    let (clap_result_file, added_options) = find_xml_file_name options in
    let options = added_options @ options  in
    let cmd = cmd ^ " " ^ (String.concat " " options) in
    let _ = try Sys.remove clap_result_file with _ -> () in
    let start = Unix.gettimeofday () in
    let (ch_stdout, ch_stdin, ch_stderr) = Unix.open_process_full cmd (Unix.environment ()) in
    let stdout_str = Std.input_all ch_stdout in
    let stderr_str = Std.input_all ch_stderr in
    let res = Unix.close_process_full (ch_stdout, ch_stdin, ch_stderr) in
    let finish = Unix.gettimeofday () in
    let run_info = (finish -. start, stdout_str, stderr_str) in
    match res with
    | WEXITED exitcode when exitcode <> 0 ->
    (
        let exit_info = "exit code " ^ (string_of_int exitcode) in
        Queue.push (exit_info, run_info) clap_results;
        if exit_info = info then
           assert_equal (exit_info = info) (true)
        else
           raise (Error_test ("Bad exit from clap: code = " ^ exit_info))
    )
    | WSIGNALED signum | WSTOPPED signum ->
    (
        let sig_info = "signal number " ^ (string_of_int signum) in
        Queue.push (sig_info, run_info) clap_results;
        if sig_info = info then
           assert_equal (sig_info = info) (true)
        else
           raise (Error_test ("Bad signal from clap: code = " ^ sig_info))
    )
    | _ ->

    let xml_result =
        try
            Xml.parse_file (clap_result_file)
        with
        | Xml.File_not_found _ ->
        (
            Queue.push ("error", run_info) clap_results;
            raise (Error_test "Error: output xml file not found")
        )
        | Xml.Error _ ->
        (
            Queue.push ("error", run_info) clap_results;
            raise (Error_test "Processing output xml file failed")
        ) in
    let is_successful = find_attrib_of_tag ("val") ("result")
                                           ([xml_result]) in
    (try Queue.push (is_successful, run_info) clap_results
     with _ -> assert false);
     assert_equal (*~msg:stdout_str*) (is_successful = info) (true)

let output_success o s = ()

let output_stdout_stderr o stdout_str stderr_str =
    Printf.fprintf o "<system-out><![CDATA[%s]]></system-out>" stdout_str;
    Printf.fprintf o "<system-err><![CDATA[%s]]></system-err>" stderr_str

let output_failure o p s msg =
(*    Printf.fprintf o "<failure type=\"%s\" message=\"%s\">\n" s msg; *)
    Printf.fprintf o "<failure type=\"%s\">\n" s;
    Printf.fprintf o "%s" msg;
    Printf.fprintf o "</failure>\n"

let output_error o p s msg =
(*    Printf.fprintf o "<error type=\"%s\" message=\"%s\">\n" s msg; *)
    Printf.fprintf o "<error type=\"%s\">\n" s;
    Printf.fprintf o "%s" msg;
    Printf.fprintf o "</error>\n"

let output_result_aux o s result =
	match result with
	| RSuccess p -> output_success o p
	| RFailure (p, msg) -> output_failure o p s msg 
	| RError (p, msg) -> output_error o p s msg

let output_result o result =
        let (s, rinfo) = Queue.pop clap_results in
	let path = result_path result in
	let name = string_of_path path in
        let (time, stdout_str, stderr_str) = rinfo in
	Printf.fprintf o "<testcase name=\"%s\" time=\"%f\">\n" name time;
	output_result_aux o s result;
        output_stdout_stderr o stdout_str stderr_str;
	Printf.fprintf o "</testcase>\n"

let rec output_results o name time timestamp hostname results =
    let num_tests = List.length results in
    let num_failures = List.length (List.filter (function RFailure _ -> true | _ -> false) results) in
    let num_errors = List.length (List.filter (function RError _ -> true | _ -> false) results) in
    Printf.fprintf o "<testsuite tests=\"%d\" failures=\"%d\" errors=\"%d\" hostname=\"%s\" name=\"%s\" time=\"%f\" timestamp=\"%s\">\n" num_tests num_failures num_errors hostname name time timestamp;
    List.iter (output_result o) results;
    Printf.fprintf o "<system-out>";
    Printf.fprintf o "</system-out>";
    Printf.fprintf o "<system-err>";
    Printf.fprintf o "</system-err>";
    Printf.fprintf o "</testsuite>\n"

let save_result_xml results options name hostname time timestamp =
    let ofile = options.topt_output_filename in
    let o = try open_out ofile
            with Sys_error _ -> ( failwith ("Creating output file `" ^ ofile ^"' failed!")) in
    output_results o name time timestamp hostname results

let save_result results options name hostname time timestamp =
    if options.topt_is_xml then
        save_result_xml results options name hostname time timestamp
    else
        save_result_txt results options

(* Create a testsuite includes tests of all file .c are read from input path *)
let _ =
begin
    let name = "test" in
    let test_data, options = get_options () in
    let testsuite_path = Sys.getcwd () in
    Sys.chdir (Filename.dirname options.topt_desc_file);
    let cmd = 
        if options.topt_include_dir <> "" then
            options.topt_clap_file ^ " ps -I " ^ options.topt_include_dir
        else
            options.topt_clap_file ^ " ps "
    in
    let cmd =
        if options.topt_temp_dir <> "" then
            cmd ^ " --temp-dir " ^ options.topt_temp_dir
        else
            cmd
    in
    let suite_list =
        List.fold_left (fun suite test ->
                            let clap_options, info = test in
                            let expected_result_str = clap_options
                                             ^ "\nExpected result: "
                                             ^ info in
                            let runcmd = (test_a_file cmd clap_options info)
                            in
                            suite @ [expected_result_str >:: runcmd])
                       [] test_data in
    let suite_list = name >::: (suite_list) in
    let start = Unix.gettimeofday () in
    let results : OUnit.test_result list = run_test_tt suite_list in
    let finish = Unix.gettimeofday () in
    Sys.chdir testsuite_path;

    let hostname = Unix.gethostname () in
    let timestamp = string_of_tm_in_iso8601 (Unix.gmtime start) in
    save_result (List.rev results) options name hostname (finish -. start) timestamp

(*
    output_results name (finish -. start) timestamp hostname results
*)
end
