(* 
    (* utils/csisatUtil *)
    Create by mai.ntt.
    mathsatUtil is the interface between Csisat solver and LogicUtil.
*)

open CsisatAst

module AstUtil = CsisatAstUtil

let rec print_expr expr = match expr with
    | Constant cst -> CsisatUtils.my_string_of_float cst
    | Variable v -> v
    | Application (sym, lst) -> sym ^ "(" ^ (CsisatUtils.string_list_cat ", " (List.map print_expr lst)) ^")"
    | Sum lst ->  "(" ^ (CsisatUtils.string_list_cat " $PLUS " (List.map print_expr lst)) ^")"
    | Coeff (co, expr) -> (CsisatUtils.my_string_of_float co) ^ " $MUL " ^ (print_expr expr)

let rec print_pred p =
    match p with
    | False -> "false"
    | True -> "true"
    | And lst -> "(" ^ (CsisatUtils.string_list_cat " $AND " (List.map print_pred lst)) ^")"
    | Or lst -> "(" ^ (CsisatUtils.string_list_cat " $OR " (List.map print_pred lst)) ^")"
    | Not p -> "not " ^ print_pred p
    | Eq (e1,e2) -> (print_expr e1) ^ " $EQ " ^ (print_expr e2)
    | Lt (e1,e2) -> (print_expr e1) ^ " $LT " ^ (print_expr e2)
    | Leq (e1,e2) -> (print_expr e1) ^ " $LEQ " ^ (print_expr e2)
    | Atom (External str) -> str
    | Atom (Internal i) -> "atom"^(string_of_int i)

let print p = print_pred p

let get_interpolant (a : predicate) (b : predicate) : predicate = (
    StatisticUtil.update_solver_invocation_number ();
    CsisatInterpolate.interpolate_with_proof a b
  )

let normalize: predicate -> predicate = AstUtil.normalize

let is_sat (pred : predicate) : bool = (
    let pred = AstUtil.simplify (normalize pred) in
    (CsisatSatPL.is_sat pred)
)

let is_valid (pred_to_check : predicate) : bool = (
    let new_pred = AstUtil.simplify (AstUtil.normalize (Not pred_to_check)) in
    if CsisatSatPL.is_sat (new_pred) = false then true
    else false
  )

(* ~~~~~~~~~~~~~~~~~~~ CHECK INTERPOLANTS ~~~~~~~~~~~~~~~~~ *)
let check_interpolants (base_preds : predicate list)
                       (interpolants : predicate list)
                       : bool = (
    assert ((List.length base_preds) - (List.length interpolants) = 1);
    let rec check_interpolant a_preds b_preds interpolants =
        assert ((List.length b_preds) > 0);
        match interpolants with
          [] -> true
        | itp :: tails -> (
            let pred_a = And a_preds
            and pred_b = And b_preds
            in
            let a_itp = Or [Not pred_a; itp]
            and b_itp = And [pred_b; itp]
            in
            if not (is_valid a_itp) then
                false
            else (
                if (is_sat b_itp) then
                    false
                else (
                    (* OK, continue checking *)
                    match b_preds with
                      [_] -> true
                    | h :: bs ->
                        check_interpolant (a_preds @ [h]) bs tails)))
    in
    match base_preds with
      []
    | [_] -> true
    | h :: t ->
        check_interpolant [h] t interpolants)

let get_interpolants (preds : predicate list)
                     (round_coeffs : bool)
                     : predicate list =
    let preds = List.map (fun pred ->
                    AstUtil.simplify (normalize pred)
                ) preds
    in
    let original_results =
        CsisatInterpolate.interpolate_with_one_proof preds
    in
    let results =
        if (round_coeffs) then (
            let rounded_results =
                List.map (CsisatLIUtils.round_coeff) original_results
            in
            if (check_interpolants preds rounded_results) then
                rounded_results
            else
                original_results)
        else
            original_results
    in
    results

let cnf: predicate -> predicate = AstUtil.cnf
let dnf: predicate -> predicate = AstUtil.dnf
  
let simplify : predicate -> predicate = AstUtil.simplify 
let simplify_expr : expression -> expression = AstUtil.simplify_expr
let remove_lit_clash : predicate -> predicate = AstUtil.remove_lit_clash
