/* utils/fxc.c */
/* Copyright (C) 2007-2012 Yonezawa Lab, The University of Tokyo */

/* Sys.command with vfork */

/* Original copyright in ocaml-3.12.1. */

/***********************************************************************/
/*                                                                     */
/*                           Objective Caml                            */
/*                                                                     */
/*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
/*                                                                     */
/*  Copyright 1996 Institut National de Recherche en Informatique et   */
/*  en Automatique.  All rights reserved.  This file is distributed    */
/*  under the terms of the GNU Library General Public License, with    */
/*  the special exception on linking described in file ../../LICENSE.  */
/*                                                                     */
/***********************************************************************/

/* Original copyright in NetBSD. */

/*
 * Copyright (c) 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "mlvalues.h"
#include "memory.h"
#include "signals.h"
/*#include "debugger.h"*/
/*#include "unixsupport.h"*/

#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

CAMLextern void caml_sys_error(value);
/*extern void uerror(char *cmdname, value arg) Noreturn;*/
/*extern char **cstringvect();*/
/*CAMLextern int caml_debugger_in_use;*/
/*CAMLextern int caml_debugger_fork_mode;*/
/*void caml_debugger_cleanup_fork(void);*/

#define SHELL "/bin/sh"

static int do_system(char *cmd);

/* It preferrs "Sys.command" interface to "Unix.create_process",
   because it is not sure the parent is suspended in Linux, where the
   schedular seemds to be called early in "execve".  It does not call
   "caml_debugger_cleanup_fork" at all.  It misses much of the
   thread-related guards in GLIBC with "_LIBC_REENTRANT".  */

CAMLprim value
edulis_command(value command)
{
    CAMLparam1 (command);
    int status, retcode;
    char *buf;
    intnat len;

    len = caml_string_length(command);
    buf = caml_stat_alloc(len + 1);
    memmove(buf, String_val(command), (len + 1));
    caml_enter_blocking_section();
    status = do_system(buf);
    caml_leave_blocking_section();
    caml_stat_free(buf);
    if (status == -1)
	caml_sys_error(command);
    if (WIFEXITED(status))
	retcode = WEXITSTATUS(status);
    else
	retcode = 255;
    CAMLreturn (Val_int(retcode));
}

static int
do_system(char *command)
{
    char * const argv[] = {"sh", "-c", command, 0};
    sigset_t nmask, omask;
    struct sigaction intrsa, quitsa, sa;
    pid_t pid;
    int cc;

    if (command == 0) {
	cc = access(SHELL, X_OK);
	return (cc == 0);
    }

    sa.sa_handler = SIG_IGN;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;

    cc = sigaction(SIGINT, &sa, &intrsa);
    if (cc == -1) {
	return -1;
    }
    cc = sigaction(SIGQUIT, &sa, &quitsa);
    if (cc == -1) {
	sigaction(SIGINT, &intrsa, 0);
	return -1;
    }

    sigemptyset(&nmask);
    sigaddset(&nmask, SIGCHLD);
    cc = sigprocmask(SIG_BLOCK, &nmask, &omask);
    if (cc == -1) {
	sigaction(SIGINT, &intrsa, 0);
	sigaction(SIGQUIT, &quitsa, 0);
	return -1;
    }

    /*__readlockenv();*/
    pid = vfork();
    if (pid == -1) {
	/*__unlockenv();*/
	sigaction(SIGINT, &intrsa, 0);
	sigaction(SIGQUIT, &quitsa, 0);
	sigprocmask(SIG_SETMASK, &omask, 0);
	return -1;
    } else if (pid == 0) {
	sigaction(SIGINT, &intrsa, 0);
	sigaction(SIGQUIT, &quitsa, 0);
	sigprocmask(SIG_SETMASK, &omask, 0);
	execve(SHELL, argv, __environ);
	_exit(127);
	return -1;
    } else {
	int pstat = 0;
	/*__unlockenv();*/
	cc = waitpid(pid, &pstat, 0);
	while (cc == -1) {
	    if (errno != EINTR) {
		pstat = -1;
		break;
	    }
	}
	sigaction(SIGINT, &intrsa, 0);
	sigaction(SIGQUIT, &quitsa, 0);
	sigprocmask(SIG_SETMASK, &omask, 0);
	if (pstat == -1) {
	    return -1;
	} else {
	    return pstat;
	}
    }
}
