(*pp camlp4o Camlp4MacroParser.cmo -D${MYMODE} *)

(* utils/logicUtil.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

(* Written by DzungDK@ANLAB Vietnam *)

(* MEMO: 
     (1) these serve as translators. 
     (2) There are some cases which are not handled yet
*)
open MathsatUtil

type solver_t = CSISAT | MATHSAT | HYBRID

let solver : solver_t ref = ref CSISAT;;

let emerg (s : string) : unit =
    LogUtil.log LogUtil.GENERAL LogUtil.EMERG s

let enable_mathsat_debug_mode () = (
    mathsat_verbosity_level := 1;
    mathsat_cmd_suffix := "")

let total_module_time = ref 0.

let round_coeffs = ref false

(** Translate MathsatUtil.predicate into CsisatUtil.predicate.
    MathsatUtil.predicate includes Constant having num form and float form,
    but CsisatUtil.predicate includes Constant having float type.
 *)
let rec pred_to_csisat_pred (input : predicate)
: Ast.predicate =
begin 
    match input with
    | True ->
        Ast.True
    | False ->
        Ast.False
    | And (pred_lst) ->
        Ast.And (preds_to_csisat_preds pred_lst)
    | Or (pred_lst) ->
        Ast.Or (preds_to_csisat_preds pred_lst)
    | Not (pred) ->
        Ast.Not (pred_to_csisat_pred pred)
    | Eq (expr1, expr2) ->
        Ast.Eq (expr_to_csisat_expr expr1,
                expr_to_csisat_expr expr2)
    | Lt (expr1, expr2) ->
        Ast.Lt (expr_to_csisat_expr expr1,
                expr_to_csisat_expr expr2)
    | Leq (expr1, expr2) ->
        Ast.Leq (expr_to_csisat_expr expr1,
                 expr_to_csisat_expr expr2)
    | Atom (atom) ->
        Ast.Atom atom
end

and preds_to_csisat_preds (preds : predicate list)
: Ast.predicate list =
begin
    List.map (fun a_pred ->
        pred_to_csisat_pred a_pred
    ) preds
end

(** Translate MathsatUtil.expression type into CsisatUtil.expression type.
    MathsatUtil.expression includes Constant having num form and float form,
    but CsisatUtil.expression includes Constant having float form.
 *)
and expr_to_csisat_expr (expr : expression)
: Ast.expression =
begin
    match expr with
    | Constant mix_cst ->
        Ast.Constant (get_float_from_mix_cst mix_cst)
    | Variable symbol ->
        Ast.Variable symbol
    | Application (fun_name, exprs) ->
        Ast.Application (fun_name, (exprs_to_csisat_exprs exprs))
    | Sum exprs ->
        Ast.Sum (exprs_to_csisat_exprs exprs)
    | Coeff (mix_cst, expr) ->
        Ast.Coeff (get_float_from_mix_cst mix_cst,
                   (expr_to_csisat_expr expr))
end

and exprs_to_csisat_exprs (exprs : expression list)
: Ast.expression list =
begin
    List.map
        (fun a_expr ->
            expr_to_csisat_expr a_expr
        ) exprs
end
  
let rec csisat_expr_to_expr (expr : Ast.expression)
: expression =
begin
    match expr with
    | Ast.Constant f ->
        Constant (create_mix_cst_from_float f) 
    | Ast.Variable syb ->
        Variable syb
    | Ast.Application (syb, exprs) -> (
        let new_exprs = csisat_exprs_to_exprs exprs in 
        Application (syb, new_exprs))
    | Ast.Sum exprs -> (
        let new_exprs = csisat_exprs_to_exprs exprs in
        (Sum new_exprs))
    | Ast.Coeff (f, expr) -> (
        let new_expr = csisat_expr_to_expr expr in
        Coeff ((create_mix_cst_from_float f), new_expr))
end

and csisat_exprs_to_exprs (exprs : Ast.expression list)
: expression list =
begin
    List.map
        (fun a_expr ->
            csisat_expr_to_expr a_expr) exprs
end

and csisat_pred_to_pred (pred : Ast.predicate)
: predicate =
begin
    match pred with
    | Ast.True -> True 
    | Ast.False -> False
    | Ast.And preds -> (
        let new_preds = csisat_preds_to_preds preds in
        And new_preds)
    | Ast.Or (preds) -> (
        let new_preds = csisat_preds_to_preds preds in
        Or new_preds)
    | Ast.Not (pred) -> (
        let new_pred = csisat_pred_to_pred pred in
        Not new_pred)
    | Ast.Eq (expr1, expr2) -> (
        let new_expr1 = csisat_expr_to_expr expr1
        and new_expr2 = csisat_expr_to_expr expr2 in
        Eq (new_expr1, new_expr2))
    | Ast.Lt (expr1, expr2) -> (
        let new_expr1 = csisat_expr_to_expr expr1
        and new_expr2 = csisat_expr_to_expr expr2 in
        Lt (new_expr1, new_expr2))
    | Ast.Leq (expr1, expr2) -> (
        let new_expr1 = csisat_expr_to_expr expr1
        and new_expr2 = csisat_expr_to_expr expr2 in
        Leq (new_expr1, new_expr2))
    | Ast.Atom atom -> Atom atom 
end

and csisat_preds_to_preds (preds : Ast.predicate list)
: predicate list =
begin
    List.map
        (fun a_pred ->
            csisat_pred_to_pred a_pred
        ) preds
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec count_atom_pred_in_pred (pred : predicate) : int = (
    match pred with
    | True
    | False -> 0
    | And preds
    | Or preds -> count_atom_pred_in_preds preds
    | Not pred1 -> count_atom_pred_in_pred pred1
    | Eq _
    | Lt _
    | Leq _
    | Atom _ -> 1
)

and count_atom_pred_in_preds (preds : predicate list) : int = (
    List.fold_left (fun atom_pred_number pred ->
        atom_pred_number + (count_atom_pred_in_pred pred)
    ) 0 preds
)

(* ~~~~~~~~~~~~~~~~~~~~~ Compute interpolant by MathSat ~~~~~~~~~~~~~~~~~~~~ *)
(** Read msat files and return predicate list.
    This function uses a parser of msat format to return predicate format.
        @param file_lst : name list of msat files
        @param vars : use to return the real names for variables and applications
        @return : predicate list of msat file list  
*)
let rec get_preds_from_msat_file ( file_lst : string list ) 
                                 ( vars : (string, string) Hashtbl.t ) 
: predicate list =
begin
    match file_lst with
    | []            -> ( [] )
    | head :: tails -> (
        let ich = try open_in head
                  with Sys_error _ -> assert false in
        try
            let lb = Lexing.from_channel ich in
            let tmp = ( Msat_dec.all Msat_lexer.token lb ) in
            let result = replace_var_by_var_predicate tmp vars in 
            close_in ich;						 
            [result] @ ( get_preds_from_msat_file tails vars )
         with  e ->  ( 
             close_in_noerr ich;
             raise e
         )
    ) 
end;;

(** Compute interpolants of predicate list by MathSat
        @param preds : predicate list need be computed interpolant
        @return : predicate list is interpolant of preds
*)
let get_interpolants_mathsat (preds : predicate list) : predicate list =
begin
    let vars : (string, classified_var_t) Hashtbl.t = Hashtbl.create 2 in
    let apps : (string, classified_app_t) Hashtbl.t = Hashtbl.create 2 in
    (* collect variables and applications from predicate list,
     until having no change about the kind of vaiables or applications*)
    while (List.exists (fun a_pred ->
                           collect_define_in_pred a_pred vars apps 
                       ) preds) 
    do () done;
    let random_file = (Filename.temp_file "" "") in
    let run_hashstring =
        !MathsatUtil.temp_dir ^ "/" ^ (Filename.basename random_file)
    in
    let input_lst =
        (write_preds_into_mathsat_files preds vars apps []
                                        (run_hashstring ^ "_input") 1)
    in
    let input_file = (String.concat " " input_lst) in
    let output_lst = get_name_list_files input_lst in
    let tmp_bool = 
        List.exists (fun a_pred -> 
                        determine_bit_pred a_pred) preds
    in
    let set_verbosity_str = "-verbosity="
                            ^ (string_of_int !mathsat_verbosity_level)
    in
    let cmd = if tmp_bool then
                  ((Filename.dirname Sys.executable_name) 
                   ^ "/../mathsat-4/bin/mathsat "
                   ^ "-interpolate="
                   ^ (run_hashstring ^ "_output")
                   ^ " "
                   ^ set_verbosity_str
                   ^ " -tsolver=euf -tcomb=decide -tsolver=bv "
                   ^ input_file
                   ^ !mathsat_cmd_suffix)
              else
                  ((Filename.dirname Sys.executable_name) 
                   ^ "/../mathsat-4/bin/mathsat "
                   ^ "-interpolate="
                   ^ (run_hashstring ^ "_output")
                   ^ " "
                   ^ set_verbosity_str
                   ^ " -tsolver=euf -tsolver=la -tcomb=decide "
                   ^ input_file
                   ^ !mathsat_cmd_suffix) 
    in
    let start_time = Unix.time () in
    Thread.join (Thread.create Fx.command cmd);
    let stop_time = Unix.time () in
    MathsatUtil.total_mathsat_time := !MathsatUtil.total_mathsat_time +. stop_time -. start_time;
    let rev_vars = Hashtbl.create (Hashtbl.length vars) in
    Hashtbl.iter (fun old_var (new_var, _) ->
        Hashtbl.add rev_vars new_var old_var) vars;
    Hashtbl.iter (fun old_app (new_app, _) ->
        Hashtbl.add rev_vars new_app old_app) apps;
    Hashtbl.clear vars;
    Hashtbl.clear apps;
    let output_preds = (get_preds_from_msat_file output_lst rev_vars) in
    (List.iter Sys.remove (input_lst @ output_lst @ [random_file]));
    output_preds
end;;

(* ~~~~~~~~~~~~~~~~~~~~~~~ COMPUTE WEAKEST PRECONDITION ~~~~~~~~~~~~~~~~~~~~~ *)
(* replace var in an expression by an other expression (the other expression  *)
(* is right_expr of an assignment)                                            *)
let rec replace_var_in_expression (expr_in_post_condition : expression) 
    (replacing_var : expression) (expr_to_replace : expression) 
: expression =
begin
	  match expr_in_post_condition with
    | Constant cst -> Constant cst
	  | Variable _ -> 
        if expr_in_post_condition = replacing_var then expr_to_replace 
        else expr_in_post_condition
    (* 20090626 *)
	  (*| Application ("addrof", _)
	  | Application ("pref", _)
	  | Application ("sref", _)
	  | Application ("srefp", _) ->
	      if expr_in_post_condition = replacing_var then expr_to_replace 
		    else expr_in_post_condition*)
	  (* temporally  process "array" *)
	  | Application ("array", expr_list) -> (*assert false*)
        if expr_in_post_condition = replacing_var then expr_to_replace 
        else begin
            let new_expr_list = ref [] in
            (* replace var in each expr in expr_list *)
            List.iter 
                (fun a_expr -> 
                    let new_expr = 
                        replace_var_in_expression (a_expr) (replacing_var) 
                          (expr_to_replace) 
                    in
                    new_expr_list := !new_expr_list @ [new_expr]
                )
                expr_list;
            (* return *)
            Application ("array", !new_expr_list)
        end
	  | Application (name_of_function_call, expr_list) -> begin
		    let new_expr_list = ref [] in
	      (* replace var in each expr in expr_list *)
		    List.iter 
		        (fun a_expr -> 
		            let new_expr = 
		                replace_var_in_expression (a_expr) (replacing_var) 
	                    (expr_to_replace) 
	              in
		            new_expr_list := !new_expr_list @ [new_expr]
            )
		        expr_list;
	      (* return *)
		    Application (name_of_function_call, !new_expr_list)
	    end
	  | Sum expr_list -> begin
		    let new_expr_list = ref [] in
	      (* replace var in each expr in expr_list *)
		    List.iter 
		        (fun a_expr -> 
		            let new_expr = 
		                replace_var_in_expression (a_expr) (replacing_var) 
	                    (expr_to_replace) 
	              in
		            new_expr_list := !new_expr_list @ [new_expr]
            )
		        expr_list;
	      (* return *)
		    Sum !new_expr_list
	    end
	  | Coeff (c, expr) -> begin
	      let new_expr = 
	          replace_var_in_expression (expr) (replacing_var) (expr_to_replace)
	      in
	      Coeff (c, new_expr)
	    end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* replace expression in an predicatee by an other expression (the other      *)
(* expression is right_expr of an assignment)                                 *)
and replace_var_in_predicate (pred_in_post_condition : predicate) 
    (replacing_var : expression) (expr_to_replace : expression) 
: predicate = 
begin
	  match pred_in_post_condition with
	  | True -> True
	  | False -> False
	  | And pred_list -> begin
		    let new_pred_list = ref [] in
		    List.iter 
		        (fun a_pred -> 
			          let new_pred = 
	                  replace_var_in_predicate (a_pred) (replacing_var) 
	                    (expr_to_replace) 
	              in
			          new_pred_list := !new_pred_list @ [new_pred]
            )
		        pred_list;
		      And !new_pred_list
        end
    | Or pred_list -> begin
        let new_pred_list = ref [] in
        List.iter 
            (fun a_pred -> 
                let new_pred : predicate = 
                    replace_var_in_predicate (a_pred) (replacing_var) 
                      (expr_to_replace) 
                in
                new_pred_list := !new_pred_list @ [new_pred]
            )
            pred_list;
        Or !new_pred_list
      end
	  | Not pred -> begin
        let new_pred = 
            replace_var_in_predicate (pred) (replacing_var) (expr_to_replace)
        in
        Not new_pred
      end
	  | Eq (expr1, expr2) -> begin 
		    let new_expr1 : expression = 
            replace_var_in_expression (expr1) (replacing_var) (expr_to_replace) 
		    in
        let new_expr2 : expression = 
            replace_var_in_expression (expr2) (replacing_var) (expr_to_replace) 
        in
		    Eq (new_expr1, new_expr2)
      end
	  | Lt (expr1, expr2) -> begin 
        let new_expr1 : expression = 
            replace_var_in_expression (expr1) (replacing_var) (expr_to_replace) 
        in
        let new_expr2 : expression = 
            replace_var_in_expression (expr2) (replacing_var) (expr_to_replace) 
        in
        Lt (new_expr1, new_expr2)
      end
	  | Leq (expr1, expr2) -> begin 
        let new_expr1 : expression = 
            replace_var_in_expression (expr1) (replacing_var) (expr_to_replace) 
        in
        let new_expr2 : expression = 
            replace_var_in_expression (expr2) (replacing_var) (expr_to_replace) 
        in
        Leq (new_expr1, new_expr2)
      end
	  | Atom i -> Atom i
end

(* ~~~~~~~~~~~~~~~~ compute weakest_precondition ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec get_weakest_precondition (post_condition : predicate)
    (edge : predicate) 
: predicate =
begin
	  (* check type of edge, if edge isn't an assignment or a list of *)
    (* assignments is assert false                                  *)
	  let wp_to_return = 
	      match edge with
		    | Eq (left_expr, right_expr) -> 
		       replace_var_in_predicate (post_condition) (left_expr) (right_expr)
		    | And pred_list ->
          begin 
			      let rev_pred_list = List.rev pred_list in 
			      let weakest_precondition = ref post_condition in
			      List.iter 
		  	        (fun a_pred -> 
			              weakest_precondition :=  
                      get_weakest_precondition (!weakest_precondition) (a_pred)
                )
			          rev_pred_list;
            (* return *)
			      !weakest_precondition
          end
		    | _ -> post_condition;
	  in
	  wp_to_return
end

let rec collect_constant_division (input : expression)
                                  (candidates : (expression * mix_const) list)
                                  : (expression * mix_const) list = (
    match input with
    | Constant _
    | Variable _ -> candidates      
    | Application (sym, args) -> (
        let new_candidates =
            if ((String.compare sym "div") == 0) then (
                assert((List.length args) = 2);
                match args with
                | [divident; divisor] -> (
                    match divisor with
                    | (Constant (n_op, Some f)) when (f > 0.) ->
                        [(divident, (n_op, Some f))]
                    | (Constant (Some n, f_op))
                      when (Num.gt_num n (Num.num_of_int 0)) ->
                        [(divident, (Some n, f_op))]
                    | _ -> [])
                | _ -> assert false)
            else
                []
        in
        List.fold_left (fun candidates expression ->
                            collect_constant_division expression candidates)
                       (candidates @ new_candidates)
                       args)
    | Sum exprs ->
        List.fold_left (fun candidates expression ->
                            collect_constant_division expression candidates)
                       candidates
                       exprs
    | Coeff (_, expr) ->
        collect_constant_division expr candidates)

let rec approximate_division (input : predicate) : predicate = (
    let make_augmented_pred const_div = (
        let divident, divisor = const_div in
        let div_expr =
            Application("div", [divident; Constant divisor])
        in
        let expr1 = Sum [divident; Constant (reverse_sign_mix_cst divisor)]
        and expr2 = Coeff (divisor, div_expr)
        in
        let lower_bound_pred = Lt (expr1, expr2)
        and upper_bound_pred = Leq (expr2, divident)
        in
        [lower_bound_pred; upper_bound_pred])
    in
    match input with
    | True
    | False -> input
    | And preds ->
        let new_preds = List.map approximate_division preds in
        And new_preds
    | Or preds ->
        let new_preds = List.map approximate_division preds in
        Or new_preds
    | Not pred ->
        let new_pred = approximate_division pred in
        Not new_pred
    | Eq (lexpr, rexpr) -> (
        let l_candidates = collect_constant_division lexpr []
        and r_candidates = collect_constant_division rexpr []
        in
        let candidates = l_candidates @ r_candidates in
        let augmented_preds =
             List.flatten (List.map make_augmented_pred candidates)
        in
        match augmented_preds with
        | [] -> input
        | _ -> And (input::augmented_preds))
    | Lt (expr1, expr2) -> (
        let candidates1 = collect_constant_division expr1 []
        and candidates2 = collect_constant_division expr2 []
        in
        let candidates = candidates1 @ candidates2 in
        let augmented_preds =
            List.flatten (List.map make_augmented_pred candidates)
        in
        match augmented_preds with
        | [] -> input
        | _ -> And (input::augmented_preds))
    | Leq (expr1, expr2) -> (
        let candidates1 = collect_constant_division expr1 []
        and candidates2 = collect_constant_division expr2 []
        in
        let candidates = candidates1 @ candidates2 in
        let augmented_preds =
            List.flatten (List.map make_augmented_pred candidates)
        in
        match augmented_preds with
        | [] -> input
        | _ -> And (input::augmented_preds))
    | Atom _ -> input)

(* ~~~~~~~~~~~~~~~~~~~ CHECK SATISFIABILITY AND VALIDITY ~~~~~~~~~~~~~~~~~ *)
let run_mathsat (solver : solver_t)
                (is_bv_preds : bool)
                (contains_big_numbers : bool)
                : bool = (
    match solver with
    | CSISAT -> false
    | MATHSAT -> true
    | HYBRID ->
        if is_bv_preds then true
        else if contains_big_numbers then true
        else false
)

let is_sat (pred : predicate) : bool = (
    let start_time = Unix.time () in 
    StatisticUtil.update_solver_invocation_number ();
    let pred =
        MathsatUtil.simple_normalize (approximate_division pred) in

    let is_bv_pred = determine_bit_pred pred in
    let contains_big_numbers = contains_big_numbers pred in
    let run_mathsat = run_mathsat !solver is_bv_pred contains_big_numbers in

    let result =
IFDEF DEBUG_MODE THEN
    print_endline ("Checking SAT of : " ^ (print_pred pred));
    let result = 
        if run_mathsat then
            MathsatUtil.is_sat_of_mathsat pred
        else CsisatUtil.is_sat (pred_to_csisat_pred pred)
    in
    print_endline ("   -->> " ^ (string_of_bool result));
    result
ELSE
    if (run_mathsat) then
        (MathsatUtil.is_sat_of_mathsat pred)
    else CsisatUtil.is_sat (pred_to_csisat_pred pred)
END
    in
    let stop_time = Unix.time () in
    total_module_time := !total_module_time +. (stop_time -. start_time);
    result
)

let is_valid (pred_to_check : predicate) : bool = (
    let start_time = Unix.time () in 
    let is_bv_pred = determine_bit_pred pred_to_check in
    let contains_big_numbers = contains_big_numbers pred_to_check in
    let run_mathsat = run_mathsat !solver is_bv_pred contains_big_numbers in

    let new_pred =
        MathsatUtil.simple_normalize (Not pred_to_check) in

    let result =
    if run_mathsat then (
        let sat = MathsatUtil.is_sat_of_mathsat new_pred in
        if sat then false else true)
    else (
        if CsisatUtil.is_sat (pred_to_csisat_pred new_pred) = false then
            true
        else false)
    in
    let stop_time = Unix.time () in
    total_module_time := !total_module_time +. (stop_time -. start_time);
    result
  )

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
module Logic = struct
  include MathsatUtil

  exception SAT_FORMULAS of (predicate list)

  (** Global variable determines whether to use MathSat or Csisat. Default is false, use Csisat*)

  let init: bool -> unit =
      fun (round : bool) -> (
          CsisatGlobal.round := round;
          round_coeffs := round)

  let set_solver (s : solver_t) : unit =
      solver := s 

  let string_of_predicate: predicate -> string = 
      fun (p: predicate) -> print_pred p
  let string_of_expression : expression -> string = 
      fun (e : expression) -> print_expr e
  
  let zero_mix_const = zero_mix_const
  let create_mix_cst_from_num = create_mix_cst_from_num
  let create_mix_cst_from_int = create_mix_cst_from_int
  let is_identical_mix_csts = MathsatUtil.is_identical_mix_csts

  let get_interpolant (a : predicate) (b : predicate) : predicate = (
      let start_time = Unix.time () in
      StatisticUtil.update_solver_invocation_number ();
      let new_a = approximate_division a
      and new_b = approximate_division b in
      let csisat_a = pred_to_csisat_pred new_a
      and csisat_b = pred_to_csisat_pred new_b in
      let result =
      (csisat_pred_to_pred (CsisatUtil.get_interpolant csisat_a csisat_b))
      in
      let stop_time = Unix.time () in
      total_module_time := !total_module_time +. (stop_time -. start_time);
      result
  )

  let get_interpolants (preds : predicate list) : predicate list =(
      let start_time = Unix.time () in
      let preds =
          List.map
              (fun pred ->
                  MathsatUtil.simple_normalize
                      (approximate_division pred)) preds
      in
      let _ =
IFDEF DEBUG_MODE THEN
      print_endline ("Getting Interpolants of");
      List.iter (fun p -> print_endline ("    * " ^ (string_of_predicate p)))
                preds;
ELSE
      ()
END
      in
      let is_bv_preds =
          List.exists (fun a_pred -> 
                           determine_bit_pred a_pred)
                      preds
      in
      let contains_big_numbers =
          List.exists contains_big_numbers preds in
      let run_mathsat = run_mathsat !solver is_bv_preds contains_big_numbers in
      let inter_preds =
          if run_mathsat then
              get_interpolants_mathsat preds
          else (
              let csisat_preds = preds_to_csisat_preds preds in
              try
                  (csisat_preds_to_preds
                      (CsisatUtil.get_interpolants csisat_preds !round_coeffs))
              with
              | CsisatAst.SAT_FORMULA f -> (
                  emerg ("Csisat results in a conflict. Formula : "
                                 ^ (CsisatFociPrinter.print_foci_predicate f));
                  emerg ("Invoking MathSat to re-calculate interpolants");                  
                  let p = And preds in
                  let b = MathsatUtil.is_sat_of_mathsat p in
                  if (b) then (
                      emerg ("Satifiabile conjunction of predicates. Cannot calculate interpolants!");
                      raise (SAT_FORMULAS preds))
                  else
                      let ps = get_interpolants_mathsat preds in
                      ps))
      in
      let stop_time = Unix.time () in
      total_module_time := !total_module_time +. (stop_time -. start_time);
      let _ =
IFDEF DEBUG_MODE THEN
      print_endline ("   >> Results");
      List.iter (fun p -> print_endline ("      * " ^ (string_of_predicate p)))
                inter_preds;
      print_endline (" END >>");
ELSE
      ()
END
      in
      inter_preds
  )

  let is_sat : predicate -> bool = is_sat
 
  let is_valid : predicate -> bool = is_valid

  let approximate_division : predicate -> predicate = approximate_division
  
  let replace_var_in_expression : expression -> expression -> expression -> expression =
      replace_var_in_expression
  let replace_var_in_predicate: predicate -> expression -> expression -> predicate =
      replace_var_in_predicate
  let get_weakest_precondition: predicate -> predicate -> predicate = 
      get_weakest_precondition
  let count_atom_pred_in_pred : predicate -> int = count_atom_pred_in_pred
  let count_atom_pred_in_preds : predicate list -> int = count_atom_pred_in_preds
  let get_total_logic_module_time : unit -> float = function () -> !total_module_time
end

