(* 
    (* utils/mathsatUtil *)
    Create by mai.ntt.
    mathsatUtil is the interface between MathSat solver and LogicUtil.
*)
open Num

module Ast = CsisatAst

(* allow to print log of mathsat in the debug mode *)
let mathsat_verbosity_level = ref 0
let mathsat_cmd_suffix = ref " > /dev/null"

let total_mathsat_time = ref 0.

let temp_dir = ref "/tmp"

(** Mix_const use for checking CsiSat and MathSat. First limit the weak-point
    of using "float_of_num" function when checking big_int by using MathSat. 
 *)
type mix_const = ((num option) * (float option))

type symbol = string
type expression = 
      Constant of mix_const (* (num option * float option) *)
    | Variable of symbol
    | Application of symbol * expression list
    | Sum of expression list
    | Coeff of mix_const * expression

type atom_type = Ast.atom_type

type predicate =
      True
    | False
    | And of predicate list
    | Or of predicate list
    | Not of predicate
    | Eq of expression * expression
    | Lt of expression * expression
    | Leq of expression * expression
    | Atom of atom_type

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let string_of_mix_cst (mix_cst : mix_const) : string =
    match mix_cst with
    | None, None -> assert false
    | Some n, _ -> string_of_num n
    | _, Some f -> CsisatUtils.my_string_of_float f

let rec print_expr expr =
    match expr with
    | Constant cst -> string_of_mix_cst cst
    | Variable v -> v
    | Application (sym, lst) ->
        sym ^ "("
        ^ (CsisatUtils.string_list_cat ", "
               (List.map print_expr lst))
        ^ ")"
    | Sum lst ->
        "("
        ^ (CsisatUtils.string_list_cat " $PLUS "
               (List.map print_expr lst))
        ^ ")"
    | Coeff (co, expr) -> (string_of_mix_cst co) ^ " $MUL " ^ (print_expr expr)

let rec print_pred p =
    match p with
    | False -> "false"
    | True -> "true"
    | And lst ->
        "("
        ^ (CsisatUtils.string_list_cat " $AND "
               (List.map print_pred lst))
        ^ ")"
    | Or lst ->
        "("
        ^ (CsisatUtils.string_list_cat " $OR "
               (List.map print_pred lst))
        ^ ")"
    | Not p -> "not " ^ print_pred p
    | Eq (e1,e2) -> (print_expr e1) ^ " $EQ " ^ (print_expr e2)
    | Lt (e1,e2) -> (print_expr e1) ^ " $LT " ^ (print_expr e2)
    | Leq (e1,e2) -> (print_expr e1) ^ " $LEQ " ^ (print_expr e2)
    | Atom (Ast.External str) -> str
    | Atom (Ast.Internal i) -> "atom"^(string_of_int i)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Returns the negation of a, assuming a is a proposition.
    If a is not a proposition, then the returned value is not normalized.
    
    Note:
        Get idea from CsisatAstUtil.contra
 *)
let contra x = match x with
    | True -> False
    | False -> True
    | Lt (e1, e2) -> Leq(e2, e1)
    | Leq (e1, e2) -> Lt(e2, e1)
    | Not e -> e
    | _ as n -> Not n

(** Normalisation of formula.

    Note:
        Get idea from CsisatAstUtil.normalize_common
 *)
let rec simple_normalize (tree : predicate) : predicate =
  match tree with
  | And ilst -> (
      let new_lst =
          List.filter (fun x -> x <> True) (
              List.fold_left
                  (fun acc x -> 
                      match x with
                      | And lst -> lst @ acc
                      | _ -> x::acc) [] (List.map simple_normalize ilst))
      in
      if List.exists (fun x -> x = False) new_lst then
          False
      else (
          match new_lst with
          | x::[] -> x
          | [] -> True
          | lst -> And lst))
  | Or ilst -> (
      let new_lst =
          List.fold_left
              (fun acc x -> 
                  match x with
                  | Or lst -> lst @ acc
                  | _ -> x::acc) [] (List.map simple_normalize ilst)
      in
      if List.exists (fun x -> x = True) new_lst then
          True
      else (
          match (List.filter (fun x -> x <> False) new_lst) with
          | x::[] -> x
          | [] -> if new_lst = [] then True else False
          | lst -> Or lst))
  | Not i -> let n = simple_normalize i in contra n
  | p -> p

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let zero_mix_const = (Some (num_of_int (0)), Some 0.)

(** Get float part from mix_const form.
 *)
let get_float_from_mix_cst (mc : mix_const) : float =
    match mc with
    | (None, None) -> assert false
    | (Some n, None) -> (
        float_of_num n)
    | (_, Some f_cst) -> f_cst

(** Get num part from mix_const form.
 *)
let get_num_from_mix_cst (mc : mix_const) : num =
    match mc with
    | (None, _) -> assert false
    | (Some n_cst, _) -> n_cst

(** Convert from a float number to a mix_const. 
    
    Note: When converting from a float number to a mix_const,
    the new mix_const has not num part. Because not having
    num_of_float function.
    
    Ex. 3.1 -> (None, Some 3.1)
 *)
let create_mix_cst_from_float (f_cst : float) : mix_const =
    (None, Some f_cst)

(** Convert from a num number to a mix_const.
    
    Ex. num 3 -> (Some num 3, Some 3.) 
 *)
let create_mix_cst_from_num (n_cst : num) : mix_const =
    (Some n_cst, Some (float_of_num n_cst))

(** Convert from a int number to a mix_const.
    
    Ex. int 3 -> (Some (num_of_int 3), Some 3.)
 *)    
let create_mix_cst_from_int (cst : int) : mix_const =
    ( Some (num_of_int cst), Some (float_of_int cst))

(** Reverse the sign for a mix_cst.
    
    Ex. (Some 3, None) -> (Some (-3), None) ...
 *)
let reverse_sign_mix_cst (mix_cst : mix_const) : mix_const =
    match mix_cst with
    | (None, None) -> assert false
    | (None, Some f) -> (None, Some (-.f))
    | (Some n, None) -> (Some (minus_num n), None)
    | (Some n, Some f) -> (Some (minus_num n), Some (-.f))

(** When a mix_const has no value return true. Otherwhile return false.
 *)
let is_nan_mix_cst (mix_cst :  mix_const) : bool = (
    match mix_cst with
    | None, None -> true
    | _ ->  false
  )

(** Check if 2 mix_const are identical or not. Focus on the value of mix_const.
 *)
let is_identical_mix_csts (mix_cst1 : mix_const) (mix_cst2 : mix_const)
: bool = (
    assert (not (is_nan_mix_cst mix_cst1));
    assert (not (is_nan_mix_cst mix_cst2));
    let n1_op, f1_op = mix_cst1
    and n2_op, f2_op = mix_cst2 in
    let f_eq =
        match f1_op, f2_op with
        | Some f1, Some f2 -> (f1 = f2)
        | _ -> false in
    let n_eq =
        match n1_op, n2_op with
        | Some n1, Some n2 -> (eq_num n1 n2)
        | _ -> false in
    let f_of_n_op1 =
        match n1_op with
        | None -> None
        | Some n -> Some (float_of_num n) in
    let f_of_n_op2 =
        match n2_op with
        | None -> None
        | Some n -> Some (float_of_num n) in
    let f_of_n_eq =
        match f_of_n_op1, f_of_n_op2 with
        | Some f1, Some f2 -> (f1 = f2)
        | _ -> false in
    ((f_eq || n_eq) || f_of_n_eq)  
  )

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Get the int format of the float format.
    If the float has int value return Some (int_of_float).
    Otherwhile return None.
 *)
let get_int_of_float (i_f : float) : int option = (
    let i = int_of_float i_f in
    let f_i = float_of_int i in
    if (f_i = i_f) then Some i
    else None
  )

(* ~~~~~~~~~~~~~~~~~ Compute Interpolant and Sat by MathSat ~~~~~~~~~~~~~~~~ *)
(** Try using MathSat to compute Sat and Interpolant.
    Inputs of MathSat functions is predicate and predicate list in
    mathsatUtil file.
 *)

type classified_var_t = string (* new name of variable after replacing special char*)                       
                     * bool    (* kind of variable : true - WORD [32] or false - REAL. 
                                  Default REAL*)
type classified_app_t = string      (* new name of application *) 
                      * (bool list) (*list of the kind of return value and parameter list.
                                      Head of the list is the kind of return value*)

(**  Categorize a expression into bit or real expression found on bit variable,
     bit application and bit operation ("and", "ior", "xor", "lsh", "rsh", 
     "bitnot")
        @param expr : the expression
        @param vars : all variables after 'collected' include of the kind
                      (true - WORD [32]; false - REAL)
        @param apps : all applications after 'collected' include of kind list
                      (true - WORD [32]; false - REAL)
                      
        @return the kind of expression (true - bit; false - real)
    Note : when this fuction is used in collection vars and apps, vars and apps
           in parameter are vars and apps collected at this time
 *)
let rec classify_expr (expr : expression)
                      (vars : (string, classified_var_t) Hashtbl.t)
                      (apps : (string, classified_app_t) Hashtbl.t)
: bool = 
begin
    match expr with
    | Constant cst -> false
    | Variable v -> (
        let (_, kind_var) = try Hashtbl.find vars v
                            with Not_found -> ("", false)
        in kind_var
    )
    | Application ("and", lst) 
    | Application ("ior", lst) 
    | Application ("xor", lst) 
    | Application ("lsh", lst) 
    | Application ("rsh", lst)
    | Application ("mod", lst)
    | Application ("bitnot", lst)  -> true
    | Application (sym, lst) -> (
        let (_, kind_lst) = try Hashtbl.find apps sym
                            with Not_found -> ("", [])
        in 
        let len = List.length kind_lst in
        if len = 0 then
            false
        else (List.hd kind_lst)
    )
    | Sum lst ->
        List.exists (fun a_expr->
            classify_expr a_expr vars apps) lst
    | Coeff (co, expr1) ->
        classify_expr expr1 vars apps
end

(** Repalace special chars in application name or variable name into '_'
    Ex : "n%2" -> "_vn_2"
 *)
let replace_special_char (str : string) : string = (
    let result =
        ExtString.String.map (fun ch ->
            if (ch >= 'a' & ch <= 'z') or (ch >= 'A' & ch <= 'Z')
                or (ch >= '0' & ch <= '9') or (ch = '_') then ch
            else '_'
        ) str in
    "_v" ^ result
)

(* ~~~~~~~~~~~~ Collect var and application for mathsat file ~~~~~~~~~~~~~~~ *)
let collect_define_in_expr (expr : expression)
                           (vars : (string, classified_var_t) Hashtbl.t)
                           (apps : (string, classified_app_t) Hashtbl.t)
                           (pre_kind : bool)
: bool = 
begin
    let rec collect (expr : expression)
                    (vars : (string, classified_var_t) Hashtbl.t)
                    (apps : (string, classified_app_t) Hashtbl.t)
                    (pre_kind : bool)
    : bool = ( 
    match expr with
    | Constant cst -> (false)
    | Variable v -> ( 
        let replaced_special_char = replace_special_char v in
        let (new_name, kind) = try Hashtbl.find vars v 
                               with Not_found -> ("", false)
        in 
        let change =
            (* variable v isn't in vars *) 
            if ((String.compare new_name "") = 0) then (
                Hashtbl.replace vars v (replaced_special_char, pre_kind);
                pre_kind
            )
            (* variable v is in vars *)
            else (
                match pre_kind, kind with
                | true, false -> (
                    Hashtbl.replace vars v (replaced_special_char, true);
                    true)
                | _ -> false    
            ) in
        change
    )
    | Application ("and", lst) 
    | Application ("ior", lst) 
    | Application ("xor", lst) 
    | Application ("lsh", lst) 
    | Application ("rsh", lst)
    | Application ("mod", lst)
    | Application ("bitnot", lst)  -> (
        List.exists (fun a_expr ->
            collect a_expr vars apps true) lst 
    )
    | Application (sym, lst) -> (
        let replaced_special_char = replace_special_char sym in
        let (_, decl_lst) = try Hashtbl.find apps sym
                            with Not_found -> ("",[])
        in
        let len = List.length decl_lst in
        let (head, pr_lst, ch) =
            (* this application isn't in apps *)
            if len = 0 then
                (pre_kind, 
                 (List.map 
                     (fun a_expr ->
                         classify_expr a_expr vars apps) 
                     lst),
                 pre_kind)
            (* this application is in apps *)
            else(
                let head = List.hd decl_lst in
                let para_lst = List.tl decl_lst in
                let pair_lst : (bool * bool) list = (* (kind_of_para, have_a_change) *)
                    try
                        List.map2
                            (fun a_expr para_kind ->
                                let kind = classify_expr a_expr vars apps in
                                if para_kind then
                                    (true, false)
                                else if kind then
                                    (true, true)
                                else (false, false)
                            ) lst para_lst
                    with _ -> assert false
                in
                let (re_para_lst, ch_lst) = List.split pair_lst in
                let ch_of_para_lst =
                    List.exists (fun a_ch -> a_ch) ch_lst
                in
                if head then (
                    (true, re_para_lst, (false || ch_of_para_lst))
                )
                else if pre_kind then
                    (true, re_para_lst, (true || ch_of_para_lst))
                else
                    (false, re_para_lst, (false || ch_of_para_lst))
            )
        in 
        Hashtbl.replace apps sym (replaced_special_char, [head] @ pr_lst);
        (try
            ch || (List.exists2
                      (fun a_expr a_para->
                          collect a_expr vars apps a_para)
                      lst pr_lst)
        with _ -> assert false)
    )
    | Sum lst ->
        List.exists (fun a_expr->
            collect a_expr vars apps pre_kind) lst
    | Coeff (co, expr1) ->
        collect expr1 vars apps pre_kind)
    in collect expr vars apps pre_kind
end

(** Collect variables and applications in a predicate and correspondingly save
    in vars and apps
    @param pred : the predicate need be collected
    @param vars : variables of predicate, include the kind of variable
                  (WORD [32] or REAL)
    @param apps : applications of predicate, include kinds of parameters and
                  return value (WORD [32] or REAL) 
                  
    @return the return value is true iff having any change about the kind of vars
            or apps (change : REAL -> WORD [32] only).
            and false in other cases
 *)
let rec collect_define_in_pred (pred : predicate) 
                               (vars : (string, classified_var_t) Hashtbl.t)
                               (apps : (string, classified_app_t) Hashtbl.t)
: bool = 
begin
    match pred with
    | False -> (false)
    | True -> (false)
    | And lst
    | Or lst -> (
        List.exists (fun a_pred ->
            collect_define_in_pred a_pred vars apps
                  ) lst
    )
    | Not p -> (
        collect_define_in_pred p vars apps
    )
    | Eq (e1,e2)
    | Lt (e1,e2)
    | Leq (e1,e2) -> (
        let kind_e1 = classify_expr e1 vars apps 
        in
        let kind_e2 = classify_expr e2 vars apps 
        in
        let pre_kind = kind_e1 || kind_e2
        in  
        ((collect_define_in_expr e1 vars apps pre_kind) 
        || (collect_define_in_expr e2 vars apps pre_kind))
    )
    | Atom (Ast.External str) -> assert false
    | Atom (Ast.Internal i) -> assert false
end

let rec print_list_vars (och : out_channel) (vars : string list) : unit =
begin
    match vars with 
    | []         -> ()
    | head :: tails -> (
        Printf.fprintf och "%s\n" ( "VAR " ^ head );
        print_list_vars och tails ;
    )
end;;

(* ~~~~~~~~~~~~~~~~~~ Print predicate into mathsat file ~~~~~~~~~~~~~~~~~~~~ *)
let rec print_expr_application (och : out_channel) 
                               (exprs : expression list)
                               (vars : (string, classified_var_t) Hashtbl.t)
                               (apps : (string, classified_app_t) Hashtbl.t)                              
                               (kind_lst : bool list)
: unit = (
    match exprs with
    | [] -> ()
    | [expr] -> (
        let head = try List.hd kind_lst
                   with _ -> assert false
        in
        print_expr_to_mathsat och expr vars apps head
    )
    | expr :: remain_exprs -> (
        let head = try List.hd kind_lst
                   with _ -> assert false
        in
        print_expr_to_mathsat och expr vars apps head;
        Printf.fprintf och "%s" ", ";
        print_expr_application och remain_exprs vars apps (List.tl kind_lst)
      )
)

and print_expr_sum (och : out_channel) 
                   (exprs : expression list)
                   (vars : (string, classified_var_t) Hashtbl.t)
                   (apps : (string, classified_app_t) Hashtbl.t)
                   (kind : bool)
: unit = (
    match exprs with
    | [] -> ()
    | [expr] -> print_expr_to_mathsat och expr vars apps kind
    | expr :: remain_exprs -> (
        print_expr_to_mathsat och expr vars apps kind;
        Printf.fprintf och "%s" " + ";
        print_expr_sum och remain_exprs vars apps kind
      )
)

and print_expr_and_bit (och : out_channel) 
                       (exprs : expression list)
                       (vars : (string, classified_var_t) Hashtbl.t)
                       (apps : (string, classified_app_t) Hashtbl.t) 
: unit = (
    match exprs with
    | []                   -> ()
    | [expr]               -> (print_expr_to_mathsat och expr vars apps true)
    | expr :: remain_exprs -> (
        print_expr_to_mathsat och expr vars apps true;
        Printf.fprintf och "%s" " && ";
        print_expr_and_bit och remain_exprs vars apps
    )
)

and print_expr_or_bit (och : out_channel) 
                      (exprs : expression list)
                      (vars : (string, classified_var_t) Hashtbl.t)
                      (apps : (string, classified_app_t) Hashtbl.t)
: unit = (
    match exprs with 
    | []                   -> ()
    | [expr]               -> (print_expr_to_mathsat och expr vars apps true)
    | expr :: remain_exprs -> (
        print_expr_to_mathsat och expr vars apps true;
        Printf.fprintf och "%s" " || ";
        print_expr_or_bit och remain_exprs vars apps 
    ) 
)

and print_expr_xor_bit (och : out_channel) 
                       (exprs : expression list)
                       (vars : (string, classified_var_t) Hashtbl.t)
                       (apps : (string, classified_app_t) Hashtbl.t)
: unit = (
    match exprs with 
    | []                   -> ()
    | [expr]               -> (print_expr_to_mathsat och expr vars apps true)
    | expr :: remain_exprs -> (
        print_expr_to_mathsat och expr vars apps true;
        Printf.fprintf och "%s" " ^^ ";
        print_expr_xor_bit och remain_exprs vars apps 
    )
)

and print_expr_lsh_bit (och :out_channel) 
                       (exprs : expression list)
                       (vars : (string, classified_var_t) Hashtbl.t)
                       (apps : (string, classified_app_t) Hashtbl.t)
: unit = (
    match exprs with
    | []                   -> ()
    | [expr]               -> (print_expr_to_mathsat och expr vars apps true)
    | expr :: remain_exprs -> (
        print_expr_to_mathsat och expr vars apps true;
        Printf.fprintf och "%s" " << ";
        print_expr_lsh_bit och remain_exprs vars apps
    ) 
) 

and print_expr_rsh_bit (och :out_channel) 
                       (exprs : expression list)
                       (vars : (string, classified_var_t) Hashtbl.t)
                       (apps : (string, classified_app_t) Hashtbl.t)
: unit = (
    match exprs with
    | []                   -> ()
    | [expr]               -> (print_expr_to_mathsat och expr vars apps true)
    | expr :: remain_exprs -> (
        print_expr_to_mathsat och expr vars apps true;
        Printf.fprintf och "%s" " >> ";
        print_expr_rsh_bit och remain_exprs vars apps
    ) 
) 

and print_expr_mod (och : out_channel)
                   (exprs : expression list)
                   (vars : (string, classified_var_t) Hashtbl.t)
                   (apps : (string, classified_app_t) Hashtbl.t)
: unit = (
    match exprs with
    | []                   -> ()
    | [expr]               -> (print_expr_to_mathsat och expr vars apps true)
    | expr :: remain_exprs -> (
        print_expr_to_mathsat och expr vars apps true;
        Printf.fprintf och "%s" " % ";
        print_expr_mod och remain_exprs vars apps
    )
)

and print_const (och : out_channel)
                (mix_cst : mix_const)
                (kind : bool) : unit = (
    assert (not (is_nan_mix_cst mix_cst));
    let print_bit_vector_of_num (och : out_channel)
                                (bv_prefix : string)
                                (n : num) : unit = (
        match n with
        | Int cst ->
            Printf.fprintf och "%s%d" bv_prefix cst
        | Big_int cst ->
            Printf.fprintf och "%s"
                (bv_prefix ^ (Big_int.string_of_big_int cst))
        | Ratio _ ->
            assert false;
    ) in

    let (n_op, f_op) = mix_cst in
    match n_op, f_op with
    | None, None -> assert false
    | Some n, _ -> (
        if (kind) then (
            if (ge_num n (num_of_string "0")) then
                print_bit_vector_of_num och "0d32_" n
            else (
                let tmp = minus_num n in
                print_bit_vector_of_num och "-0d32_" (tmp)
              )
          )
        else (
            match n with
            | Int cst -> Printf.fprintf och "%12d" cst
            | Big_int cst ->
                Printf.fprintf och "%30s" (Big_int.string_of_big_int cst)
            | Ratio _ ->
                Printf.fprintf och "%30.30f" (float_of_num n);)
      )
    | _, Some f -> (
        if (kind) then (
            let i =
                match get_int_of_float f with
                | None -> assert false
                | Some i_f -> i_f in
            if ( i >= 0 ) then (
                Printf.fprintf och "%s%d" "0d32_" i;
            )
            else (
                let tmp = - i in
                Printf.fprintf och "%s%d" "-0d32_" (tmp)
            );)
        else (
            Printf.fprintf och "%30.30f" (f);)
      )
)

(** Write a expression into a channel (a msat file).
        @param och : output channel used for msat input file
        @param expr : the expression need be writen
        @param kind : the kind of expression (true - bit; false - real)
        
        @return : unit
 *)
and print_expr_to_mathsat (och : out_channel) 
                          (expr : expression)
                          (vars : (string, classified_var_t) Hashtbl.t)
                          (apps : (string, classified_app_t) Hashtbl.t) 
                          (kind : bool)
: unit = (
    match expr with
    | Constant mix_cst ->
        print_const och mix_cst kind
    | Variable v -> (
        let replaced_special_char = replace_special_char v in
        Printf.fprintf och "%s" replaced_special_char
    )
    | Application ("and", andb_lst) -> (
        let len = List.length andb_lst in
        if len = 0 then ()
        else (
            Printf.fprintf och "%s"  ("(");
            print_expr_and_bit och andb_lst vars apps;
            Printf.fprintf och "%s"  (")");)
		)
    | Application ("ior", orb_lst) -> (
        let len = List.length orb_lst in
        if len = 0 then ()
        else (
            Printf.fprintf och "%s"  ("(");
            print_expr_or_bit och orb_lst vars apps;
            Printf.fprintf och "%s"  (")");)
		)
    | Application ("xor", xorb_lst) -> (
        let len = List.length xorb_lst in
        if len = 0 then ()
        else (
            Printf.fprintf och "%s"  ("(");
            print_expr_xor_bit och xorb_lst vars apps;
            Printf.fprintf och "%s"  (")");)
		)
    | Application ("lsh", lsh_lst) -> (
        let len = List.length lsh_lst in
        if len = 0 then ()
        else (
            Printf.fprintf och "%s"  ("(");
            print_expr_lsh_bit och lsh_lst vars apps;
            Printf.fprintf och "%s"  (")");)
		)
    | Application ("rsh", rsh_lst) -> (
        let len = List.length rsh_lst in
        if len = 0 then ()
        else (     
            Printf.fprintf och "%s"  ("(");
            print_expr_rsh_bit och rsh_lst vars apps;
            Printf.fprintf och "%s"  (")");)
		)
    | Application ("bitnot", lst)  -> (
        let len = List.length lst in
        if ( len = 1 ) then (
            Printf.fprintf och "%s" ("(~~");
            print_expr_to_mathsat och ( List.hd lst ) vars apps true;
            Printf.fprintf och "%s" (")");  
        )
        else ( assert false )
    )
    | Application ("mod", mod_lst) -> (
        let len = List.length mod_lst in
        if len = 0 then ()
        else (      
            Printf.fprintf och "%s"  ("(");
            print_expr_mod och mod_lst vars apps;
            Printf.fprintf och "%s"  (")");)
    )
    | Application (sym, lst) -> (
        let (name_app, decl_lst) =
            try Hashtbl.find apps sym
            with Not_found -> assert false
        in
        let para_lst = try List.tl decl_lst
                       with _ -> assert false
        in
        if (List.length lst) > 0 then (
            if (List.length lst <> List.length para_lst) then assert false
            else (
                Printf.fprintf och "%s" (name_app ^ "(");
                print_expr_application och lst vars apps para_lst;
                Printf.fprintf och "%s"  (")");)    
        )
        else Printf.fprintf och "%s" name_app;
    )
    | Sum lst -> (
        let len = List.length lst in
        if len = 0 then ()
        else (      
            Printf.fprintf och "%s"  ("(");
            print_expr_sum och lst vars apps kind;
            Printf.fprintf och "%s"  (")");)
    )
    | Coeff (co, expr1) -> (
        print_const och co kind;         
        Printf.fprintf och "%s" " * ";
        print_expr_to_mathsat och expr1 vars apps kind;
    )
)

let rec print_pred_and (och : out_channel) 
                       (preds : predicate list)
                       (vars : (string, classified_var_t) Hashtbl.t)
                       (apps : (string, classified_app_t) Hashtbl.t)
: unit = (
    match preds with
    | [] -> ()
    | [pred] -> print_pred_all och pred vars apps
    | pred :: remain_preds -> (
        print_pred_all och pred vars apps;
        Printf.fprintf och "%s\n" " & ";
        print_pred_and och remain_preds vars apps
      )
)

and print_pred_or (och : out_channel)
                  (preds : predicate list)
                  (vars : (string, classified_var_t) Hashtbl.t)
                  (apps : (string, classified_app_t) Hashtbl.t)
: unit = (
    match preds with
    | [] -> ()
    | [pred] -> print_pred_all och pred vars apps
    | pred :: remain_preds -> (
        print_pred_all och pred vars apps ;
        Printf.fprintf och "%s\n" " | ";
        print_pred_or och remain_preds vars apps
      )
)

(** Write a predicate into a output channel (a msat file).
        @param och : output channel used for input msat file
        @param pred : predicate need be writen
        @param vars : all of variables of predicate, include kind of 
                      variable (WORD [32] or REAL)
        @param apps : all of application of predicate, include kind of 
                      parameters and return value (WORD [32] or REAL)
                      
        @return : unit   
 *)
and print_pred_all (och : out_channel) 
                   (pred : predicate)
                   (vars : (string, classified_var_t) Hashtbl.t)
                   (apps : (string, classified_app_t) Hashtbl.t)
: unit = (
    let print_comp_expr (e1 : expression)
                        (e2 : expression)
                        (vars : (string, classified_var_t) Hashtbl.t)
                        (apps : (string, classified_app_t) Hashtbl.t)
                        (kind : bool)
                        (comp : string)
    : unit = (
        Printf.fprintf och "%s" ( "(" );
        print_expr_to_mathsat och e1 vars apps kind;
        Printf.fprintf och "%s" (" " ^ comp ^ " ");
        print_expr_to_mathsat och e2 vars apps kind;
        Printf.fprintf och "%s\n" ( ")" );
    ) in
    match pred with
    | False ->
        Printf.fprintf och "%s\n" ( "false" );
    | True ->
        Printf.fprintf och "%s\n" ( "true" );
    | And lst -> (
        let len = List.length lst in
        if len = 0 then ()
        else (      
            Printf.fprintf och "%s" ( "(" );
            print_pred_and och lst vars apps;
            Printf.fprintf och "%s\n" ( ")" );)
    )
    | Or lst -> (
        let len = List.length lst in
        if len = 0 then ()
        else (
            Printf.fprintf och "%s" ( "(" );
            print_pred_or och lst vars apps;
            Printf.fprintf och "%s\n" ( ")" );)
    )
    | Not p -> (
        Printf.fprintf och "%s" ( "!(" );
        print_pred_all och p vars apps;
        Printf.fprintf och "%s\n" ( ")" );
    )
    | Eq (e1,e2) -> (
        let kind_e1 = classify_expr e1 vars apps
        in
        let kind_e2 = classify_expr e2 vars apps 
        in
        if (kind_e1 || kind_e2) then
            print_comp_expr e1 e2 vars apps true "="
        else 
            print_comp_expr e1 e2 vars apps false "="
    )
    | Lt (e1,e2) -> (
        let kind_e1 = classify_expr e1 vars apps 
        in
        let kind_e2 = classify_expr e2 vars apps 
        in
        if (kind_e1 || kind_e2) then
            print_comp_expr e1 e2 vars apps true "<"
        else 
            print_comp_expr e1 e2 vars apps false "<"
    )
    | Leq (e1,e2) -> (
        let kind_e1 = classify_expr e1 vars apps  
        in
        let kind_e2 = classify_expr e2 vars apps  
        in
        if (kind_e1 || kind_e2) then
            print_comp_expr e1 e2 vars apps true "<="
        else 
            print_comp_expr e1 e2 vars apps false "<="
    )
    | Atom (Ast.External str)	-> assert false
    | Atom (Ast.Internal i)   -> assert false
)

let print_pred_to_mathsat (och : out_channel) 
                          (pred : predicate)
                          (vars : (string, classified_var_t) Hashtbl.t)
                          (apps : (string, classified_app_t) Hashtbl.t)
: unit = (
    Printf.fprintf och "%s" ( "FORMULA " );
    print_pred_all och pred vars apps;
)

(** Prepare the kind of variable to write into msat file.
    @param kind : the kind of variable
    @return if kind is true will write WORD [32].
            If kind is false will write REAL 
 *)
let rec create_var_decl (kind : bool) : string = 
begin
    if kind then
        "WORD [32]"
    else "REAL"
end

(** Prepare to write the declaration of a application into msat file
        @param decl_lst : informations of return value and parameters. 
                          The head of decl_lst is the kind of return value, and
                          remain list is correspondingly parameter list
        @return : a string is the declaration of application
    Ex : [true] -> WORD [32]
         [true, false, true] -> REAL * WORD [32] -> WORD [32]   
 *)
let create_func_decl (decl_lst : bool list): string
= begin
    let rec create_para_lst (para_lst : bool list)
    : string = (
        match para_lst with
        | [] -> ""
        | [a_para] -> 
            if a_para then
                "WORD [32] -> "
            else "REAL -> "
        | para::remain_paras ->
            if para then
                "WORD [32] * " ^ (create_para_lst remain_paras)    
            else "REAL * " ^ (create_para_lst remain_paras)          
    ) in
    match decl_lst with
    | [] -> assert false
    | head::tail -> (
        if head then
            (create_para_lst tail) ^ "WORD [32]"
        else (create_para_lst tail) ^ "REAL"
    )
end

(** Write informations (formula informations and variable informations) of a 
    predicate into a msat file.
        @param fileout : the name of a msat file
        @param pred : the predicate need be writen
        
        @return : unit
 *)     
let write_into_a_msat_file (fileout : string) (pred : predicate)
: unit = (
    let och = try open_out fileout
              with Sys_error _ -> assert false
    in
    let vars : (string, classified_var_t) Hashtbl.t = Hashtbl.create 2 in
    let apps : (string, classified_app_t) Hashtbl.t = Hashtbl.create 2 in
    (* collect variables and applications from predicate,
     until having no change about the kind of vaiables or applications*)
    while (collect_define_in_pred pred vars apps) do () done;
    let vars_to_write = Hashtbl.fold (fun name (new_name, kind) new_names ->
                            new_names @ [new_name ^ " : " ^ (create_var_decl kind)]
                        ) vars [] in
    let apps_to_write = Hashtbl.fold (fun name (new_name, decl_lst) new_names ->
                            new_names @
                                [new_name ^ " : " ^ (create_func_decl decl_lst)]
                        ) apps [] in
    print_list_vars och (vars_to_write @ apps_to_write);
    print_pred_to_mathsat och pred vars apps;
    Hashtbl.clear vars;
    Hashtbl.clear apps;
    try (
        close_out och;)
    with Sys_error _ -> assert false;
)

(** Write informations (formula informations and variable informations) of 
    predicate list into a msat file list
        @param preds : predicate list. Every predicate is writen into a msat file
        @param vars : all variable in predicate list
        @param apps : all application in predicate list
        @param intmp_lst : name list of writen file list
        @param f_order : the number of file, use to distinguish these files.
        
        @return : name list of all files
 *)   
let rec write_preds_into_mathsat_files (preds : predicate list) 
                                       (vars : (string, classified_var_t) Hashtbl.t) 
                                       (apps : (string, classified_app_t) Hashtbl.t)
				                               (intmp_lst : string list)
                                       (prefix : string)
                                       (f_order : int) 
: string list =
begin
    match preds with
    | [] -> ( intmp_lst )
    | head :: tails -> (
        let input_file = (prefix ^ "." ^ (string_of_int f_order) ^ ".msat") in
        let och = try open_out input_file
                  with Sys_error _ -> assert false
        in
        let vars_to_write = Hashtbl.fold (fun name (new_name, kind) new_names ->
                                new_names @ [new_name ^ " : " ^ (create_var_decl kind)]
                                         ) vars [] in
        let apps_to_write = Hashtbl.fold (fun name (new_name, decl_lst) new_names ->
                                new_names @
                                    [new_name ^ " : " ^ (create_func_decl decl_lst)]
                                          ) apps [] in
        print_list_vars och (vars_to_write @ apps_to_write);
        print_pred_to_mathsat och head vars apps;
        try (
            close_out och;
            write_preds_into_mathsat_files tails vars apps 
                                           (intmp_lst @ [input_file]) 
                                           (prefix)
                                           (f_order + 1);
        )
        with Sys_error _ -> assert false;			  								
    )
end

(* ~~~~~~~~~~~~~~~~ Construct predicates from mathsat file ~~~~~~~~~~~~~~~~~~ *)
let rec replace_var_by_var_expression (expr : expression) 
                                      (vars : (string, string) Hashtbl.t)
: expression =
begin
    match expr with
    | Constant cst -> Constant cst
    | Variable v -> (
        let lst = Hashtbl.find_all vars v in
        if List.length lst = 1 then (
            Variable ( List.hd lst )
        )
        else ( assert false ) 
    )
    | Application ("bitnot", expr_list) -> (
        let new_expr_list = ref [] in
        List.iter 
            ( fun a_expr -> 
                let new_expr = replace_var_by_var_expression a_expr vars in
                    new_expr_list := !new_expr_list @ [new_expr]
            )
            expr_list;
        Application ("bitnot", !new_expr_list)
    )   
    | Application ("and", expr_list) -> (
        let new_expr_list = ref [] in
        List.iter 
            ( fun a_expr -> 
                let new_expr = replace_var_by_var_expression a_expr vars in
                    new_expr_list := !new_expr_list @ [new_expr]
            )
            expr_list;
        Application ("and", !new_expr_list)
    ) 
    | Application ("ior", expr_list) -> (
        let new_expr_list = ref [] in
        List.iter 
            ( fun a_expr -> 
                let new_expr = replace_var_by_var_expression a_expr vars in
                    new_expr_list := !new_expr_list @ [new_expr]
            )
            expr_list;
        Application ("ior", !new_expr_list)
    )  
    | Application ("xor", expr_list) -> (
        let new_expr_list = ref [] in
        List.iter 
            ( fun a_expr -> 
                let new_expr = replace_var_by_var_expression a_expr vars in
                    new_expr_list := !new_expr_list @ [new_expr]
            )
            expr_list;
        Application ("xor", !new_expr_list)
    )
    | Application ("lsh", expr_list) -> (
        let new_expr_list = ref [] in
        List.iter 
            ( fun a_expr -> 
                let new_expr = replace_var_by_var_expression a_expr vars in
                    new_expr_list := !new_expr_list @ [new_expr]
            )
            expr_list;
        Application ("lsh", !new_expr_list)
    ) 
    | Application ("rsh", expr_list) -> (
        let new_expr_list = ref [] in
        List.iter 
            ( fun a_expr -> 
                let new_expr = replace_var_by_var_expression a_expr vars in
                    new_expr_list := !new_expr_list @ [new_expr]
            )
            expr_list;
        Application ("rsh", !new_expr_list)
    )
    | Application ("mod", expr_list) -> (
        let new_expr_list = ref [] in
        List.iter 
            ( fun a_expr -> 
                let new_expr = replace_var_by_var_expression a_expr vars in
                    new_expr_list := !new_expr_list @ [new_expr]
            )
            expr_list;
        Application ("mod", !new_expr_list)
    )   
    | Application ( sym, expr_list ) -> (
        let new_expr_list = ref [] in
        List.iter 
            ( fun a_expr -> 
                let new_expr = replace_var_by_var_expression a_expr vars in
                    new_expr_list := !new_expr_list @ [new_expr]
            )
            expr_list;
        let new_syms = Hashtbl.find_all vars sym in
        assert ((List.length new_syms) = 1);
        let new_sym = List.hd new_syms in
        Application (new_sym, !new_expr_list)
      )
    | Sum expr_list -> (
        let new_expr_list = ref [] in
        (* replace var in each expr in expr_list *)
        List.iter 
            (fun a_expr -> 
                let new_expr = replace_var_by_var_expression a_expr vars in
                new_expr_list := !new_expr_list @ [new_expr]
            ) expr_list;
        Sum !new_expr_list
    )
    | Coeff (c, expr) -> (
        let new_expr = replace_var_by_var_expression expr vars in
        Coeff (c, new_expr)
    )
end;;

(** Return the real names for variables and predicates which was replaced 
    special chars (in 'replace_special_char' function) 
 *)
let rec replace_var_by_var_predicate ( pred : predicate ) 
                                     ( vars : (string, string) Hashtbl.t ) 
: predicate = 
begin
    match pred with
    | True -> True
    | False -> False
    | And pred_list -> (
        let new_pred_list = ref [] in
        List.iter 
            ( fun a_pred -> 
                let new_pred = replace_var_by_var_predicate a_pred vars in
                new_pred_list := !new_pred_list @ [new_pred]
            )
            pred_list;
        And !new_pred_list
    )
    | Or pred_list -> (
        let new_pred_list = ref [] in
        List.iter 
            (fun a_pred -> 
                let new_pred = replace_var_by_var_predicate a_pred vars in
			          new_pred_list := !new_pred_list @ [new_pred]
            )
            pred_list;
        Or !new_pred_list
    )
    | Not pred -> (
        let new_pred = replace_var_by_var_predicate pred vars in
        Not new_pred
    )
    | Eq (expr1, expr2) -> ( 
        let new_expr1 : expression = 
            replace_var_by_var_expression expr1 vars 
        in
        let new_expr2 : expression = 
            replace_var_by_var_expression expr2 vars 
        in
		    Eq (new_expr1, new_expr2)
    )
    | Lt (expr1, expr2) -> ( 
        let new_expr1 : expression =  
            replace_var_by_var_expression expr1 vars 
		    in
        let new_expr2 : expression = 
            replace_var_by_var_expression expr2 vars 
        in
        Lt (new_expr1, new_expr2)
    )
    | Leq (expr1, expr2) -> ( 
        let new_expr1 : expression = 
            replace_var_by_var_expression expr1 vars 
        in
        let new_expr2 : expression = 
            replace_var_by_var_expression expr2 vars 
        in
        Leq (new_expr1, new_expr2)
     )
     | Atom i -> assert false
end;;

(* ~~~~~~~~~~~ Determine a predicate to be bit vector kind or not ~~~~~~~~~~ *)
let rec determine_bit_expr (expr : expression)
: bool = (
     match expr with
    | Constant cst -> false
    | Variable v -> false
    | Application ("and", lst) 
    | Application ("ior", lst) 
    | Application ("xor", lst) 
    | Application ("lsh", lst) 
    | Application ("rsh", lst)
    | Application ("mod", lst)
    | Application ("bitnot", lst)  -> true
    | Application (sym, lst) ->
        List.exists 
            (fun a_expr -> 
                determine_bit_expr a_expr
            ) lst
    | Sum lst -> 
        List.exists 
            (fun a_expr -> 
                determine_bit_expr a_expr
            ) lst
    | Coeff (co, expr1) -> 
        determine_bit_expr expr1        
) 

let rec determine_bit_pred (pred : predicate)
: bool = (
    match pred with
    | False 
    | True  -> false
    | And lst
    | Or lst ->
        List.exists 
            (fun a_pred -> 
                determine_bit_pred a_pred
            ) lst
    | Not p -> 
            determine_bit_pred p 
    | Eq (e1,e2)
    | Lt (e1,e2)
    | Leq (e1,e2) -> 
        determine_bit_expr e1 || determine_bit_expr e2  
    | Atom (Ast.External str) -> assert false
    | Atom (Ast.Internal i) -> assert false  
)

let big_number_threshold = 10000.

let rec contains_big_numbers_expr (expr : expression)
: bool = (
    match expr with
    | Constant cst -> (
        let f = get_float_from_mix_cst cst in
        f > big_number_threshold)
    | Variable v -> false
    | Application (_, lst)
    | Sum lst -> 
        List.exists contains_big_numbers_expr lst
    | Coeff (cst, expr) ->
        let f = get_float_from_mix_cst cst in
        (f > big_number_threshold) || (contains_big_numbers_expr expr)
) 

let rec contains_big_numbers (pred : predicate)
: bool = (
    match pred with
    | False
    | True -> false
    | And lst
    | Or lst ->
        List.exists 
            (fun a_pred -> 
                contains_big_numbers a_pred
            ) lst
    | Not p -> 
            contains_big_numbers p 
    | Eq (e1,e2)
    | Lt (e1,e2)
    | Leq (e1,e2) -> 
        contains_big_numbers_expr e1 || contains_big_numbers_expr e2
    | Atom (Ast.External str) -> assert false
    | Atom (Ast.Internal i) -> assert false  
)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Get name list of msat output files from name list of msat input files 
        @param in_lst : name list of msat input files
        @return : name list of msat output files 
    Ex : ip1.msat -> op1.msat  
*)
let rec get_name_list_files (in_lst : string list) : string list =
begin
    match in_lst with 
    | []            -> (assert false)
    | [x]           -> ([])   
    | head :: tails -> (
        let (bol, op_file) =
            (ExtString.String.replace ~str:head ~sub:"input" ~by:"output")
        in 
        if (bol) then (
            [op_file] @ ( get_name_list_files tails)
        )
        else (assert false)
    )
end;;

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ CHECK VALID ~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(** Read a msat result file to analyse result information (sat, unsat or error)  
 *)
let read_sat_file (filename : string) : bool = (
    let ich = try open_in filename
              with Sys_error _ -> assert false in
    let lines = ref [] in
    let _ =
    try
        while true do
            let line = ExtString.String.strip ~chars:"\t\n " (input_line ich) in
            if line <> "" then
                lines := line :: !lines;
	      done
    with End_of_file -> close_in ich in
    try
        match List.hd !lines with
        | "sat" -> true
        | "unsat" -> false
        | _ -> assert false
    with Failure _ -> assert false
)

(** Compute sat for a pred by MathSat  
 *)
let is_sat_of_mathsat (pred : predicate) : bool = (
    if pred = False then false
    else if pred = True then true
    else (
    let random_file = (Filename.temp_file "" "") in
    let run_hashstring = !temp_dir ^ "/" ^ (Filename.basename random_file)
    in
    let input_file = (run_hashstring ^ "_input.msat")
    and output_file = (run_hashstring ^ "_output.msat") in
    let tmp_bool = determine_bit_pred pred in
    write_into_a_msat_file input_file pred;
    let set_verbosity_str = "-verbosity="
                            ^ (string_of_int !mathsat_verbosity_level)
    in
    let cmd = if tmp_bool then
                  ((Filename.dirname Sys.executable_name)
                   ^ "/../mathsat-4/bin/mathsat "
                   ^ set_verbosity_str
                   ^ " -tcomb=ack -solve -tsolver=euf -tsolver=bv "
                   ^ input_file ^ " > " ^ output_file)
              else
                  ((Filename.dirname Sys.executable_name)
                   ^ "/../mathsat-4/bin/mathsat "
                   ^ set_verbosity_str
                   ^ " -tcomb=ack -solve -tsolver=euf -tsolver=la "
                   ^ input_file ^ " > " ^ output_file) 
    in
    let start_time = Unix.time () in
    Thread.join (Thread.create Fx.command cmd);
    let stop_time = Unix.time () in
    total_mathsat_time := !total_mathsat_time +. stop_time -. start_time;
    let sat = read_sat_file output_file in
    (Sys.remove input_file);
    (Sys.remove output_file);
    (Sys.remove random_file);
    sat)
)

