#ifdef __ONLY_FOR_MODELCHECKING
#include <assert.h>
#include <vmm-mc.h>
#else
#include <kernel/page_fault.h>
#include <kernel/guest.h>
#include <kernel/guest_paging.h>
#include <kernel/logger.h>
#include <kernel/panic.h>
#endif

#if 0
#ifdef __ONLY_FOR_MODELCHECKING
#include <assert.h>
#include <vmm-mc.h>
#else
#include <kernel/guest_paging.h>
#include <kernel/cpu.h>
#include <kernel/guest.h>
#include <kernel/tempref.h>
#include <kernel/logger.h>
#include <kernel/profile.h>
#include <kernel/percpu.h>
#include <string.h>
#include <power2.h>
#include <assert.h>
#endif
#endif

#ifndef __ONLY_FOR_MODELCHECKING
#ifdef PROFILE
static profile_id_t profile_flush_gpage_all;
static profile_id_t profile_flush_gpage;
static profile_id_t profile_flush_gpage_at;
static profile_id_t profile_update_gpage;
#endif
#endif

typedef struct {
	uint top;
	uint end;
} page_range_t;

static page_range_t global_page_range AT_PERCPU;
static page_range_t non_global_page_range AT_PERCPU;


static inline void
page_range_init(page_range_t *p)
{
	p->top = NUM_PDE*NUM_PTE;
	p->end = 0;
}


static inline void
page_range_update(page_range_t *p,uint top,uint len)
{
	uint end = top + len;
	if (top < p->top) p->top = top;
	if (end > p->end) p->end = end;
}

static inline void
page_range_update_by_1(page_range_t *p,uint top)
{
	if (top < p->top) p->top = top;
	if (top >= p->end) p->end = top + 1;
}


static inline bool
page_range_exists(page_range_t *p)
{
	return p->top < p->end;
}


#if 0
static void
print_page_range(page_range_t *p)
{
	printf("0x%08x ... 0x%08x\n",p->top*PAGE_SIZE,p->end*PAGE_SIZE - 1);
}
#endif


static inline page_entry_t
get_guest_pde(uint pde_index)
{
	vmm_assert(pde_index < NUM_PDE);
	page_directory_t *dir = (page_directory_t *)TEMPREF_MAP_G(cpu.guest.cr[3] & CR3_BASE_MASK,sizeof(page_directory_t));
	page_entry_t pde = dir->entries[pde_index];
	tempref_unmap((addr_t)dir,sizeof(page_directory_t));
	return pde;
}


static inline page_entry_t
get_guest_pte(page_entry_t pde,uint pte_index)
{
	vmm_assert(pde & PAGE_ENTRY_P);
	vmm_assert(!(pde & PAGE_ENTRY_PS));
	page_table_t *table = (page_table_t *)TEMPREF_MAP_G(pde & PAGE_MASK,sizeof(page_table_t));
	page_entry_t pte = table->entries[pte_index];
	tempref_unmap((addr_t)table,sizeof(page_table_t));
	return pte;
}


void
guest_paging_init()
{
	if (i_am_bsp()) {
		PROFILE_NEW(flush_gpage_all);
		PROFILE_NEW(flush_gpage);
		PROFILE_NEW(flush_gpage_at);
		PROFILE_NEW(update_gpage);
	}
	non_global_page_range.top = 0;
	non_global_page_range.end = VMM_TOP >> PAGE_SHIFT;
}


/**
 * ゲストのページングが無効の場合のページディレクトリを作成します。
 **/
void
setup_guest_no_paging(void)
{
	vmm_assert(!guest_paging_is_enabled());
	uint vmm_top = VMM_TOP >> PAGE_SHIFT;
	paging_unmap_pages(&cpu.paging,0,vmm_top,false);

	uint n = gpmap_region_size(&cpu.config->guest.gpmap);
	uint i;
	for (i = 0 ; i < n ; ++i) {
		uint gtop = gpmap_nth_region_gtop(&cpu.config->guest.gpmap,i);
		uint len = gpmap_nth_region_len(&cpu.config->guest.gpmap,i);
		uint pos;
		for (pos = 0 ; pos < len && gtop + pos < vmm_top ; ++pos) {
			uint gpos = gtop + pos;
			uint ppos = gpmap_nth_region_ppage_at(&cpu.config->guest.gpmap,i,pos);
			paging_map_pages(&cpu.paging,
					 gpos,
					 ppos,
					 1,
					 PAGE_ENTRY_RW | PAGE_ENTRY_A | PAGE_ENTRY_D,
					 false);
		}
	}
	flush_tlb();
}


/**
 * ゲストからコピーしたページテーブルを無効化します。
 **/
void
flush_guest_paging(bool include_global) PROFILE_FUNCTION(include_global?profile_flush_gpage_all:profile_flush_gpage,void)
{
	print_detail1("flush_guest_paging(include_global = %d)\n",include_global);

	if (include_global && page_range_exists(&global_page_range)) {
		/*
		 * グローバル指定されたページテーブルを無効化します。
		 */
		uint top = global_page_range.top;
		uint end = global_page_range.end;
		vmm_assert(top < end);
		vmm_assert(end <= VMM_TOP >> PAGE_SHIFT);

		/*
		 * グローバル指定された個々のページについてtlbをフラッシュします。
		 */
		page_entry_t *pte = &(&cpu.page_tables[0].entries[0])[top];
		uint i;
		for (i = top ; i < end ; ) {
			if (is_p2aligned(i,NUM_PTE)) {
				page_entry_t pde = cpu.page_directory.entries[i/NUM_PTE];
				if (pde & PAGE_ENTRY_P) {
					if (pde & PAGE_ENTRY_PS) {
						if (pde & PAGE_ENTRY_G)
							flush_tlb_at(i*PAGE_SIZE);
						i += NUM_PTE;
						pte += NUM_PTE;
						continue;
					}
				} else {
					i += NUM_PTE;
					pte += NUM_PTE;
					continue;
				}
			}
			if ((*pte & PAGE_ENTRY_P) && (*pte & PAGE_ENTRY_G))
				flush_tlb_at(i*PAGE_SIZE);
			++i;
			++pte;
		}

		/*
		 * ページディレクトリのみをゼロクリアします。
		 */
		memsetl(&cpu.page_directory.entries[top/NUM_PTE],0,(end + NUM_PTE - 1)/NUM_PTE - top/NUM_PTE);
		page_range_init(&global_page_range);
	}
	if (page_range_exists(&non_global_page_range)) {
		/*
		 * グローバル指定されていないページテーブルを無効化します。
		 */
		uint top = non_global_page_range.top;
		uint end = non_global_page_range.end;
		vmm_assert(top < end);
		vmm_assert(end <= VMM_TOP >> PAGE_SHIFT);

		/*
		 * ページディレクトリのみをゼロクリアします。
		 */
		memsetl(&cpu.page_directory.entries[top/NUM_PTE],0,(end + NUM_PTE - 1)/NUM_PTE - top/NUM_PTE);
		page_range_init(&non_global_page_range);
	}
	flush_tlb();
} PROFILE_FUNCTION_END_VOID


/**
 * ゲストからコピーしたページテーブルの指定アドレスに対応するエントリー
 * を無効化します。
 **/
void
flush_guest_paging_at(addr_t vaddr) PROFILE_FUNCTION(profile_flush_gpage_at,void)
{
	print_detail1("flush_guest_paging_at(vaddr = 0x%08x)\n",vaddr);
	vmm_assert(guest_paging_is_enabled());
	guest_assert(!(VMM_TOP <= vaddr && vaddr < VMM_END));

	uint pde_index = vaddr >> L2_PAGE_SHIFT;
	page_entry_t *pde = &cpu.page_directory.entries[pde_index];
	if (*pde & PAGE_ENTRY_P) {
		if (*pde & PAGE_ENTRY_PS) {
			*pde = 0;
		} else {
			page_entry_t gpde = get_guest_pde(pde_index);
			if (gpde & PAGE_ENTRY_PS) {
				*pde = 0;
			} else {
				page_entry_t *pte = &cpu.page_tables[0].entries[vaddr >> L1_PAGE_SHIFT];
				*pte = 0;
			}
		}
	}
	flush_tlb_at(vaddr);
} PROFILE_FUNCTION_END_VOID


/**
 * 指定アドレスのページテーブルをゲストからコピーします。
 **/
bool
update_guest_paging_at(addr_t vaddr) PROFILE_FUNCTION(profile_update_gpage,bool)
{
	print_detail2("update_guest_paging_at(vaddr = 0x%08x)\n",vaddr);
	vmm_assert(guest_paging_is_enabled());
	guest_assert(vaddr < VMM_TOP);

	/*
	 * ページディレクトリエントリーの処理
	 */
	uint pde_index = vaddr >> L2_PAGE_SHIFT;
	page_entry_t gpde = get_guest_pde(pde_index);
	if (!(gpde & PAGE_ENTRY_P))
		/*
		 * ページが存在しない
		 */
		return false;
	page_entry_t *ppde = &cpu.page_directory.entries[pde_index];
	if (!(*ppde & PAGE_ENTRY_P)) {
#ifdef GUEST_PAGING_USE_4M_PAGE
		bool guest_page_is_straight(uint gpos,uint len)
		{
			if (len > 0) {
				gpmap_t *gpmap = &cpu.config->guest.gpmap;
				uint pprev = gpmap_gpos2p(gpmap,gpos++);
				for (--len ; len > 0 ; --len,++gpos) {
					uint ppos = gpmap_gpos2p(gpmap,gpos);
					if (ppos != pprev + 1)
						return false;
				}
			}
			return true;
		}
		if ((gpde & PAGE_ENTRY_PS) &&
		    (gaddr2p(gpde & L2_PAGE_MASK) & L2_PAGE_OFFSET_MASK) == 0 &&
		    guest_page_is_straight((gpde & L2_PAGE_MASK) >> PAGE_SHIFT,NUM_PTE)) {
			/*
			 * ゲスト用の空間で物理ページが4M境界でかつ連続
			 * しているので4Mページを使用する。
			 */
			print_debug("use guest 4M page at 0x%08x\n",gpde & L2_PAGE_MASK);
			*ppde = gaddr2p(gpde & L2_PAGE_MASK);
			*ppde |= gpde & PAGE_OFFSET_MASK;
			page_range_update(&global_page_range,pde_index*NUM_PTE,NUM_PTE);
			return true;
		}
#endif
		/*
		 * 4Mページが指定されていても物理ページが4M境界になかっ
		 * たり連続していなかったりするので4Kページに変換する。
		 */
		*ppde = paging_get_table_paddr(&cpu.paging,&cpu.page_tables[pde_index]);
		*ppde |= gpde & PAGE_OFFSET_MASK & ~PAGE_ENTRY_PS;

		/*
		 * flush_guest_pagingではページディレクトリのみをクリア
		 * しているので、ここでページテーブルをクリアする。
		 */
		memsetl(&cpu.page_tables[pde_index].entries[0],0,NUM_PTE);
	}

	/*
	 * ページテーブルエントリーの処理
	 */
	uint pte_index = (vaddr & L2_PAGE_OFFSET_MASK) >> PAGE_SHIFT;
	page_entry_t *ppte = &cpu.page_tables[pde_index].entries[pte_index];
	vmm_assert(gpde & PAGE_ENTRY_P);
	if (gpde & PAGE_ENTRY_PS) {
		/*
		 * 4Kページに変換された4Mページ
		 */
		*ppte = gaddr2p((gpde & L2_PAGE_MASK) + PAGE_SIZE*pte_index);
		*ppte |= gpde & L1_PAGE_OFFSET_MASK & ~PAGE_ENTRY_PS;
	} else {
		/*
		 * ゲストのページテーブルエントリーをコピー
		 */
		page_entry_t gpte = get_guest_pte(gpde,pte_index);
		if (!(gpte & PAGE_ENTRY_P))
			return false;
		*ppte = gaddr2p(gpte & PAGE_MASK);
		*ppte |= gpte & PAGE_OFFSET_MASK;
	}
	page_range_update_by_1((*ppte & PAGE_ENTRY_G)?&global_page_range:&non_global_page_range,pde_index*NUM_PTE + pte_index);
	return true;
} PROFILE_FUNCTION_END

void
page_fault(regs_t *regs)
{
	u32 cr2 = read_cr2();
	print_detail2("%p: page fault at %p (error_code=0x%08x)\n",(void *)regs->eip,(void *)cr2,regs->error_code);
	
	if ((regs->error_code & PF_ERR_P) == 0 && guest_paging_is_enabled() && update_guest_paging_at(cr2))
		/*
		 * ゲスト側にページ情報が存在したのでページング情報を更
		 * 新して再実行する。
		 */
		return;

	if (SEGSEL_RPL(regs->cs) == VMM_RING) {
		print_panic("********************************\n");
		call_panic(print_regs(regs));
		print_panic("cr2 = %p\n",(void *)cr2);
		PANIC("page fault in vmm kernel.\n");
#warning ページアウトされたゲストのページへのアクセスに対応すべき？
	}


#ifndef __ONLY_FOR_MODELCHECKING
	/*
	 * ゲストに配送する
	 */
	guest_assert(cr2 < VMM_TOP || VMM_END <= cr2);
	cpu.guest.cr[2] = cr2;
	setup_upcall(regs,regs->vector,true,regs->error_code);
#endif
}

