/*
  Simple ring buffer implementation

  expected result: SAFE
 */

#include <assert.h>

#define RBUF_MAXLEN 8192

/*@ 
    $ensures ($zero (*ptr));
*/
void* memset ( void* ptr, int value, int num );

/*@
  $requires (dividend >= 0 && divisor > 0);
 */
static int round(int dividend, int divisor)
{
	while (dividend >= divisor)
		dividend -= divisor;

	return dividend;
}

struct rbuf {
	char buff[RBUF_MAXLEN];
	int rp, len;
};

static void rbuf_init(struct rbuf *rbuf)
{
	memset(rbuf, 0, sizeof(*rbuf));
}

static const char *rbuf_rp(struct rbuf *rbuf)
{
	return rbuf->buff + rbuf->rp;
}

static char *rbuf_wp(struct rbuf *rbuf)
{
	return rbuf->buff + round(rbuf->rp + rbuf->len, RBUF_MAXLEN);
}

/*@
  $ensures($result >= 0);
  $ensures($result <= rbuf->len);
 */
static int rbuf_readable(const struct rbuf *rbuf)
{
	return RBUF_MAXLEN - rbuf->rp < rbuf->len ?
		RBUF_MAXLEN - rbuf->rp
		: rbuf->len;
}

/*@
  $ensures($result >= 0);
  $ensures($result <= RBUF_MAXLEN - rbuf->len);
 */
static int rbuf_writable(const struct rbuf *rbuf)
{
	int wp = round(rbuf->rp + rbuf->len, RBUF_MAXLEN);

	return RBUF_MAXLEN - rbuf->len < RBUF_MAXLEN - wp ?
		RBUF_MAXLEN - rbuf->len :
		RBUF_MAXLEN - wp;
}

/*@
  $requires (0 <= len);
  $requires (len <= rbuf_writable(rbuf));
 */
static void rbuf_write(struct rbuf *rbuf, int len)
{
	rbuf->len += len;
}

/*@
  $requires (0 <= len);
  $requires (len <= rbuf_readable(rbuf));
 */
static void rbuf_read(struct rbuf *rbuf, int len)
{
	rbuf->rp += len;
	rbuf->len -= len;
	while (rbuf->rp >= RBUF_MAXLEN)
		rbuf->rp -= RBUF_MAXLEN;

	if (rbuf->len == 0)
		rbuf->rp = 0;
}

static void rbuf_test(void)
{
	struct rbuf rbuf;

	rbuf_init(&rbuf);

	/*@
	  $assert(rbuf_readable(&rbuf) == 0);
	  $assert(rbuf_writable(&rbuf) == RBUF_MAXLEN);
	*/

	rbuf_write(&rbuf, 1);

	/*@
	  $assert(rbuf_readable(&rbuf) == 1);
	  $assert(rbuf_writable(&rbuf) == RBUF_MAXLEN-1);
	*/

	rbuf_read(&rbuf, 1);

	/*@
	  $assert(rbuf_readable(&rbuf) == 0);
	  $assert(rbuf_writable(&rbuf) == RBUF_MAXLEN);
	*/

	rbuf_write(&rbuf, RBUF_MAXLEN-1);

	/*@
	  $assert(rbuf_readable(&rbuf) == RBUF_MAXLEN-1);
	  $assert(rbuf_writable(&rbuf) == 1);
	*/

	rbuf_write(&rbuf, 1);

	/*@
	  $assert(rbuf_readable(&rbuf) == RBUF_MAXLEN);
	  $assert(rbuf_writable(&rbuf) == 0);
	*/

	rbuf_read(&rbuf, RBUF_MAXLEN);

	/*@
	  $assert(rbuf_readable(&rbuf) == 0);
	  $assert(rbuf_writable(&rbuf) == RBUF_MAXLEN);
	*/
}

int main()
{
	rbuf_test();
}
